/*
 * File         : adlib_BlockRange.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:06:49 1997
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1998/01/08 21:35:19 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_BlockRange.h"
#include "adlib.h"
#include "iostream.h"
#include "jctable.h"

extern jfieldID hCPPObjID;
extern jctable<Dimension> adlib_Dimension_Table;
extern jctable<Range> adlib_Range_Table;

Range *adlib_Range_ResolvePtr(JNIEnv *, jobject);

/*ResolvePtr
 */
BlockRange * adlib_BlockRange_ResolvePtr(JNIEnv *env, jobject jthis) {
  return((BlockRange *)adlib_Range_ResolvePtr(env, jthis));
}

/*
 * Class:     adlib_BlockRange
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_BlockRange_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
};

/*
 * Class:     adlib_BlockRange
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_BlockRange_clean
  (JNIEnv *, jclass);

Layout int2emu(int i){
  switch (i) {
  case 0: 
    return (LAY_STANDARD);
    break;
  case 1: 
    return (LAY_PACKED);
    break;
  case 2: 
    return (LAY_ADAPT);
    break;
  case 3: 
    return (LAY_EXTENDED);
    break;
  default:
    break;
  }

  return(LAY_STANDARD);
}

/*
 * Class:     adlib_BlockRange
 * Method:    constructor
 * Signature: (Ladlib/Range;II)V
 */
JNIEXPORT void JNICALL Java_adlib_BlockRange_constructor__Ladlib_Range_2II
  (JNIEnv *env, jobject jthis,jobject ker,jint blockSize,jint layout)
{
  BlockRange *cppBlockRange=new BlockRange(*adlib_Range_ResolvePtr(env,ker),blockSize,int2emu(layout));

  int pos = env->GetIntField(jthis, hCPPObjID);

  delete (adlib_Range_Table.ObjPtr[pos]);

//  adlib_Range_ObjPtr[pos]=cppBlockRange;
  adlib_Range_Table.ObjPtr[pos]=cppBlockRange;
  
  //printf("the range pointer is %x\n", cppBlockRange); //debug

  if (adlib_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

}

/*
 * Class:     adlib_BlockRange
 * Method:    constructor
 * Signature: (ILadlib/Range;I)V
 */
JNIEXPORT void JNICALL Java_adlib_BlockRange_constructor__ILadlib_Range_2I
  (JNIEnv * env, jobject jthis, jint extent, jobject ker, jint layout)
{

  BlockRange *cppBlockRange=new BlockRange(extent, *adlib_Range_ResolvePtr(env,ker),int2emu(layout));

  int pos = env->GetIntField(jthis, hCPPObjID);

  delete (adlib_Range_Table.ObjPtr[pos]);

//  adlib_Range_ObjPtr[pos]=cppBlockRange;
  adlib_Range_Table.ObjPtr[pos]=cppBlockRange;
 
  if (adlib_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

};
