/*
 * File         : adlib_DAD.C
 * Author       : Guansong Zhang
 * Created      : Fri Oct 24 20:52:55 1997
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1998/01/13 18:22:40 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_DAD.h"

#include "adlib.h"

Group * adlib_Group_ResolvePtr(JNIEnv *, jobject);
Range * adlib_Range_ResolvePtr(JNIEnv *, jobject);
Stride * adlib_Stride_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;

#include "jctable.h"
jctable<DAD> adlib_DAD_Table;

/*ResolvePtr
 */
DAD * adlib_DAD_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_DAD_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_DAD
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_DAD_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<DAD> *tempadlib_DAD_Table=new jctable<DAD>;
  adlib_DAD_Table=*tempadlib_DAD_Table;
  delete tempadlib_DAD_Table;
};

/*
 * Class:     adlib_DAD
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_DAD_clean
  (JNIEnv *, jclass);

/*
 * Class:     adlib_DAD
 * Method:    constructor
 * Signature: (Ladlib/Group;)V
 */

JNIEXPORT void JNICALL Java_adlib_DAD_constructor__Ladlib_Group_2
  (JNIEnv * env, jobject jthis, jobject jGroup)
{
  Group * groupPtr = adlib_Group_ResolvePtr(env, jGroup); 

  // this is for rank-0 array
  DAD *cppDAD=new DAD(0, *groupPtr);
  long pos = adlib_DAD_Table.NewHandle(cppDAD);

  if (adlib_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     adlib_DAD
 * Method:    constructor
 * Signature: (Ladlib/Group;[Ladlib/Range;[Ladlib/Stride;)V
 */
JNIEXPORT void JNICALL Java_adlib_DAD_constructor__Ladlib_Group_2_3Ladlib_Range_2_3Ladlib_Stride_2
  (JNIEnv * env, jobject jthis, jobject jGroup, jobjectArray jRangeArray, jobjectArray jStrideArray)
{
  Group * groupPtr = adlib_Group_ResolvePtr(env, jGroup); 

  // the rank of Range and Stride will be the same.

  int rank=(env->GetArrayLength(jRangeArray));

  // Range and Stride should have default constructor

  Range * rangeLst = new Range [rank];
  Stride * strideLst = new Stride [rank];

  for (int i = 0; i<rank; i++) {
    jobject jRange=env->GetObjectArrayElement(jRangeArray, i);
    *(rangeLst+i)=(*adlib_Range_ResolvePtr(env, jRange));

    jobject jStride=env->GetObjectArrayElement(jStrideArray, i);
    *(strideLst+i)=(*adlib_Stride_ResolvePtr(env, jStride));
  }

  DAD *cppDAD=new DAD(rank, *groupPtr, rangeLst, strideLst);
  long pos= adlib_DAD_Table.NewHandle(cppDAD);

  //printf("DAD: %d\n",pos);

  if (adlib_DAD_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }


  //  AdlibPrintf("group address is %x\n", groupPtr); // debug

  //is this a problem?
  //delete [] rangeLst;
  //delete [] strideLst;

  //adlib_DAD_Table.IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

};

/*
 * Class:     adlib_DAD
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_DAD_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=adlib_DAD_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    // first delete two arrays for the object

    delete adlib_DAD_Table.ObjPtr[thecppobj];
    adlib_DAD_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     adlib_DAD
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_adlib_DAD_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  DAD * dad=adlib_DAD_ResolvePtr(env, jthis);

  AdlibPrintf("DAD %s, Handle %d, Address %x ", cname, pos, dad);

  printf("Rank %d, (Group %x), RangeList %x, StrideList %x\n", 
	 dad->rank, &dad->group, dad->ranges, dad->strides);

  delete [] cname;

};
