/*
 * File         : adlib_Group.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 20 09:45:42 1997
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1998/01/12 17:43:00 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_Group.h"
#include "adlib.h"
#include "apgStack.h"

extern jfieldID hCPPObjID;

Procs * adlib_Procs_ResolvePtr(JNIEnv *, jobject);
Location * adlib_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Group> adlib_Group_Table;

/*ResolvePtr
 */
Group * adlib_Group_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_Group_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_Group
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Group_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Group> *tempadlib_Group_Table=new jctable<Group>;
  adlib_Group_Table=*tempadlib_Group_Table;
  delete tempadlib_Group_Table;
};

/*
 * Class:     adlib_Group
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Group_clean
  (JNIEnv *, jclass)
{
//  delete adlib_Group_ObjPtr;
//  delete adlib_Group_ObjRef;
};

/*
 * Class:     adlib_Group
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Group_constructor
  (JNIEnv *env, jobject jthis)
{
  Group *cppGroup=new Group();
  long pos=adlib_Group_Table.NewHandle(cppGroup);

  if (adlib_Group_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  // adlib_Group_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     adlib_Group
 * Method:    constructor
 * Signature: (Ladlib/Procs;)V
 */
/*
JNIEXPORT void JNICALL Java_adlib_Group_constructor__Ladlib_Procs_2
  (JNIEnv *env, jobject jthis,jobject p)
{
  Group *cppGroup=new Group(*adlib_Procs_ResolvePtr(env,p));
  long pos=adlib_Group_Table.Newhandle(cppGroup);

  if (adlib_Group_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

 // adlib_Group_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);
};
*/

/*
 * Class:     adlib_Group
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Group_finalize
  (JNIEnv * env, jobject jthis)
{
  int thecppobj=env->GetIntField(jthis, hCPPObjID);
  int rc=adlib_Group_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete adlib_Group_Table.ObjPtr[thecppobj];
    adlib_Group_Table.ObjPtr[thecppobj]=NULL;
  }

};

/*
 * Class:     adlib_Group
 * Method:    on
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_adlib_Group_on
  (JNIEnv * env, jobject jthis)
{
  Group * cppGroup=adlib_Group_ResolvePtr(env, jthis); 

  apgStack.push(&apg);
  apg = *cppGroup;

  int temp = cppGroup->member();
  //printf("testing on: %d\n", temp); // debug
  
  if (temp)
    return (JNI_TRUE);
  else
    return (JNI_FALSE);

  return(JNI_TRUE);
};

/*
 * Class:     adlib_Group
 * Method:    no
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Group_no
  (JNIEnv * env, jobject jthis)
{  
  apgStack.pop(&apg);
};


/*
 * Class:     adlib_Group
 * Method:    id
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Group_id
  (JNIEnv * env, jobject jthis)
{
  return (adlib_Group_ResolvePtr(env, jthis))->id();
}

/*
 * Class:     adlib_Group
 * Method:    restrict
 * Signature: (Ladlib/Location;)Ladlib/Group;
 */
JNIEXPORT jobject JNICALL Java_adlib_Group_restrict
  (JNIEnv *env, jobject jthis, jobject i)
{
  //Group *newCPPgroupPtr=new Group();

  Group CPPgroup = (*adlib_Group_ResolvePtr(env, jthis))/(*adlib_Location_ResolvePtr(env,i));

  //*newCPPgroupPtr=CPPgroup;

  //long pos=adlib_Group_NewHandle(newCPPgroupPtr);

  jclass ccGroup=NULL;
  ccGroup=env->FindClass("adlib/Group");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccGroup, "<init>", "()V");
  jobject jNode=env->NewObject(ccGroup, handleConstructorID);

  Group *newCPPgroupPtr=adlib_Group_ResolvePtr(env,jNode);

  *newCPPgroupPtr=CPPgroup;
  return(jNode);
 
}

/*
 * Class:     adlib_Group
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_adlib_Group_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Group %s, Handle %d, Address %x\n", cname, pos, adlib_Group_ResolvePtr(env, jthis));

  adlib_Group_ResolvePtr(env, jthis)->show();

  delete [] cname;

};

