/*
 * File         : adlib_LocBlocksIndex.C
 * Author       : Guansong Zhang
 * Created      : Wed Jan  7 14:42:39 1998
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1998/01/08 21:35:44 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include "adlib_LocBlocksIndex.h"

#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;
#include "jctable.h"

extern jctable<Location> adlib_Location_Table;

Location * adlib_Location_ResolvePtr(JNIEnv *, jobject);
Range * adlib_Range_ResolvePtr(JNIEnv *, jobject);

/*ResolvePtr
 */
LocBlocksIndex * adlib_LocBlocksIndex_ResolvePtr(JNIEnv *env, jobject jthis) {
  return((LocBlocksIndex *)adlib_Location_ResolvePtr(env, jthis));
}

/*
 * Class:     adlib_LocBlocksIndex
 * Method:    constructor
 * Signature: (Ladlib/Range;)V
 */
JNIEXPORT void JNICALL Java_adlib_LocBlocksIndex_constructor
  (JNIEnv * env, jobject jthis, jobject jRange)
{
  LocBlocksIndex * cppLocBlocksIndex = new LocBlocksIndex(*adlib_Range_ResolvePtr(env, jRange));

  int pos = env->GetIntField(jthis, hCPPObjID);

  delete (adlib_Location_Table.ObjPtr[pos]);
  
  adlib_Location_Table.ObjPtr[pos]=cppLocBlocksIndex;

  if (adlib_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }
  
};

/*
 * Class:     adlib_LocBlocksIndex
 * Method:    beginLocBlk
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_LocBlocksIndex_beginLocBlk
  (JNIEnv * env, jobject jthis)
{
  adlib_LocBlocksIndex_ResolvePtr(env, jthis)->beginLocBlk();
};


/*
 * Class:     adlib_LocBlocksIndex
 * Method:    test
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_adlib_LocBlocksIndex_test
  (JNIEnv * env, jobject jthis)
{
  if (adlib_LocBlocksIndex_ResolvePtr(env, jthis)->test())
    return (JNI_TRUE);
  else
    return (JNI_FALSE);
};

/*
 * Class:     adlib_LocBlocksIndex
 * Method:    nextLocBlk
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_LocBlocksIndex_nextLocBlk
  (JNIEnv * env, jobject jthis)
{
  adlib_LocBlocksIndex_ResolvePtr(env, jthis)->nextLocBlk();
};
