/*
 * File         : adlib_Location.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:03:03 1997
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 1998/01/13 18:22:41 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_Location.h"

#include "adlib.h"

#include "apgStack.h"

extern jfieldID hCPPObjID;

Range * adlib_Range_ResolvePtr(JNIEnv *, jobject);
Dimension * adlib_Dimension_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Location> adlib_Location_Table;


/*ResolvePtr
 */
Location * adlib_Location_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_Location_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_Location
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Location_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Location> *tempadlib_Location_Table=new jctable<Location>;
  adlib_Location_Table=*tempadlib_Location_Table;
  delete tempadlib_Location_Table;
};

/*
 * Class:     adlib_Location
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Location_clean
  (JNIEnv *, jclass)
{
//  delete adlib_Location_ObjPtr;
//  delete adlib_Location_ObjRef;
};

/*
 * Class:     adlib_Location
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Location_constructor__
  (JNIEnv * env, jobject jthis)
{
  Location *cppLocation=new Location();
  long pos=adlib_Location_Table.NewHandle(cppLocation);

  if (adlib_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //AdlibPrintf("adlib_Location_AllocLength %d\n",adlib_Location_AllocLength ); // debug 
  //AdlibPrintf("new location handle is %d\n", pos); // debug

  //adlib_Location_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);
};

/*
 * Class:     adlib_Location
 * Method:    constructor
 * Signature: (Ladlib/Range;I)V
 */
JNIEXPORT void JNICALL Java_adlib_Location_constructor__Ladlib_Range_2I
  (JNIEnv *env, jobject jthis, jobject x, jint glb)
{
  Location *cppLocation=new Location(*adlib_Range_ResolvePtr(env,x),glb);
  long pos=adlib_Location_Table.NewHandle(cppLocation);

  if (adlib_Location_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Location_IncRefCntr(pos);

  env->SetIntField(jthis, hCPPObjID, pos);
}

/*
 * Class:     adlib_Location
 * Method:    ta
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Location_ta
  (JNIEnv *env, jobject jthis)
{
  apgStack.pop(&apg) ;
}

/*
 * Class:     adlib_Location
 * Method:    at
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_adlib_Location_at
  (JNIEnv * env, jobject jthis)
{
  apgStack.push(&apg);

  Location * location=adlib_Location_ResolvePtr(env, jthis);

  apg.restrict(location->dim);

  if (location->dim.crd()==location->crd)
    return (JNI_TRUE);
  else
    return (JNI_FALSE);

  return(JNI_TRUE);
};

/*
 * Class:     adlib_Location
 * Method:    dim
 * Signature: ()Ladlib/Dimension;
 */
JNIEXPORT jobject JNICALL Java_adlib_Location_dim
  (JNIEnv * env, jobject jthis)
{
  Dimension cppDim = (adlib_Location_ResolvePtr(env, jthis)->dim);

  jclass ccDimension=NULL;
  ccDimension=env->FindClass("adlib/Dimension");

  if (env->ExceptionOccurred()) return 0;

  jmethodID handleConstructorID = env->GetMethodID(ccDimension, "<init>", "()V");
  jobject jDimension=env->NewObject(ccDimension, handleConstructorID);

  Dimension * newCPPdimPtr=adlib_Dimension_ResolvePtr(env, jDimension);

  *newCPPdimPtr=cppDim;

  return(jDimension);
};


/*
 * Class:     adlib_Location
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Location_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=adlib_Location_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete adlib_Location_Table.ObjPtr[thecppobj];
    adlib_Location_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     adlib_Location
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_adlib_Location_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Location %s, Handle %d, Address %x\n", cname, pos, adlib_Location_ResolvePtr(env, jthis));

  delete [] cname;

};
