/*
 * File         : adlib_Range.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:06:49 1997
 * Revision     : $Revision: 1.9 $
 * Updated      : $Date: 1998/01/13 18:22:41 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_Range.h"
#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;
#include "jctable.h"
extern jctable<Dimension> adlib_Dimension_Table;

Dimension * adlib_Dimension_ResolvePtr(JNIEnv *, jobject);
Stride *adlib_Stride_ResolvePtr(JNIEnv *, jobject);
Location * adlib_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Range> adlib_Range_Table;

/*ResolvePtr
 */
Range * adlib_Range_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_Range_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_Range
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Range_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Range> *tempadlib_Range_Table=new jctable<Range>;
  adlib_Range_Table=*tempadlib_Range_Table;
  delete tempadlib_Range_Table;
};

/*
 * Class:     adlib_Range
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Range_clean
  (JNIEnv *, jclass);

/*
 * Class:     adlib_Range
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Range_finalize
  (JNIEnv * env, jobject jthis)
{
  int thecppobj=env->GetIntField(jthis, hCPPObjID);
  int rc=adlib_Range_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete adlib_Range_Table.ObjPtr[thecppobj];
    adlib_Range_Table.ObjPtr[thecppobj]=NULL;
  }
}

/*
 * Class:     adlib_Range
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Range_constructor__
  (JNIEnv *env, jobject jthis)
{
  Range *cppRange= new Range();
  long pos=adlib_Range_Table.NewHandle(cppRange);

  if (adlib_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Range_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

}

/*
 * Class:     adlib_Range
 * Method:    constructor
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_adlib_Range_constructor__I
  (JNIEnv * env, jobject jthis, jint extend)
{
  Range *cppRange= new Range(extend);
  long pos=adlib_Range_Table.NewHandle(cppRange);

  if (adlib_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);
};



/*
 * Class:     adlib_Range
 * Method:    constructor
 * Signature: (Ladlib/Dimension;)V
 */
JNIEXPORT void JNICALL Java_adlib_Range_constructor__Ladlib_Dimension_2
  (JNIEnv *env, jobject jthis, jobject dim) 
{
  //  (new Range(*adlib_Dimension_ResolvePtr(env,dim)))->show();
  Range *cppRange= new Range(*adlib_Dimension_ResolvePtr(env,dim));
  long pos=adlib_Range_Table.NewHandle(cppRange);
  
  //adlib_Dimension_ResolvePtr(env,dim)->show();
  //adlib_Range_ObjPtr[pos]->show(); //debug;

  if (adlib_Range_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Range_IncRefCntr(pos);
 
  env->SetIntField(jthis, hCPPObjID, pos);

}

/*
 * Class:     adlib_Range
 * Method:    dim
 * Signature: (I)Ladlib/Dimension;
 */
JNIEXPORT jobject JNICALL Java_adlib_Range_dim
  (JNIEnv *env, jobject jthis)
{

  //Dimension *newCPPdimPtr=new Dimension();

  Dimension CPPdim = adlib_Range_ResolvePtr(env, jthis)->dim();
  
  //*newCPPdimPtr=CPPdim;

  //long pos=adlib_Dimension_NewHandle(newCPPdimPtr);

  //printf("dim pos %ld, address %x\n", pos, adlib_Dimension_ObjPtr[pos]);
  //printf("is this dim? crd %d\n", adlib_Dimension_ObjPtr[pos]->crd());

  jclass ccDimension=NULL;
  ccDimension=env->FindClass("adlib/Dimension");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccDimension, "<init>", "()V");
  jobject jNode=env->NewObject(ccDimension, handleConstructorID);

  Dimension * newCPPdimPtr=adlib_Dimension_ResolvePtr(env, jNode);

  *newCPPdimPtr=CPPdim;

  return(jNode);
};

/*
 * Class:     adlib_Range
 * Method:    subrng
 * Signature: (III)Ladlib/Range;
 */
JNIEXPORT jobject JNICALL Java_adlib_Range_subrng
  (JNIEnv *env, jobject jthis, jint extent, jint base, jint stride)
{
  Range CPPrange = adlib_Range_ResolvePtr(env, jthis)->subrng(extent,base,stride);

  jclass ccRange=NULL;
  ccRange=env->FindClass("adlib/Range");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccRange, "<init>", "()V");
  jobject jNode=env->NewObject(ccRange, handleConstructorID);

  Range * newCPPrangePtr=adlib_Range_ResolvePtr(env, jNode);

  *newCPPrangePtr=CPPrange;

  return(jNode);
  
}

/*
 * Class:     adlib_Range
 * Method:    size
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_size
  (JNIEnv * env, jobject jthis)
{
  int temp = adlib_Range_ResolvePtr(env, jthis)->size();
  //printf("testing on: %d\n", temp);
  //cout<<"testing on: " << temp <<"\n";
  return(temp);
};

/*
 * Class:     adlib_Range
 * Method:    lev
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_lev
  (JNIEnv * env, jobject jthis)
{
	return(adlib_Range_ResolvePtr(env, jthis)->lev());
}

/*
 * Class:     adlib_Range
 * Method:    bas
 * Signature: ()I
 */
JNIEXPORT  jint JNICALL Java_adlib_Range_bas
  (JNIEnv * env, jobject jthis)
{
  return(adlib_Range_ResolvePtr(env, jthis)->bas());  
}

/*
 * Class:     adlib_Range
 * Method:    str
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_str
  (JNIEnv *env, jobject jthis)
{
  return(adlib_Range_ResolvePtr(env, jthis)->str());
}

/*
 * Class:     adlib_Range
 * Method:    ker_glb
 * Signature: (I[I)I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_ker_1glb
  (JNIEnv *env, jobject jthis, jint glb, jintArray sub)
{
  int size=(env->GetArrayLength(sub));
  int * cppsub=new int [size];

  jboolean isCopy;
  jint * jsub;

  jsub=env->GetIntArrayElements(sub, &isCopy);

  for (int i=0; i<size; i++)
    *(cppsub+i)=*(jsub+i);

  env->ReleaseIntArrayElements(sub, jsub, 0);

  return(adlib_Range_ResolvePtr(env, jthis)->ker_glb(glb,cppsub));
}

/*
 * Class:     adlib_Range
 * Method:    width
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_width
  (JNIEnv *env, jobject jthis)
{
  return(adlib_Range_ResolvePtr(env, jthis)->width());
}

/*
 * Class:     adlib_Range
 * Method:    volume
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Range_volume
  (JNIEnv *env, jobject jthis)
{
  return(adlib_Range_ResolvePtr(env, jthis)->volume());
}

/*
 * Class:     adlib_Range
 * Method:    stride
 * Signature: (I)Ladlib/Stride;
 */
JNIEXPORT jobject JNICALL Java_adlib_Range_stride
  (JNIEnv *env, jobject jthis, jint size)
{
  //Stride *newCPPstridePtr=new Stride();

  int ccsize=size;
  Stride CPPstride = adlib_Range_ResolvePtr(env, jthis)->stride(ccsize);
 
  //printf("range address %x, stride size %d\n",adlib_Range_ResolvePtr(env, jthis), ccsize); //debug

  //*newCPPstridePtr=CPPstride;

  //long pos=adlib_Stride_NewHandle(newCPPstridePtr);

  jclass ccStride=NULL;
  ccStride=env->FindClass("adlib/Stride");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccStride, "<init>", "()V");
  jobject jNode=env->NewObject(ccStride, handleConstructorID);

  jfieldID jsize=env->GetFieldID(ccStride,"size","I");
  env->SetIntField(jNode,jsize,ccsize);

  Stride *newCPPstridePtr=adlib_Stride_ResolvePtr(env,jNode);
  *newCPPstridePtr=CPPstride;

  return(jNode);
}

/*
 * Class:     adlib_Range
 * Method:    location
 * Signature: (I)Ladlib/Location;
 */
JNIEXPORT jobject JNICALL Java_adlib_Range_location
   (JNIEnv * env, jobject jthis, jint i)
{
 
  Location CPPlocation = (*adlib_Range_ResolvePtr(env, jthis))(i);
  
  //*newCPPstridePtr=CPPstride;
  
  //long pos=adlib_Stride_NewHandle(newCPPstridePtr);
  
  jclass ccLocation=NULL;
  ccLocation=env->FindClass("adlib/Location");
  if (env->ExceptionOccurred()) return 0;
  jmethodID handleConstructorID = env->GetMethodID(ccLocation, "<init>", "()V");
  jobject jNode=env->NewObject(ccLocation, handleConstructorID);
  
  Location *newCPPlocationPtr=adlib_Location_ResolvePtr(env,jNode);
  
  *newCPPlocationPtr=CPPlocation;
  
  return(jNode);
   
};

/*
 * Class:     adlib_Range
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_adlib_Range_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Range %s, Handle %d, Address %x\n", cname, pos, adlib_Range_ResolvePtr(env, jthis));

  adlib_Range_ResolvePtr(env, jthis)->show();
  
  delete [] cname;

};
