/*
 * File         : adlib_Shift.C
 * Author       : Guansong Zhang
 * Created      : Tue Nov  4 21:00:00 1997
 * Revision     : $Revision: 1.3 $
 * Updated      : $Date: 1998/01/08 21:35:46 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_Shift.h"
#include "adlib.h"

DAD * adlib_DAD_ResolvePtr(JNIEnv *, jobject);

extern jfieldID hCPPObjID;
#include "jctable.h"
jctable<Shift> adlib_Shift_Table;

/*ResolvePtr
 */
Shift * adlib_Shift_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_Shift_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_Shift
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_init
  (JNIEnv * env, jclass jthis)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");  
  jctable<Shift> *tempadlib_Shift_Table=new jctable<Shift>;
  adlib_Shift_Table=*tempadlib_Shift_Table;
  delete tempadlib_Shift_Table;
};

/*
 * Class:     adlib_Shift
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_clean
  (JNIEnv *, jclass);

Mode int2mode(int i)
{
  switch (i) {
  case 0:
    return CYCL;
    break;

  case 1:
    return EDGE;
    break;

  }

  return CYCL;
}

/*
 * Class:     adlib_Shift
 * Method:    constructor
 * Signature: (Ladlib/DAD;Ladlib/DAD;IIII)V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_constructor
  (JNIEnv *env, jobject jthis, jobject dDAD, jobject sDAD, jint len, jint shift, jint dim, jint mode)
{
  DAD * dDADPtr = adlib_DAD_ResolvePtr(env, dDAD); 
  DAD * sDADPtr = adlib_DAD_ResolvePtr(env, sDAD); 

  Shift *cppShift=new Shift(dDADPtr, sDADPtr, len, shift, dim, int2mode(mode));
  long pos=adlib_Shift_Table.NewHandle(cppShift);

  if (adlib_Shift_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Shift_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);  

};

/*
 * Class:     adlib_Shift
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_finalize
  (JNIEnv * env, jobject jthis) 
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=adlib_Shift_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete adlib_Shift_Table.ObjPtr[thecppobj];
    adlib_Shift_Table.ObjPtr[thecppobj]=NULL;
  }

};

/*
 * Class:     adlib_Shift
 * Method:    execute
 * Signature: ([IJ[IJ)V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_execute___3IJ_3IJ
  (JNIEnv * env, jobject jthis, jintArray dst, jlong dOffset, jintArray src, jlong sOffset)
{
  jboolean isCopy;

  int dLength=env->GetArrayLength(dst);
  jint * jDA=env->GetIntArrayElements(dst, &isCopy);
    
  int sLength=env->GetArrayLength(src);
  jint * jSA=env->GetIntArrayElements(src, &isCopy);

  /*
  {//debug
    int i;

    for (i=0; i<dLength; i++) {
      printf("dst %d: %d | ", i, *(jDA+i));
    }
    printf("\n");
    
    for (i=0; i<sLength; i++) {
      printf("src %d: %d | ", i, *(jSA+i));
    }
    printf("\n");
  }
  */

  adlib_Shift_ResolvePtr(env, jthis)->execute(jDA+dOffset, jSA+sOffset); 

  //printf("recorded remap at %x; dOffset %d; sOffset %d\n", adlib_Shift_ResolvePtr(env, jthis), dOffset, sOffset);  //debug

  /*
  {//debug
    int i;

    for (i=0; i<dLength; i++) {
      printf("dst- %d: %d | ", i, *(jDA+i));
    }
    printf("\n");
    
    for (i=0; i<sLength; i++) {
      printf("src- %d: %d | ", i, *(jSA+i));
    }
    printf("\n");
  }
  */

  env->ReleaseIntArrayElements(dst, jDA, 0);
  env->ReleaseIntArrayElements(src, jSA, JNI_ABORT);
};

/*
 * Class:     adlib_Shift
 * Method:    execute
 * Signature: ([FJ[FJ)V
 */
JNIEXPORT void JNICALL Java_adlib_Shift_execute___3FJ_3FJ
  (JNIEnv * env, jobject jthis, jfloatArray dst, jlong dOffset, jfloatArray src, jlong sOffset)
{
  jboolean isCopy;

  int dLength=env->GetArrayLength(dst);
  jfloat * jDA=env->GetFloatArrayElements(dst, &isCopy);
    
  int sLength=env->GetArrayLength(src);
  jfloat * jSA=env->GetFloatArrayElements(src, &isCopy);

  adlib_Shift_ResolvePtr(env, jthis)->execute(jDA+dOffset, jSA+sOffset); 

  env->ReleaseFloatArrayElements(dst, jDA, 0);
  env->ReleaseFloatArrayElements(src, jSA, JNI_ABORT);
};
