/*
 * File         : adlib_Stride.C
 * Author       : Guansong Zhang
 * Created      : Mon Oct 13 12:06:49 1997
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1998/01/08 21:35:47 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1997
 */

#include "adlib_Stride.h"
#include "adlib.h"
#include "iostream.h"

extern jfieldID hCPPObjID;

Location * adlib_Location_ResolvePtr(JNIEnv *, jobject);

#include "jctable.h"
jctable<Stride> adlib_Stride_Table;

/*ResolvePtr
 */
Stride * adlib_Stride_ResolvePtr(JNIEnv *env, jobject jthis) {
  return(adlib_Stride_Table.ObjPtr[env->GetIntField(jthis, hCPPObjID)]);
}

/*
 * Class:     adlib_Stride
 * Method:    init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_init
  (JNIEnv * env, jclass thisClass)
{
  jclass InterfaceObject_class=env->FindClass("adlib/InterfaceObject");
  hCPPObjID=env->GetFieldID(InterfaceObject_class, "hCPPObj", "I");
  jctable<Stride> *tempadlib_Stride_Table=new jctable<Stride>;
  adlib_Stride_Table=*tempadlib_Stride_Table;
  delete tempadlib_Stride_Table;
};

/*
 * Class:     adlib_Stride
 * Method:    clean
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_clean
  (JNIEnv *, jclass);

/*
 * Class:     adlib_Stride
 * Method:    constructor
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_constructor__
  (JNIEnv *env, jobject jthis)
{
  Stride *cppStride=new Stride();
  long pos=  adlib_Stride_Table.NewHandle(cppStride);

  //printf("Stride: %d\n",pos);

  if (adlib_Stride_Table.ObjPtr[pos]==NULL) {
    jclass cError=env->FindClass("java/lang/OutOfMemeoryError");
    env->ThrowNew(cError,"adlib memory allocation failed");
    return;
  }

  //adlib_Stride_IncRefCntr(pos);
  
  env->SetIntField(jthis, hCPPObjID, pos);

}

/*
 * Class:     adlib_Stride
 * Method:    constructor
 * Signature: (Ladlib/Stride;)V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_constructor__Ladlib_Stride_2
  (JNIEnv *, jobject, jobject);


/*
 * Class:     adlib_Stride
 * Method:    finalize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_finalize
  (JNIEnv * env, jobject jthis)
{
  long thecppobj=env->GetIntField(jthis, hCPPObjID);
  long rc=adlib_Stride_Table.DecRefCntr(thecppobj);
  if (rc==0) {
    delete adlib_Stride_Table.ObjPtr[thecppobj];
    adlib_Stride_Table.ObjPtr[thecppobj]=NULL;
  }
};

/*
 * Class:     adlib_Stride
 * Method:    len
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_adlib_Stride_len
  (JNIEnv *env, jobject jthis)
{
  return(adlib_Stride_ResolvePtr(env, jthis)->len());
}

/*
 * Class:     adlib_Stride
 * Method:    offset
 * Signature: (Ladlib/Location;)I
 */
JNIEXPORT jint JNICALL Java_adlib_Stride_offset
  (JNIEnv *env, jobject jthis, jobject i)
{
  return(adlib_Stride_ResolvePtr(env, jthis)->offset(*adlib_Location_ResolvePtr(env,i)));
}

/*
 * Class:     adlib_Stride
 * Method:    show
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_adlib_Stride_show
  (JNIEnv * env, jobject jthis, jstring name)
{
  jclass Stride_class=env->FindClass("adlib/Stride");
  jfieldID sizeID=env->GetFieldID(Stride_class, "size", "I");;

  int size=env->GetIntField(jthis, sizeID);

  int length=env->GetStringUTFLength(name);

  char * cname=new char[length+1];

  const char * jnamePtr=env->GetStringUTFChars(name, 0);

  for (int i=0; i<length; i++)
    cname[i]=*(jnamePtr+i);
  cname[length]='\0';

  env->ReleaseStringUTFChars(name, jnamePtr);

  int pos =  env->GetIntField(jthis, hCPPObjID);

  AdlibPrintf("Stride %s, Handle %d, Address %x, Size %d\n", cname, pos, adlib_Stride_ResolvePtr(env, jthis), size);

  delete [] cname;

};
