
// Operations on block distributed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRangeRep : public RangeRep {
public :
  BlockRangeRep(RangeRep* _kernel, const int _blockSize) :
          RangeRep(_blockSize * _kernel->size(),
                   _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize) {
    kernel->incr() ;
  }

  ~BlockRangeRep() {
    kernel->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  Format format() const ;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;

MEMORY_POOL_DECS
} ;

//wjp add 1999.4.8
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GenBlockRangeRep': GenBlock-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
class GenBlockRangeRep : public RangeRep {
public :
  GenBlockRangeRep(RangeRep* _kernel, const int _blockSize, const int _extent, 
          const int _rank, int* _index) :
          RangeRep(_extent,
                   _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize),rank(_rank),index(_index) {
    kernel->incr() ;
  }

  ~GenBlockRangeRep() {
    kernel->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  Format format() const ;
//wjp add 1999.5.9 for gen_block
  int proc_num() const ;
  int* index_()  const;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;
  int rank ;
  int* index ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockSubRangeRep : public RangeRep {
public :
  BlockSubRangeRep(const int _extent,
		   const int _base, const int _stride,
		   BlockRangeRep* _parent,
		   RangeRep* _kernel, const int _blockSize) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize),
	  base(_base), stride(_stride), parent(_parent) {
    parent->incr() ;

//AdlibPrintf("BlockSubRangeRep::base=%d,stride=%d,extent=%d,tmp_hi=%d,blockSize=%d\n",base,stride,extent,base+stride*(extent-1),blockSize);
    if(stride > 0) {
      ker_lo = base / blockSize ;
      ker_hi = (base + stride * (extent - 1)) / blockSize ;
    }
    else {
      ker_hi = base / blockSize ;
      ker_lo = (base + stride * (extent - 1)) / blockSize ;
    }
//AdlibPrintf("BlockSubRangeRep::ker_lo=%d,ker_hi=%d\n",ker_lo,ker_hi);
  }

  ~BlockSubRangeRep() {
    parent->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* tem() ;
  int bas()  const ;
  int str()  const ;

  Format format() const ;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  RangeRep* subker(int* ker_bas, int* ker_str) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;
  int base, stride ;
  BlockRangeRep* parent ;

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;

//wjp add 1999.4.8
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GenBlockSubRangeRep': Genblock-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class GenBlockSubRangeRep : public RangeRep {
public :
  GenBlockSubRangeRep(const int _extent,
		   const int _base, const int _stride,
		   GenBlockRangeRep* _parent, RangeRep* _kernel, 
                   const int _blockSize, const int _rank, int* _index) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize), rank(_rank), index(_index),
	  base(_base), stride(_stride), parent(_parent) {
    parent->incr() ;

    int i,acc_ind,tmp_hi;
    if (extent == 0 && base == 0) tmp_hi = base;
    else tmp_hi = base + stride * (extent -1);
//AdlibPrintf("GenBlockSubRangeRep::rank=%d,base=%d,stride=%d,extent=%d,tmp_hi=%d\n",rank,base,stride,extent,tmp_hi);
    acc_ind = 0;
    if(stride > 0) {
      for (i=0;i<rank;i++) {
        if ((base >= acc_ind) && (base < acc_ind + index[i]))
             ker_lo = i;
        if ((tmp_hi >= acc_ind) && (tmp_hi < acc_ind + index[i]))
             ker_hi = i;
        acc_ind += index[i];
      }
      if (base >= acc_ind) ker_lo = rank;
      if (tmp_hi >= acc_ind) ker_hi = rank;
    }
    else {
      for (i=0;i<rank;i++) {
        if ((base >= acc_ind) && (base < acc_ind + index[i]))
             ker_hi = i;
        if ((tmp_hi >= acc_ind) && (tmp_hi < acc_ind + index[i]))
             ker_lo = i;
        acc_ind += index[i];
      }
      if (base >= acc_ind) ker_hi = rank;
      if (tmp_hi >= acc_ind) ker_lo = rank;
    }
//AdlibPrintf("GenBlockSubRangeRep::ker_lo=%d,ker_hi=%d\n",ker_lo,ker_hi);
//    if(stride > 0) {
//      ker_lo = base / blockSize ;
//      ker_hi = (base + stride * (extent - 1)) / blockSize ;
//    }
//    else {
//      ker_hi = base / blockSize ;
//      ker_lo = (base + stride * (extent - 1)) / blockSize ;
//    }
  }

  ~GenBlockSubRangeRep() {
    parent->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* tem() ;
  int bas()  const ;
  int str()  const ;

  Format format() const ;
//wjp add 1999.5.9 for gen_block
  int proc_num() const ;
  int* index_()  const ;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  RangeRep* subker(int* ker_bas, int* ker_str) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;
  int base, stride ;
  GenBlockRangeRep* parent ;
  int rank; //wjp add 1999.4.8
  int* index; //wjp add 1999.4.8

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockRangeRep', `PackedBlockSubRangeRep':            //
//      Block-distributed ranges yielding arrays with ``packed''       //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockRangeRep : public BlockRangeRep {
public :
  PackedBlockRangeRep(RangeRep* _kernel, const int _blockSize) :
          BlockRangeRep(_kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;
} ;


class PackedBlockSubRangeRep : public BlockSubRangeRep {
public :
  PackedBlockSubRangeRep(const int _extent,
                         const int _base, const int _stride,
                         BlockRangeRep* _parent,
                         RangeRep* _kernel, const int _blockSize) :
        BlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize),
        packing(_stride > 0 ? _stride : -_stride) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

    // added by Zheng Gengbin
  ~PackedBlockSubRangeRep() {};

protected :
  int packing ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedStrideRep'.                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedStrideRep : public StrideRep {
public :
  PackedStrideRep(StrideRep* _kernel, const int _length, const int _packing) :
          StrideRep(_kernel, _length), packing(_packing) {}

  ~PackedStrideRep() {}

  int disp(const int sub) const ;
  int disp_step(const int sub_stp) const ;

protected :
  int packing ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedBlockRangeRep : public BlockRangeRep {
public :
  ExtendedBlockRangeRep(RangeRep* _kernel, const int _blockSize,
                        const int _wlo, const int _whi) :
          BlockRangeRep(_kernel, _blockSize), wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int disp(const int sub) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int wlo_() const { return wlo; }
  int whi_() const { return whi; }
 
    // added by Zheng Gengbin
  ~ExtendedBlockRangeRep() {};

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


class ExtendedBlockSubRangeRep : public BlockSubRangeRep {
public :
  ExtendedBlockSubRangeRep(const int _extent,
                           const int _base, const int _stride,
			   BlockRangeRep* _parent,
			   RangeRep* _kernel, const int _blockSize,
                           const int _wlo, const int _whi) :
    BlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize),
    wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int wlo_() const { return wlo; }
  int whi_() const { return whi; }

    // added by Zheng Gengbin
  ~ExtendedBlockSubRangeRep() {};
protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


//wjp add 1999.5.9 for gen_block
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedGenBlockRangeRep', `ExtendedGenBlockSubRangeRep':        //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedGenBlockRangeRep : public GenBlockRangeRep {
public :
  ExtendedGenBlockRangeRep(RangeRep* _kernel, const int _blockSize,
                        const int _wlo, const int _whi, const int _extent, const int _rank, int* _index) :
          GenBlockRangeRep(_kernel, _blockSize, _extent, _rank, _index),
          wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int disp(const int sub) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int wlo_() const { return wlo; }
  int whi_() const { return whi; }
 
    // added by Zheng Gengbin
  ~ExtendedGenBlockRangeRep() {};

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


class ExtendedGenBlockSubRangeRep : public GenBlockSubRangeRep {
public :
  ExtendedGenBlockSubRangeRep(const int _extent,
                           const int _base, const int _stride,
			   GenBlockRangeRep* _parent,
			   RangeRep* _kernel, const int _blockSize,
                           const int _wlo, const int _whi, const int _rank, int* _index) :
    GenBlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize, _rank, _index),
    wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int wlo_() const { return wlo; }
  int whi_() const { return whi; }

    // added by Zheng Gengbin
  ~ExtendedGenBlockSubRangeRep() {};
protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedStrideRep'.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedStrideRep : public StrideRep {
public :
  ExtendedStrideRep(StrideRep* _kernel, const int _length, const int _wlo) :
          StrideRep(_kernel, _length), wlo(_wlo) {}

  ~ExtendedStrideRep() {}

  int disp(const int sub) const ;

protected :
  int wlo ;

MEMORY_POOL_DECS
} ;


//wjp add 1999.5.9
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedGenStrideRep'.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedGenStrideRep : public StrideRep {
public :
  ExtendedGenStrideRep(StrideRep* _kernel, const int _length, const int _wlo) :
          StrideRep(_kernel, _length), wlo(_wlo) {}

  ~ExtendedGenStrideRep() {}

  int disp(const int sub) const ;

protected :
  int wlo ;

MEMORY_POOL_DECS
} ;


// -------------------------------------------------------------
// Things to do
//
//   Revert to `BlockSubRangeRep' derived from `BlockRangeRep'.

