/*
$Id: close.c,v 1.2 1995/11/05 18:00:15 ssmore Exp $

File Name : close.c

Entry Point(s) :

(1) PASSION_close   prototype in passion.h

Modification History :

Module Description                                      Date         Author
------------------------------------------------------------------------------
(1) PASSION_close                                     2/20/95        ssmore
(2) PASSION_close                                     2/27/95        ssmore
(3) PASSION_close				      3/13/98        Zheng

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_CLOSE_C
#define PASSION_CLOSE_C

/*#define DEBUG*/

#include <stdio.h>

#include <malloc.h>

#include "pfile.h"

extern PSTACK pop();

extern int reuse_fd;

/*
  Name : PASSION_close

  Function : This function is used to close the Local Array File (LAF) or the 
	     global file.

  Parameters : 
    Input :
      PFilePtr - pointer to PFILE data structure returned by the PASSION_open
		 call
    Output :
      none
    Input/Ooutput :
      none

  Return Value :
    0, if successful
    -1, if not successful

  Algorithm :

    if invalid PFILE pointer
      return error 
    endif

    close the file
    if not successful
      de-allocate space for the parallel file data structure (PFILE)
      return error
    endif

    de-allocate space for the parallel file data structure (PFILE)

    return success

  Side Effects :
    none

  Shortcomings :
    (1) treats LAF and global file in the same way.

*/
int PASSION_close(PFILE *PFilePtr)
{

  if (PFilePtr == (PFILE *)0) 
  {
#ifdef DEBUG
    fprintf(stderr,"PASSION : (close) Invalid File Pointer\n");
#endif
    return -1;
  }

#ifdef HAVE_MPIO
  MPI_File_close(&PFilePtr->share->Fd);
#else
  if (close(PFilePtr->share->Fd) == -1)
  {
#ifdef DEBUG
    fprintf(stderr,"PASSION : (close) System Error\n");
#endif
    free((void *)PFilePtr);
    return -1;
  }
#endif               /* HAVE_MPIO */

#ifdef PASSION_TRACE
	if (PFilePtr->trace_fd != -1)
	{
		close(PFilePtr->trace_fd);
	}
#endif

#ifndef DEBUG
#ifdef HAVE_MPIO
  	{
  	char prefix[500];
  	strcpy(prefix,"ufs:");
  	strcpy(prefix+4, PFilePtr->share->filename);
	MPI_File_delete(prefix, MPI_INFO_NULL);
	}
#else
	remove (PFilePtr->share->filename);
#endif
#endif

  free((void *)PFilePtr->share);
  free((void *)PFilePtr);

  return 0;
}

int LAF_close(PFILE *PFilePtr)
{
  if (reuse_fd)
  {
#ifdef HAVE_MPIO
     MPI_File_set_size(PFilePtr->share->Fd, 0); 
#endif
       /* save for reuse. */
     push(PFilePtr->share->Fd, PFilePtr->share->filename); 
     free((void *)PFilePtr->share);
     free((void *)PFilePtr);
  }
  else {
     PASSION_close(PFilePtr);
  }
}

int LAF_close_all()
{
    PSTACK sptr;
    LAF_File fd;
    char *filename;

    while (!empty()) {
	sptr = pop();
        fd = sptr->fd;
        filename = sptr->filename;
	free(sptr);
#ifdef HAVE_MPIO
  	MPI_File_close(&fd);
  	{
  	char prefix[500];
  	strcpy(prefix,"ufs:");
  	strcpy(prefix+4, filename);
#ifndef DEBUG
	MPI_File_delete(prefix, MPI_INFO_NULL);
#endif
	}
#else
        if (close(fd) == -1)
	      fprintf(stderr,"PASSION : (close) System Error\n");
	else {
#ifndef DEBUG
	remove (filename);
#endif
 	}
#endif
    }
}

#endif  /* PASSION_CLOSE_C */
