/*
$Id: passion.h,v 1.15 1995/11/04 03:46:07 ssmore Exp $

File Name : passion.h
Author    : ssmore

Modification History :

Module Description                                Date     Author
------------------------------------------------------------------------------
(1) Interface to the PASSION_read_section         2/27/95  ssmore
(2) Interface to the PASSION_write_section        2/27/95  ssmore
(3) Interface to the PASSION_read_prefetch        3/06/95  ssmore
(4) Interface to the PASSION_ReuseInit            3/07/95  ssmore

Notes :

*/

#ifndef PASSION_H
#define PASSION_H

#include "define.h"

#include "pfile.h"
/*
#include "ocad.h"
#include "gad.h"
#include "prefetch.h"
#include "reuse.h"
#include "fpassion.h"
*/

OCAD *PASSION_malloc_OCAD(int dimensions, int Storage);
void PASSION_free_OCAD(OCAD *OCADp);
int PASSION_fill_OCAD(OCAD* OCADp, int *size, int distribution[][2],
                      int *nprocs, int *ocla_size, int icla_size[][2], 
											int overlap[][2], int elemsize);

PFILE *PASSION_open(char *FileName, unsigned int HeaderSize);
PFILE *PASSION_create(char *FileName, unsigned int HeaderSize, OCAD *OCADp);
PFILE *PASSION_gcreate(char *FileName, unsigned int HeaderSize, GAD *GADp);
int    PASSION_close(PFILE *PFilePtr);

int PASSION_read_header(PFILE *PFilePtr, char *HBuf);
int PASSION_write_header(PFILE *PFilePtr, char *HBuf);

int PASSION_read(PFILE *PFilePtr, OCAD *OCADp, char *Array);
int PASSION_write(PFILE *PFilePtr, OCAD *OCADp, char *Array);

int PASSION_write_section(PFILE *PFilePtr, OCAD *OCADp, char *Array, 
													int *Index, int AccessArray[][4]);
int PASSION_gwrite_section(PFILE *PFilePtr, OCAD *OCADp, char *Array, 
													int *Index, int AccessArray[][4]);
int PASSION_read_section(PFILE *PFilePtr, OCAD *OCADp, char *Array, 
												 int *Index, int AccessArray[][4]);
int PASSION_gread_section(PFILE *PFilePtr, OCAD *OCADp, char *Array, 
												 int *Index, int AccessArray[][4]);

PREFETCH *PASSION_read_prefetch(PFILE *PFilePtr, OCAD *OCADp, 
                    char *Array, int *Index, int AccessArray[][4]);
PREFETCH *PASSION_gread_prefetch(PFILE *PFilePtr, OCAD *OCADp, 
                    char *Array, int *Index, int AccessArray[][4]);
int PASSION_prefetch_wait(PREFETCH *PREFETCHp);

REUSE *PASSION_reuse_init(PFILE *PFilePtr, OCAD *OCADp);
void PASSION_reuse_end(REUSE *REUSEp);
int PASSION_read_reuse(REUSE *REUSEp, char *Array, int *Index, 
											 int AccessArray[][4]);

GAD *PASSION_malloc_GAD(int dimensions, int Storage);
void PASSION_free_GAD(GAD *GAD);
int PASSION_fill_GAD(GAD* GADp, int *glb_size, int *nprocs,  int *icla_size, 
		      int elemsize);
int PASSION_global_read( PFILE *PFilePtr, GAD *GADp, char *A,
int *index, int AccessArray[][4], int nprocs);
int PASSION_global_write( PFILE *PFilePtr, GAD *GADp, char *A,
int *index, int AccessArray[][4], int nprocs);

/* performance measuring routines */
#ifdef DEBUG
unsigned long count_bytes_read(unsigned long bytes);
unsigned long count_read_ops(unsigned long nops);
unsigned long count_write_ops(unsigned long nops);
unsigned long count_bytes_write(unsigned long bytes);
#endif

void PASSION_init(int *argc, char ***argv);
void PASSION_end();

#endif  /* PASSION_H */
