/*
$Id: rd_sbs_1d.c,v 1.2 1995/10/17 02:15:24 ssmore Exp $

File Name : rd_sbs_1d.c
Author    : srshah

Entry Point(s) :

(1) passion_read_block_section_1D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_SBS_1D
#define PASSION_RD_SBS_1D

/*#define DEBUG*/

#include <stdio.h>

#include <memory.h>

/*
#include "ocad.h"
#include "pfile.h"
*/
#include "passion_base.h"

/*
  Name : passion_read_block_section_1D

  Function : This function is used to read a block from a strided section 
	for 1-D array.

  Parameters : 
    Input :
      fd       - LAF descriptor
		  StartIndex - start of the section
			EndIndex - end of the section
			stride  - stride in the section
			block_size - size of the block to be read from a strided section
		  elemsize - size of each array element in bytes
			headersize - size of the header
			icla_size - size of the icla in no. of elements
			Index    - index in the array where to put data
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		if there is no stride in the section
			read the unstrided section (passion_usection_1D)
		else
			calculate the icla size
			allocate space equal to ICLA size for temporary storage

      calculate no. of slabs to read
			slabs to read = size of the section / icla size
			elements read = 0
			array offset = index
			slab offset = 0
			for each slab to be read
        slab start index = start index + elements read
				slab end index   = slab start index + icla size - 1
				read the slab into the temporary staorage (passion_read_usection_1D)
				elements read = elements read + icla size
				do forever
					select the element in the slab at the slab offset
					put the element in the array at array offset
					array offset = array offset + 1
					if slab offset >  (slab offset + stride) % icla size
						break
          else
					  slab offset =  (slab offset + stride) % icla size
          endif
				enddo
			endfor
			if elements read < size of the section
				size of the slab = size of the section - elements read
        slab start index = start index + elements read
				slab end index   = slab start index + size of the slab - 1
													 
				read the slab into the temporary staorage (passion_read_usection_1D)
				do forever
					select the element in the slab at the slab offset
					put the element in the array at array offset
					array offset = array offset + 1
					if slab offset >  (slab offset + stride) % size of the slab
						break
          else
					  slab offset =  (slab offset + stride) % size of the slab
          endif
				enddo
			endif
			free the slab memory
		endif
  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

PREFETCH *passion_pread_block_section_1D(LAF_File fd, char *Array, int ArrayIndex, 
                                  int StartIndex, int stride, int DataLen, 
                                  int elemsize, int headersize, int icla_size)
{
int SectionSize, Slabs2Read, ElementsRead;
int ArrayOffset, SlabOffset, i;
int SlabStartIndex, SlabEndIndex, SlabSize;
int PrevSlabElem;  /* to keep track of the no. of current-block elements 
				 	read in the previous slab */ 
char *temp_buffer;
PREFETCH *PREFETCHp;

        int block_size = 1;

        if (DataLen <= 0) return (PREFETCH *)0;

	if ((stride == 1) || (stride == block_size) || DataLen == 1 )
	{
		/* there is no stride in the section practically */
    		return passion_pread_usection_1D(fd, Array, ArrayIndex, StartIndex,
					     DataLen, elemsize, headersize);
	}
	else
	{
		/* Too many reads will result hence not supported */
		fprintf(stderr, "PASSION : (read prefetch 1d) Too many asynch reads will result.\n");
		return (PREFETCH *)0;
	}
	return (PREFETCH *)0;
}

#endif  /* PASSION_RD_SBS_1D */
