/*
$Id: rd_sbs_2d.c,v 1.1 1995/08/09 01:38:49 choudhar Exp $

File Name : rd_sbs_2d.c
Author    : srshah

Entry Point(s) :

(1) passion_read_block_section_2D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_SBS_2D
#define PASSION_RD_SBS_2D

#include <memory.h>

#include "passion_base.h"
/*
#include "ocad.h"
#include "pfile.h"
*/

/*#define DEBUG*/

#include <stdio.h>

/*
  Name : passion_read_block_section_2D

  Function : This function is used to read a block-strided section for 2-D
	array.

  Parameters : 
    Input :
      fd       - LAF descriptor
		  StartIndex0 - start row of the section
		  StartIndex1 - start column of the section
			EndIndex0 - end row of the section
			EndIndex1 - end column of the section
			stride0  - row stride in the section
			stride1  - column stride in the section
			block_size0 - block size in the row dimension
			block_size1 - bolck size in the column dimension
		  elemsize - size of each array element in bytes
			headersize - size of the header
			icla_size - size of the icla in no. of elements
			Index0    - row index in the array where to put data
			Index1    - column index in the array where to put data
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

	if no stride in the column
	  if no stride in the row
			read 2-D unstrided section
		else
			you are reading non-consecutive strips of unstrided data
      section start = start index1
      section end   = end   index1
			array row     = index0
			for each row of interest in the section
				compute the offset of the start of the row from the start of the
				file and call it new file header
				new file header = row * size of ocla row * size of each element
				                  + file header
				array index   = (array row * size of icla row) + index1
        read the row as unstrided section of 1-D array 
				(passion_read_usection_1d)
				array row = array row + 1
			endfor
		endif
  else
		its a general 2-D section
    section start = start index1
    section end   = end   index1
 		stride        = stride1
		array row     = index0
		for each row of interest in the section
			compute the offset of the start of the row from the start of the
			file and call it new file header
			new file header = row * size of ocla row * size of each element
			                  + file header
			array index   = (array row * size of icla row) + index1
      read the row as regular section of 1-D array (passion_read_section_1d)
			array row = array row + 1
		endfor
	endif

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

PREFETCH *passion_pread_block_section_2D(LAF_File fd, char *Array, int Index, int StartIndex, 
				  int u0, int u1, int DataLen0, int DataLen1, 
				  int elemsize, int headersize, int icla_size, 
                                  int icla_rowsize, int ocla_rowsize)
{
int i, j, fetch_size, SectionStart, SectionEnd;
int NewHeaderSize, ArrayIndex, ArrayRow;
PREFETCH *PREFETCHp, *PREFETCHp1;

int block_size0 = 1;
int block_size1 = 1;

	if (DataLen0 <= 0 || DataLen1 <= 0) {
#ifdef DEBUG
	fprintf(stderr,"passion_pread_block_section_2D: DataLen is zero.\n");
#endif
	   return (PREFETCH *)0;
        }

	if ((u0 == 1) || (block_size1 == u0))
	{
	  if ((u1 == ocla_rowsize) || (block_size0 == u1))
		{
			/* read 2-D unstrided section */
	       	return passion_pread_usection_2D(fd, Array, Index, StartIndex, u0, u1, 
				   DataLen0, DataLen1, elemsize, headersize, 
				   icla_rowsize, ocla_rowsize);
		}
		else
		{
			/* you are reading non-consecutive strips of unstrided data */
      			SectionStart = StartIndex;
			ArrayRow     = 0;
			PREFETCHp    = (PREFETCH *)0;
			for (i = 0; i < DataLen1; i++)
			{
				if ((DataLen1 - i + 1) < block_size0) {
					fetch_size = DataLen1 - i + 1;
					}
				else {
					fetch_size = block_size0;
					}

				for (j=0; j<fetch_size; j++) {
				 	  /* compute the offset of the start of the row from the start of the
					file and call it new file header */
					ArrayIndex   = (ArrayRow * icla_rowsize) + Index;
        				  /* read the row as unstrided section of 1-D array */
        				PREFETCHp1=passion_pread_usection_1D(fd,
					     Array, ArrayIndex, SectionStart, 
					     DataLen0, elemsize, headersize);
					if (PREFETCHp1 == (PREFETCH *)0)
					{
					   fprintf(stderr, "PASSION : (read prefetch) Read error.\n");
					   PASSION_prefetch_wait(PREFETCHp);
					   return (PREFETCH *)0;
					}
					PREFETCHp = passion_mergePREFETCH(PREFETCHp, PREFETCHp1);
					if (PREFETCHp == (PREFETCH *)0)
					{
					   fprintf(stderr, "PASSION : (read prefetch) merge error.\n");
					   return (PREFETCH *)0;
					}
					SectionStart = SectionStart + u1;
					ArrayRow = ArrayRow + 1;
				} /* end of for */
			}
		}
	}
  else
	{
		/* its a general 2-D section */
		/* Too many reads will result */
		fprintf(stderr, "PASSION : (read prefetch) Too many reads.\n");
		return (PREFETCH *)0;
	}
	return PREFETCHp;
}

#endif  /* PASSION_RD_SBS_1D */
