/*
$Id: rd_us_1d.c,v 1.1 1995/06/21 15:34:36 ssmore Exp $

File Name : rd_us_1d.c
Author    : ssmore

Entry Point(s) :

(1) passion_read_usection_1D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_US_1D
#define PASSION_RD_US_1D

#include "passion_base.h"

/*#define DEBUG*/

#include <stdio.h>

/*
  Name : passion_read_usection_1D

  Function : This function is used to read an unstrided section for 1-D
	array.

  Parameters : 
    Input :
      fd - file descriptor for the LAF
			Index - index in the array where to start putting data
			StartIndex - index where the section starts
			EndIndex - index where the section ends
      elemsize - size of each element
			headersize - size of the header at the start of the file
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    calculate the offset in the file where the section starts.
		file offset = header size + start index * elemsize
		calculate how many bytes to read
		datalen = (endindex - startindex + 1) * elemsize
		calculate where to place the data in the buffer
		array offset = index * elemsize

		read the data from the file (passion_read)

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

/*
note:

*/
PREFETCH *passion_pread_usection_1D(LAF_File fd, char *Array, unsigned int Index, 
                             int StartIndex, int DataLen, 
                             int elemsize, int headersize)
{
unsigned int FileOffset, ArrayOffset ;
PREFETCH *PREFETCHp;

    /* calculate the offset in the file where the section starts. */
		FileOffset = headersize + (StartIndex * elemsize);

		/* calculate how many bytes to read */
		DataLen = DataLen * elemsize;

                /* calculate where to place the data in the buffer */
                ArrayOffset = Index * elemsize;

if ((PREFETCHp = passion_mallocPREFETCH(1)) == (PREFETCH *)0)
                {
                        fprintf(stderr, "PASSION : (read prefetch) Out of memory.\n");
                        return (PREFETCH *)0;
                }

    (PREFETCHp->Ids)[0] = passion_pread(fd, FileOffset, Array + ArrayOffset, DataLen);
    if ((PREFETCHp->Ids)[0] < 0)
		{
#ifdef DEBUG
	fprintf(stderr,"passion_pread_usection_1D error: fd:%d, FileOffset:%d, ArrayOffset:%d, DataLen:%d\n",fd,FileOffset,ArrayOffset,DataLen);
#endif
			PASSION_prefetch_wait(PREFETCHp);
			return (PREFETCH *)0;
		}
/*
#ifdef DEBUG
	fprintf(stderr,"passion_pread_usection_1D: FileOffset:%d, ArrayOffset:%d, DataLen:%d\n",FileOffset,ArrayOffset,DataLen);
{
int i;
int *data = (int *)(Array + ArrayOffset);
fprintf(stderr,"buffer: ");
for (i=0; i< DataLen/elemsize; i++) fprintf(stderr,"%d ",data[i]);
fprintf(stderr,"\n");
}
#endif
*/

    return PREFETCHp;
}

#endif  /* PASSION_RD_US_1D */
