/*
$Id: rd_us_2d.c,v 1.6 1995/09/14 14:47:33 ssmore Exp $

File Name : rd_us_2d.c
Author    : ssmore

Entry Point(s) :

(1) passion_read_usection_2D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_US_2D
#define PASSION_RD_US_2D

#include "passion_base.h"
/*
#include "ocad.h"
#include "pfile.h"
*/

#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

/*
  Name : passion_read_usection_2D

  Function : This function is used to read an unstrided section for 2-D
	array.

  Parameters : 
    Input :
      fd  - file descriptor for LAF
			Index0 - row index in the array
			Index1 - column index in the array
			StartIndex0 - starting row index in the section
			StartIndex1 - starting column index in the section
		  EndIndex0   - ending row index in the section
			EndIndex1   - ending column index in the array
			elemsize    - size of each element of the array
		  headersize  - size of the LAF header
			ocla_rowsize - size of a row in the OCLA in number of elements
			icla_rowsize - size of a row in the ICLA in number of elements
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    calculate the offset in the file where the section starts.
		file offset = header size + startindex1 * rowsize + startindex2 *
		              elementsize

		calculate where to place the data in the buffer
		array offset = ((index1 * icla_rowsize) + index2) * elemsize

		if the section spans the second dimension
			you can read the data in one read
		  calculate how many bytes to read
		  datalen = (startindex1 - endindex1 + 1) * size of second dimension
			* elemsize
		  read the data from the file (passion_read)
		else
			read part of one row at a time
		  calculate how many bytes to read on each read call
		  datalen = (startindex2 - endindex2 + 1) * element size
			for each row in the section
		    read the data from the file (passion_read)
				file offset = file offset + ocla_rowsize * element size
				array offset = array offset + icla_rowsize * element size
			endfor
		endif

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

/*  
note:

*/
PREFETCH *passion_pread_usection_2D(LAF_File fd, char *Array, int ArrayIndex, int StartIndex, 
			     int u0, int u1, int DataLen0, int DataLen1, 
			     int elemsize, int headersize, 
			     int icla_size_0, int ocla_size_0)
{
unsigned int FileOffset, DataLen, i, Id;
int NoOfIds;
PREFETCH *PREFETCHp;

      /* calculate the offset in the file where the section starts. */
    FileOffset = headersize + StartIndex*elemsize ;

    if (DataLen0 == ocla_size_0) 
    {
		  /* calculate how many bytes to read */
	DataLen = DataLen0  * DataLen1 * elemsize;

	/* the data can be read in one read call and hence only one
	prefetch id is needed */
	NoOfIds = 1;
	if ((PREFETCHp = passion_mallocPREFETCH(NoOfIds)) == (PREFETCH *)0)
	{
		#ifdef DEBUG
		fprintf(stderr, "PASSION : (read prefetch) Out of memory.\n");
		#endif
		return (PREFETCH *)0;
	}
	(PREFETCHp->Ids)[0] = passion_pread(fd, FileOffset, Array + ArrayIndex, DataLen);
        if ((PREFETCHp->Ids)[0] < 0)
	{
		#ifdef DEBUG
		fprintf(stderr, "PASSION : (read prefetch) Asynch read error.\n");
		#endif
		/* wait for the already started async calls and free the
		prefetch structure */
		PASSION_prefetch_wait(PREFETCHp);
		return (PREFETCH *)0;
	}
    }
    else
    {
		  /* calculate how many bytes to read on each read call */
	DataLen = DataLen0 * elemsize;

	NoOfIds = DataLen1;

	if ((PREFETCHp = passion_mallocPREFETCH(NoOfIds)) == (PREFETCH *)0)
	{
		#ifdef DEBUG
			fprintf(stderr, "PASSION : (read prefetch) Out of memory.\n");
		#endif
		return (PREFETCH *)0;
	}

        Id = 0;
	for (i = 0; i < DataLen1; i++)
	{
        	(PREFETCHp->Ids)[Id] = passion_pread(fd, FileOffset, Array + ArrayIndex, DataLen);
        	if ((PREFETCHp->Ids)[Id] < 0)
		{
			#ifdef DEBUG
			fprintf(stderr, "PASSION : (read prefetch) Asynch read error.\n");
			#endif
		        /* wait for the already started async calls and free the
		        prefetch structure */
			PASSION_prefetch_wait(PREFETCHp);
			return (PREFETCH *)0;
		}
        	Id = Id + 1;
		FileOffset = FileOffset + u1 * elemsize;
		ArrayIndex = ArrayIndex + icla_size_0 * elemsize;
	}
    }
    return PREFETCHp;
}

#endif  /* PASSION_RD_US_2D */







