/*
$Id: read_cur.c,v 1.9 1995/10/16 16:41:32 ssmore Exp $

File Name : read_cur.c
Author    : ssmore

Entry Point(s) :

(1) passion_read_cur   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_READ_CUR_C
#define PASSION_READ_CUR_C

#ifdef PARAGON
#include "nx.h"       /* This include file is PARAGON specific */
#else
#ifdef DELTA
#include "mesh.h"       /* This include file is DELTA specific */
#else
#ifdef IPSC
#include "cube.h"       /* This include file is IPSC specific */
#else
#ifdef PIOFS
#include <unistd.h>
#else
#ifdef AIX
#include <aio.h>
#else
/* its an error if the preprocessor comes here but how to flag it ?*/
#endif
#endif
#endif
#endif
#endif

/* for debug */
/*#define DEBUG*/

#include <stdio.h>

#ifdef SP_PIOFS_MPI
#include <unistd.h>
#endif
#ifdef SP_AFS_MPI
#include <unistd.h>
#endif

#include "passion_base.h"

#ifdef DEBUG
/*#include "passion.h" */

unsigned long count_bytes_read(unsigned long bytes)
{
static long bytes_read=0;

	bytes_read += bytes;
	return bytes_read;
}

unsigned long count_read_ops(unsigned long nops)
{
static long read_ops=0;

	read_ops += nops;
	return read_ops;
}
#endif

/*
  Name : passion_read_cur

  Function : This function is used to read data from the LAF. It starts
	reading data at the current file offset.

  Parameters : 
    Input :
			FileDescriptor - File descriptor of the LAF
			DataLen        - Amount of data in bytes
    Output :
			Buffer         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		read the data from the LAF

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

extern int c_ad_FromPointer() ;

int passion_pread_cur(LAF_File fd, char *buffer, unsigned int buflen)
{
  int numRead;

#ifdef PIOFS
  fprintf(stderr, "PASSION : (read prefetch) Unavailable under PIOFS on SP2.\n");
#else
#ifdef SP_PIOFS_MPI
  if (read(fd, (void *)buffer, (size_t)buflen) != (ssize_t)buflen)
  {
    return -1;
	}
#else
#ifdef SP_AFS_MPI
  if (read(fd, (char *)buffer, (int)buflen) != (int)buflen)
  {
    return -1;
	}
#else
#ifdef DELTA
  /*
    return value of the cread is not checked
  */
  return (int)iread(fd, buffer, buflen);
#else
#ifdef PARAGON
  /*
    return value of the cread is not checked
  */
  return (int)iread(fd, buffer, buflen);
#else
#ifdef IPSC
  /*
    return value of the cread is not checked
  */
  return (int)iread(fd, buffer, buflen);
#else
#ifdef HAVE_MPIO
  MPIO_Request request;
  MPI_File_iread(fd, buffer, buflen, MPI_BYTE, &request);
  return c_ad_FromPointer(request);
#else
#ifdef AIX
  {
  aio_handle_t aiocbp = (aio_handle_t)malloc(sizeof(struct aiocb));
  aiocbp->aio_buf = buffer;
  aiocbp->aio_nbytes = buflen;
  aiocbp->aio_whence = SEEK_CUR;
  aiocbp->aio_offset = 0;
  aio_read(fd, aiocbp);
  return c_ad_FromPointer(aiocbp->aio_handle);
  }
#else
#ifdef UNIX
  fprintf(stderr, "PASSION : (read prefetch) Unavailable under NFS on UNIX.\n");
  return -1;
#else
	/* no architecture defined */
  return -1;
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifdef DEBUG
	count_bytes_read((unsigned long)buflen);
	count_read_ops((unsigned long)1);
#endif
	return 0;
}

#endif  /* PASSION_READ_CUR_C */
