/*
$Id: wr_sbs_1d.c,v 1.2 1995/09/02 15:31:00 ssmore Exp $

File Name : wr_ss_1d.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_section_1D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WR_SS_1D
#define PASSION_WR_SS_1D

#include "passion_base.h"

#include <stdio.h>

#include <memory.h>

/*
#include "ocad.h"
*/
#include "pfile.h"

#define DEBUG

/*
  Name : passion_write_section_1D

  Function : This function is used to write a block strided section for 1-D
	array.

  Parameters : 
    Input :
      PFilePtr - pointer to the parallel file
      OCADp    - pointer to the OCAD structure
			Array         - The pointer to the data buffer
			Index    - index where to put data
			AccessArray - section specifier
			OldFile - non-zero if writing to an existing file
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		if there is no stride in the section
			write the unstrided section (passion_usection_1D)
		else
			calculate the icla size
			allocate space equal to ICLA size for temporary storage

      calculate no. of slabs to write
			slabs to write = size of the section / icla size
			elements written = 0
			array offset = index
			slab offset = 0
			for each slab to be written
        slab start index = start index + elements written
				slab end index   = slab start index + icla size - 1
				read in the slab
				do forever
					select the element in the array at array offset
					put the element in the slab at the slab offset
					array offset = array offset + 1
					if (slab offset + stride) >= icla_size
					  slab offset =  (slab offset + stride) % icla size
						break
          else
					  slab offset =  (slab offset + stride) % icla size
          endif
				enddo
				write the slab into the temporary staorage (passion_write_usection_1D)
				elements written = elements written + icla size
			endfor
			if elements written < size of the section
				size of the slab = size of the section - elements written
        slab start index = start index + elements written
				read in the slab
				do forever
					select the element in the array at array offset
					put the element in the slab at the slab offset
					array offset = array offset + 1
					if (slab offset + stride) >= size of the slab
					  slab offset =  (slab offset + stride) % size of the slab
						break
          else
					  slab offset =  (slab offset + stride) % size of the slab
          endif
				enddo
				slab end index   = slab start index + size of the slab - 1
													 
				write the slab into the temporary staorage (passion_write_usection_1D)
			endif
			free the slab memory
		endif
  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

#ifdef RUNTIME_PROFILE

unsigned long count_seiving(unsigned long nops)
{
extern long seiving;

	seiving += nops;
	return seiving;
}

unsigned long count_nonseiving(unsigned long nops)
{
extern long nonseiving;

	nonseiving += nops;
	return nonseiving;
}

#endif

#ifndef ABS
#define ABS(x)  ((x)>=0?(x):-(x))
#endif

int passion_write_block_section_1D(LAF_File fd, char *Array, int ArrayIndex, 
                                  int StartIndex, int stride, int DataLen, 
                                  int elemsize, int headersize, int icla_size, 
			 	  int OldFile)
{
int SectionSize, Slabs2Write, ElementsWritten;
int ArrayOffset, SlabOffset, i, FileOffset;
int SlabStartIndex, SlabEndIndex, SlabSize;
char *temp_buffer;
int PrevSlabElem=0;  /* to keep track of the no. of current-block elements
											read in the previous slab */
	 /* make sure stride0 is positive */
	int stride0 = ABS(stride);

	int block_size = 1;

	if (DataLen <= 0) {
#ifdef DEBUG
fprintf(stderr,"passion_write_block_section_1D: DataLen is (less than) zero.\n");
#endif
	   return 0;
        }

	if ((stride0 < 2) || (stride0 == block_size) || DataLen == 1)
	{
		if (stride < 0) DataLen = -DataLen;
		/* there is no stride in the section */
    		if (passion_write_usection_1D(fd, Array, ArrayIndex, StartIndex, 
					DataLen, elemsize, headersize))
    		{
			return -1; 
		}
	}
	else
	{
		/* the section is strided */
                int seiving = 1;
                if (stride > MAX_STRIDE)  seiving = 0;
		if (stride > icla_size && seiving)   icla_size = 1000000;
/*		if (stride > icla_size || stride > MAX_STRIDE) */
/*		if (stride > icla_size || (stride > MAX_STRIDE && stride < 1300)) */
	        if ( seiving == 0) 
		{
			/* sieving is of no use, each array element has to be written
			individually */
/*
printf("passion_write_block_section_1D: Stride:%d is bigger than icla size:%d.\n",stride, icla_size);
*/
		  /*SlabStartIndex = StartIndex;*/

		  FileOffset = headersize + StartIndex * elemsize;
		  ArrayOffset = ArrayIndex * elemsize;
		  SlabSize = elemsize * stride;
		  for (i=0; i<DataLen ; i++)
		  {
			/* write one element a time */
/*
                    if (passion_write_usection_1D(fd, Array, ArrayIndex+i, SlabStartIndex, 
					     1, elemsize, headersize))
*/
    		    if (passion_write(fd, FileOffset, Array+ArrayOffset, elemsize))
		    {
#ifdef DEBUG
      fprintf(stderr,"PASSION : (passion_write_block_section_1D)Error(unseiving). \n");
#endif
			return -1;
		    }  
		    ArrayOffset += elemsize;
		    FileOffset += SlabSize; 
		  }

#ifdef RUNTIME_PROFILE
		  count_nonseiving((unsigned long)1);     
#endif
		  return 0;
/*
printf("passion_write_block_section_1D: Stride:%d is bigger than the icla size:%d. This case is not yet programmed.\n",stride, icla_size);
			return -1;
*/
		}

    		temp_buffer = (char *)malloc(icla_size * elemsize);
		if (temp_buffer == (char *)0)
		{
#ifdef DEBUG
      fprintf(stderr,"PASSION : (write section 1D) Not enough memory\n");
#endif
      			return -1;
		}

    		  /* calculate no. of full slabs to write */
		SectionSize = (DataLen-1) * stride0 + 1;
		Slabs2Write = SectionSize / icla_size;

		ElementsWritten = 0;
		ArrayOffset = ArrayIndex;
		SlabOffset = 0;

		for (i = 0; i < Slabs2Write; i++)
		{
      			SlabStartIndex = StartIndex + ElementsWritten;
			SlabEndIndex   = SlabStartIndex + icla_size - 1;
			if (OldFile)
			{
                if (passion_read_usection_1D(fd, temp_buffer, 0, SlabStartIndex, 
					     icla_size, elemsize, headersize))
        		  {
/*		in this case, we donot return because read may occur beyond file size */
/*
			  	free(temp_buffer);
			  	return -1;
*/
			  }
			}
			while (1)
			{
				if (PrevSlabElem) {
					memcpy(temp_buffer, Array + (ArrayOffset * elemsize), 
								 (block_size - PrevSlabElem)*elemsize);
					ArrayOffset = ArrayOffset + block_size - PrevSlabElem;
					PrevSlabElem = 0;
					}
				else {
					if ((SlabOffset + block_size) <= icla_size) {
						memcpy(temp_buffer + (SlabOffset * elemsize), Array + 
									 (ArrayOffset * elemsize), (block_size * elemsize));
						ArrayOffset += block_size;
						}
					else {
						PrevSlabElem = icla_size - SlabOffset;
						memcpy(temp_buffer + (SlabOffset * elemsize), Array + 
									(ArrayOffset * elemsize), (PrevSlabElem * elemsize));
						ArrayOffset += PrevSlabElem;
						}
					}
				if ((SlabOffset + stride) >= icla_size)
        			{
					/* the slab is over */
					if (PrevSlabElem)
						SlabOffset = -1*PrevSlabElem;
					else
			    	                SlabOffset =  (SlabOffset + stride) % icla_size;
					break;
        			}
        			else
        			{
				  SlabOffset =  (SlabOffset + stride) % icla_size;
        			}
			}

      if (passion_write_usection_1D(fd, temp_buffer, 0, SlabStartIndex, 
					icla_size, elemsize, headersize))
      {
#ifdef DEBUG
      fprintf(stderr,"PASSION : (passion_write_block_section_1D)Error \n");
#endif
				free(temp_buffer);
				return -1;
			}
			ElementsWritten = ElementsWritten + icla_size;
    }
	if (ElementsWritten < SectionSize)
	{
		SlabSize = SectionSize - ElementsWritten;
      		SlabStartIndex = StartIndex + ElementsWritten;
		SlabEndIndex   = SlabStartIndex + SlabSize - 1;

		if (OldFile)
		{
        	   if (passion_read_usection_1D(fd, temp_buffer, 0, SlabStartIndex, 
					SlabSize, elemsize, headersize))
        	   {
/*
			  	free(temp_buffer);
			  	return -1;
*/
		   }
		}
      		while (1)
		{
			if (PrevSlabElem) {
				memcpy(temp_buffer, Array + (ArrayOffset * elemsize), 
							 (block_size - PrevSlabElem)*elemsize);
				ArrayOffset = ArrayOffset + block_size - PrevSlabElem;
				PrevSlabElem = 0;
			}
			else {
				if ((SlabOffset + block_size) <= SlabSize) {
					memcpy(temp_buffer + (SlabOffset * elemsize), Array + 
								 (ArrayOffset * elemsize), (block_size * elemsize));
					ArrayOffset += block_size;
				}
				else {
					PrevSlabElem = SlabSize - SlabOffset;
					memcpy(temp_buffer + (SlabOffset * elemsize), Array + 
								(ArrayOffset * elemsize), (PrevSlabElem * elemsize));
					ArrayOffset += PrevSlabElem;
				}
			}
			if ((SlabOffset + stride) >= SlabSize)
        		{
				/* the slab is over */
				if (PrevSlabElem)
					SlabOffset = -1*PrevSlabElem;
				else
			    		SlabOffset =  (SlabOffset + stride) % SlabSize;
				break;
        		}
        		else
        		{
				  SlabOffset =  (SlabOffset + stride) % SlabSize;
        		}
		} /* end of while */

      		if (passion_write_usection_1D(fd, temp_buffer, 0, SlabStartIndex, 
					SlabSize, elemsize, headersize))
      		{
#ifdef DEBUG
      fprintf(stderr,"PASSION : (passion_write_block_section_1D)Error \n");
#endif
				free(temp_buffer);
				return -1;
		}
	}
	free(temp_buffer);
	}

#ifdef RUNTIME_PROFILE
	count_seiving((unsigned long)1);     
#endif

	return 0;
}

#endif  /* PASSION_WR_SS_1D */
