/*
$Id: rd_sbs_1d.c,v 1.2 1995/10/17 02:15:24 ssmore Exp $

File Name : rd_sbs_1d.c
Author    : srshah

Entry Point(s) :

(1) passion_read_block_section_1D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_SBS_1D
#define PASSION_RD_SBS_1D

/*#define DEBUG*/

#ifdef DEBUG
#include <stdio.h>
#endif

#include <memory.h>

/*
#include "ocad.h"
#include "pfile.h"
*/
#include "passion_base.h"

/*
  Name : passion_read_block_section_1D

  Function : This function is used to read a block from a strided section 
	for 1-D array.

  Parameters : 
    Input :
      fd       - LAF descriptor
		  StartIndex - start of the section
			EndIndex - end of the section
			stride  - stride in the section
			block_size - size of the block to be read from a strided section
		  elemsize - size of each array element in bytes
			headersize - size of the header
			icla_size - size of the icla in no. of elements
			Index    - index in the array where to put data
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		if there is no stride in the section
			read the unstrided section (passion_usection_1D)
		else
			calculate the icla size
			allocate space equal to ICLA size for temporary storage

      calculate no. of slabs to read
			slabs to read = size of the section / icla size
			elements read = 0
			array offset = index
			slab offset = 0
			for each slab to be read
        slab start index = start index + elements read
				slab end index   = slab start index + icla size - 1
				read the slab into the temporary staorage (passion_read_usection_1D)
				elements read = elements read + icla size
				do forever
					select the element in the slab at the slab offset
					put the element in the array at array offset
					array offset = array offset + 1
					if slab offset >  (slab offset + stride) % icla size
						break
          else
					  slab offset =  (slab offset + stride) % icla size
          endif
				enddo
			endfor
			if elements read < size of the section
				size of the slab = size of the section - elements read
        slab start index = start index + elements read
				slab end index   = slab start index + size of the slab - 1
													 
				read the slab into the temporary staorage (passion_read_usection_1D)
				do forever
					select the element in the slab at the slab offset
					put the element in the array at array offset
					array offset = array offset + 1
					if slab offset >  (slab offset + stride) % size of the slab
						break
          else
					  slab offset =  (slab offset + stride) % size of the slab
          endif
				enddo
			endif
			free the slab memory
		endif
  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

#ifndef ABS
#define ABS(x)  ((x)>=0?(x):-(x))
#endif

int passion_read_block_section_1D(LAF_File fd, char *Array, int ArrayIndex, 
                                  int StartIndex, int stride, int DataLen, 
                                  int elemsize, int headersize, int icla_size)
{
int SectionSize, Slabs2Read, ElementsRead;
int ArrayOffset, SlabOffset, i, FileOffset;
int SlabStartIndex, SlabEndIndex, SlabSize;
int PrevSlabElem;  /* to keep track of the no. of current-block elements */
char *temp_buffer;

int block_size = 1;

	 /* make sure stride0 is positive */
	int stride0 = ABS(stride);

        if (DataLen <= 0) return -1; 

	if ((stride0 == 1) || (stride0 == block_size) || DataLen == 1 )
	{
		if (stride < 0) DataLen = -DataLen;
		/* there is no stride in the section practically */
    		if (passion_read_usection_1D(fd, Array, ArrayIndex, StartIndex,
					     DataLen, elemsize, headersize))
    		{
#ifdef DEBUG
    	fprintf(stderr,"PASSION : (passion_read_block_section_1D) usection error: ArrayIndex:%d,StartIndex:%d,DataLen:%d,headersize:%d\n", ArrayIndex, StartIndex, DataLen,headersize);
#endif
			return -1;
		}
	}
	else
	{
		/* the section is strided */
		if (stride0 > icla_size || stride0 > MAX_STRIDE)
		{
			/* sieving is of no use, each array element has to be read
			individually */
/*
printf("passion_read_block_section_1D: Stride:%d is bigger than icla size:%d.\n",stride, icla_size);
*/
		  FileOffset = headersize + StartIndex * elemsize;
		  ArrayOffset = ArrayIndex * elemsize;
		  SlabSize = elemsize * stride;
		  for (i=0; i<DataLen ; i++)
		  {
			/* read one element a time */
/*
                    if (passion_read_usection_1D(fd, Array, ArrayIndex+i, 
					SlabStartIndex, 1, elemsize, headersize))
*/
    		    if (passion_read(fd, FileOffset, Array+ArrayOffset, elemsize))
		    {
#ifdef DEBUG
      fprintf(stderr,"PASSION : (passion_read_block_section_1D)Error \n");
#endif
			return -1;
		    }  
		    ArrayOffset += elemsize;
		    FileOffset += SlabSize; 
		  }
		  return 0;
/*
printf("passion_read_block_section_1D: Stride:%d is bigger than the icla size:%d. This case is not yet programmed.\n",stride, icla_size);
			return -1;
*/
		}
    		temp_buffer = (char *)malloc(icla_size * elemsize);
		if (temp_buffer == (char *)0)
		{
#ifdef DEBUG
     	fprintf(stderr,"PASSION : (read section 1D) Not enough memory\n");
#endif
      			return -1;
		}

    		  /* calculate no. of full slabs to read */
		SectionSize = (DataLen-1) * stride0 + 1;         /* ??? */
		Slabs2Read = SectionSize / icla_size;

		ElementsRead = 0;
		ArrayOffset = ArrayIndex;
		SlabOffset = 0;

		for (i = 0; i < Slabs2Read; i++)
		{
     			SlabStartIndex = StartIndex + ElementsRead;
			SlabEndIndex   = SlabStartIndex + icla_size - 1;

      			if (passion_read_usection_1D(fd, temp_buffer, 0,
				 SlabStartIndex, icla_size, elemsize, headersize))
      			{
#ifdef DEBUG
      			fprintf(stderr,"PASSION : (read section) ustride read error1.\n");
#endif
				free(temp_buffer);
				return -1;
			}

			ElementsRead = ElementsRead + icla_size;
			PrevSlabElem = 0;
			while (1)
			{
				if (SlabOffset < 0) {
					memcpy(Array + (ArrayOffset * elemsize), 
						 temp_buffer, 
						 (block_size + SlabOffset) * elemsize);
					ArrayOffset += (block_size + SlabOffset);
				}
				else {
					if ((SlabOffset + block_size) <= icla_size) {
						memcpy(Array + (ArrayOffset * elemsize),
							temp_buffer + (SlabOffset * elemsize),
							(block_size * elemsize) );
						ArrayOffset += block_size;
					}
					else {
						PrevSlabElem = icla_size - SlabOffset;
						memcpy(Array + (ArrayOffset * elemsize), 
							 temp_buffer + (SlabOffset * elemsize), 
						   (PrevSlabElem * elemsize) );
						ArrayOffset += PrevSlabElem;
					}
				}
				if ((SlabOffset + stride) >= icla_size)
        			{
					/* the slab is over */
					if (PrevSlabElem)
						SlabOffset = -1*PrevSlabElem;
					else
				    		SlabOffset =  (SlabOffset + stride) % icla_size;
					break;
        			}
        			else
        			{
				  	SlabOffset =  (SlabOffset + stride) % icla_size;
        			}
			}
    		}	/*  for */

		if (ElementsRead < SectionSize)
		{
			SlabSize = SectionSize - ElementsRead;
      			SlabStartIndex = StartIndex + ElementsRead;

      			if (passion_read_usection_1D(fd, temp_buffer, 0 ,
				 SlabStartIndex, SlabSize, elemsize, headersize))
      			{
#ifdef DEBUG
      			fprintf(stderr,"PASSION : (read section) ustride read error2: SlabStartIndex:%d,SlabSize:%d.\n",SlabStartIndex,SlabSize);
#endif
				free(temp_buffer);
				return -1;
			}
      			while (1)
			{
			    if (SlabOffset < 0) {
				memcpy(Array + (ArrayOffset * elemsize), 
						 temp_buffer, 
						 (block_size + SlabOffset) * elemsize);
				ArrayOffset += (block_size + SlabOffset);
			    }
			    else {
				if ((SlabOffset + block_size) <= SlabSize) {
					memcpy(Array + (ArrayOffset * elemsize),
							temp_buffer + (SlabOffset * elemsize),
							(block_size * elemsize) );
					ArrayOffset += block_size;
				}
				else {
					PrevSlabElem = SlabSize - SlabOffset;
					memcpy(Array + (ArrayOffset * elemsize), 
						 temp_buffer + (SlabOffset * elemsize), 
					   (PrevSlabElem * elemsize) );
					ArrayOffset += PrevSlabElem;
				}
			    }  /* end of else */
			    if ((SlabOffset + stride) >= SlabSize)
        		    {
				/* the slab is over */
				if (PrevSlabElem)
					SlabOffset = -1*PrevSlabElem;
				else
				    	SlabOffset =  (SlabOffset + stride) % SlabSize;
				break;
        		    }
        		    else
        		    {
				  SlabOffset =  (SlabOffset + stride) % SlabSize;
        		    }
			} /* end of while */
    		}
		free(temp_buffer);
	}
	return 0;
}

/* for debugging */
void print_array(char *Array, int icla_size) 
{
int i;
	for(i=0; i< icla_size; i++)
			printf("%.0f\t", (float)((double *)Array)[i]);
	printf("\n");
}

#endif  /* PASSION_RD_SBS_1D */
