/*
$Id: rd_us_3d.c,v 1.2 1995/09/13 21:06:55 ssmore Exp $

File Name : rd_us_3d.c
Author    : ssmore

Entry Point(s) :

(1) passion_read_usection_3D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_RD_US_3D
#define PASSION_RD_US_3D

#include <stdio.h>

#include "passion_base.h"

/*#define DEBUG*/


/*
  Name : passion_read_usection_3D

  Function : This function is used to read an unstrided section to 3-D
	array.

  Parameters : 
    Input :
      fd  - file descriptor for LAF
			Index0 - row index in the array
			Index1 - column index in the array
			Index2 - z index in the array
			StartIndex0 - starting row index in the section
			StartIndex1 - starting column index in the section
			StartIndex2 - starting z index in the section
		  EndIndex0   - ending row index in the section
			EndIndex1   - ending column index in the array
			EndIndex2   - ending z index in the array
			elemsize    - size of each element of the array
		  headersize  - size of the LAF header
			ocla_size_2 - size of dimension 2 in the OCLA in number of elements
			ocla_size_3 - size of dimension 3 in the OCLA in number of elements
			icla_size_2 - size of dimension 2 in the ICLA in number of elements
			icla_size_3 - size of dimension 3 in the ICLA in number of elements
    Output :
			Array         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    calculate the offset in the file where the section starts.
		file offset = header size + ((startindex0 * ocla_size_2 + startindex1)
		* ocla_size_3 + startindex2) * elementsize

		calculate where to put the data in the buffer
		array offset = ((index0 * icla_size_2 + index1) * icla_size_3 + index2) 
									 * elementsize

    if the section spans the third dimension
		  if the section spans the second dimension
			  you can read the data in one read
		    calculate how many bytes to read
		    datalen = (endindex0 - startindex0 + 1) * size of second dimension
			            * size of the third dimension * elemsize
		    read the data from the file (passion_read)
			else
			  read one plane at a time
		    calculate how many bytes to read on each read call
		    datalen = (endindex1 - startindex1 + 1) * size of the third dim 
									* element size
			  for each plane in the section
		      read the data from the file (passion_read)
				  file offset = file offset + ocla_size_2 * ocla_size_3 * element size
				  array offset = array offset + icla_size_2 * icla_size3 * element size
			  endfor
			endif
		else
			read in the z dimension
			datalen = (endindex2 - startindex2 + 1) * element size
			for each row
		    file offset = header size + ((ocla row * ocla_size_2 + startindex1)
		                  * ocla_size_3 + startindex2) * elementsize
		    array offset = ((icla row * icla_size_2 + index1) * icla_size_3 
										 	 + index2) * elementsize
				for each column
		      read the data from the file (passion_read)
				  file offset = file offset + ocla_size_3 * element size
				  array offset = array offset + icla_size3 * element size
				endfor
      endfor
		endif

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

/* note: for unstride section
ocla_size_0=u1           ocla_size_1=u2
DataLen0=icla_size_0     DataLen1=icla_size_1
*/
int passion_read_usection_3D(LAF_File fd, char *Array, int ArrayIndex, 
			     int StartIndex, int u0, int u1, int u2,
                             int DataLen0, int DataLen1, int DataLen2,
			     int elemsize, int headersize, 
                             int icla_size_0, int icla_size_1,
                             int ocla_size_0, int ocla_size_1)
{
unsigned int FileOffset, DataLen, i, j, k;
int  ArrayOffset;
float *data = (float*)Array;

    /* calculate the offset in the file where the section starts. */
		FileOffset = headersize + StartIndex * elemsize;

		/* calculate where to put the data in the buffer */

    if ((DataLen0 == ocla_size_0) && (DataLen0 == icla_size_0))
		{
      if ((DataLen1 == ocla_size_1) && (DataLen1 == icla_size_1))
		  {
		    /* calculate how many bytes to read */
		    DataLen = DataLen0 * DataLen1 * DataLen2 * elemsize; 

        if (passion_read(fd, FileOffset, Array + ArrayIndex * elemsize , DataLen))
		    {
#ifdef DEBUG
fprintf(stderr,"passion_read_usection_3D(unstride) error: FileOffset:%d, ArrayIndex:%d, DataLen:%d\n",FileOffset,ArrayIndex,DataLen);
#endif
		    	return -1;
		    }
		  }
		  else
		  {
		    /* calculate how many bytes to read on each write call */
		    DataLen = DataLen1 * ocla_size_0 * elemsize;

			  for (i = 0; i < DataLen2; i++)
			  {
          if (passion_read(fd, FileOffset, Array + ArrayIndex * elemsize , DataLen))
		      {
#ifdef DEBUG
fprintf(stderr,"passion_read_usection_3D (plane)error: FileOffset:%d, ArrayIndex:%d, DataLen:%d error\n",FileOffset, ArrayIndex, DataLen);
#endif
		      	return -1;
		      }
				  FileOffset = FileOffset + u2*elemsize;
				  ArrayIndex = ArrayIndex + icla_size_0 * icla_size_1 ;
			  }
		  }
    }
		else
		{
			/* read in the z dimension */
			DataLen = DataLen0 * elemsize;
			k = 0;
/*
printf("DataLen0:%d,DataLen1:%d,DataLen2:%d\n",DataLen0,DataLen1,DataLen2);
printf("u0:%d,u1:%d,u2:%d,elemsize:%d\n",u0,u1,u2,elemsize);
printf("icla_size_0:%d,icla_size_1:%d\n",icla_size_0,icla_size_1);
printf("ocla_size_0:%d,ocla_size_1:%d\n",ocla_size_0,ocla_size_1);
*/
		  for (i = 0; i < DataLen2; i++)
		  {
		    FileOffset = headersize + (StartIndex + i * u2) * elemsize;
/*		    ArrayOffset = (k * icla_size_1 * icla_size_0 ) * elemsize + ArrayIndex; */
		    ArrayOffset = (k * DataLen0 * DataLen1) + ArrayIndex;
		    for (j = 0; j < DataLen1; j++)
		    {
          if (passion_read(fd, FileOffset, Array + ArrayOffset * elemsize, DataLen))
		      {
#ifdef DEBUG
fprintf(stderr,"passion_read_usection_3D error: StartIndex:%d,FileOffset:%d, ArrayOffset:%d, DataLen:%d error\n",StartIndex,FileOffset, ArrayOffset, DataLen);
#endif
		      	return -1;
		      }
			  FileOffset = FileOffset + u1 * elemsize;
			  ArrayOffset = ArrayOffset + DataLen0 ;
		     }
			k = k + 1;
			}
		}

    return 0;
}

#endif  /* PASSION_RD_US_3D */
