/*
$Id: read_cur.c,v 1.9 1995/10/16 16:41:32 ssmore Exp $

File Name : read_cur.c
Author    : ssmore

Entry Point(s) :

(1) passion_read_cur   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_READ_CUR_C
#define PASSION_READ_CUR_C

#ifdef PARAGON
#include "nx.h"       /* This include file is PARAGON specific */
#else
#ifdef DELTA
#include "mesh.h"       /* This include file is DELTA specific */
#else
#ifdef IPSC
#include "cube.h"       /* This include file is IPSC specific */
#else
#ifdef PIOFS
#include <unistd.h>
#else
/* its an error if the preprocessor comes here but how to flag it ?*/
#endif
#endif
#endif
#endif

/* for debug */
/*#define DEBUG*/

#ifdef DEBUG
#include <stdio.h>
#endif

#ifdef SP_PIOFS_MPI
#include <unistd.h>
#endif
#ifdef SP_AFS_MPI
#include <unistd.h>
#endif

#include "pfile.h"

#ifdef RUNTIME_PROFILE

double count_bytes_read(unsigned long bytes)
{
extern double bytes_read;

	bytes_read += (double)bytes;
	return bytes_read;
}

unsigned long count_read_ops(unsigned long nops)
{
extern unsigned int read_ops;

	read_ops += nops;
	return read_ops;
}
#endif

/*
  Name : passion_read_cur

  Function : This function is used to read data from the LAF. It starts
	reading data at the current file offset.

  Parameters : 
    Input :
			FileDescriptor - File descriptor of the LAF
			DataLen        - Amount of data in bytes
    Output :
			Buffer         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		read the data from the LAF

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_read_cur(LAF_File fd, char *buffer, unsigned int buflen)
{
  int numRead;

#ifdef PIOFS
  if (read(fd, (void *)buffer, (size_t)buflen) != (ssize_t)buflen)
  {
    return -1;
	}
#else
#ifdef SP_PIOFS_MPI
  if (read(fd, (void *)buffer, (size_t)buflen) != (ssize_t)buflen)
  {
    return -1;
	}
#else
#ifdef SP_AFS_MPI
  if (read(fd, (char *)buffer, (int)buflen) != (int)buflen)
  {
    return -1;
	}
#else
#ifdef DELTA
  /*
    return value of the cread is not checked
  */
  cread(fd, buffer, buflen);
#else
#ifdef PARAGON
  /*
    return value of the cread is not checked
  */
  cread(fd, buffer, buflen);
#else
#ifdef IPSC
  /*
    return value of the cread is not checked
  */
  cread(fd, buffer, buflen);
#else
#ifdef HAVE_MPIO
  MPI_Status status;
  MPI_File_read(fd, buffer, buflen, MPI_BYTE, &status);
/*
  MPIO_Request request;
  MPI_File_iread(fd, buffer, buflen, MPI_BYTE, &request);
  MPIO_Wait(&request, &status);
*/
#else
#if (defined(UNIX) || defined(AIX))
  numRead = read(fd, (char *)buffer, (int)buflen);
  if (numRead == -1) 
  {
#ifdef DEBUG
fprintf(stderr,"passion_read_cur: error: buflen:%d, read:%d\n",buflen,numRead);
#endif
    return -1;
  }
  if (numRead < (int)buflen) 
  {
#ifdef DEBUG
fprintf(stderr,"passion_read_cur: Warning: buflen:%d, read:%d\n",buflen,numRead);
#endif
    return 0;
  }

#else
	/* no architecture defined */
  return -1;
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifdef RUNTIME_PROFILE
	count_bytes_read((unsigned long)buflen); 
	count_read_ops((unsigned long)1);
#endif
	return 0;
}

#endif  /* PASSION_READ_CUR_C */
