/*
$Id: seek_pread.c,v 1.1 1995/06/21 15:34:47 ssmore Exp $

File Name : seek_pread.c
Author    : ssmore

Entry Point(s) :

(1) passion_seek_pread   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_SEEK_PREAD_C
#define PASSION_SEEK_PREAD_C

#include "passion_base.h"

/*
  Name : passion_pread

  Function : This function is used to read data asynchronously from the LAF. 
	It first seeks to the specified offset wrt the start of the LAF and then 
	starts reading the data.

  Parameters : 
    Input :
			FileDescriptor - File descriptor of the LAF
			Offset         - Offset from the start of the LAF, the read starts
											 here
			DataLen        - Amount of data in bytes
    Output :
			Buffer         - The pointer to the data buffer
    Input/Output :
      none

  Return Value : 
    0+, if successful, the prefetch identifier
		-1, otherwise

  Algorithm :

    seek to the offset (passion_seek)
		read the data asynchronously from the LAF (passion_pread_cur)

  Side Effects :
    (1) Modifies the file pointer in the LAF.

  Shortcomings :
    (1) Since passion_pread_cur is NOT supported on PIOFS on SP2, thus
		function is also unavailable on PIOFS.

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_pread(LAF_File fd, unsigned long int offset, char *buffer, 
								 unsigned int buflen)
{
  if (passion_seek(fd, offset)) return -1;
  return  passion_pread_cur(fd, buffer, buflen);
}

#endif  /* PASSION_SEEK_PREAD_C */
