/*
$Id: seek_write.c,v 1.1 1995/06/21 15:34:52 ssmore Exp $

File Name : seek_write.c
Author    : ssmore

Entry Point(s) :

(1) passion_seek_write   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_SEEK_WRITE_C
#define PASSION_SEEK_WRITE_C

#include "passion_base.h"

/*
  Name : passion_seek_write

  Function : This function is used to write data to the LAF. It first
	seeks to the specified offset wrt the start of the LAF and then starts
	writing the data.

  Parameters : 
    Input :
			FileDescriptor - File descriptor of the LAF
			Offset         - Offset from the start of the LAF, the read starts
											 here
			Buffer         - The pointer to the data buffer
			DataLen        - Amount of data in bytes
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    seek to the offset (passion_seek)
		write the data to the LAF (passion_write_cur)

  Side Effects :
    (1) Modifies the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_write(LAF_File fd, unsigned long int offset, char *buffer, 
								 unsigned int buflen)
{
  if (passion_seek(fd, offset)) return -1;
  if (passion_write_cur(fd, buffer, buflen)) return -1;
	return 0;
}

#endif  /* PASSION_SEEK_WRITE_C */
