/*
Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
(2) The implementation assumes that the array is stored in the row major
    format.
*/

#define DEBUG

#ifndef PASSION_VAL_SECT_C
#define PASSION_VAL_SECT_C

#include <stdio.h>

#include "passion_base.h"

/*
  Name : passion_validate_section

  Function : This function is used to validate a section request against
						 the OCAD information

  Parameters : 
    Input :
      OCADp       - Pointer to the OCAD structure
      AccessArray - The section to be validated
			Index       - Index inside the ICLA where the section is to be
			              placed
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
    -1, if not successful

  Algorithm :

    if OCAD pointer is not specified
      return error 
    endif

    to be filled in

    return success

  Side Effects :
    none

  Shortcomings :
    none

  Modifications :

  Description                                         Date       Author
  ------------------------------------------------------------------------------
*/
int passion_validate_section(int rank, int start, int *stride, int *number, 
                             int headersize, int *ocla_size, int icla_size)
{
int i;
int offset;
int ocla;
int read_num;

     if (start < 0)  {
	fprintf(stderr,"PASSION : (validate section) Invalid Start:%d Section\n",start);
 	return -1;
     }

     for (i=0; i<rank; i++) 
     {
	  /* consider the case stride is negative. */
/*
	if (stride[i] < 0 || number[i] < 0) {
		stride[i] = stride[i]<0? -stride[i] : stride[i];
		number[i] = number[i]<0? -number[i] : number[i];
		start = start - stride[i] * number[i];
	}
*/
	if (number[i] > ocla_size[i])
	{
          fprintf(stderr,"PASSION : (validate section) Out Of OCLA:%d > %d.\n",number[i],ocla_size[i]);
          return -1; 
	}
     }

/*
     ocla=1;
     for (i=0; i<rank; i++)
     {
	ocla *= ocla_size[i];
     }
     ocla = ocla + headersize;
     offset=start;
     for (i=0; i<rank; i++)
     {
	offset += stride[i]*(number[i]-1);
     }
     if (offset > ocla) {
        fprintf(stderr,"PASSION : (validate section) Out Of Section:%d > %d (headersize:%d)\n",offset,ocla,headersize);
        return -1; 
     }
*/

     read_num = 1;
     for (i=0; i<rank; i++)
     {
	read_num *= number[i];
     }
     if (read_num > icla_size) {
        fprintf(stderr,"PASSION : (validate section) Invalid Read(%d) more than ICLA size(%d).\n",read_num,icla_size);
 	return -1;
     }
     
     return 0;
}


#endif  /* PASSION_VAL_SECT_C */
