/*
$Id: wr_sbs_2d.c,v 1.2 1995/09/02 15:31:03 ssmore Exp $

File Name : wr_ss_2d.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_section_2D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WR_SBS_2D
#define PASSION_WR_SBS_2D

#include <stdio.h>

#include <memory.h>

#include "passion_base.h"
/*
#include "ocad.h"
#include "pfile.h"
*/

/*#define DEBUG*/

/*
  Name : passion_write_section_2D

  Function : This function is used to write a strided section to 2-D
	array.

  Parameters : 
    Input :
      fd       - LAF descriptor
			Array         - The pointer to the data buffer
		  StartIndex0 - start row of the section
		  StartIndex1 - start column of the section
			EndIndex0 - end row of the section
			EndIndex1 - end column of the section
			stride0  - row stride in the section
			stride1  - column stride in the section
		  elemsize - size of each array element in bytes
			headersize - size of the header
			icla_size - size of the icla in no. of elements
			Index0    - row index in the array where to put data
			Index1    - column index in the array where to put data
			OldFile   - non-zero if writing to an existing file
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

	if no stride in the column
	  if no stride in the row
			write 2-D unstrided section
		else
			you are writing non-consecutive strips of unstrided data
      section start = start index1
      section end   = end   index1
			array row     = index0
			for each row in the section
				compute the offset of the start of the row from the start of the
				file and call it new file header
				new file header = row * size of ocla row * size of each element
				                  + file header
				array index   = (array row * size of icla row) + index1
        write the row as unstrided section of 1-D array 
				(passion_write_usection_1d)
				array row = array row + 1
			endfor
		endif
  else
		its a general 2-D section
    section start = start index1
    section end   = end   index1
 		stride        = stride1
		array row     = index0
		for each row in the section
			compute the offset of the start of the row from the start of the
			file and call it new file header
			new file header = row * size of ocla row * size of each element
			                  + file header
			array index   = (array row * size of icla row) + index1
      write the row as regular section of 1-D array (passion_write_section_1d)
			array row = array row + 1
		endfor
	endif

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_write_block_section_2D(LAF_File fd, char *Array, int Index, int StartIndex, 
				  int u0, int u1, int DataLen0, int DataLen1, 
				  int elemsize, int headersize, int icla_size, 
                                  int icla_rowsize, int ocla_rowsize,
				  int OldFile)
{
int i, j, fetch_size, SectionStart, SectionEnd;
int NewHeaderSize, ArrayIndex, ArrayRow;

int block_size0 = 1;
int block_size1 = 1;

	if (DataLen0 <= 0 || DataLen1 <= 0) {
#ifdef DEBUG
fprintf(stderr,"passion_write_block_section_2D: DataLen is (less than) zero.\n");
#endif
	   return 0;
        }

	if ((u0 == 1) || (u0 == block_size1))
	{
	  if ((u1 == ocla_rowsize) || (u1 == block_size0))
	  {
		/* write 2-D unstrided section */
              if (passion_write_usection_2D(fd, Array, Index, StartIndex, u0, u1, 
				DataLen0, DataLen1, elemsize, headersize, 
				icla_rowsize, ocla_rowsize))
	      {
			return -1;
	      }
	  }
	  else
	      {
		/* you are writing non-consecutive strips of unstrided data */
		ArrayRow     = 0;
		for (i = 0; i < DataLen1; i++)
		{
			if ((DataLen1 - i + 1) < block_size0) {
				fetch_size = DataLen1 - i + 1;
			}
			else {
				fetch_size = block_size0;
			}

			for (j=0; j<fetch_size; j++) {

				/* compute the offset of the start of the row from the start of the file and call it new file header */
				NewHeaderSize = (i*u1) * elemsize + headersize;
				ArrayIndex   = (ArrayRow * icla_rowsize) + Index;

 	                        /* write the row as unstrided section of 1-D array */
	        		if (passion_write_usection_1D(fd, Array, ArrayIndex, 
						StartIndex, DataLen0, elemsize, 
						NewHeaderSize))
 	       			{
					return -1;
				}
				ArrayRow = ArrayRow + 1;
			}
		}  /* for */
	   }
	}
  else
	{
		/* its a general 2-D section */
	   ArrayRow     = 0;
	   for (i = 0; i < DataLen1; i++)
	   {
		if ((DataLen1 - i + 1) < block_size0) {
			fetch_size = DataLen1 - i + 1;
		}
		else {
			fetch_size = block_size0;
		}
		for (j=0; j<fetch_size; j++) {

			/* compute the offset of the start of the row from the start of the file and call it new file header */
			NewHeaderSize = (i*u1) * elemsize + headersize;
			ArrayIndex   = (ArrayRow * icla_rowsize) + Index;

      			/* write the row as unstrided section of 1-D array */
      			if (passion_write_block_section_1D(fd, Array, ArrayIndex, 
					   StartIndex , u0, DataLen0,elemsize,
					   NewHeaderSize, icla_size, OldFile))
      			{
					return -1;
			}
			ArrayRow = ArrayRow + 1;
		}
	   }  /* for */
	}
	return 0;
}

#endif  /* PASSION_WR_SBS_1D */
