/*
$Id: wr_us_1d.c,v 1.1 1995/06/21 15:35:06 ssmore Exp $

File Name : wr_us_1d.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_usection_1D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WR_US_1D
#define PASSION_WR_US_1D

#include "passion_base.h"

/*#define DEBUG*/

#ifdef DEBUG
#include <stdio.h>
#endif

/*
  Name : passion_write_usection_1D

  Function : This function is used to write an unstrided section for 1-D
	array.

  Parameters : 
    Input :
      fd - file descriptor for the LAF
			Array         - The pointer to the data buffer
			Index - index in the array where to start getting data
			StartIndex - index where the section starts
			EndIndex - index where the section ends
      elemsize - size of each element
			headersize - size of the header at the start of the file
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    calculate the offset in the file where the section starts.
		file offset = header size + start index * elemsize
		calculate how many bytes to write
		datalen = (endindex - startindex + 1) * elemsize
		calculate where to get the data in the buffer
		array offset = index * elemsize

		write the data to the file (passion_write)

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

/* note:
the Index, StartIndex, DataLen is of elemsize.
*/
int passion_write_usection_1D(LAF_File fd, char *Array, unsigned int Index, 
                             int StartIndex, int DataLen, 
                             int elemsize, int headersize)
{
unsigned int FileOffset, ArrayOffset ;
int ascend = 1;

		 /* if DataLen is negative, means read in inverse */
		if (DataLen < 0) {
			DataLen = -DataLen;
			StartIndex = StartIndex - DataLen + 1;
			ascend = 0;
		}

    		/* calculate the offset in the file where the section starts. */
		FileOffset = headersize + (StartIndex * elemsize);

		/* calculate how many bytes to write */
		DataLen = DataLen * elemsize;

		/* calculate where to get the data in the buffer */
		ArrayOffset = Index * elemsize;

#ifdef DEBUG
fprintf(stderr,"passion_write_usection_1D: FileOffset:%d, ArrayOffset:%d, DataLen:%d\n",FileOffset,ArrayOffset,DataLen);
#endif

		if ( ascend == 0) {
		    	reverseBuf(Array + ArrayOffset, DataLen, elemsize);
		}

		if (passion_write(fd, FileOffset, Array + ArrayOffset, DataLen))
		{
			return -1;
		}

    		return 0;
}

#endif  /* PASSION_WR_US_1D */
