/*
$Id: wr_us_2d.c,v 1.2 1995/09/13 21:15:38 ssmore Exp $

File Name : wr_us_2d.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_usection_2D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WR_US_2D
#define PASSION_WR_US_2D

#include "passion_base.h"
/*
#include "ocad.h"
#include "pfile.h"
*/

/*
  Name : passion_write_usection_2D

  Function : This function is used to write an unstrided section to 2-D
	array.

  Parameters : 
    Input :
      fd  - file descriptor for LAF
			Index0 - row index in the array
			Index1 - column index in the array
			StartIndex0 - starting row index in the section
			StartIndex1 - starting column index in the section
		  EndIndex0   - ending row index in the section
			EndIndex1   - ending column index in the array
			elemsize    - size of each element of the array
		  headersize  - size of the LAF header
			ocla_rowsize - size of a row in the OCLA in number of elements
			icla_rowsize - size of a row in the ICLA in number of elements
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		ocla_rowsize = no. of elements in the 2nd dimension (ocla) * elementsize
		icla_rowsize = no. of elements in the 2nd dimension (ocla) * elementsize
 
    calculate the offset in the file where the section starts.
		file offset = header size + startindex1 * rowsize + startindex2 *
		              elementsize

		calculate where to get the data in the buffer
		array offset = ((index1 * icla_rowsize) + index2) * elemsize

		if the section spans the second dimension
			you can write the data in one write
		  calculate how many bytes to write
		  datalen = (startindex1 - endindex1 + 1) * size of second dimension
			* elemsize
		  write the data to the file (passion_write)
		else
			write part of one row at a time
		  calculate how many bytes to write on each write call
		  datalen = (startindex2 - endindex2 + 1) * element size
			for each row in the section
		    write the data to the file (passion_write)
				file offset = file offset + ocla_rowsize * element size
				array offset = array offset + icla_rowsize * element size
			endfor
		endif

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_write_usection_2D(LAF_File fd, char *Array, int ArrayIndex, int StartIndex, 
			     int u0,int u1, int DataLen0, int DataLen1, 
                             int elemsize, int headersize, 
                             int icla_rowsize, int ocla_rowsize)
{
unsigned int FileOffset, ArrayOffset, DataLen, i;

	/* calculate the offset in the file where the section starts. */
	FileOffset = headersize + StartIndex * elemsize; 

         /* calculate where to place the data in the buffer */
        ArrayOffset = ArrayIndex * elemsize;

        if ((DataLen0 == ocla_rowsize) && (DataLen0 == icla_rowsize))
	{
		  /* calculate how many bytes to write */
		DataLen = DataLen0 * DataLen1 * elemsize;

                if (passion_write(fd, FileOffset, Array + ArrayOffset, DataLen))
		{
		  	return -1;
		}
	}
	else
	{
		  /* calculate how many bytes to write on each write call */
		DataLen = DataLen0 * elemsize;

		for (i = 0; i < DataLen1; i++)
		{
                   if (passion_write(fd, FileOffset, Array + ArrayOffset, DataLen))
		   {
		    	return -1;
		   }
		   FileOffset = FileOffset + u1 * elemsize;
		   ArrayOffset = ArrayOffset + icla_rowsize * elemsize;
		}
	}

    return 0;
}

#endif  /* PASSION_WR_US_2D */
