/*
$Id: wr_us_3d.c,v 1.2 1995/09/13 21:15:40 ssmore Exp $

File Name : wr_us_3d.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_usection_3D   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WR_US_3D
#define PASSION_WR_US_3D

#include "passion_base.h"

/*#define DEBUG*/

#include <stdio.h>
/*
  Name : passion_write_usection_3D

  Function : This function is used to write an unstrided section to 3-D
	array.

  Parameters : 
    Input :
      fd  - file descriptor for LAF
			Array  - The pointer to the data buffer
			Index0 - row index in the array
			Index1 - column index in the array
			Index2 - z index in the array
			StartIndex0 - starting row index in the section
			StartIndex1 - starting column index in the section
			StartIndex2 - starting z index in the section
		  EndIndex0   - ending row index in the section
			EndIndex1   - ending column index in the array
			EndIndex2   - ending z index in the array
			elemsize    - size of each element of the array
		  headersize  - size of the LAF header
			ocla_size_2 - size of dimension 2 in the OCLA in number of elements
			ocla_size_3 - size of dimension 3 in the OCLA in number of elements
			icla_size_2 - size of dimension 2 in the ICLA in number of elements
			icla_size_3 - size of dimension 3 in the ICLA in number of elements
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

    calculate the offset in the file where the section starts.
		file offset = header size + ((startindex0 * ocla_size_2 + startindex1)
		* ocla_size_3 + startindex2) * elementsize

		calculate where to get the data in the buffer
		array offset = ((index0 * icla_size_2 + index1) * icla_size_3 + index2) 
									 * elementsize

    if the section spans the third dimension
		  if the section spans the second dimension
			  you can write the data in one write
		    calculate how many bytes to write
		    datalen = (endindex0 - startindex0 + 1) * size of second dimension
			            * size of the third dimension * elemsize
		    write the data to the file (passion_write)
			else
			  write one plane at a time
		    calculate how many bytes to write on each write call
		    datalen = (endindex1 - startindex1 + 1) * size of the third dim 
									* element size
			  for each plane in the section
		      write the data to the file (passion_write)
				  file offset = file offset + ocla_size_2 * ocla_size_3 * element size
				  array offset = array offset + icla_size_2 * icla_size3 * element size
			  endfor
			endif
		else
			write in the z dimension
			datalen = (endindex2 - startindex2 + 1) * element size
			for each row
		    file offset = header size + ((ocla row * ocla_size_2 + startindex1)
		                  * ocla_size_3 + startindex2) * elementsize
		    array offset = ((icla row * icla_size_2 + index1) * icla_size_3 
										 	 + index2) * elementsize
				for each column
		      write the data to the file (passion_write)
				  file offset = file offset + ocla_size_3 * element size
				  array offset = array offset + icla_size3 * element size
				endfor
      endfor
		endif

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_write_usection_3D(LAF_File fd, char *Array, int ArrayIndex, 
			      int StartIndex, int u0, int u1, int u2, 
			      int DataLen0, int DataLen1, int DataLen2, 
			      int elemsize, int headersize, 
			      int icla_size_0, int icla_size_1,
			      int ocla_size_0, int ocla_size_1)
{
unsigned int FileOffset, ArrayOffset, DataLen, i, j, k;

    /* calculate the offset in the file where the section starts. */
		FileOffset = headersize + StartIndex * elemsize;

		/* calculate where to get the data in the buffer */

    if ((DataLen0 == ocla_size_0) && (DataLen0 == icla_size_0))
		{
      if ((DataLen1 == ocla_size_1) && (DataLen1 == icla_size_1))
		  {
		    /* calculate how many bytes to write */
		    DataLen = DataLen0 * DataLen1 * DataLen2 * elemsize;


        if (passion_write(fd, FileOffset, Array + ArrayIndex * elemsize, DataLen))
		    {
#ifdef DEBUG
fprintf(stderr,"passion_write_usection_3D(unstride): Error FileOffset:%d, ArrayIndex:%d, DataLen:%d\n",FileOffset, ArrayIndex, DataLen);
#endif
		    	return -1;
		    }
		  }
		  else
		  {
		    /* calculate how many bytes to write on each write call */
		    DataLen = DataLen1 * ocla_size_0 * elemsize;

			  for (i = 0; i < DataLen2; i++)
			  {
          if (passion_write(fd, FileOffset, Array + ArrayIndex * elemsize, DataLen))
		      {
		      	return -1;
		      }
				  FileOffset = FileOffset + u2*elemsize;
				  ArrayIndex = ArrayIndex + icla_size_0 * icla_size_1 ;
			  }
		  }
    }
		else
		{
			/* write in the z dimension */
			DataLen = DataLen0 * elemsize;
			k = 0;
		  for (i = 0; i < DataLen2; i++)
		  {
		    FileOffset = headersize + (StartIndex +i* u2) * elemsize;
		    ArrayOffset = (k * icla_size_1 * icla_size_0) + ArrayIndex;
		    for (j = 0; j < DataLen1; j++)
		    {
          if (passion_write(fd, FileOffset, Array + ArrayOffset * elemsize, DataLen))
		      {
#ifdef DEBUG
fprintf(stderr,"passion_write_usection_3D: StartIndex:%d,FileOffset:%d, ArrayOffset:%d, DataLen:%d error\n",StartIndex,FileOffset, ArrayOffset, DataLen);
#endif
		      	return -1;
		      }
			  FileOffset = FileOffset + u1 * elemsize;
			  ArrayOffset = ArrayOffset + icla_size_0 ;
			}
				k = k + 1;
			}
		}

    return 0;
}

#endif  /* PASSION_WR_US_3D */
