/*
$Id: write_cur.c,v 1.5 1995/11/04 03:53:49 ssmore Exp $

File Name : write_cur.c
Author    : ssmore

Entry Point(s) :

(1) passion_write_cur   prototype in passion_base.h

Modification History :

Module Description                                      Date      Author
------------------------------------------------------------------------------

Notes :

(1) Modifications to the entry points interface should be reflected in the 
    prototypes declarations.
*/

#ifndef PASSION_WRITE_CUR_C
#define PASSION_WRITE_CUR_C

#ifdef PARAGON
#include "nx.h"       /* This include file is PARAGON specific */
#else
#ifdef DELTA
#include "mesh.h"       /* This include file is DELTA specific */
#else
#ifdef IPSC
#include "cube.h"       /* This include file is IPSC specific */
#else
#ifdef PIOFS
#include <unistd.h>
#else
/* its an error if the preprocessor comes here but how to flag it ?*/
#endif
#endif
#endif
#endif
#ifdef SP_PIOFS_MPI
#include <unistd.h>
#endif
#ifdef SP_AFS_MPI
#include <unistd.h>
#endif

#include "pfile.h"

#ifdef RUNTIME_PROFILE

double count_bytes_write(unsigned long bytes)
{
extern  double bytes_write;

	bytes_write += (double)bytes;
	return bytes_write;
}

unsigned long count_write_ops(unsigned long nops)
{
extern unsigned int write_ops;

	write_ops += nops;
	return write_ops;
}
#endif

/*
  Name : passion_write_cur

  Function : This function is used to write data to the LAF. It starts
	writing data at the current file offset.

  Parameters : 
    Input :
			FileDescriptor - File descriptor of the LAF
			Buffer         - The pointer to the data buffer
			DataLen        - Amount of data in bytes
    Output :
			none
    Input/Output :
      none

  Return Value : 
    0, if successful
		-1, otherwise

  Algorithm :

		wriet the data to the LAF

  Side Effects :
    (1) Advances the file pointer in the LAF.

  Shortcomings :
    none

  Modifications :

  Description                                       Date          Author
  ------------------------------------------------------------------------------
*/

int passion_write_cur(LAF_File fd, char *buffer, unsigned int buflen)
{

#ifdef PIOFS
  if (write(fd, (void *)buffer, (size_t)buflen) != (ssize_t)buflen)
  {
    return -1;
	}
#else
#ifdef SP_AFS_MPI
	if (write(fd, (char *)buffer, (int)buflen) != (int)buflen)
	{
		return -1;
	}
#else
#ifdef SP_PIOFS_MPI
  if (write(fd, (void *)buffer, (size_t)buflen) != (ssize_t)buflen)
  {
    return -1;
	}
#else
#ifdef DELTA
  /*
    return value of the cread is not checked
  */
  cwrite(fd, buffer, buflen);
#else
#ifdef PARAGON
  /*
    return value of the cread is not checked
  */
  cwrite(fd, buffer, buflen);
#else
#ifdef IPSC
  /*
    return value of the cread is not checked
  */
  cwrite(fd, buffer, buflen);
#else
#ifdef HAVE_MPIO
  MPI_Status status;
  MPI_File_write(fd, buffer, buflen, MPI_BYTE, &status);
/*
  MPIO_Request request;
  MPI_File_iwrite(fd, buffer, buflen, MPI_BYTE, &request);
  MPIO_Wait(&request, &status);
*/
#else
#if (defined(UNIX) || defined(AIX))
  if (write(fd, (char *)buffer, (int)buflen) != (int)buflen)
  {
    return -1;
	}
#else
	/* no known architecture */
  return -1;
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

#ifdef RUNTIME_PROFILE
	count_bytes_write((unsigned long)buflen);     
	count_write_ops((unsigned long)1);
#endif

	return 0;
}

#endif  /* PASSION_WRITE_CUR_C */
