
// for untyped block operations

// Common utilities for constructing schedules.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"  // debug
#include "passion_base.h"	// Zheng Gengbin
#include "platform.h"
#include "threads.h"
#include "block.h"
#include "adtags.h"
#include "schedule.h"
#include "oc_block.h"


extern align_t statBuf [(STATIC_BUFFER_SIZE + ALIGN_SZ - 1) / ALIGN_SZ] ;

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//	oc_gatherOffBuf		oc_scatterOffBuf
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void oc_gatherOffBuf(PFILE* Pfile, int* offsets, int count, void* buf, 
                                                       int elementLen) 
{
      char* cbuf  = (char*) buf  ;
      int strs = 1;
      int exts = 1;
      for(int i = 0 ; i < count ; i++) {
//AdlibPrintf("oc_gatherOffBuf:offsets:%d, strs:%d, exts:%d.\n", *offsets, strs,exts);
	oc_blockGather(Pfile, *offsets, &strs, &exts, 
		   	1, cbuf, elementLen, elementLen) ;
        cbuf += elementLen ;
        offsets++ ;
      }
}

void oc_scatterOffBuf(PFILE* Pfile, int* offsets, int count, void* buf,
							int elementLen) 
{
      char* cbuf  = (char*) buf  ;
      int strs = 1;
      int exts = 1;
      for(int i = 0 ; i < count ; i++) {
//AdlibPrintf("oc_scatterOffBuf:offsets:%d, strs:%d, exts:%d.\n", *offsets, strs,exts);
	oc_blockScatter(Pfile, *offsets, &strs, &exts, 
		   1, cbuf, elementLen, elementLen) ;
        cbuf += elementLen ;
        offsets++ ;
      }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//	oc_copyMblkMblk		
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// for both IN_CORE & OUT_OF_CORE
void oc_copyMblkMblk(void* dstBase, MultiBlock dstBlock,
                     void* srcBase, MultiBlock srcBlock, 
	  	     const int elementLen) 
{

  // `MultiBlock' to `MultiBlock' copy.  The blocks must be the same shape.

  int dstStatus = dstBlock[MBLK_STA];
  int srcStatus = srcBlock[MBLK_STA];

  int dstIclaSize = dstBlock[MBLK_INS];
  int srcIclaSize = srcBlock[MBLK_INS];
  int icla_size   = min(dstIclaSize, srcIclaSize);

  const int rank      = dstBlock [MBLK_RNK] ;
  int dstOffset       = dstBlock[MBLK_OFF];
  int srcOffset       = srcBlock[MBLK_OFF];
  int* dstStrs        = dstBlock + MBLK_STR ;
  int* srcStrs        = srcBlock + MBLK_STR ;
  int* exts           = dstStrs  + rank ;
  int dstMajor        = dstBlock[MBLK_MAJ];
  int srcMajor        = srcBlock[MBLK_MAJ];


  if (dstStatus == IN_CORE && srcStatus == IN_CORE) {
	copyMblkMblk(dstBase, dstBlock, srcBase, srcBlock, elementLen);
  }
  else {
        int org_storage = defaultStorageMode;
	defaultStorageMode = dstMajor;

        void *buffer;
	int useSrc=0;
	  // source
        if (srcStatus == OUT_OF_CORE)  {
		PFILE *srcPfile = (PFILE*)srcBase;
		if (dstStatus == IN_CORE && dstBlock[MBLK_CTG]) 
		{
  	           char* dst = (char*) dstBase + elementLen * dstBlock [MBLK_OFF];
		   buffer = (void *)dst;
		   oc_blockGather(srcPfile, srcOffset, srcStrs, exts, 
		   	rank, (char *)buffer, elementLen, icla_size) ;
		   defaultStorageMode = org_storage;
		   return;
		}
		else {
		   int vol=1;
                   for (int k=0;k<rank; k++) vol = vol*exts[k];
		   buffer = malloc(vol * elementLen);
		   if(buffer == NULL) AdlibError("oc_copyMblkMblk: Memory exhausted!\n");
		   oc_blockGather(srcPfile, srcOffset, srcStrs, exts, 
		   	rank, (char *)buffer, elementLen, icla_size) ;
		}
        }
        else {	// SRC is IN_CORE
  	  	char* src = (char*) srcBase + elementLen * srcBlock [MBLK_OFF] ;
	  	if (srcBlock [MBLK_CTG]) {
		   buffer =  (void *)src;
		   useSrc = 1;
	  	}
	  	else  {
	  	   int dstStrs[10];
	  	   if (defaultStorageMode == COLUMN_MAJOR || rank < 2)
          	   {
	     	     dstStrs[0]=1;
	     	     for (int i=1; i<rank; i++) dstStrs[i] = dstStrs[i-1]*exts[i-1];
          	   }
	  	   else {
	     	     dstStrs[rank-1]=1;
	     	     for (int i=rank-2; i>=0; i--) dstStrs[i] = dstStrs[i+1]*exts[i+1];
          	   }
		   int vol=1;
                   for (int k=0;k<rank; k++) vol = vol*exts[k];
		   buffer = malloc(vol * elementLen);
		   if(buffer == NULL) AdlibError("oc_copyMblkMblk: Memory exhausted!\n");
	  	   blockCopy((char*)buffer, dstStrs, src, srcStrs, exts, rank, 
			                                          elementLen);
		   if (srcMajor != dstMajor)
			buffer_storage_convert((char *)buffer, rank, exts, elementLen, srcMajor, dstMajor);
          	}
        }

	  // destination
        if (dstStatus == OUT_OF_CORE) {
	  PFILE *dstPfile = (PFILE*)dstBase;
	  oc_blockScatter(dstPfile, dstOffset, dstStrs, exts, 
		   	rank, (char *)buffer, elementLen, icla_size) ;
        }
        else {
	  int srcStrs[10];
	  if (defaultStorageMode == COLUMN_MAJOR || rank < 2)
          {
	     srcStrs[0]=1;
	     for (int i=1; i<rank; i++) srcStrs[i] = srcStrs[i-1]*exts[i-1];
          }
	  else {
	     srcStrs[rank-1]=1;
	     for (int i=rank-2; i>=0; i--) srcStrs[i] = srcStrs[i+1]*exts[i+1];
          }
  	  char* dst = (char*) dstBase + elementLen * dstBlock [MBLK_OFF] ;
	  blockCopy(dst, dstStrs, (char*)buffer, srcStrs,
               exts, rank, elementLen);
        }

	defaultStorageMode = org_storage;
	if (!useSrc) free(buffer);
  }
}

// copy while type conversion
void oc_copyTypeMblkMblk(void* dstBase, MultiBlock dstBlock, int dstType,
                     void* srcBase, MultiBlock srcBlock, int srcType)
{

  // `MultiBlock' to `MultiBlock' copy.  The blocks must be the same shape.
  int srcElementLen = typeLen(srcType);
  int dstElementLen = typeLen(dstType);

  int dstStatus = dstBlock[MBLK_STA];
  int srcStatus = srcBlock[MBLK_STA];

  int dstIclaSize = dstBlock[MBLK_INS];
  int srcIclaSize = srcBlock[MBLK_INS];
  int icla_size   = min(dstIclaSize, srcIclaSize);

  const int rank      = dstBlock [MBLK_RNK] ;
  int dstOffset       = dstBlock[MBLK_OFF];
  int srcOffset       = srcBlock[MBLK_OFF];
  int* dstStrs        = dstBlock + MBLK_STR ;
  int* srcStrs        = srcBlock + MBLK_STR ;
  int* exts           = dstStrs  + rank ;

  int vol = 1;
  for (int i=1; i<rank; i++) vol = vol * exts[i];

  void *buffer = malloc(vol * srcElementLen);
  if(buffer == NULL) AdlibError("Memory exhausted!\n");
  // source
  if (srcStatus == OUT_OF_CORE)  {
	PFILE *srcPfile = (PFILE*)srcBase;
	oc_blockGather(srcPfile, srcOffset, srcStrs, exts, 
	   	rank, (char *)buffer, srcElementLen, icla_size) ;
  }
  else {
  	char* src = (char*) srcBase + srcElementLen * srcBlock [MBLK_OFF] ;
	int dstStrs[10];
	dstStrs[0]=1;
	for (int i=1; i<rank; i++) dstStrs[i] = dstStrs[i-1]*exts[i-1];
	blockCopy((char*)buffer, dstStrs, src, srcStrs, exts, rank, srcElementLen);
  }

  // type conversion
  char *dstbuf = (char *)malloc(vol * dstElementLen);
  TypeConversion(dstbuf, (char *)buffer, vol, dstType, srcType);
  free(buffer);

  // destination
  if (dstStatus == OUT_OF_CORE) {
	PFILE *dstPfile = (PFILE*)dstBase;
	oc_blockScatter(dstPfile, dstOffset, dstStrs, exts, 
		   	rank, (char *)dstbuf, dstElementLen, icla_size) ;
  }
  else {
        int srcStrs[10];
	srcStrs[0]=1;
	for (int i=1; i<rank; i++) srcStrs[i] = srcStrs[i-1]*exts[i-1];
  	char* dst = (char*) dstBase + dstElementLen * dstBlock [MBLK_OFF] ;
	blockCopy(dst, dstStrs, (char*)dstbuf, srcStrs,
               exts, rank, dstElementLen);
  }
  free(dstbuf);
}
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//	oc_blockGather		oc_gatherMblkBuf
//	oc_blockScatter		oc_scatterMblkBuf
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#include <time.h>
#include <unistd.h>

// storage mode converter from storage0 => storage1
void buffer_storage_convert(char *array, int rank, int *number, int elementlen, 
			    int storage0, int storage1)
{
    if (storage0  == storage1) return;
    if (rank == 1) return;
    if (rank == 2 && (number[0] == 1 || number[1] == 1)) return; 


//AdlibPrintf("convert from %d => %d (%d)... \n",storage0,storage1,rank);

    int i,j,k;
    int offset0=0, offset1=0;
    int nums = 1;
    for (i = 0; i<rank; i++) nums *= number[i];
    char *buf = (char *)malloc(nums * elementlen);

    int strs[3], exts[3];
/*
int *data = (int*) array;
AdlibPrintf("Before ");
for (k=0; k<nums; k++) printf(" %d",data[k]);
printf("\n");
*/

    switch (rank)
    {
    case 2:
      {
	if (storage0 == COLUMN_MAJOR) {
		exts[0]=number[1]; exts[1]=number[0];
		strs[0]=number[1]; 
	}
	else {
		exts[0]=number[0]; exts[1]=number[1];
		strs[0]=number[0]; 
	}
        for (i=0; i<exts[0]; i++)
	{
	  int offset10 = offset1;
	  for (j=0; j<exts[1]; j++)
	  {
	      memcpy(buf + offset1, array + offset0, elementlen);
	      offset0 += elementlen;
	      offset1 += strs[0]*elementlen;
	  }
	  offset1 = offset10 + elementlen;
	}
      }
      break;
    case 3:
	{
	for (i=0; i<number[0]; i++)
         for (j=0; j<number[1]; j++)
          for (k=0; k<number[2]; k++)
	   if (storage0 == COLUMN_MAJOR)
	      memcpy(buf + (i*number[2]*number[1] + j*number[2] + k)*elementlen, array + (k*number[0]*number[1] + j*number[0] + i)*elementlen , elementlen);
	   else
	      memcpy(buf + (k*number[0]*number[1] + j*number[0] + i)*elementlen, array + (i*number[2]*number[1] + j*number[2] + k)*elementlen , elementlen);
	}
	break;
    }

/*
data = (int*) buf;
AdlibPrintf("After ");
for (k=0; k<nums; k++) printf(" %d",data[k]);
printf("\n");
*/

    memcpy(array, buf , nums * elementlen);
    free(buf);

}


char* oc_blockGather(PFILE* Pfile, int dstOffset, int* strs, int* exts, 
		   int rank, char* Buffer, int elementLen, int IclaSize) 
{
//   if (rank == 1 && exts[0] == 1) strs[0]=1;
   if (Pfile->share->Status == PASSION_NEW)
        return NULL;

//AdlibPrintf("PFILE:%d,Gather:storage:%d,rank:%d,strs0:%d, strs1:%d, exts0:%d, exts1:%d,ocla_size0:%d,ocla_size1:%d.\n",Pfile->share->Fd,Pfile->share->storage,rank,strs[0],strs[1], exts[0], exts[1], Pfile->ocla_size[0],Pfile->ocla_size[1]);

   ProfBeginRead();

   int storage = Pfile->share->storage;

   switch (rank)
   {
   case 0: 
	passion_read_block_section_1D(Pfile->share->Fd, Buffer, 0, 
                                         dstOffset, 1, 1,  
                                         elementLen, Pfile->HeaderSize, 
					 IclaSize);
	break;
   case 1:
	passion_read_block_section_1D(Pfile->share->Fd, Buffer, 0, 
                                         dstOffset, strs[0], exts[0],  
                                         elementLen, Pfile->HeaderSize, 
	 				 IclaSize);
	break;
   case 2:
//AdlibPrintf("gather %d %d %d %d %d %d %d %d %d\n",strs[0],strs[1],exts[0],exts[1],IclaSize,elementLen,storage,COLUMN_MAJOR,ROW_MAJOR);
	if (storage == COLUMN_MAJOR)
           passion_read_block_section_2D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[0], strs[1], exts[0], exts[1],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[0], Pfile->ocla_size[0]);
	else
           passion_read_block_section_2D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[1], strs[0], exts[1], exts[0],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[1], Pfile->ocla_size[1]);
	break;
   case 3:
//AdlibPrintf("gather %d %d %d %d %d %d %d %d %d\n",strs[0],strs[1],strs[2],exts[0],exts[1],exts[2],IclaSize,storage,elementLen);
	if (storage == COLUMN_MAJOR)
           passion_read_block_section_3D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[0], strs[1], strs[2],
					exts[0], exts[1], exts[2],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[0], exts[1],
					Pfile->ocla_size[0],Pfile->ocla_size[1]);
	else
           passion_read_block_section_3D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[2], strs[1], strs[0],
					exts[2], exts[1], exts[0],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[2], exts[1],
					Pfile->ocla_size[2],Pfile->ocla_size[1]);
	break;
   }

   buffer_storage_convert(Buffer, rank, exts, elementLen, storage, defaultStorageMode);

/*
int *data = (int *)Buffer;
if (rank == 2)
{
printf("oc_blockGather:");
for (int k=0; k<exts[0]*exts[1]; k++) printf("%d ",data[k]);
printf("\n");
}
*/

   ProfEndRead();

   return Buffer;
}

//wjp add 1998-3-8
void oc_gatherMblkBuf(PFILE* Pfile, MultiBlock block, void* buffer,
                   const int elementLen) {


    char* Buffer = (char *)buffer;
    int StartIndex = block [MBLK_OFF];

    const int rank      = block [MBLK_RNK] ;
    int* strs           = block + MBLK_STR ;
    int* exts           = strs  + rank ;
    int IclaSize        = block[MBLK_INS];


//AdlibPrintf("StartIndex:%d, rank:%d, IclaSize:%d.\n", StartIndex, rank, IclaSize);

    oc_blockGather(Pfile, StartIndex, strs, exts, rank, Buffer, elementLen, IclaSize); 

}

char* oc_blockScatter(PFILE* Pfile, int dstOffset, int* strs, int* exts, 
		   int rank, char* Buffer, int elementLen, int IclaSize) 
{

   int oldfile;
   if (Pfile->share->Status == PASSION_NEW) oldfile=0;
   else oldfile=1;

/*
int *data = (int *)Buffer;
if (rank == 2)
{
printf("oc_blockScatter:");
for (int k=0; k<exts[0]*exts[1]; k++) printf("%d ",data[k]);
printf("\n");
}
*/
//AdlibPrintf("oc_blockScatter: %d,%d,%d,%d\n",data[0],data[1],data[2],data[3]);
//AdlibPrintf("Scatter:storage:%d,rank:%d,strs0:%d, strs1:%d, exts0:%d, exts1:%d, ocla_size0:%d,ocla_size1:%d.\n",Pfile->share->storage,rank,strs[0],strs[1], exts[0], exts[1], Pfile->ocla_size[0],Pfile->ocla_size[1]);

   ProfBeginWrite();

   int storage = Pfile->share->storage;

   buffer_storage_convert(Buffer, rank, exts, elementLen, defaultStorageMode, storage);


/*
int* xtt=(int*)Buffer;
int tmp,i;
for( i=0;i<25;i++)
{tmp=*xtt;
AdlibPrintf("WWW %d %d \n",tmp,i);
xtt++;
}
*/

   switch (rank)
   {
   case 0:
	passion_write_block_section_1D(Pfile->share->Fd, Buffer, 0, 
                                         dstOffset, 1, 1,  
                                         elementLen, Pfile->HeaderSize, 
					 IclaSize, oldfile);
	break;
   case 1:
	passion_write_block_section_1D(Pfile->share->Fd, Buffer, 0, 
                                         dstOffset, strs[0], exts[0],  
                                         elementLen, Pfile->HeaderSize, 
					 IclaSize, oldfile);
	break;
   case 2:
//AdlibPrintf("write rank %d %d %d %d %d,%d,%d \n",strs[0],strs[1],exts[0],exts[1],IclaSize,elementLen,storage);

	if (storage == COLUMN_MAJOR)
           passion_write_block_section_2D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[0], strs[1], exts[0], exts[1],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[0], Pfile->ocla_size[0], 
					oldfile);
 	else
           passion_write_block_section_2D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[1], strs[0], exts[1], exts[0],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[1], Pfile->ocla_size[1], 
					oldfile);
	break;
   case 3:
//AdlibPrintf("write rank %d %d %d %d %d,%d,%d %d\n",strs[0],strs[1],strs[2],exts[0],exts[1],exts[2],storage,COLUMN_MAJOR);
	if (storage == COLUMN_MAJOR)
           passion_write_block_section_3D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[0], strs[1], strs[2],
					exts[0], exts[1], exts[2],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[0], exts[1],
					Pfile->ocla_size[0],Pfile->ocla_size[1],
					oldfile);
	else
           passion_write_block_section_3D(Pfile->share->Fd, Buffer, 0, dstOffset,
					strs[2], strs[1], strs[0],
					exts[2], exts[1], exts[0],
					elementLen, Pfile->HeaderSize, 
					IclaSize, exts[2], exts[1],
					Pfile->ocla_size[2],Pfile->ocla_size[1],
					oldfile);
	break;
   }
   Pfile->share->Status = PASSION_OLD;

   ProfEndWrite();

   return Buffer;
}

//wjp add 1998-3-8
void oc_scatterMblkBuf(PFILE* Pfile, MultiBlock block, void* buffer,
                    const int elementLen) {

char* Buffer = (char *)buffer;
int StartIndex = block [MBLK_OFF];

    const int rank      = block [MBLK_RNK] ;
    int* strs           = block + MBLK_STR ;
    int* exts           = strs  + rank ;
    int IclaSize        = block[MBLK_INS];


    oc_blockScatter(Pfile, StartIndex, strs, exts, rank, Buffer, elementLen, IclaSize); 

//AdlibPrintf("oc_scatterMblkBuf:StartIndex:%d, rank:%d, IclaSize:%d.\n", StartIndex, rank, IclaSize);
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//	oc_blockZero	oc_blockAssign 	  oc_blockCopy		
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void oc_blockZero(PFILE *Pfile, int Offset, int* strs, int* exts, int rank, int IclaSize) 
{
   int i;

   int vol = 1;
   for (i=0; i<rank; i++) vol = vol * exts[i];
   int elementLen = sizeof(int);
   
   SparseArray<MessQueue> Messages(4);
   MessQueue* queue = Messages.getElement(1) ;
   addMessageBlock(queue->messages, queue->blocks, Offset, strs, 0, 
		   elementLen * vol, exts, rank, OUT_OF_CORE, IclaSize, Pfile->share->storage);

   int* buffer  =  new int [IclaSize] ; 
   if (buffer == NULL)
	AdlibError("oc_blockZero: Out of memory");
   for (i=0; i< IclaSize; i++) buffer[i]=0;

   List<MessQueue>* record = Messages.list ;
   if (record) {
         const int numMes = record->messages.size() ;
         Mess* message = record->messages.data() ;
         MultiBlock block = record->blocks.data() ;

         for (int mes=0; mes < numMes; mes++) {
	     oc_scatterMblkBuf(Pfile, block, (void*)buffer, elementLen);

             nextBlock(block) ;
             message++;
         }
  }  
  delete [] buffer ;
}

template<class T>
void oc_blockAssign(PFILE *Pfile, int Offset, int* strs, int* exts, int rank, T val, int IclaSize) 
{
   int i;

   int vol = 1;
   for (i=0; i<rank; i++) vol = vol * exts[i];
   int elementLen = sizeof(T);
   
   SparseArray<MessQueue> Messages(4);
   MessQueue* queue = Messages.getElement(1) ;
   addMessageBlock(queue->messages, queue->blocks,
           Offset, strs, 0, elementLen * vol, exts, rank, OUT_OF_CORE, IclaSize);

   T* buffer  =  new T [IclaSize] ; 
   if (buffer == NULL)
	AdlibError("oc_blockAssign: Out of memory");
   for (i=0; i< IclaSize; i++) buffer[i]=val;

   List<MessQueue>* record = Messages.list ;
   if (record) {

         const int numMes = record->messages.size() ;
         Mess* message = record->messages.data() ;
         MultiBlock block = record->blocks.data() ;

         int mes;
         for (mes=0; mes < numMes; mes++) {

	     oc_scatterMblkBuf(Pfile, block, (void*)buffer, elementLen);

             nextBlock(block) ;
             message++;
         }

  }  
  delete [] buffer ;

}

void oc_blockCopy(void* dst, int dstOffset, int* dstStrs, Status dstStatus,
		         void* src, int srcOffset, int* srcStrs, Status srcStatus,
                         int* exts, int rank, int elementLen, int IclaSize) 
{
   int vol=1;
   for (int i=0; i<rank; i++) vol = vol*exts[i];
   SparseArray<MessQueue> dstMessages(4);
   MessQueue* queue = dstMessages.getElement(1) ;
   addMessageBlock(queue->messages, queue->blocks,
           dstOffset, dstStrs, 0, elementLen * vol, exts, rank, dstStatus, IclaSize);
   SparseArray<MessQueue> srcMessages(4);
   queue = srcMessages.getElement(1) ;
   addMessageBlock(queue->messages, queue->blocks,
           srcOffset, srcStrs, 0, elementLen * vol, exts, rank, srcStatus, IclaSize);

   List<MessQueue>* dstRecord = dstMessages.list ;
   List<MessQueue>* srcRecord = srcMessages.list ;
   if (dstRecord) {

         const int numMes = dstRecord->messages.size() ;
         Mess* message = dstRecord->messages.data() ;
         MultiBlock dstblock = dstRecord->blocks.data() ;
         MultiBlock srcblock = srcRecord->blocks.data() ;

         int mes;
         for (mes=0; mes < numMes; mes++) {

	     oc_copyMblkMblk(dst, dstblock, src, srcblock, elementLen) ;
             nextBlock(srcblock) ;
             nextBlock(dstblock) ;
             message++;
         }

   }  
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T>
inline void type_coerce(S* dst, T* src, int num)
{
	for (int i=0; i<num; i++)   dst[i] = (S)src[i];
}

void TypeConversion(char *dst, char *src, int num, int dstType, int srcType)
{
   switch (dstType) 
   {
   case ADLIB_INTEGER: {
	switch (srcType) {
 	case ADLIB_INTEGER:
		type_coerce( (int *)dst, (int *)src, num); break;
 	case ADLIB_REAL:
		type_coerce( (int *)dst, (float *)src, num); break;
 	case ADLIB_DOUBLE:
		type_coerce( (int *)dst, (double *)src, num); break;
 	case ADLIB_COMPLEX:
		type_coerce( (int *)dst, (COMPLEX *)src, num); break;
 	case ADLIB_BYTE:
		type_coerce( (int *)dst, (char *)src, num); break;
 	case ADLIB_WORD:
		type_coerce( (int *)dst, (short *)src, num); break;
	default:
		AdlibError("TypeConversion Error.");
        }
        break;
     }
   case ADLIB_REAL: {
	switch (srcType) {
 	case ADLIB_INTEGER:
		type_coerce( (float *)dst, (int *)src, num); break;
 	case ADLIB_REAL:
		type_coerce( (float *)dst, (float *)src, num); break;
 	case ADLIB_DOUBLE:
		type_coerce( (float *)dst, (double *)src, num); break;
 	case ADLIB_COMPLEX:
		type_coerce( (float *)dst, (COMPLEX *)src, num); break;
 	case ADLIB_BYTE:
		type_coerce( (float *)dst, (char *)src, num); break;
	default:
		AdlibError("TypeConversion Error.");
        }
        break;
     }
   case ADLIB_DOUBLE: {
	switch (srcType) {
 	case ADLIB_INTEGER:
		type_coerce( (double *)dst, (int *)src, num); break;
 	case ADLIB_REAL:
		type_coerce( (double *)dst, (float *)src, num); break;
 	case ADLIB_DOUBLE:
		type_coerce( (double *)dst, (double *)src, num); break;
 	case ADLIB_COMPLEX:
		type_coerce( (double *)dst, (COMPLEX *)src, num); break;
 	case ADLIB_BYTE:
		type_coerce( (double *)dst, (char *)src, num); break;
	default:
		AdlibError("TypeConversion Error.");
        }
        break;
     }
   case ADLIB_COMPLEX: {
	switch (srcType) {
 	case ADLIB_INTEGER:
		type_coerce( (COMPLEX *)dst, (int *)src, num); break;
 	case ADLIB_REAL:
		type_coerce( (COMPLEX *)dst, (float *)src, num); break;
 	case ADLIB_DOUBLE:
		type_coerce( (COMPLEX *)dst, (double *)src, num); break;
 	case ADLIB_COMPLEX:
		type_coerce( (COMPLEX *)dst, (COMPLEX *)src, num); break;
 	case ADLIB_BYTE:
		type_coerce( (COMPLEX *)dst, (char *)src, num); break;
	default:
		AdlibError("TypeConversion Error.");
        }
        break;
     }
   case ADLIB_BYTE: {
	switch (srcType) {
 	case ADLIB_INTEGER:
		type_coerce( (char *)dst, (int *)src, num); break;
 	case ADLIB_REAL:
		type_coerce( (char *)dst, (float *)src, num); break;
 	case ADLIB_DOUBLE:
		type_coerce( (char *)dst, (double *)src, num); break;
 	case ADLIB_COMPLEX:
		type_coerce( (char *)dst, (COMPLEX *)src, num); break;
 	case ADLIB_BYTE:
		return;
	default:
		AdlibError("TypeConversion Error.");
        }
        break;
     }
   case ADLIB_LOGICAL:
	type_coerce( (int *)dst, (int *)src, num); break;
   case ADLIB_CHARACTER:
	type_coerce( (char *)dst, (char *)src, num); break;
   default:
	AdlibError("TypeConversion Error.");
   }
}

