
#include <stdio.h>

#include <stdlib.h>

#include <unistd.h>

#include <sys/times.h>

#include <unistd.h>
#include <limits.h>

#include "MPIinclude.h"

#include "globals.h"

#define MAXNUM 5 


ProfileInfo  *Prof;
unsigned int  read_ops, write_ops;      // used in io
double          bytes_read, bytes_write;  
unsigned int  seiving, nonseiving;      // write (seiving)

ProfileInfo::ProfileInfo()
{
//  CLKTCK = sysconf(_SC_CLK_TCK);
  gethostname (name, 25);
  n = 0;
  send = recv = 0.0;
  io_r = io_w = 0.0;
  aio_r = 0.0;
  read_ops = write_ops = 0;
  bytes_read = bytes_write = 0;
  seiving = nonseiving = 0;
  remap_time = 0;
  remap_count = 0;
}

void ProfileInfo::StartProfile(int mode_)
{
  send = recv = 0.0;
  io_r = 0.0;
  io_w = 0.0;
  aio_r = 0.0;
  n = 0;
  read_ops = write_ops = 0;
  bytes_read = bytes_write = 0;
  seiving = nonseiving = 0;
  remap_time = 0;
  remap_count = 0;
  mode = mode_;
}

void ProfileInfo::EndProfile()
{
  display();
}

void ProfileInfo::setMode(int mode_)
{
  mode = mode_;
}

int ProfileInfo::getMode()
{
  return mode;
}

void ProfileInfo::pushTime()
{
  time[n] = MPI_Wtime();
  n++;
  if (n>MAX_TIMER) AdlibError("Out of Timer.");
}

double ProfileInfo::popTime()
{
  n--;
  if (n<0) AdlibError("Timer Abnormal.");
  double elapse = MPI_Wtime() - time[n];
  return elapse;
}

void ProfileInfo::BeginSend()
{
  pushTime();
}

void ProfileInfo::EndSend()
{
  send = send + popTime();
}

void ProfileInfo::BeginRecv()
{
  pushTime();
}

void ProfileInfo::EndRecv()
{
  recv = recv + popTime();
}

void ProfileInfo::BeginRead()
{
  pushTime();
}

void ProfileInfo::EndRead()
{
  io_r += popTime();
}

void ProfileInfo::BeginARead()
{
  pushTime();
}

void ProfileInfo::EndARead()
{
  aio_r += popTime();
}

void ProfileInfo::BeginWrite()
{
  pushTime();
}

void ProfileInfo::EndWrite()
{
  io_w += popTime();
}

void ProfileInfo::BeginRemap()
{
  pushTime();
  remap_count ++;
}

void ProfileInfo::EndRemap()
{
  remap_time += popTime();
}

void ProfileInfo::display()
{
  int procId = AdlibLocId();
  fprintf(stdout," ========================= %s(P%d) == Begin ==========================\n",name, procId);
  fprintf(stdout,"(%s)(P%d) Communication time(S/R): %f / %f(%f).\n", name, procId, send, recv, send+recv);
  fprintf(stdout,"(%s)(P%d) I/O time: %f / %f AIO: %f (%f).\n", name, procId, io_r,io_w,aio_r,io_r+io_w+aio_r);
  fprintf(stdout,"(%s)(P%d) I/O count: %d / %d.\n", name, procId, read_ops,write_ops);
  fprintf(stdout,"(%s)(P%d) I/O(bytes): %fK / %fK(%fK).\n", name, procId, bytes_read/1000.0,bytes_write/1000.0,(bytes_read+bytes_write)/1000.0);
  fprintf(stdout,"(%s)(P%d) I/O seiveing: %d / %d .\n", name, procId, seiving, nonseiving);
  if (mode >= 1) {
  fprintf(stdout,"(%s)(P%d) remap time: %f remap count: %d.\n", name, procId, remap_time, remap_count);
  }
  fprintf(stdout," ========================= %s(P%d) == End   ==========================\n",name, procId);
}

long ProfileInfo::timer(int num)
{

  static struct tms actTimeBuf[MAXNUM+1];
  static clock_t wallTimeBuf[MAXNUM+1];

  static int timeNum=0;

  static int flag;

  if ((num) == 0 ) {
    timeNum=0;
    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);  
  } 
  else {
    timeNum++;

    if (timeNum > MAXNUM ) {
      printf("Maximum time is %d\n", (timeNum-1));
      exit(1);
    }

    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);

    if (timeNum < (num)) {
      return wallTimeBuf[timeNum]-wallTimeBuf[0];
    }
    else {
      return wallTimeBuf[timeNum]-wallTimeBuf[timeNum-(num)];
    }
  }

}


void AdlibTimer(int num)
{

  static long CLKTCK = sysconf(_SC_CLK_TCK);

  static struct tms actTimeBuf[MAXNUM+1];
  static clock_t wallTimeBuf[MAXNUM+1];

  static int timeNum=0;

  static char name[25];

  static int procId, flag;

//  MPI_Comm_rank(MPI_COMM_WORLD, &procId) ;
  procId = AdlibLocId();

  // printf("....%p %d\n", num, (num));

  if ((num) == 0 ) {
    gethostname (name, 25);
    timeNum=0;
    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);  
  } 
  else {
    timeNum++;

    if (timeNum > MAXNUM ) {
      printf("Maximum time is %d\n", (timeNum-1));
      exit(1);
    }

    wallTimeBuf[timeNum]=times(& actTimeBuf[timeNum]);

    //printf (" %d - %d \n",timeNum, timeNum-(*num) );
    
    //printf ( " %ld ----- %ld \n", wallTimeBuf[timeNum], wallTimeBuf[timeNum-(*num)] );
    
    //printf ( " %ld -- \n", CLKTCK);

    //printf ( " %ld -- %ld \n", actTimeBuf[timeNum].tms_utime, actTimeBuf[timeNum].tms_stime );

    if (timeNum < (num)) {
      printf ("%s(P%d) == Rtime: %f, Utime: %f, CUtime: %f, Stime: %f, CStime: %f\n",
	      name, procId,
	      (float)(wallTimeBuf[timeNum]-wallTimeBuf[0])/CLKTCK,
	      (float)(actTimeBuf[timeNum].tms_utime-actTimeBuf[0].tms_utime)/CLKTCK,
	      (float)(actTimeBuf[timeNum].tms_cutime-actTimeBuf[0].tms_cutime)/CLKTCK,
	      (float)(actTimeBuf[timeNum].tms_stime-actTimeBuf[0].tms_stime)/CLKTCK,
	      (float)(actTimeBuf[timeNum].tms_cstime-actTimeBuf[0].tms_cstime)/CLKTCK);
    }
    else {
      printf ("%s(P%d) == Rtime: %f, Utime: %f, CUtime: %f, Stime: %f, CStime: %f\n",
              name, procId,
              (float)(wallTimeBuf[timeNum]-wallTimeBuf[timeNum-(num)])/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_utime-actTimeBuf[timeNum-(num)].tms_utime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cutime-actTimeBuf[timeNum-(num)].tms_cutime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_stime-actTimeBuf[timeNum-(num)].tms_stime)/CLKTCK,
              (float)(actTimeBuf[timeNum].tms_cstime-actTimeBuf[timeNum-(num)].tms_cstime)/CLKTCK);
    }
  }

}

