
#define PCRC_INTEGER    1
#define PCRC_REAL       2
#define PCRC_DOUBLE     3
#define PCRC_CHARACTER  4
#define PCRC_COMPLEX    5
#define PCRC_LOGICAL    6
#define PCRC_BYTE       7
#define PCRC_WORD       8

/* external C++ functions */
const void *ad_ToPointer(int idx);
int ad_FromPointer(const void* ptr);
void setCollapse(Range& rng, const int _g_lb, const int _g_ub);

/* external C functions */
extern "C" {
   void pcrc_oc_init(int *dad, int *icla_size);
   int pcrc_same_dad_check(int *dad1, int *dad2);
}

// assign numeric type res to pcrcfDAD's i element
template<class T>
inline void numeric_assign(pcrcfDAD *resPtr , int i , T res)
{
 	switch ( resPtr->elementType)
  	{
	case PCRC_BYTE: {
		((char *)resPtr->data)[i] = (char)res; 
		break;
		}
	case PCRC_WORD: {
		((short *)resPtr->data)[i] = (short)res; 
		break;
		}
	case PCRC_LOGICAL:
  	case PCRC_INTEGER: {
		((int *)resPtr->data)[i] = (int)res; 
		break;
		}
  	case PCRC_REAL: {
		((float *)resPtr->data)[i] = (float)res; 
		break;
		}
  	case PCRC_DOUBLE: {
		((double *)resPtr->data)[i] = (double)res; 
		break;
		}
  	case PCRC_COMPLEX: {
		((COMPLEX *)resPtr->data)[i] = (COMPLEX)res; 
		break;
		}
	default:  {
		fprintf(stderr, "** RUNTIME ERROR **  Error in type of numeric_assign.\n");
	 	}
  	}
}

// assign scalar numeric type res to all the elements of pcrcfDAD
template<class T>
inline void scalar_assign(pcrcfDAD *resPtr , T res)
{
  switch (resPtr->rank)
  {
  case 0: {
	if (resPtr->group.member()) {
	numeric_assign(resPtr , 0 , res); 
	} 
	break;
    }
  case 1:  {
	LocBlocksIndex i(resPtr->ranges[0]);
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
		numeric_assign(resPtr , off0_bas+off0_stp*l , res); 
	  }
	}
	break;
    }
  case 2: {
	LocBlocksIndex i(resPtr->ranges[0]),j(resPtr->ranges[1]) ;
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
	     for (j.beginLocBlk(); j.test(); j.nextLocBlk() ) {
	  	int off1_bas = resPtr->str(1).offset(j);
	  	int off1_stp = resPtr->str(1).step(j);
	  	for (int m=0; m<j.count; m++) {
			numeric_assign(resPtr , off0_bas+off0_stp*l + off1_bas+off1_stp*m , res); 
		}
	     }
	  }
	}
	break;
    }
  case 3: {
	LocBlocksIndex i(resPtr->ranges[0]),j(resPtr->ranges[1]),k(resPtr->ranges[2]) ;
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
	     for (j.beginLocBlk(); j.test(); j.nextLocBlk() ) {
	  	int off1_bas = resPtr->str(1).offset(j);
	  	int off1_stp = resPtr->str(1).step(j);
	  	for (int m=0; m<j.count; m++) {
		    for (k.beginLocBlk(); k.test(); k.nextLocBlk() ) {
	  		int off2_bas = resPtr->str(2).offset(k);
	  		int off2_stp = resPtr->str(2).step(k);
			for (int kk=0; kk<k.count; kk++) {
			    numeric_assign(resPtr , off0_bas+off0_stp*l + off1_bas+off1_stp*m + off2_bas+off2_stp*kk , res); 
			}
		   }
		}
	     }
	  }
	}
	break;
    }
  default: {
	printf("scalar_assign: cannot handle !\n");
    }
  }
}


/*
template<class T>
class ScalarCopy {
public :
  ScalarCopy(const DAD* _dst, T _val, int len=sizeof(T)) :
           dst(_dst), val(_val), rank(src->rnk()), elementLen(len) {

    exts     = new int [rank] ;
    dstSteps = new int [rank] ;
  }

  ~ScalarCopy() {
    delete [] dstSteps ;
    delete [] exts ;
  }

    // added by Zheng Gengbin
  void oc_execute(void* dstBase, T _val) {	// for out of core
    if(src->grp().member())
      oc_loop(dstBase, 0, _val, 0) ;
  }

private :
  const DAD* dst ;
  const T    val;
  int rank ;
  int *exts, *dstSteps;
  int elementLen;

    // Zheng Gengbin
  void oc_loop(void* dstBase, int dstOffset, T _val, const int r) {
    if(r == rank)
     oc_blockAssign(dstBase, dstOffset, dstSteps, dst->status, 
		  exts, rank, elementLen, _val, dst->icla_size);
    else {
      Stride dstStr = dst->str(r) ;

      LocBlocksIndex i(dst->rng(r)) ;
      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	exts     [r] = i.count ;
	dstSteps [r] = dstStr.step(i) ;

	oc_loop(dstBase, dstOffset + dstStr.offset(i), 
		_val, r + 1) ;
      }
    }
  }
} ;
*/
