#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"

extern "C" {
   void pcrc_matmul(void *res0, int *dad_res, void *src1, int *dad_src1, void *src2, int *dad_src2);
}

//*************************************************************************//

template<class U>
inline void  temp_matmul( int *dad_res,U type1,int *dad_src1, int *dad_src2)
{
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;
  pcrcfDAD* src1Ptr    = (pcrcfDAD*) ad_ToPointer(*dad_src1) ;
  pcrcfDAD* src2Ptr    = (pcrcfDAD*) ad_ToPointer(*dad_src2) ;

  Section2<U>  dstsect(resPtr->ranges[0],resPtr->ranges[1],resPtr->group,resPtr->strides[0],resPtr->strides[1],(U *)resPtr->data);	
  dstsect.status = resPtr->status;
  dstsect.icla_size = resPtr->icla_size;
  dstsect.major = resPtr->major;
  switch ( src1Ptr->elementType)
  {
  case PCRC_LOGICAL:
	{
  	Section2<int>  src1sect(src1Ptr->ranges[0],src1Ptr->ranges[1],src1Ptr->group,src1Ptr->strides[0],src1Ptr->strides[1],(int *)src1Ptr->data);	
  	Section2<int>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(int *)src2Ptr->data);	
	src1sect.status = src1Ptr->status;
	src1sect.icla_size = src1Ptr->icla_size;
	src1sect.major = src1Ptr->major;
	src2sect.status = src2Ptr->status;
	src2sect.icla_size = src2Ptr->icla_size;
	src2sect.major = src2Ptr->major;
  	matmul(dstsect,src1sect,src2sect);
	break;
	}
  case PCRC_INTEGER: {
 	Section2<int>  src1sect(src1Ptr->ranges[0],src1Ptr->ranges[1],src1Ptr->group,src1Ptr->strides[0],src1Ptr->strides[1],(int *)src1Ptr->data);	
	src1sect.status = src1Ptr->status;
	src1sect.icla_size = src1Ptr->icla_size;
	src1sect.major = src1Ptr->major;
	switch (src2Ptr->elementType)
        {
	case PCRC_INTEGER: {
  		Section2<int>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(int *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_REAL:  {
  		Section2<float>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(float *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_DOUBLE: {
  		Section2<double>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(double *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break;  }
	case PCRC_COMPLEX: {
  		Section2<COMPLEX>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(COMPLEX *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	default:
		fprintf(stderr, "** RUNTIME ERROR ** error in MATMUL.\n");
     	}
	break;
      }
  case PCRC_REAL: {
  	Section2<float>  src1sect(src1Ptr->ranges[0],src1Ptr->ranges[1],src1Ptr->group,src1Ptr->strides[0],src1Ptr->strides[1],(float *)src1Ptr->data);	
	src1sect.status = src1Ptr->status;
	src1sect.icla_size = src1Ptr->icla_size;
	src1sect.major = src1Ptr->major;
	switch (src2Ptr->elementType)
        {
	case PCRC_INTEGER: {
  		Section2<int>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(int *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_REAL: {
  		Section2<float>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(float *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_DOUBLE: {
  		Section2<double>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(double *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_COMPLEX: {
  		Section2<COMPLEX>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(COMPLEX *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	default:
		fprintf(stderr, "** RUNTIME ERROR ** error in MATMUL.\n");
     	}
	break;
      }
  case PCRC_DOUBLE: {
 	Section2<double>  src1sect(src1Ptr->ranges[0],src1Ptr->ranges[1],src1Ptr->group,src1Ptr->strides[0],src1Ptr->strides[1],(double *)src1Ptr->data);
	src1sect.status = src1Ptr->status;
	src1sect.icla_size = src1Ptr->icla_size;
	src1sect.major = src1Ptr->major;
	switch (src2Ptr->elementType)
        {
	case PCRC_INTEGER: {
  		Section2<int>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(int *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_REAL: {
  		Section2<float>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(float *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_DOUBLE: {
  		Section2<double>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(double *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_COMPLEX: {
  		Section2<COMPLEX>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(COMPLEX *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	default:
		fprintf(stderr, "** RUNTIME ERROR ** error in MATMUL.\n");
     	}
	break;
      }
  case PCRC_COMPLEX: {
  	Section2<COMPLEX>  src1sect(src1Ptr->ranges[0],src1Ptr->ranges[1],src1Ptr->group,src1Ptr->strides[0],src1Ptr->strides[1],(COMPLEX *)src1Ptr->data);
	src1sect.status = src1Ptr->status;
	src1sect.icla_size = src1Ptr->icla_size;
	src1sect.major = src1Ptr->major;
	switch (src2Ptr->elementType)
        {
	case PCRC_INTEGER: {
  		Section2<int>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(int *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_REAL: {
  		Section2<float>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(float *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_DOUBLE: {
  		Section2<double>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(double *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	case PCRC_COMPLEX: {
  		Section2<COMPLEX>  src2sect(src2Ptr->ranges[0],src2Ptr->ranges[1],src2Ptr->group,src2Ptr->strides[0],src2Ptr->strides[1],(COMPLEX *)src2Ptr->data);	
		src2sect.status = src2Ptr->status;
		src2sect.icla_size = src2Ptr->icla_size;
		src2sect.major = src2Ptr->major;
  		matmul(dstsect,src1sect,src2sect);
		break; }
	default:
		fprintf(stderr, "** RUNTIME ERROR ** error in MATMUL.\n");
     	}
	break;
      }
  default: {
	fprintf(stderr, "** RUNTIME ERROR ** error in MATMUL.\n");
      }
  }
}

void pcrc_matmul(void *res0, int *dad_res, void *src1, int *dad_src1, void *src2, int *dad_src2)
{
  pcrcfDAD* src1Ptr    = (pcrcfDAD*) ad_ToPointer(*dad_src1) ;
  pcrcfDAD* src2Ptr    = (pcrcfDAD*) ad_ToPointer(*dad_src2) ;
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;

  // compare the result and the input , make sure they are not equal
  // not fully compared !!
  pcrcfDAD *tmpPtr;
  int tmpDad=*dad_res;
  void *newdata;

  int same_array = 0;
  if (src1Ptr->memdata == resPtr->memdata || src2Ptr->memdata == resPtr->memdata)
	same_array = 1;
  else if (resPtr->status == OUT_OF_CORE) {
        if (src1Ptr->status == OUT_OF_CORE && src1Ptr->pfile->share == resPtr->pfile->share || 
            src2Ptr->status == OUT_OF_CORE && src2Ptr->pfile->share == resPtr->pfile->share)
	     same_array = 1;
  }

  if ( same_array )
  {
	  // create a tmp dad to hold result.
	int in_core_size = (resPtr->status == OUT_OF_CORE)?  resPtr->icla_size: resPtr->size;
	newdata = malloc(in_core_size * resPtr->elementLen);
	tmpPtr = new pcrcfDAD(newdata, *resPtr);
	tmpDad = ad_FromPointer(tmpPtr);
	if (resPtr->status == OUT_OF_CORE)
	   pcrc_oc_init(&tmpDad, &in_core_size);
  }

  switch ( resPtr->elementType)
  {
  case PCRC_LOGICAL:
  case PCRC_INTEGER: {
  	temp_matmul(&tmpDad, (int)0, dad_src1, dad_src2); 
	break;
	}
  case PCRC_REAL: {
  	temp_matmul(&tmpDad, (float)0, dad_src1, dad_src2);
	break;
	}
  case PCRC_DOUBLE: {
  	temp_matmul(&tmpDad, (double)0, dad_src1, dad_src2);
	break;
	}
  case PCRC_COMPLEX: {
  	temp_matmul(&tmpDad, (COMPLEX)0, dad_src1, dad_src2);
	break;
	}
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in MATMUL. \n");
  }

  // free up
  if (tmpDad != *dad_res) {
        Copy<int> schedule(resPtr, tmpPtr) ;
        schedule.oc_execute(resPtr->data, tmpPtr->data) ;
	free(newdata);
	delete tmpPtr;
  }
}

