
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"

extern "C" {
   void pcrc_maxloc(void *res0, int *dad_res,void *src0, int *dad_src);
   void pcrc_msk_maxloc(void *res0, int *dad_res,void *src0, int *dad_src,void *msk0, int *dad_msk);
   void pcrc_dim_maxloc(void *res0, int *dad_res,void *src0, int *dad_src, int *dim);
   void pcrc_dim_msk_maxloc(void *res0, int *dad_res,void *src0,int *dad_src, int *dim,void *msk0, int *dad_msk);
}

//*************************************************************************//

template<class T>
inline void temp_maxloc( void *res0,int *dad_res,void *src0,int *dad_src,T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);

  reduceMode = 1;
  int *loc;
  int *tmploc;
  T val;

  int rank=src->rank;
  loc = (int *)malloc(rank * sizeof(int) );
  tmploc = (int *)malloc(rank * sizeof(int) );

  Maxloc<T> schedule1(src) ;
  schedule1.execute(&val, loc, (T *)src->data) ;

  reduceMode = 0;
/*
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXLOC \n");
	break;
  case 1:
      {   
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
        maxloc(&val, loc, srcsect);
        break;
       }  
  case 2:
      {   
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
        maxloc(&val,loc ,srcsect);
        break;
       }  
  case 3:
      {   
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
        maxloc(&val,loc , srcsect);
        break;
       }  
  }
*/
  for (int i=0; i<rank ; i++)
	loc[i]++;
//  res->data=loc;
  // root group
  pcrcfDAD *tmp = new pcrcfDAD(loc , 1, sizeof(int),1 , *(Group *)ad_ToPointer(2) );
  Range x ;
  setCollapse(x, 1, rank) ;
  tmp->addRange(0, x) ;

  Remap schedule(res, tmp, tmp->elementLen) ;
  schedule.execute(res->data, tmp->data) ;

  delete tmp;
}

void pcrc_maxloc( void *res0,int *dad_res, void *src0,int *dad_src)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_maxloc( res0, dad_res,src0,dad_src, (int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_maxloc( res0, dad_res,src0,dad_src, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_maxloc( res0, dad_res,src0,dad_src, (short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_maxloc( res0, dad_res,src0,dad_src, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_maxloc( res0, dad_res,src0,dad_src, (double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in MAXLOC \n");
  }

}

//*************************************************************************//

template<class T>
inline void temp_msk_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, void *msk0,int *dad_msk,T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk   = (pcrcfDAD*) ad_ToPointer(*dad_msk); 

  int *loc;
  T val;

  int rank=src->rank;
  loc = (int *)malloc(rank * sizeof(int) );

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;
    MaxlocMsk<T> schedule2(src, tmpptr) ;
    schedule2.execute(&val, loc, (T *)src->data, (int *)tmpptr->data) ;

    delete tmpptr ;
    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    MaxlocMsk<T> schedule2(src, msk) ;
    schedule2.execute(&val, loc, (T *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }
/*
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXLOC \n");
	break;
  case 1:
      {   
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
        maxloc(&val, loc, srcsect,msksect);
        break;
       }  
  case 2:
      {   
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
        maxloc(&val,loc ,srcsect, msksect);
        break;
       }  
  case 3:
      {   
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
        maxloc(&val,loc , srcsect, msksect);
        break;
       }  
  }
*/
  for (int i=0; i<rank ; i++)
	loc[i]++;
//  res->data=loc;
  // root group
  pcrcfDAD *tmp = new pcrcfDAD(loc , 1, sizeof(int),1 , *(Group *)ad_ToPointer(2) );
  Range x ;
  setCollapse(x, 1, rank) ;
  tmp->addRange(0, x) ;

  Remap schedule(res, tmp, res->elementLen) ;
  schedule.execute(res->data, tmp->data) ;
}

void pcrc_msk_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_msk_maxloc( res0, dad_res,src0,dad_src,msk0,dad_msk, (int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_msk_maxloc( res0, dad_res,src0,dad_src,msk0,dad_msk, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_msk_maxloc( res0, dad_res,src0,dad_src,msk0,dad_msk, (short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_msk_maxloc( res0, dad_res,src0,dad_src,msk0,dad_msk, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_msk_maxloc( res0, dad_res,src0,dad_src,msk0,dad_msk, (double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in msk MAXLOC \n");
  }
}

//*************************************************************************//

template<class T>
inline void temp_dim_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, int *dim, T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0=*dim-1;
  int rank=src->rank;

//xh

  pcrcfDAD* valptr = new pcrcfDAD(src->rank-1,*src);
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
  tmpptr->elementLen = sizeof(int);
  
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     valptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0); 
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;

      valptr->rngs[counter]=src->rngs[i] ;
      valptr->strs[counter]=src->strs[i] ;

      if(tmpptr->status == OUT_OF_CORE) {

         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
         valptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
       }
      counter++;
   }
  }
  
    int tmpDAD = ad_FromPointer(tmpptr);
    int valDAD = ad_FromPointer(valptr);

    if(tmpptr->status==IN_CORE) {
      tmpptr->data = new char [tmpptr->elementLen * src->size ] ;	
      valptr->data = new char [src->elementLen * src->size ] ;
     } 
    else 
      {  pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 
         pcrc_oc_init(&valDAD,&(valptr->icla_size));
      }

  MaxlocDim<T> schedule(valptr,tmpptr, src, dim0) ;
  schedule.execute((T *)valptr->data, (int *)tmpptr->data, (T *)src->data) ;

  Remap schedule1(res,tmpptr,tmpptr->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  
  delete valptr ;

/*
  int size=1;
  for (i=0; i< res->rnk() ; i++)
	size = size * res->rng(i).size() ;
  T *valdata = (T *)malloc( size * sizeof(T) );
  pcrcfDAD* val = new pcrcfDAD(valdata, *res);

  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXLOC \n");
	break;
  case 1:
      {   
	Section0<T>  valsect(val->group,(T *)val->data);
	Section0<int>  ressect(res->group,(int *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
        MaxlocDim<T> schedule(&valsect, &ressect, &srcsect, dim0) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat()) ;
        break;
       }  
  case 2:
      {   
	Section1<T>  valsect(val->ranges[0],val->group,val->strides[0],(T *)val->data);
	Section1<int>  ressect(res->ranges[0],res->group,res->strides[0],(int *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
        MaxlocDim<T> schedule(&valsect, &ressect, &srcsect, dim0) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat()) ;
        break;
       }  
  case 3:
      {   
	Section2<T>  valsect(val->ranges[0],val->ranges[1],val->group,val->strides[0],val->strides[1],(T *)val->data);
	Section2<int>  ressect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(int *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
        MaxlocDim<T> schedule(&valsect, &ressect, &srcsect, dim0) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat()) ;
        break;
       }  
  }
  for (i = 0; i< res->size ; i++ )
	((int*)res->data)[i] ++;
  free(valdata);
*/
}

void pcrc_dim_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, int *dim)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_dim_maxloc( res0, dad_res,src0,dad_src,dim, (int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_dim_maxloc( res0, dad_res,src0,dad_src,dim, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_maxloc( res0, dad_res,src0,dad_src,dim, (short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_dim_maxloc( res0, dad_res,src0,dad_src,dim, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_maxloc( res0, dad_res,src0,dad_src,dim, (double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim MAXLOC \n");
  }
}

//*************************************************************************//

template<class T>
inline void temp_dim_msk_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk,T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk   = (pcrcfDAD*) ad_ToPointer(*dad_msk); 
  int dim0=*dim-1;
  int rank=src->rank;

//xh
  pcrcfDAD* valptr = new pcrcfDAD(src->rank-1,*src);
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
  tmpptr->elementLen = sizeof(int);
  
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     valptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0); 
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;

      valptr->rngs[counter]=src->rngs[i] ;
      valptr->strs[counter]=src->strs[i] ;

      if(tmpptr->status == OUT_OF_CORE) {

         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
         valptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
       }
      counter++;
   }
  }
  
    int tmpDAD = ad_FromPointer(tmpptr);
    int valDAD = ad_FromPointer(valptr);

    if(tmpptr->status==IN_CORE) {
      tmpptr->data = new char [tmpptr->elementLen * src->size ] ;	
      valptr->data = new char [src->elementLen * src->size ] ;
     } 
    else 
      {  pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 
         pcrc_oc_init(&valDAD,&(valptr->icla_size));
      }


    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        MaxlocDimMsk<T> schedule(valptr,tmpptr,src,dim0,msktmpptr);
        schedule.execute((T*)valptr->data,(int*)tmpptr->data,(T*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       MaxlocDimMsk<T> schedule(valptr,tmpptr,src,dim0,msk);
       schedule.execute((T*)valptr->data,(int*)tmpptr->data,(T*)src->data,(int*)msk->data);

     }



  Remap schedule1(res,tmpptr,tmpptr->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  
  delete valptr ;


/*
  int i;

  int size=1;
  for (i=0; i< res->rnk() ; i++)
	size = size * res->rng(i).size() ;
  T *valdata = (T *)malloc( size * sizeof(T) );
  pcrcfDAD* val = new pcrcfDAD(valdata, *res);

  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXLOC \n");
	break;
  case 1:
      {   
	Section0<T>  valsect(val->group,(T *)val->data);
	Section0<int>  ressect(res->group,(int *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
        MaxlocDimMsk<T> schedule(&valsect, &ressect, &srcsect, dim0, &msksect) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  case 2:
      {   
	Section1<T>  valsect(val->ranges[0],val->group,val->strides[0],(T *)val->data);
	Section1<int>  ressect(res->ranges[0],res->group,res->strides[0],(int *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
        MaxlocDimMsk<T> schedule(&valsect, &ressect, &srcsect, dim0, &msksect) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  case 3:
      {   
	Section2<T>  valsect(val->ranges[0],val->ranges[1],val->group,val->strides[0],val->strides[1],(T *)val->data);
	Section2<int>  ressect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(int *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
        MaxlocDimMsk<T> schedule(&valsect, &ressect, &srcsect, dim0, &msksect) ;
	schedule.execute(valsect.dat(), ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  }
  for (i = 0; i< res->size ; i++ )
	((int*)res->data)[i] ++;
  free(valdata); 
*/
}

void pcrc_dim_msk_maxloc(void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_dim_msk_maxloc( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_dim_msk_maxloc( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(short)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_msk_maxloc( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_dim_msk_maxloc( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_msk_maxloc( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim-msk MAXLOC \n");
  }
}

