
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"


extern "C" {
   void pcrc_maxval(void *res0, int *dad_res,void *src0, int *dad_src);
   void pcrc_msk_maxval(void *res0,int *dad_res, void *src0,int *dad_src,void *msk0, int *dad_msk);
   void pcrc_dim_maxval(void *res0, int *dad_res, void *src0,int *dad_src, int *dim);
   void pcrc_dim_msk_maxval( void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk);
}

//*************************************************************************//

template<class T>
inline  void temp_maxval(void *res0, int *dad_res,void *src0,int *dad_src,T type)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);

  reduceMode = 1;

  T result;
  Maxval<T> schedule(src) ;
  schedule.execute(&result, (T *)src->data) ;

  reduceMode = 0;
/*
  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXVAL \n");
	break;
  case 1:
      {   
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
        result= maxval(srcsect);
        break;
       }  
  case 2:
      {   
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
        result= maxval(srcsect);
        break;
       }  
  
  case 3:
      {   
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
        result= maxval(srcsect);
        break;
       }  
  }
*/
  scalar_assign(resPtr,result);
}

 void pcrc_maxval(void *res0, int *dad_res,void *src0,int *dad_src)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_maxval( res0, dad_res,src0,dad_src,(int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_maxval( res0, dad_res,src0,dad_src,(char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_maxval( res0, dad_res,src0,dad_src,(short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_maxval( res0, dad_res,src0,dad_src,(float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_maxval( res0, dad_res,src0,dad_src,(double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in MAXVAL \n");
  }
}

//*************************************************************************//

template<class T>
inline void  temp_msk_maxval(void *res0,int *dad_res, void *src0,int *dad_src, void *msk0,int *dad_msk,T type)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  T res;

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;

    MaxvalMsk<T> schedule(src, tmpptr) ;
    schedule.execute(&res, (T *)src->data, (int *)tmpptr->data) ;

    delete tmpptr ;

    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    MaxvalMsk<T> schedule(src, msk) ;
    schedule.execute(&res, (T *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }

/*
  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXVAL \n");
	return;
  case 1:
      {   
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
//        return maxval((Section_<1, T>)srcsect,(Section_<1, int>)msksect);
        MaxvalMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&res, srcsect.dat(), msksect.dat()) ;
        break;
       }  
  case 2:
      {   
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
//        return maxval((Section_<2, T>)srcsect,(Section_<2, int>)msksect);
        MaxvalMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&res, srcsect.dat(), msksect.dat()) ;
        break;
       }  
  
  case 3:
      {   
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
//        return maxval((Section_<3, T>)srcsect,(Section_<3, int>)msksect);
        MaxvalMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&res, srcsect.dat(), msksect.dat()) ;
        break;
       }  
  }
*/
  scalar_assign(resPtr,res);
}

void  pcrc_msk_maxval(void *res0,int *dad_res, void *src0,int *dad_src, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_msk_maxval( res0, dad_res,src0,dad_src,msk0,dad_msk,(int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_msk_maxval( res0, dad_res,src0,dad_src,msk0,dad_msk,(char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_msk_maxval( res0, dad_res,src0,dad_src,msk0,dad_msk,(short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_msk_maxval( res0, dad_res,src0,dad_src,msk0,dad_msk,(float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_msk_maxval( res0, dad_res,src0,dad_src,msk0,dad_msk,(double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in msk MAXVAL \n");
  }
}

//*************************************************************************//

template<class T>
inline  void temp_dim_maxval( void *res0,int *dad_res, void *src0,int *dad_src, int *dim, T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0    = *dim -1;

//xh add
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
 
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;
      if(tmpptr->status == OUT_OF_CORE)
         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
   }
  }

    int tmpDAD = ad_FromPointer(tmpptr);

    if(tmpptr->status==IN_CORE)
      tmpptr->data = new char [src->elementLen * src->size ] ;	
    else  pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 

  MaxvalDim<T> schedule(tmpptr, src, dim0) ;
  schedule.execute((T *)tmpptr->data, (T *)src->data) ;

  Remap schedule1(res,tmpptr,src->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  

/*
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0=*dim-1;

  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXVAL \n");
	break;
  case 1:
      {   
	Section0<T>  ressect(res->group,(T *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
        MaxvalDim<T> schedule(&ressect, &srcsect,dim0) ;
	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
       }  
  case 2:
      {   
	Section1<T>  ressect(res->ranges[0],res->group,res->strides[0],(T *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
        MaxvalDim<T> schedule(&ressect, &srcsect, dim0) ;
	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
       }  
  case 3:
      {   
	Section2<T>  ressect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(T *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
        MaxvalDim<T> schedule(&ressect, &srcsect, dim0) ;
	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
       }  
  }
*/
}

 void pcrc_dim_maxval( void *res0,int *dad_res, void *src0,int *dad_src, int *dim)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_dim_maxval( res0, dad_res,src0,dad_src,dim,(int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_dim_maxval( res0, dad_res,src0,dad_src,dim,(char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_maxval( res0, dad_res,src0,dad_src,dim,(short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_dim_maxval( res0, dad_res,src0,dad_src,dim,(float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_maxval( res0, dad_res,src0,dad_src,dim,(double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim MAXVAL \n");
  }
}

//*************************************************************************//

template<class T>
 inline void temp_dim_msk_maxval( void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk,T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  int dim0=*dim-1;

//xh
    pcrcfDAD* restmpptr = new pcrcfDAD(src->rank-1,*src);
 
    int counter = 0;
    for (int i=0; i<src->rank ; i++) {
    if (i == dim0) {
       restmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     }
    else {
      restmpptr->rngs[counter]=src->rngs[i] ;
      restmpptr->strs[counter]=src->strs[i] ;
      if(restmpptr->status == OUT_OF_CORE)
         restmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
      }
    }
   int restmpDAD = ad_FromPointer(restmpptr);

   if(restmpptr->status==IN_CORE)
      restmpptr->data = new char [src->elementLen * src->size ] ;
   else  pcrc_oc_init(&restmpDAD,&(restmpptr->icla_size)); 
   

    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        MaxvalDimMsk<T> schedule(restmpptr,src,dim0,msktmpptr);
        schedule.execute((T*)restmpptr->data,(T*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       MaxvalDimMsk<T> schedule(restmpptr,src,dim0,msk);
       schedule.execute((T*)restmpptr->data,(T*)src->data,(int*)msk->data);
     
     }

     Remap schedule1(res,restmpptr,src->elementLen);
     schedule1.execute(res->data,restmpptr->data);
   
     delete restmpptr;
      
/*
  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of MAXVAL \n");
	break;
  case 1:
      {   
	Section0<T>  ressect(res->group,(T *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
        MaxvalDimMsk<T> schedule(&ressect, &srcsect, dim0, &msksect) ;
	schedule.execute(ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  case 2:
      {   
	Section1<T>  ressect(res->ranges[0],res->group,res->strides[0],(T *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
        MaxvalDimMsk<T> schedule(&ressect, &srcsect, dim0, &msksect) ;
	schedule.execute(ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  case 3:
      {   
	Section2<T>  ressect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(T *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
        MaxvalDimMsk<T> schedule(&ressect, &srcsect,dim0, &msksect) ;
	schedule.execute(ressect.dat(), srcsect.dat(), msksect.dat()) ;
        break;
       }  
  }
*/
}

 void pcrc_dim_msk_maxval( void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	temp_dim_msk_maxval( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(int)0 );
	break;
	}
  case PCRC_BYTE: {
	temp_dim_msk_maxval( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_msk_maxval( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(short)0 );
	break;
	}
  case PCRC_REAL: {
	temp_dim_msk_maxval( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_msk_maxval( res0, dad_res,src0,dad_src,dim,msk0,dad_msk,(double)0 );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim-msk MAXVAL \n");
  }
}


