#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"

extern "C" {
   void pcrc_count(void *res0, int *dad_res , void *src0 , int *dad_src );
   void pcrc_msk_count (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk);
   void pcrc_dim_count(void *res0, int *dad_res , void *src0 , int *dad_src ,int *dim);
   void pcrc_dim_msk_count(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk);

   void pcrc_cshift0(void *res0, int* dad, void *src0, int* dad_source, int* shifts, int* dim) ;
   void pcrc_cshift01(void *res0,int* dad, void *src0, int* dad_source, int* shifts) ;

   void pcrc_integer_cmplx(void *res0 , int *dad_res , int *x,int *y);
   void pcrc_real_cmplx(void *res0 , int *dad_res , float *x,float *y);
   void pcrc_double_cmplx(void *res0 , int *dad_res , double *x,double *y);

   void pcrc_lbound(void *res0 , int *dad_res , void *src0, int *dad_src);
   void pcrc_dim_lbound(void *res0 , int *dad_res , void *src0, int *dad_src, int *dim);

   void pcrc_ubound(void *res0 , int *dad_res , void *src0, int *dad_src);
   void pcrc_dim_ubound(void *res0 , int *dad_res , void *src0, int *dad_src, int *dim);

   void pcrc_size(void *res0, int *dad_res, void *src0, int *dad_src);
   void pcrc_dim_size(void *res0, int *dad_res, void *src0, int *dad_src, int *dim);

   void pcrc_allocated(void *res0, int *dad_res, void *src0, int *dad_src);

   void pcrc_any(void *res0, int *dad_res, void *src0, int *dad_src);
   void pcrc_msk_any (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk);
   void pcrc_dim_any(void *res0, int *dad_res, void *src0, int *dad_src, int *dim);
   void pcrc_dim_msk_any(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk);

   void pcrc_all(void *res0, int *dad_res, void *src0, int *dad_src);
   void pcrc_msk_all (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk);
   void pcrc_dim_all(void *res0, int *dad_res, void *src0, int *dad_src, int *dim);
   void pcrc_dim_msk_all(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk);

   void pcrc_reshape(void *res0, int *dad_res, void *src, int *dad_src, void *shp, int *dad_shp);

   void pcrc_proc_id();
}



//*************************************************************************//

 void pcrc_count(void *res0, int *dad_res , void *src0 , int *dad_src )
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  res=0;

  reduceMode = 1;

  Count  schedule(src) ;
  schedule.execute(&res, (int *)src->data) ;

  reduceMode = 0;
 
  scalar_assign(resPtr,res);
/*
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  res=0;

  int rank=src->rnk();
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section1<int>  srcsect(src->ranges[0],src->group,src->strides[0],(int *)src->data);
	res= count(srcsect);
	break;
        }
  case 2:
        {
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
	res = count(srcsect);
        break;
        }
  case 3:
        {
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
	res =  count(srcsect);
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM , NOT SUPPORTED. \n");
	break;
  }
*/
}


void pcrc_msk_count (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk){

  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int result;

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;

    CountMsk schedule(src, tmpptr) ;
    schedule.execute(&result, (int *)src->data, (int *)tmpptr->data) ;
 
    delete tmpptr ;

    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    CountMsk schedule(src, msk) ;
    schedule.execute(&result, (int *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }
   scalar_assign(resPtr,result);
 
}


 void pcrc_dim_count(void *res0, int *dad_res , void *src0 , int *dad_src ,int *dim)
{

  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0    = *dim -1;

//xh add
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
 
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;
      if(tmpptr->status == OUT_OF_CORE)
         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
   }
  }

    int tmpDAD = ad_FromPointer(tmpptr);

    if(tmpptr->status==IN_CORE)
      tmpptr->data = new char [src->elementLen * src->size ] ;	
    else  pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 

  CountDim schedule(tmpptr, src, dim0) ;
  schedule.execute((int *)tmpptr->data, (int *)src->data) ;


  Remap schedule1(res,tmpptr,src->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  

/*
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0    = *dim -1;

  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section0<int>  dstsect(res->group,(int *)res->data);
	Section1<int>  srcsect(src->ranges[0],src->group,src->strides[0],(int *)src->data);
	CountDim schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  case 2:
	{
	Section1<int>  dstsect(res->ranges[0],res->group,res->strides[0],(int *)res->data);
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
	CountDim schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  case 3:
	{
	Section2<int>  dstsect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(int *)res->data);
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
	CountDim schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  }
*/
}

 void pcrc_dim_msk_count(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk){

  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  int  dim0   = *dim -1;

//xh
    pcrcfDAD* restmpptr = new pcrcfDAD(src->rank-1,*src);
    int counter = 0;
    for (int i=0; i<src->rank ; i++) {
    if (i == dim0) {
       restmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     }
    else {
      restmpptr->rngs[counter]=src->rngs[i] ;
      restmpptr->strs[counter]=src->strs[i] ;
      if(restmpptr->status == OUT_OF_CORE)
         restmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
      }
    }
   int restmpDAD = ad_FromPointer(restmpptr);

   if(restmpptr->status==IN_CORE)
      restmpptr->data = new char [src->elementLen * src->size ] ;
   else   pcrc_oc_init(&restmpDAD,&(restmpptr->icla_size)); 
   

    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        CountDimMsk schedule(restmpptr,src,dim0,msktmpptr);
        schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       CountDimMsk schedule(restmpptr,src,dim0,msk);
       schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msk->data);

     }
     Remap schedule1(res,restmpptr,src->elementLen);
     schedule1.execute(res->data,restmpptr->data);
   
     delete restmpptr;

}

//*************************************************************************//

void pcrc_cshift0(void *res0,int* dad, void *src0,int* dad_source, int* shifts, int* dim) {
  pcrcfDAD* ptr     = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr  = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  void* _data = new char [srcPtr->elementLen * ptr->icla_size ] ; 
  pcrcfDAD* tmpptr = new pcrcfDAD(_data, *ptr) ;
  tmpptr->elementLen = srcPtr->elementLen ;
  tmpptr->elementType = srcPtr->elementType ;
  int tmpDAD = ad_FromPointer(tmpptr);
  if (ptr->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

  Remap schedule1(tmpptr, srcPtr, srcPtr->elementLen) ;
  schedule1.execute(tmpptr->data, srcPtr->data) ;

  int srcRank = srcPtr->rank;

  if(srcRank == 0)
	fprintf(stderr," ** RUNTIME ERROR ** Error in rank of array to shift \n");
  else {
        Shift schedule(ptr, tmpptr, ptr->elementLen, *shifts, *dim -1, CYCL) ;
        schedule.execute(ptr->data, tmpptr->data) ;
  }

  delete tmpptr ;
/*
  int srcRank = srcPtr->rank;

  if(srcRank == 0)
	fprintf(stderr," ** RUNTIME ERROR ** Error in rank of array to shift \n");
  else {
        Shift schedule(ptr, srcPtr, ptr->elementLen, *shifts, *dim -1, CYCL) ;
        schedule.execute(ptr->data, srcPtr->data) ;
  }
*/
}	


void pcrc_cshift01(void *res0,int* dad, void *src0,int* dad_source, int* shifts) {
  pcrcfDAD* ptr     = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr  = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  void* _data = new char [srcPtr->elementLen * ptr->icla_size ] ; 
  pcrcfDAD* tmpptr = new pcrcfDAD(_data, *ptr) ;
  tmpptr->elementLen = srcPtr->elementLen ;
  tmpptr->elementType = srcPtr->elementType ;
  int tmpDAD = ad_FromPointer(tmpptr);
  if (ptr->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

  Remap schedule1(tmpptr, srcPtr, srcPtr->elementLen) ;
  schedule1.execute(tmpptr->data, srcPtr->data) ;

  int srcRank = srcPtr->rank;

  if(srcRank == 0)
        fprintf(stderr," ** RUNTIME ERROR ** Error in rank of array to shift \n");
  else {
        Shift schedule(ptr, tmpptr, ptr->elementLen, *shifts, 0, CYCL) ;
        schedule.execute(ptr->data, tmpptr->data) ;
  }

  delete tmpptr ;
/*
  int srcRank = srcPtr->rank;

  if(srcRank == 0)
	fprintf(stderr," ** RUNTIME ERROR ** Error in rank of array to shift \n");
  else {
        Shift schedule(ptr, srcPtr, ptr->elementLen, *shifts, 0, CYCL) ;
        schedule.execute(ptr->data, srcPtr->data) ;
  }
*/
}	


//*************************************************************************//

void pcrc_integer_cmplx(void *res0 , int *dad_res , int *x,int *y)
{
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  COMPLEX cmp(*x,*y);

  for (int i = 0; i< resPtr->size ; i++ )
		((COMPLEX*)resPtr->data)[i] = cmp; 
}

void pcrc_real_cmplx(void *res0 , int *dad_res , float *x,float *y)
{
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  COMPLEX cmp(*x,*y);

  for (int i = 0; i< resPtr->size ; i++ )
		((COMPLEX*)resPtr->data)[i] = cmp; 
}

void pcrc_double_cmplx(void *res0 , int *dad_res , double *x,double *y)
{
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  COMPLEX cmp(*x,*y);

  for (int i = 0; i< resPtr->size ; i++ )
		((COMPLEX*)resPtr->data)[i] = cmp; 
}

//*************************************************************************//

void pcrc_lbound(void *res0 , int *dad_res , void *src0, int *dad_src)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  int rank = srcPtr->rnk();
  int *data = (int *)malloc(rank * sizeof(int) );

  for (int i=0; i<rank; i++)
  {
	data[i] = srcPtr->ranges[i].g_lb();
  }
  // root group
  pcrcfDAD *tmp = new pcrcfDAD(data , 1, sizeof(int),1 , *(Group *)ad_ToPointer(2) );
  Range x ;
  setCollapse(x, 1, rank) ;
  tmp->addRange(0, x) ;

  Remap schedule(ptr, tmp, ptr->elementLen) ;
  schedule.execute(ptr->data, tmp->data) ;
}

void pcrc_dim_lbound(void *res0 , int *dad_res , void *src0, int *dad_src, int *dim)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  int rank = srcPtr->rnk();

  if ( *dim > rank )
  {
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of LBOUND. \n");
	return;
  }

  *((int*)(ptr->data)) = srcPtr->ranges[*dim-1].g_lb();
}

void pcrc_ubound(void *res0 , int *dad_res , void *src0, int *dad_src)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  int rank = srcPtr->rnk();
  int *data = (int *)malloc(rank * sizeof(int) );

  for (int i=0; i<rank; i++)
  {
  	int extent = srcPtr->ranges[i].size();
  	int g_lb  = srcPtr->ranges[i].g_lb();
	data[i] = extent - 1 + g_lb;
  }
  // root group
  pcrcfDAD *tmp = new pcrcfDAD(data , 1, sizeof(int),1 , *(Group *)ad_ToPointer(2) );
  Range x ;
  setCollapse(x, 1, rank) ;
  tmp->addRange(0, x) ;

  Remap schedule(ptr, tmp, ptr->elementLen) ;
  schedule.execute(ptr->data, tmp->data) ;
}

void pcrc_dim_ubound(void *res0 , int *dad_res , void *src0, int *dad_src, int *dim)
{
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  int rank = srcPtr->rnk();

  if ( *dim > rank )
  {
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of UBOUND. \n");
	return;
  }
 
  int extent = srcPtr->ranges[*dim-1].size();
  int g_lb  = srcPtr->ranges[*dim-1].g_lb();

  *((int*)(resPtr->data)) = extent - 1 + g_lb ;
}

//*************************************************************************//

void pcrc_size(void *res0, int *dad_res, void *src0, int *dad_src)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int rank=src->rnk();
  int res=1;
  for (int i=0; i<rank; i++)
    res *= src->rngs[i].size();
  scalar_assign(resPtr,res);
}

void pcrc_dim_size(void *res0, int *dad_res, void *src0, int *dad_src, int *dim)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr = (pcrcfDAD*) ad_ToPointer(*dad_res);

  if ( *dim > src->rank )
  {
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SIZE. \n");
	return;
  }
 
  int res = src->rngs[*dim -1].size();
  scalar_assign(resPtr,res);
}

//*************************************************************************//

void pcrc_allocated(void *res0, int *dad_res, void *src0, int *dad_src)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr = (pcrcfDAD*) ad_ToPointer(*dad_res);

  int res = src->memHandle != -1 ? 1:0;
  scalar_assign(resPtr,res);
}

//*************************************************************************//

void pcrc_any(void *res0, int *dad_res, void *src0, int *dad_src)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  res=0;

//xh

  reduceMode = 1;

  Any schedule(src) ;
  schedule.execute(&res, (int *)src->data) ;

  reduceMode = 0;

  scalar_assign(resPtr,res);
/*
  int rank=src->rnk();
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ANY \n");
	break;
  case 1:
	{
	Section1<int>  srcsect(src->ranges[0],src->group,src->strides[0],(int *)src->data);
  	Any schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
	break;
        }
  case 2:
        {
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
  	Any schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
        break;
        }
  case 3:
        {
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
  	Any schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ANY , NOT SUPPORTED. \n");
	break;
  }
*/
}

void pcrc_msk_any (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk)
{
//xh
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int result;

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;

    AnyMsk schedule(src, tmpptr) ;
    schedule.execute(&result, (int *)src->data, (int *)tmpptr->data) ;
 
    delete tmpptr ;

    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    AnyMsk schedule(src, msk) ;
    schedule.execute(&result, (int *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }
scalar_assign(resPtr,result);
}



void pcrc_dim_any(void *res0, int *dad_res, void *src0, int *dad_src, int *dim)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  dim0 = *dim -1 ;


//xh add
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
 
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;
      if(tmpptr->status == OUT_OF_CORE)
         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
   }
  }

    int tmpDAD = ad_FromPointer(tmpptr);

//**xh
    if(tmpptr->status==IN_CORE)
      tmpptr->data = new char [src->elementLen * src->size ] ;	
    else   pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 

  AnyDim schedule(tmpptr, src, dim0) ;
  schedule.execute((int *)tmpptr->data, (int *)src->data) ;


  Remap schedule1(res,tmpptr,src->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  


/*
  int rank=src->rnk();
  switch (rank)
  {
  case 0:
  case 1:
	{
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ANY \n");
	break;
        }
  case 2:
        {
	Section1<int>  ressect(resPtr->ranges[0],resPtr->group,resPtr->strides[0],(int *)resPtr->data);
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
	AnyDim schedule(&ressect, &srcsect, dim0);
  	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
        }
  case 3:
        {
	Section2<int>  ressect(resPtr->ranges[0],resPtr->ranges[1],resPtr->group,resPtr->strides[0],resPtr->strides[1],(int *)resPtr->data);
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
	AnyDim schedule(&ressect, &srcsect, dim0);
  	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ANY , NOT SUPPORTED. \n");
	break;
  }
 */
}

void pcrc_dim_msk_any(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk)
{

  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  int  dim0   = *dim -1;

//xh
    pcrcfDAD* restmpptr = new pcrcfDAD(src->rank-1,*src);
    int counter = 0;
    for (int i=0; i<src->rank ; i++) {
    if (i == dim0) {
       restmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     }
    else {
      restmpptr->rngs[counter]=src->rngs[i] ;
      restmpptr->strs[counter]=src->strs[i] ;
      if(restmpptr->status == OUT_OF_CORE)
         restmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
      }
    }
   int restmpDAD = ad_FromPointer(restmpptr);

   if(restmpptr->status==IN_CORE)
      restmpptr->data = new char [src->elementLen * src->size ] ;
   else   pcrc_oc_init(&restmpDAD,&(restmpptr->icla_size)); 
   

    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        AnyDimMsk schedule(restmpptr,src,dim0,msktmpptr);
        schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       AnyDimMsk schedule(restmpptr,src,dim0,msk);
       schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msk->data);
     
     }
     Remap schedule1(res,restmpptr,src->elementLen);
     schedule1.execute(res->data,restmpptr->data);
   
     delete restmpptr;
}

//*************************************************************************//


void pcrc_all(void *res0, int *dad_res, void *src0, int *dad_src)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  res=0;

//xh
  reduceMode = 1;

  All schedule(src) ;
  schedule.execute(&res, (int *)src->data) ;

  reduceMode = 0;

  scalar_assign(resPtr,res);

/*
  int rank=src->rnk();
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ALL \n");
	break;
  case 1:
	{
	Section1<int>  srcsect(src->ranges[0],src->group,src->strides[0],(int *)src->data);
  	All schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
	break;
        }
  case 2:
        {
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
  	All schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
        break;
        }
  case 3:
        {
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
  	All schedule(&srcsect) ;
  	schedule.execute(&res, srcsect.dat()) ;
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ALL , NOT SUPPORTED. \n");
	break;
  }
*/
}

void pcrc_msk_all (void* res0, int* dad_res, void* src0, int* dad_src,
                     void* msk0, int* dad_msk)
{

  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int result;

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;

    AllMsk schedule(src, tmpptr) ;
    schedule.execute(&result, (int *)src->data, (int *)tmpptr->data) ;
 
    delete tmpptr ;

    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    AllMsk schedule(src, msk) ;
    schedule.execute(&result, (int *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }
scalar_assign(resPtr,result);
}

void pcrc_dim_all(void *res0, int *dad_res, void *src0, int *dad_src, int *dim)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int  dim0 = *dim -1 ;

//xh add
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1,*src);
 
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;
      if(tmpptr->status == OUT_OF_CORE)
         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
   }
  }

    int tmpDAD = ad_FromPointer(tmpptr);

    if(tmpptr->status==IN_CORE)
      tmpptr->data = new char [src->elementLen * src->size ] ;	
    else   pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 

  AllDim schedule(tmpptr, src, dim0) ;
  schedule.execute((int *)tmpptr->data, (int *)src->data) ;

  Remap schedule1(res,tmpptr,src->elementLen);
  schedule1.execute(res->data,tmpptr->data);

  delete tmpptr ;  

 
/*
  int rank=src->rnk();
  switch (rank)
  {
  case 0:
  case 1:
	{
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ALL \n");
	break;
        }
  case 2:
        {
	Section1<int>  ressect(resPtr->ranges[0],resPtr->group,resPtr->strides[0],(int *)resPtr->data);
	Section2<int>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(int *)src->data);
	AllDim schedule(&ressect, &srcsect, dim0);
  	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
        }
  case 3:
        {
	Section2<int>  ressect(resPtr->ranges[0],resPtr->ranges[1],resPtr->group,resPtr->strides[0],resPtr->strides[1],(int *)resPtr->data);
	Section3<int>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(int *)src->data);
	AllDim schedule(&ressect, &srcsect, dim0);
  	schedule.execute(ressect.dat(), srcsect.dat()) ;
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of ALL , NOT SUPPORTED. \n");
	break;
  }
*/
}

void pcrc_dim_msk_all(void *res0,int *dad_res,void *src0, int *dad_src,
                     int *dim,void* msk0,int *dad_msk)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  int  dim0   = *dim -1;

//xh
    pcrcfDAD* restmpptr = new pcrcfDAD(src->rank-1,*src);
    int counter = 0;
    for (int i=0; i<src->rank ; i++) {
    if (i == dim0) {
       restmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     }
    else {
      restmpptr->rngs[counter]=src->rngs[i] ;
      restmpptr->strs[counter]=src->strs[i] ;
      if(restmpptr->status == OUT_OF_CORE)
         restmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
      }
    }
   int restmpDAD = ad_FromPointer(restmpptr);

   if(restmpptr->status==IN_CORE)
      restmpptr->data = new char [src->elementLen * src->size ] ;
   else  pcrc_oc_init(&restmpDAD,&(restmpptr->icla_size)); 
   
    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        AllDimMsk schedule(restmpptr,src,dim0,msktmpptr);
        schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       AllDimMsk schedule(restmpptr,src,dim0,msk);
       schedule.execute((int*)restmpptr->data,(int*)src->data,(int*)msk->data);

     }

     Remap schedule1(res,restmpptr,src->elementLen);
     schedule1.execute(res->data,restmpptr->data);
   
     delete restmpptr;
}

//*************************************************************************//

void pcrc_reshape(void *res0, int *dad_res, void *src, int *dad_src, void *shp, int *dad_shp)
{
  pcrcfDAD* resPtr = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* shpPtr = (pcrcfDAD*) ad_ToPointer(*dad_shp);

//  reshape(resPtr, srcPtr);
}

//*************************************************************************//

// return the process id.
void pcrc_proc_id(int *data)
{
        *data =  AdlibLocId();
}


