#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"

extern "C" {
   void pcrc_sum(void *res0, int *dad_res,void *src0,int *dad_src);
   void pcrc_msk_sum(void *res0, int *dad_res ,void *src0,int *dad_src, void *msk0,int *dad_msk);
   void pcrc_dim_sum(void *res0,int *dad_res,void *src0, int *dad_src , int *dim) ;
   void pcrc_dim_msk_sum(void *res0,int *dad_res,void *src0, int *dad_src, int *dim,void *msk0, int *dad_msk);
}

/*
// local copy array
template<class T>
inline void assign_array(pcrcfDAD *resPtr , T *srcDat)
{
  switch (resPtr->rank)
  {
  case 0: {
	if (resPtr->group.member()) {
	numeric_assign(resPtr , 0 , *srcDat); 
	} 
	break;
    }
  case 1:  {
	LocBlocksIndex i(resPtr->ranges[0]);
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
		int idx = off0_bas+off0_stp*l;
		numeric_assign(resPtr , idx, srcDat[idx]); 
	  }
	}
	break;
    }
  case 2: {
	LocBlocksIndex i(resPtr->ranges[0]),j(resPtr->ranges[1]) ;
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
	     for (j.beginLocBlk(); j.test(); j.nextLocBlk() ) {
	  	int off1_bas = resPtr->str(1).offset(j);
	  	int off1_stp = resPtr->str(1).step(j);
	  	for (int m=0; m<j.count; m++) {
			int idx = off0_bas+off0_stp*l + off1_bas+off1_stp*m;
			numeric_assign(resPtr , idx, srcDat[idx]); 
		}
	     }
	  }
	}
	break;
    }
  case 3: {
	LocBlocksIndex i(resPtr->ranges[0]),j(resPtr->ranges[1]),k(resPtr->ranges[2]) ;
	for (i.beginLocBlk(); i.test(); i.nextLocBlk() ) {
	  int off0_bas = resPtr->str(0).offset(i);
	  int off0_stp = resPtr->str(0).step(i);
	  for (int l=0; l<i.count; l++) {
	     for (j.beginLocBlk(); j.test(); j.nextLocBlk() ) {
	  	int off1_bas = resPtr->str(1).offset(j);
	  	int off1_stp = resPtr->str(1).step(j);
	  	for (int m=0; m<j.count; m++) {
		    for (k.beginLocBlk(); k.test(); k.nextLocBlk() ) {
	  		int off2_bas = resPtr->str(2).offset(k);
	  		int off2_stp = resPtr->str(2).step(k);
			for (int kk=0; kk<k.count; kk++) {
			    int idx = off0_bas+off0_stp*l + off1_bas+off1_stp*m + off2_bas+off2_stp*kk;
			    numeric_assign(resPtr , idx, srcDat[idx]); 
			}
		   }
		}
	     }
	  }
	}
	break;
    }
  default: {
	printf("scalar_assign: cannot handle !\n");
    }
  }
}

// copy data from srcPtr to resPtr , these two dad have same data mapping.
inline void local_data_copy(pcrcfDAD *resPtr,pcrcfDAD *srcPtr)
{
	int i;
 	switch ( srcPtr->elementType)
  	{
  	case PCRC_BYTE: {
		char *res = (char*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	case PCRC_WORD: {
		short *res = (short*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	case PCRC_INTEGER: {
		int *res = (int*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	case PCRC_REAL: {
		float *res = (float*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	case PCRC_DOUBLE: {
		double *res = (double*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	case PCRC_COMPLEX: {
		COMPLEX *res = (COMPLEX*)srcPtr->data;
		assign_array(resPtr,res);
		break;
		}
  	}
}
*/


//*************************************************************************//

template<class T>
inline void  temp_sum( void *res0, int *dad_res,void *src0,int *dad_src, T type)
{
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  T  res=0;

  reduceMode = 1;

  Sum<T> schedule(src) ;
  schedule.execute(&res, (T *)src->data) ;

  reduceMode = 0;

/*
  int rank=src->rnk();
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data,src->status, src->icla_size);
	res= sum(srcsect);
	break;
        }
  case 2:
        {
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data,src->status, src->icla_size);
	res= sum(srcsect);
        break;
        }
  case 3:
        {
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data,src->status, src->icla_size);
	res =  sum(srcsect);
        break;
        }
  default:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM , NOT SUPPORTED. \n");
	break;
  }
*/
//  *(int*)res0 = res;
//  for (int i = 0; i< resPtr->size ; i++ )
//	numeric_assign(resPtr , i , res);
  scalar_assign(resPtr,res);
}

void  pcrc_sum(void *res0, int *dad_res, void *src0,int *dad_src)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_BYTE: {
	temp_sum( res0, dad_res,src0,dad_src, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_sum( res0, dad_res,src0,dad_src, (short)0 );
	break;
	}
  case PCRC_INTEGER: {
	temp_sum( res0, dad_res,src0,dad_src, (int)0 );
	break;
	}
  case PCRC_REAL: {
	temp_sum( res0, dad_res,src0,dad_src, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_sum( res0, dad_res,src0,dad_src, (double)0 );
	break;
	}
  case PCRC_COMPLEX: {
	temp_sum( res0, dad_res,src0,dad_src, COMPLEX(0,0) );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in SUM \n");
  }
}

//*************************************************************************//

template<class T>
inline void  temp_msk_sum( void *res0, int *dad_res,void *src0,int *dad_src,void
 *msk0,int *dad_msk, T type)
{
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  T result;

  if(!pcrc_same_dad_check(dad_src, dad_msk)) {
    void* _data = new char [msk->elementLen * src->icla_size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *src) ;
    tmpptr->elementLen = msk->elementLen ;
    tmpptr->elementType = msk->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (src->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, msk, msk->elementLen) ;
    schedule1.execute(tmpptr->data, msk->data) ;

    reduceMode = 1;

    SumMsk<T> schedule(src, tmpptr) ;
    schedule.execute(&result, (T *)src->data, (int *)tmpptr->data) ;
 
    delete tmpptr ;

    reduceMode = 0;
  }
  else {
    reduceMode = 1;
    SumMsk<T> schedule(src, msk) ;
    schedule.execute(&result, (T *)src->data, (int *)msk->data) ;
    reduceMode = 0;
  }

/*
  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
        SumMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&result, srcsect.dat(), msksect.dat()) ;
        break;
        }
  case 2:
        {
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
        SumMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&result, srcsect.dat(), msksect.dat()) ;
        break;
        }
  case 3:
        {
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
        SumMsk<T> schedule(&srcsect, &msksect) ;
        schedule.execute(&result, srcsect.dat(), msksect.dat()) ;
        break;
        }
       
  }
*/
//  *(T *)res0 = result;
//  for (int i = 0; i< resPtr->size ; i++ )
//	numeric_assign(resPtr , i , result);

  scalar_assign(resPtr,result);
}

void pcrc_msk_sum(void *res0, int *dad_res ,void *src0,int *dad_src, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_BYTE: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, (short)0 );
	break;
	}
  case PCRC_INTEGER: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, (int)0 );
	break;
	}
  case PCRC_REAL: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, (double)0 );
	break;
	}
  case PCRC_COMPLEX: {
	temp_msk_sum( res0, dad_res,src0,dad_src,msk0,dad_msk, COMPLEX(0,0) );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in msk SUM \n");
  }
}

//*************************************************************************//

template<class T>
inline void  temp_dim_sum( void *res0, int *dad_res,void *src0,int *dad_src,int *dim,T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  int dim0    = *dim -1;

//wjp add
  pcrcfDAD* tmpptr = new pcrcfDAD(src->rank-1, *src) ;
  
  int counter = 0;
  for (int i=0; i<src->rank ; i++) {
   if (i == dim0) {
     tmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
   }
   else {
      tmpptr->rngs[counter]=src->rngs[i] ;
      tmpptr->strs[counter]=src->strs[i] ;
      if(tmpptr->status == OUT_OF_CORE)
         tmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
   }
  }
  
   int tmpDAD = ad_FromPointer(tmpptr);

//  tmpptr->data = new char [src->elementLen * src->size ] ; 
  //**xh
    if(tmpptr->status==IN_CORE)
      tmpptr->data = new char [src->elementLen * src->size ] ;
    else 
      {  pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size)); 
        //PFILE* tmpfile = (PFILE*)tmpptr->pfile;
        //PFILE* srcfile = (PFILE*)src->pfile;
        //AdlibPrintf("tmpptr->status=OUT_CORE,%s %s %d %d\n",tmpfile->share->fi
      }

  
  
  SumDim<T> schedule(tmpptr, src, dim0) ;
  schedule.execute((T *)tmpptr->data, (T *)src->data) ;

//AdlibPrintf("near end\n");
  Remap schedule1(res, tmpptr, src->elementLen) ;
  schedule1.execute(res->data, tmpptr->data) ;

//AdlibPrintf("almost end %d\n",src->size);

/*
int* tt=(int*)tmpptr->data;
int tmp;
for (int i=0;i<src->size;i++)
{tmp=*tt;
AdlibPrintf("DDD %d %d\n",tmp,i);
tt++;
}

 tt=(int*)res->data;
for (int i=0;i<src->size;i++)
{tmp=*tt;
AdlibPrintf("xDDD %d %d\n",tmp,i);
tt++;
}
*/
  delete tmpptr ;  

AdlibPrintf("finish\n");
/*
  res->group = src->group ;

  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section0<T>  dstsect(res->group,(T *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	SumDim<T> schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  case 2:
	{
	Section1<T>  dstsect(res->ranges[0],res->group,res->strides[0],(T *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	SumDim<T> schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  case 3:
	{
	Section2<T>  dstsect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(T *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	SumDim<T> schedule(&dstsect, &srcsect, dim0) ;
	schedule.execute(dstsect.dat(), srcsect.dat()) ;
	}
	break;
  }
*/
}

void pcrc_dim_sum(void *res0,int *dad_res, void *src0,int *dad_src , int *dim) 
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_BYTE: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, (short)0 );
	break;
	}
  case PCRC_INTEGER: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, (int)0 );
AdlibPrintf("comeout\n");
	break;
	}
  case PCRC_REAL: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, (double)0 );
	break;
	}
  case PCRC_COMPLEX: {
	temp_dim_sum( res0, dad_res,src0,dad_src,dim, COMPLEX(0,0) );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim SUM \n");
  }
}

//*************************************************************************//

template<class T>
inline void  temp_dim_msk_sum( void *res0, int *dad_res,void *src0,int *dad_src,int *dim,void *msk0, int *dad_msk, T type)
{
  pcrcfDAD* res    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  pcrcfDAD* src    = (pcrcfDAD*) ad_ToPointer(*dad_src);
  pcrcfDAD* msk    = (pcrcfDAD*) ad_ToPointer(*dad_msk);
  int  dim0   = *dim -1;

//xh
    pcrcfDAD* restmpptr = new pcrcfDAD(src->rank-1,*src);
    int counter = 0;
    for (int i=0; i<src->rank ; i++) {
    if (i == dim0) {
       restmpptr->addScalar(src->rngs[dim0].g_lb(),*src,dim0);
     }
    else {
      restmpptr->rngs[counter]=src->rngs[i] ;
      restmpptr->strs[counter]=src->strs[i] ;
      if(restmpptr->status == OUT_OF_CORE)
         restmpptr->pfile->ocla_size[counter] = src->pfile->ocla_size[counter];
      counter++;
      }
    }
   int restmpDAD = ad_FromPointer(restmpptr);

   if(restmpptr->status==IN_CORE)
      restmpptr->data = new char [src->elementLen * src->size ] ;
    else 
      {  pcrc_oc_init(&restmpDAD,&(restmpptr->icla_size)); 
        //PFILE* restmpfile = (PFILE*)restmpptr->pfile;
        //AdlibPrintf("restmpptr->status=OUT_CORE,%s \n",restmpfile->share->filename);
      }
   

    if(!pcrc_same_dad_check(dad_src, dad_msk)) {
        void* _data = new char [msk->elementLen * src->icla_size ] ; 
        pcrcfDAD* msktmpptr = new pcrcfDAD(_data, *src) ;
        msktmpptr->elementLen = msk->elementLen ;
        msktmpptr->elementType = msk->elementType ;
        int msktmpDAD = ad_FromPointer(msktmpptr);
        if (src->status == OUT_OF_CORE) pcrc_oc_init(&msktmpDAD,&(msktmpptr->icla_size));

        Remap schedule1(msktmpptr, msk, msk->elementLen) ;
        schedule1.execute(msktmpptr->data, msk->data) ;

        SumDimMsk<T> schedule(restmpptr,src,dim0,msktmpptr);
        schedule.execute((T*)restmpptr->data,(T*)src->data,(int*)msktmpptr->data);

        delete[] _data;
        delete msktmpptr;

       }
     else {

       SumDimMsk<T> schedule(restmpptr,src,dim0,msk);
       schedule.execute((T*)restmpptr->data,(T*)src->data,(int*)msk->data);

     }

     Remap schedule1(res,restmpptr,src->elementLen);
     schedule1.execute(res->data,restmpptr->data);
   
     delete restmpptr;
AdlibPrintf("finish\n");
/*  int rank=src->rank;
  switch (rank)
  {
  case 0:
    	fprintf(stderr, "** RUNTIME ERROR **  Error in rank of SUM \n");
	break;
  case 1:
	{
	Section0<T>  dstsect(res->group,(T *)res->data);
	Section1<T>  srcsect(src->ranges[0],src->group,src->strides[0],(T *)src->data);
	Section1<int>  msksect(msk->ranges[0],msk->group,msk->strides[0],(int *)msk->data);
	SumDimMsk<T> schedule(&dstsect, &srcsect, dim0, &msksect) ;
	schedule.execute(dstsect.dat(), srcsect.dat(), msksect.dat()) ;
	}
	break;
  case 2:
	{
	Section1<T>  dstsect(res->ranges[0],res->group,res->strides[0],(T *)res->data);
	Section2<T>  srcsect(src->ranges[0],src->ranges[1],src->group,src->strides[0],src->strides[1],(T *)src->data);
	Section2<int>  msksect(msk->ranges[0],msk->ranges[1],msk->group,msk->strides[0],msk->strides[1],(int *)msk->data);
	SumDimMsk<T> schedule(&dstsect, &srcsect, dim0, &msksect) ;
	schedule.execute(dstsect.dat(), srcsect.dat(), msksect.dat()) ;
	}
	break;
  case 3:
	{
	Section2<T>  dstsect(res->ranges[0],res->ranges[1],res->group,res->strides[0],res->strides[1],(T *)res->data);
	Section3<T>  srcsect(src->ranges[0],src->ranges[1],src->ranges[2],src->group,src->strides[0],src->strides[1],src->strides[2],(T *)src->data);
	Section3<int>  msksect(msk->ranges[0],msk->ranges[1],msk->ranges[2],msk->group,msk->strides[0],msk->strides[1],msk->strides[2],(int *)msk->data);
	SumDimMsk<T> schedule(&dstsect, &srcsect, dim0, &msksect) ;
	schedule.execute(dstsect.dat(), srcsect.dat(), msksect.dat()) ;
	}
	break;
  }
*/
}

void pcrc_dim_msk_sum(void *res0,int *dad_res, void *src0,int *dad_src, int *dim, void *msk0,int *dad_msk)
{
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_src) ;

  switch ( srcPtr->elementType)
  {
  case PCRC_BYTE: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, (char)0 );
	break;
	}
  case PCRC_WORD: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, (short)0 );
	break;
	}
  case PCRC_INTEGER: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, (int)0 );
	break;
	}
  case PCRC_REAL: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, (float)0 );
	break;
	}
  case PCRC_DOUBLE: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, (double)0 );
	break;
	}
  case PCRC_COMPLEX: {
	temp_dim_msk_sum( res0, dad_res,src0,dad_src,dim,msk0,dad_msk, COMPLEX(0,0) );
	break;
	}
  default:
	fprintf(stderr, "** RUNTIME ERROR **  un-supported type in dim-msk SUM \n");
  }
}

