#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "intrinname.h"

#include "intrinsic.h"
#include "admacros.h"

extern "C" {
   void pcrc_transpose(void *res0, int* dad,void * src0, int* dad_source ) ;

   void  pcrc_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
/*
   void  pcrc_int_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
   void  pcrc_float_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
   void  pcrc_double_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
   void  pcrc_complex_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
   void  pcrc_bool_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2);
*/

}


//*********************************************************************//

/*
#define _TRANSPOSE(type) 	\
  	Section2<type>  dstsect(ptr->ranges[0],ptr->ranges[1],ptr->group,ptr->strides[0],ptr->strides[1],(type *)ptr->data);	\
  	Section2<type>  srcSect(srcPtr->ranges[0],srcPtr->ranges[1],srcPtr->group,srcPtr->strides[0],srcPtr->strides[1],(type *)srcPtr->data);	\
  	transpose(dstsect,srcSect);

void pcrc_transpose(void *res0,int* dad, void *src0,int* dad_source) 
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  pcrcfDAD *tmpPtr = ptr;
  void *newdata;
  if (srcPtr->data == ptr->data)
  {
	// create a tmp dad to hold result.
	newdata = malloc(ptr->size * ptr->elementLen);
	ptr = new pcrcfDAD(newdata, *ptr);
  }

  switch ( srcPtr->elementType)
  {
  case PCRC_INTEGER: {
	_TRANSPOSE(int);
  	if (tmpPtr != ptr) {
	  Copy<int> schedule(tmpPtr, ptr) ;
          schedule.execute((int*)tmpPtr->data, (int*)ptr->data) ;
	}
	break;
	}
  case PCRC_REAL: {
	_TRANSPOSE(float);
  	if (tmpPtr != ptr) {
	  Copy<float> schedule(tmpPtr, ptr) ;
          schedule.execute((float*)tmpPtr->data, (float*)ptr->data) ;
	}
	break;
	}
  case PCRC_DOUBLE: {
	_TRANSPOSE(double);
  	if (tmpPtr != ptr) {
	  Copy<double> schedule(tmpPtr, ptr) ;
          schedule.execute((double*)tmpPtr->data, (double*)ptr->data) ;
	}
	break;
	}
  case PCRC_COMPLEX: {
	_TRANSPOSE(COMPLEX);
  	if (tmpPtr != ptr) {
	  Copy<COMPLEX> schedule(tmpPtr, ptr) ;
          schedule.execute((COMPLEX*)tmpPtr->data, (COMPLEX*)ptr->data) ;
	}
	break;
	}
  }
  // free up
  if (tmpPtr != ptr) {
	free(newdata);
	delete ptr;
  }
}
*/

void pcrc_transpose(void *res0,int* dad, void *src0,int* dad_source) 
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

/*
  if (srcPtr->data == ptr->data) {
	void *newdata = malloc(ptr->size * ptr->elementLen);
     if (ptr->status == OUT_OF_CORE) {
         DAD *tmpPtr = new pcrcfDAD(newdata, *ptr);
         ptr->pfile->share->refcount--;
         pcrc_oc_init(tmpPtr, &ptr->icla_size);
     }
     else
         ptr->data = newdata ;
  }
  else
*/

  pcrcfDAD *tmpPtr = ptr;
  void *newdata;
  if (srcPtr->data == ptr->data)
  {
	// create a tmp dad to hold result.
	newdata = malloc(ptr->icla_size * ptr->elementLen);
	ptr = new pcrcfDAD(newdata, *ptr);	// DAD copy
  }

  pcrcfDAD* trans = new pcrcfDAD(srcPtr->memdata, *srcPtr);
  trans->rngs[0]=srcPtr->rngs[1];
  trans->strs[0]=srcPtr->strs[1];
  trans->rngs[1]=srcPtr->rngs[0];
  trans->strs[1]=srcPtr->strs[0];
    // ???
//  trans->major = srcPtr->major==COLUMN_MAJOR ? ROW_MAJOR : COLUMN_MAJOR;
  if (trans->status == OUT_OF_CORE) {
     trans->pfile->share->storage = srcPtr->major==COLUMN_MAJOR ? ROW_MAJOR : COLUMN_MAJOR;
     int tmp = trans->pfile->ocla_size[0];
     trans->pfile->ocla_size[0] = trans->pfile->ocla_size[1];
     trans->pfile->ocla_size[1] = tmp;
  }

  Remap schedule(ptr, trans, srcPtr->elementLen) ;
  schedule.execute(ptr->data, srcPtr->data) ;
  delete trans;

     // copy data back
  if (tmpPtr != ptr) {
	Copy<int> schedule(tmpPtr, ptr, ptr->elementLen) ;
        schedule.oc_execute(tmpPtr->data, ptr->data) ;
	free(newdata);
	delete ptr;
  }

    // change back
  if (srcPtr->status == OUT_OF_CORE) {
     srcPtr->pfile->share->storage = srcPtr->major;
     int tmp = srcPtr->pfile->ocla_size[0];
     srcPtr->pfile->ocla_size[0] = srcPtr->pfile->ocla_size[1];
     srcPtr->pfile->ocla_size[1] = tmp;
  }
}

//*************************************************************************//

template<class S, class T, class U>
inline void  pcrc_temp_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2,S type1,T type2,U type3)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  S res=0;

  if(!pcrc_same_dad_check(dad1, dad2)) {
   printf("different shape!\n");
    void* _data = new char [ptr2->elementLen * ptr2->size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *ptr1) ;
    tmpptr->elementLen = ptr2->elementLen ;
    tmpptr->elementType = ptr2->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (ptr1->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, ptr2, ptr2->elementLen) ;
    schedule1.execute(tmpptr->data, ptr2->data) ;

    reduceMode = 1;

    DotProduct<S, T, U> schedule(ptr1, tmpptr) ;
    schedule.execute(&res, (T *)ptr1->data, (U *)tmpptr->data) ;

    reduceMode = 0;
    delete tmpptr ;
  }
  else {
   printf("same shape!\n");
    reduceMode = 1;
    DotProduct<S, T, U> schedule(ptr1, ptr2) ;
    schedule.execute(&res, (T *)ptr1->data, (U *)ptr2->data) ;
    reduceMode = 0;
  }

  scalar_assign(resPtr,res);
}

void  pcrc_temp_booldot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int res=0;

  if(!pcrc_same_dad_check(dad1, dad2)) {
    void* _data = new char [ptr2->elementLen * ptr2->size ] ; 
    pcrcfDAD* tmpptr = new pcrcfDAD(_data, *ptr1) ;
    tmpptr->elementLen = ptr2->elementLen ;
    tmpptr->elementType = ptr2->elementType ;
    int tmpDAD = ad_FromPointer(tmpptr);
    if (ptr1->status == OUT_OF_CORE) pcrc_oc_init(&tmpDAD,&(tmpptr->icla_size));

    Remap schedule1(tmpptr, ptr2, ptr2->elementLen) ;
    schedule1.execute(tmpptr->data, ptr2->data) ;

    reduceMode = 1;
 
    BoolDotProduct schedule(ptr1, tmpptr) ;
    schedule.execute(&res, (int *)ptr1->data, (int *)tmpptr->data) ;

    reduceMode = 0;
    delete tmpptr ;
  }
  else {
    reduceMode = 1;
    BoolDotProduct schedule(ptr1, ptr2) ;
    schedule.execute(&res, (int *)ptr1->data, (int *)ptr2->data) ;
    reduceMode = 0;
  }

  scalar_assign(resPtr,res);
}

void  pcrc_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);

  switch (ptr1->elementType){
  case PCRC_INTEGER:{
      switch (ptr2->elementType){
       case PCRC_INTEGER:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(int)0,(int)0,(int)0) ;
         break ;
       }
       case PCRC_REAL:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(float)0,(int)0,(float)0) ;
         break ;
       }
       case PCRC_DOUBLE:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(double)0,(int)0,(double)0) ;
         break ;
       }
       case PCRC_COMPLEX:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),(int)0,COMPLEX(0,0)) ;
         break ;
       }
       default:
	 fprintf(stderr, "** RUNTIME ERROR ** cannot handle in DOT_PRODUCT.\n");
      }
      break ;
  }
  case PCRC_REAL:{
      switch (ptr2->elementType){
       case PCRC_INTEGER:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(float)0,(float)0,(int)0) ;
         break ;
       }
       case PCRC_REAL:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(float)0,(float)0,(float)0) ;
         break ;
       }
       case PCRC_DOUBLE:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(double)0,(float)0,(double)0) ;
         break ;
       }
       case PCRC_COMPLEX:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),(float)0,COMPLEX(0,0)) ;
         break ;
       }
       default:
	 fprintf(stderr, "** RUNTIME ERROR ** cannot handle in DOT_PRODUCT.\n");
      }
      break ;
  }
  case PCRC_DOUBLE:{
      switch (ptr2->elementType){
       case PCRC_INTEGER:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(double)0,(double)0,(int)0) ;
         break ;
       }
       case PCRC_REAL:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(double)0,(double)0,(float)0) ;
         break ;
       }
       case PCRC_DOUBLE:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,(double)0,(double)0,(double)0) ;
         break ;
       }
       case PCRC_COMPLEX:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),(double)0,COMPLEX(0,0)) ;
         break ;
       }
       default:
	 fprintf(stderr, "** RUNTIME ERROR ** cannot handle in DOT_PRODUCT.\n");
      }
      break ;
  }
  case PCRC_COMPLEX:{
      switch (ptr2->elementType){
       case PCRC_INTEGER:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),COMPLEX(0,0),(int)0) ;
         break ;
       }
       case PCRC_REAL:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),COMPLEX(0,0),(float)0) ;
         break ;
       }
       case PCRC_DOUBLE:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),COMPLEX(0,0),(double)0) ;
         break ;
       }
       case PCRC_COMPLEX:{
         pcrc_temp_dot_product(res0,dad_res,add1,dad1,add2,dad2,COMPLEX(0,0),COMPLEX(0,0),COMPLEX(0,0)) ;
         break ;
       }
       default:
	 fprintf(stderr, "** RUNTIME ERROR ** cannot handle in DOT_PRODUCT.\n");
      }
      break ;
  }
  case PCRC_LOGICAL:{
      switch (ptr2->elementType){
       case PCRC_LOGICAL:{
         pcrc_temp_booldot_product(res0,dad_res,add1,dad1,add2,dad2) ;
         break ;
       }
      }
      break ;
  }
  default: {
      printf("dot_product: unknown data type!\n");
  }
  } //end switch ptr1->elementType
}


// old code

/*
void  pcrc_int_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)

{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  int res;

  Section1<int>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr1->data);
  // sect2 must same data mapping as sect1
  Section1<int>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(int *)ptr2->data);
//  Section1<int>  sect2(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr2->data);

  res= dotProduct(sect1,sect2);
  printf("res=%d\n",res);

  scalar_assign(resPtr,res);
}

   void  pcrc_float_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  float res;
 
  switch (ptr1->elementType){

  case PCRC_INTEGER:{
       Section1<int>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr1->data);
       Section1<float>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(float *)ptr2->data);
       res= dotProduct(sect1,sect2);
       break;
                     }

  case PCRC_REAL:{
       Section1<float>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(float *)ptr1->data);
       switch (ptr2->elementType){
         case PCRC_INTEGER:{
              Section1<int>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(int *)ptr2->data);
              res= dotProduct(sect1,sect2);
			break;
                            }
         case PCRC_REAL:{
              Section1<float>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(float *)ptr2->data);
              res= dotProduct(sect1,sect2);
                        break;
                         }
      }
   }
 }

  scalar_assign(resPtr,res);
}

   void  pcrc_double_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  float res;

  switch (ptr1->elementType){

  case PCRC_INTEGER:{
       Section1<int>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr1->data);
       Section1<double>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(double *)ptr2->data);
       res= dotProduct(sect1,sect2);
                     }
  case PCRC_REAL:{
       Section1<float>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(float *)ptr1->data);
       Section1<double>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(double *)ptr2->data);
       res= dotProduct(sect1,sect2);
                     }
  case PCRC_DOUBLE:{
       Section1<double>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(double *)ptr1->data);

       switch (ptr2->elementType){
           case PCRC_INTEGER:{
                Section1<int>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(int *)ptr2->data);

                res= dotProduct(sect1,sect2);
                             }
           case PCRC_REAL:{
                Section1<float>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(float *)ptr2->data);
                res= dotProduct(sect1,sect2);
                          }
           case PCRC_DOUBLE:{
                Section1<double>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(double *)ptr2->data);
                res= dotProduct(sect1,sect2);
                     }
         }
      }
  }

  scalar_assign(resPtr,res);

}

   void  pcrc_complex_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  COMPLEX res;
 
  switch (ptr1->elementType){

  case PCRC_INTEGER:{
  	Section1<int>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr1->data);
  	Section1<COMPLEX>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(COMPLEX *)ptr2->data);
   	res= dotProduct(sect1,sect2);
                     }
  case PCRC_REAL:{
	  Section1<float>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(float *)ptr1->data);
	  Section1<COMPLEX>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(COMPLEX *)ptr2->data);
	   res= dotProduct(sect1,sect2);
                 }
  case PCRC_DOUBLE:{
  	Section1<double>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(double *)ptr1->data);
	  Section1<COMPLEX>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(COMPLEX *)ptr2->data);
	   res= dotProduct(sect1,sect2);
                     }
  case PCRC_COMPLEX:{
  	Section1<COMPLEX>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(COMPLEX *)ptr1->data);
  	switch (ptr2->elementType){
	case PCRC_INTEGER:{
		Section1<int>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(int *)ptr2->data);
		res= dotProduct(sect1,sect2);
                     }
  	case PCRC_REAL:{
  		Section1<float>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(float *)ptr2->data);
   		res= dotProduct(sect1,sect2);
                     }
  	case PCRC_DOUBLE:{
  		Section1<double>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(double *)ptr2->data);
   		res= dotProduct(sect1,sect2);
                     }
  	case PCRC_COMPLEX:{
  		Section1<COMPLEX>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(COMPLEX *)ptr2->data);
   		res= dotProduct(sect1,sect2);
                     }
  	}
        }
  }
  scalar_assign(resPtr,res);

}

   void  pcrc_bool_dot_product(void * res0,int *dad_res,int *add1,int *dad1,int * add2,int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1);
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2);
  pcrcfDAD* resPtr    = (pcrcfDAD*) ad_ToPointer(*dad_res);
  COMPLEX res;

  Section1<int>  sect1(ptr1->ranges[0],ptr1->group,ptr1->strides[0],(int *)ptr1->data);
  Section1<int>  sect2(ptr2->ranges[0],ptr2->group,ptr2->strides[0],(int *)ptr2->data);

  res= boolDotProduct(sect1,sect2);
  scalar_assign(resPtr,res);
}

*/

