#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>

#define PASSION_CCPLUSPLUS
#include "passion_base.h"

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"
#include "transformations.h"

#include "extname.h"

#ifdef PROFILELIB
        #define _RUNTIME
        #include "headers.h"
#endif

//#define NEED_ADDITIONAL_BUFFER

#ifdef NEED_ADDITIONAL_BUFFER
#define ADD_BUFFER_SIZE  1024*1024
#endif

  //  Zheng Gengbin
extern int runtimeDebug;
extern int storageMode;
extern int defaultStorageMode;

extern "C" {
   int c_ad_FromPointer(const void* ptr) ;   // for c interface
   void *c_ad_ToPointer(int idx) ;
   void c_ad_RmPointer(int idx) ;

   void pcrc_init(int *storage, int *debugFlag) ;
   void pcrc_finalize() ;

   int pcrc_new_group_grid(int* rank, int* shape) ;
   int pcrc_sub_group_grid(int* rank, int* l, int* u, int* s, int* grp_grid);
   void pcrc_delete_group(int* grp) ;

   int pcrc_new_range_distribute(int* g_lb, int* g_ub,
                                  int* distribution, int* grp_grid, int *dim) ;
//wjp add 1999.4.8
   int pcrc_gen_range_distribute(int* g_lb, int* g_ub, int* rank, int* index,
                                  int* distribution, int* grp_grid, int *dim) ;
   int pcrc_new_range_collapse(int* g_lb, int* g_ub) ;
   int pcrc_new_range_align(int* g_lb, int* g_ub,
                             int* offset, int* stride, int* rng_parent) ;
   int pcrc_new_range_copy(int* rng_parent) ;
   int pcrc_new_range_loop(int* g_lb, int* g_ub, int* g_stride,
                            int* offset, int* stride, int* rng_parent) ;
   void pcrc_delete_range(int* rng) ;

   int pcrc_new_array_data(void* data,
                        int* element_type, int* element_size,
                        int* rank, int* grp, int *major);
   void pcrc_set_array_distribute(int* dad, int* dim, int* g_lb, int* g_ub,
                                   int* distribution,
                                   int* grp_grid, int *grid_dim) ;
   void pcrc_set_array_collapse(int* dad, int* dim,
                                 int* g_lb, int* g_ub) ;
   void pcrc_set_array_align(int* dad, int* dim,
                              int* g_lb, int* g_ub,
                              int* offset, int* stride,
                              int* gl, int* gu,
                              int* rng_parent) ;
   void pcrc_set_array_copy(int* dad, int* dim, int* rng_parent) ;
   void pcrc_set_array_loop(int* dad, int* dim,
                             int* g_lb, int* g_ub, int* g_stride,
                             int* offset, int* stride, int* rng_parent) ;
   
   int pcrc_new_array_section(int* rank, int* dad_parent) ;
   void pcrc_set_array_triplet(int* dad, int* dim,
                                int* lb, int* ub, int* stride,
                                int* dad_parent, int* dim_parent) ;
   void pcrc_set_array_scalar(int* dad,
                               int* i,
                               int* dad_parent, int* dim_parent) ;

   int pcrc_new_array_copy(void* data, int* dad_parent) ;

   void pcrc_delete_array(int* dad) ;

   int pcrc_group(int* dad) ;
   int pcrc_range(int* dad, int* dim) ;
   int pcrc_stride(int* dad, int* dim) ;
   int pcrc_mem_stride(int* dad, int* dim) ;

   void pcrc_set_type(int *dad, int *type, int *len);
   int pcrc_local_size(int* dad) ;

   void pcrc_reset_array_base(int* dad, void* array, int *mh) ;
   void pcrc_reset_in_core_base(int* dad, void* data, int *mh) ;

   int pcrc_on(int* grp) ;

   int pcrc_glob(int* rng, int* l_i) ;
   int pcrc_loc(int* rng, int* g_i) ;

   void pcrc_loop_bounds(int* rng, int* l_lb, int* l_ub, int* l_stride) ;

   void pcrc_coef(int * dad, int *dim, int *l, int *u, int *s, int *a, int *b,
                   int *amount, int *m, int *c);
   void pcrc_offset(int *dad, int *dim, int *subscript, int *v);

   void pcrc_reset_push_bounds(int* dad, int* dim, int* lb, 
 			       int* lb_parent, int* ub_parent,
                                int* stride_parent) ;
   void pcrc_reset_pop_bounds(int* dad, int* dim,
                               int* lb, int* ub, int* stride) ;

   void pcrc_shift(int* dad, int* dad_source, int* dim, int* amount) ;
   void pcrc_tshift(void* data, int* dad_source, int* dim, int* amount) ;
   void pcrc_remap(int* dad, int* dad_source) ;
   void pcrc_remap_transpose(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7);
   void pcrc_gather_remap(int* dad, int* dad_source) ;
   void pcrc_scatter_remap(int* dad, int* dad_source) ;
   void pcrc_gather(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7,int *d8);
   void pcrc_scatter(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7,int *d8);
   void pcrc_write_halo(int* dad, int* wlo, int* whi) ;
//wjp add 1999.6.8
   void pcrc_write_back_halo(int* dad, int* wlo, int* whi) ;

   int pcrc_communication_detection(int* dad, int* dad_src,int *dim, 
                            int *l0, int *u0, int *s0, int *index0, 
			    int *l1, int *u1, int *s1, int *index1, int *amount);

   int pcrc_number_of_processors();
   void pcrc_proc_id();

   // dad query functions
   int pcrc_array_memdata(int *dad);

   int pcrc_same_data_mapping(int *dad1, int *dad2);
   int pcrc_same_dad_check(int *dad1, int *dad2);

   int pcrc_distcode(int *dad,int *dim);
   int pcrc_int_space_index(int *dad,int *addr2);
   int pcrc_double_space_index(int *dad,int *addr2);
   int pcrc_byte_space_index(int *dad,int *addr2);
   int pcrc_word_space_index(int *dad,int *addr2);

   // global environment
   void pcrc_set_global_major(int *major);

   void pcrc_set_io_process(int *io_grp);

// out of core interface
   int  pcrc_in_core_size(int *dad);
   void pcrc_oc_reset(int *dad);
   void pcrc_oc_read(int *dad, void *array, int *start, int *stride, int *number);
   void pcrc_oc_write(int *dad, void *array, int *start, int *stride, int *number);

// profile
   void pcrc_begin_prof(int *mode_);
   void pcrc_end_prof();
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Global state.                                                       //
//                                                                     //
// Go straight into the Fortran run-time library to pick up the values //
// of `argc' and `argv'.                                               //
//                                                                     //
// There doesn't seem to be a sensible, portable alternative           //
// (this is probably a bug in the MPI definition).                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

extern "C" {
  extern void pcrc_numargc(int *);
  extern void pcrc_args(int *,char *);
}

extern "C" {
  extern int pcrc_iargc();
  extern void pcrc_getarg(int *, char *);
}

static void getArgs(int &argidx, char *argv[])
{
    argidx = 0;
    char *q;
    int argcnt;
    
//    pcrc_numargc(&argcnt);
    argcnt = pcrc_iargc();
    while (argidx <= argcnt)
    {
        q = (char *)malloc(200);
        /* args Fortran subroutine */
        fflush(stdout);
//        pcrc_args(&argidx,q);
        pcrc_getarg(&argidx, q);
        argv[argidx] = q;
        q = (char *)index(argv[argidx],' ');
        *q = '\0';
        ++argidx;
    }
}

#ifdef SOLARIS

extern "C" {
  extern int __xargc ;
  extern char** __xargv ;
}

static void _AdlibInit() {
  AdlibInit(__xargc, __xargv) ;
}

#endif

#ifdef AIX

extern "C" {
  extern int p_xargc ;
  extern char** p_xargv ;
}

static void _AdlibInit() {
  AdlibInit(p_xargc, p_xargv) ;
}

/*
static void _AdlibInit() {
  int argidx;
  char * argv[200];
  char **argvp = argv;
  getArgs(argidx, argv);
  AdlibInit(argidx, argvp) ;
}
*/

#endif

#ifdef NAG

extern "C" {
  extern int f90_argc ;
  extern char** f90_argv ;
}

static void _AdlibInit() {
  AdlibInit(f90_argc, f90_argv) ;
}

#endif

#ifdef EPC

extern "C" {
  extern int xargc ;
  extern char** xargv ;
}

static void _AdlibInit() {
  AdlibInit(xargc, xargv) ;
}

#endif

#ifdef MIPS

extern "C" {
  extern int f77argc ;
  extern char** f77argv ;
}

static void _AdlibInit() {
  AdlibInit(f77argc, f77argv) ;
}

#endif

#ifdef DEC

extern "C" {
  extern int for__l_argc ;
  extern char** for__a_argv ;
}

static void _AdlibInit() {
  AdlibInit(for__l_argc, for__a_argv) ;
}

#endif

#ifdef F2C

extern "C" {
  extern int xargc ;
  extern char** xargv ;
}

static void _AdlibInit() {
  AdlibInit(xargc, xargv) ;
}

#endif

#ifdef HPUX

static void _AdlibInit() {
  int argidx;
  char * argv[200];
  char **argvp = argv;
  getArgs(argidx, argv);
  AdlibInit(argidx, argvp) ;
}

#endif

// more generally
#ifdef UNIX

static void _AdlibInit() {
  int argidx;
  char * argv[200];
  char **argvp = argv;
  getArgs(argidx, argv);
  AdlibInit(argidx, argvp) ;
}

#endif

static Group* rootPtr ;


// static int ad_FromPointer(const void* ptr) ;		zheng.gb
int ad_FromPointer(const void* ptr) ;
static Procs* getProcs(int rank, int shape [], int sub_ids [] = 0) ;

#ifdef NEED_ADDITIONAL_BUFFER
void *userBuf;
#endif

void pcrc_init(int *storage, int *debugFlag) {
  _AdlibInit() ;

  ad_FromPointer(&apg) ;     // should return 1 as global 

  ad_FromPointer(new Group(*getProcs(0, (int*) 0, (int*) 0))) ;  // should return 2 as root processor

  ad_FromPointer(new Group(*getProcs(0, (int*) 0, (int*) 0))) ;  // should return 3 as IO processor

  #ifdef PROFILELIB
  {
        int nProcID=AdlibLocId(), nProcNum=AdlibNProcs();
        ProfBegin(nProcID, nProcNum);
  }
  #endif

  runtimeDebug = *debugFlag;
  defaultStorageMode = *storage;
  reduceMode = 0;//wjp add

  ProfInit();

    // try to add more Buffer
#ifdef NEED_ADDITIONAL_BUFFER
  int buffersize;
  MPI_Pack_size(ADD_BUFFER_SIZE, MPI_BYTE, MPI_COMM_WORLD, &buffersize);
  buffersize += MPI_BSEND_OVERHEAD;
  userBuf = malloc(buffersize);
  MPI_Buffer_attach(userBuf, buffersize);
#endif
}

void pcrc_finalize() {
  #ifdef PROFILELIB
  {
        int nProcID=AdlibLocId(), nProcNum=AdlibNProcs();
        ProfBegin(nProcID, nProcNum);
  }
  #endif

    // delete all temp LAFs
  LAF_close_all();

    // free up additional buffer
#ifdef NEED_ADDITIONAL_BUFFER
  free(userBuf);
#endif

  AdlibFinalize() ;
}

void pcrc_begin_prof(int *mode_)
{
	ProfBegin(*mode_);
}

void pcrc_end_prof()
{
	ProfEnd();
}

// add by Zheng Gengbin
int pcrc_number_of_processors()
{
	return AdlibNProcs();
}


#include "detect_comm.h"
inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//  Routines to convert to and from pointers.                          //
//                                                                     //
//  This code is lifted, with minimal modification, from               //
//  `mpich/src/util/ptrcvt.c'.  It is a dirty fix for the fact that    //
//  pointers and int's are different lengths on the DEC Alpha.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Ad_PtrToIdx {
  int idx;
  const void *ptr;
  Ad_PtrToIdx *next;
} ;

#define MAX_PTRS 10000

static Ad_PtrToIdx PtrArray[MAX_PTRS];
static Ad_PtrToIdx *avail=0;
static int      DoInit = 1;

static void ad_InitPointer() {
  int  i;
  for (i=0; i<MAX_PTRS-1; i++) {
    PtrArray[i].next = PtrArray + i + 1;
    PtrArray[i].idx  = i;
  }
  PtrArray[MAX_PTRS-1].next = 0;
  /* Don't start with the first one, whose index is 0. That could
     break some code. */
  avail   = PtrArray + 1;
}

// static 	zheng.gb
const void *ad_ToPointer(int idx) {
  if (DoInit) {
    DoInit = 0;
    ad_InitPointer();
  }
  if (idx < 0 || idx >= MAX_PTRS) {
    fprintf(stderr, "** RUNTIME ERROR **  "
            "Could not convert index %d into a pointer\n", idx );
    exit(1);
  }
  if (idx == 0) return (void *)0;
  return PtrArray[idx].ptr;
}

int ad_FromPointer(const void* ptr) {
  int      idx;
  Ad_PtrToIdx *newP;
  if(DoInit) {
    DoInit = 0;
    ad_InitPointer();
  }
  if(!ptr) return 0;
  if(avail) {
    newP	      = avail;
    avail	      = avail->next;
    newP->next	      = 0;
    idx		      = newP->idx;
    PtrArray[idx].ptr = ptr;
    return idx;
  }
  // This isn't the right thing to do, but it isn't too bad
  fprintf(stderr, "** RUNTIME ERROR **  Pointer conversions exhausted\n" );
  exit(1);
  // Some systems may complain here about the lack of a return.
  return 0;
}

void ad_RmPointer(int idx) {
  int myrank;
  if (DoInit) {
    DoInit = 0;
    ad_InitPointer();
    }
  if(idx < 0 || idx >= MAX_PTRS) {
    fprintf(stderr, "** RUNTIME ERROR **  "
            "Could not convert index %d into a pointer\n", idx );
    exit(1);
    }
  if (idx == 0) return;
  if (PtrArray[idx].next) {
    /* In-use pointers NEVER have next set */
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    fprintf(stderr, "** RUNTIME ERROR **  "
	    "[%d] Error in recovering Fortran pointer; already freed\n", 
	    myrank );
    exit(1);
  }
  PtrArray[idx].next = avail;
  PtrArray[idx].ptr  = 0;
  avail              = PtrArray + idx;
}


// Zheng Gengbin
// for C interface
int c_ad_FromPointer(const void* ptr) {
   return ad_FromPointer(ptr);
}

void *c_ad_ToPointer(int idx) {
   return (void *)ad_ToPointer(idx);
}

void c_ad_RmPointer(int idx) {
   ad_RmPointer(idx);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Processor grid.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// `processArrays' table is used to ensure that if several process
// grids are defined with the same shape they refer to the same
// `Procs' structure.  It shouldn't be confused with the handle-to-pointer
// translation table `PtrArray'.

// Construction and destruction.

struct Ad_ProcessArrayRecord {
  int rank ;
  int shape [7] ;
  Procs* procs ;
  int* sub_ids ; //add by wjp
} ;

#define MAX_PROCESS_ARRAYS 100

static Ad_ProcessArrayRecord processArrays [MAX_PROCESS_ARRAYS] ;
static int numProcessArrays = 0 ;

// // sub_ids default is NULL
static Procs* getProcs(int rank, int shape [], int *sub_ids ) {

  // Scan `processArrays' table for pre-existing process array of the
  // same shape.  Return that if it exists.

  for(int slot = 0 ; slot < numProcessArrays ; slot++) {
    Ad_ProcessArrayRecord& processArray = processArrays [slot] ;
    if(rank == processArray.rank) {
       int i = 0 ;
       while (i < rank && shape [i] == processArray.shape [i]) i++ ;
//      if(i == rank) return processArray.procs;
       if(i == rank) {
          if(processArray.sub_ids && sub_ids) {
             int nu = 1;
             for(int j = 0; j < rank; j++) nu *= shape[j] ;
             int k = 0 ;
             while (k < nu && processArray.sub_ids[k] == sub_ids[k]) k++ ;
             if(k == nu) return processArray.procs;
          }
          else if (!processArray.sub_ids && !sub_ids)
		return processArray.procs;
//        AdlibPrintf("find processArray! slot=%d\n",slot);
       }
    }
  }

  // Otherwise create a new `Procs'

  if(numProcessArrays == MAX_PROCESS_ARRAYS) {
    fprintf(stderr, "** RUNTIME ERROR **  Process array records exhausted\n") ;
    exit(1) ;
    return 0 ;
  }
  else {
   int i ;
   if(sub_ids == NULL) {

    // Check size of processor array is number of available processors.

    int size = 1 ;
    for (i = 0 ; i < rank ; i++)
      size *= shape [i] ;

    int asize = apg.size() ;

    if(size != asize && size != 1) {
      fprintf(stderr, "** RUNTIME ERROR **  "
              "Processor array requires %d processes---only %d available\n",
              size, asize) ;
      exit(1) ;
    }
   }

    // Create new process array, and add it to `processArrays' table.

    Ad_ProcessArrayRecord& processArray = processArrays [numProcessArrays] ;

    processArray.rank = rank ;
    for (i = 0 ; i < rank ; i++) 
      processArray.shape [i] = shape [i] ;

    processArray.procs = new Procs(rank, shape, sub_ids) ;
//    AdlibPrintf("creat processArray! rank=%d\n",rank);
//    if (rank>0) AdlibPrintf("shape[0]=%d\n",shape[0]);
//wjp add
    processArray.sub_ids = sub_ids ;

    numProcessArrays++ ;

    return processArray.procs ;
  }
}

int pcrc_new_group_grid(int* rank, int* shape) {
  int handle =  ad_FromPointer(new Group(*getProcs(*rank, shape))) ;
//AdlibPrintf("procedure start!\n");

  #ifdef PROFILELIB
  {
    ProfNewGroupGrid(handle, *rank, shape);
  }
  #endif
 
  return handle;
}

int pcrc_sub_group_grid(int* rank, int* l, int* u, int* s, int* grp_grid) {
  Group* grpPtr = (Group*) ad_ToPointer(*grp_grid) ;

//check whether the parameters are available
  if(grpPtr->prc()->rnk() != *rank) {
      fprintf(stderr, "sub_group's RANK is different from parent_group's");
      exit(1) ;
  }

  if(grpPtr->check_sub(*rank, l, u)) {
      fprintf(stderr, "sub_group's parameters aren't available!");
      exit(1) ;
  }

  int* shape = new int [*rank] ;
  int* inde = new int [*rank] ;
  int num = 1;
  for(int i=0; i<*rank; i++) {
      num *= max((u[i]-l[i]+1+s[i]-1)/s[i], 0) ;
      inde[i] = 0;
      shape[i] = max((u[i]-l[i]+1+s[i]-1)/s[i], 0) ;
  }
  int* sub_ids = new int [num] ;

  int g_index = 0 ;
  int rk = 0 ;
  grpPtr->sub_id(*rank, &g_index, sub_ids, l, u, s, inde, &rk) ;

/*
AdlibPrintf("sub_id:");
for (int k=0; k<num; k++) printf("%d ",sub_ids[k]);
printf("\n");
*/
  int handle =  ad_FromPointer(new Group(*getProcs(*rank, shape, sub_ids))) ;

  #ifdef PROFILELIB
  {
    ProfNewGroupGrid(handle, *rank, shape);
  }
  #endif

  delete [] shape ;
  delete [] inde ;

  return handle;
}

void pcrc_delete_group(int* grp) {
  delete (Group*) ad_ToPointer(*grp) ;

  ad_RmPointer(*grp) ;
}

int pcrc_on(int* grp) {
  Group* p = (Group*) ad_ToPointer(*grp) ;
//AdlibPrintf("loop start\n");
  return p->member() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Initialization of ranges                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void setDistribute(Range& rng, const int _g_lb, const int _g_ub,
                   const int distribution, Group* grpPtr, const int dim) {

  Dimension d = grpPtr->prc()->dim(dim - 1) ;
  const int extent = _g_ub - _g_lb + 1 ;
  switch(distribution) {
    case 1 :
      rng = BlockRange(extent, d) ;
      break ;
    case 2 :
      rng = CyclicRange(extent, d) ;
      break ;
  }

  rng.attrib()   = distribution ; //wjp add 1999.5.9
  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

//wjp add 1999.4.8
void setGenDistribute(Range& rng, const int _g_lb, const int _g_ub, const int rank,
         int* index, const int distribution, Group* grpPtr, const int dim) {

  Dimension d = grpPtr->prc()->dim(dim - 1) ;
  const int extent = _g_ub - _g_lb + 1 ;
  rng = GenBlockRange(rank, index, extent, d) ;

  rng.attrib()   = distribution ; //wjp add 1999.5.9
  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setCollapse(Range& rng, const int _g_lb, const int _g_ub) {

  const int extent = _g_ub - _g_lb + 1 ;
  rng = Range(extent) ;

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setAlign(Range& rng, const int _g_lb, const int _g_ub,
              const int a, const int b, const int gl, const int gu,
              Range parent) {

  const int extent = _g_ub - _g_lb + 1 ;
  const int base   = (a * _g_lb + b - parent.g_lb()) / parent.g_stride() ;
  const int stride = a / parent.g_stride() ;

  if (stride == 0) printf("setAlign:stride = 0!!\n");	// zheng.gb

//printf("setAlign: a,b:%d,%d,base:%d,stride:%d\n",a,b,base,stride);

  if(gl == 0 && gu == 0) {
    rng = parent.subrng(extent, base, stride) ;
  }
  else {
    // Create range with SHADOW layout.

    const int parBas = parent.bas() ;
    const int parStr = parent.str() ;

    if (parent.show_attr() == 3) {
       GenBlockRange x(parent.proc_num(), parent.index_(), parent.tem().size(), parent.dim(), LAY_EXTENDED, gl, gu) ; 
       rng = x.subrng(extent, parBas + parStr * base, parStr * stride) ;
    }
    else
    {
       BlockRange x(parent.tem().size(), parent.dim(), LAY_EXTENDED, gl, gu) ;
       rng = x.subrng(extent, parBas + parStr * base, parStr * stride) ;
    }
  }

  // Zheng Gengbin  1998/2/2  ; make sure communication_detecion() will get correct tem().g_stride()
  rng.tem().g_lb() = parent.g_lb();

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = 1 ;
}

void setLoop(Range& rng, const int _g_lb, const int _g_ub, const int _g_stride,
             const int a, const int b, Range parent) {

  const int extent = (_g_ub - _g_lb) / _g_stride + 1 ;
  const int base   = (a * _g_lb + b - parent.g_lb()) / parent.g_stride() ;
  const int stride = a * _g_stride / parent.g_stride() ;
  rng = parent.subrng(extent, base, stride) ;

  rng.g_lb()     = _g_lb ;
  rng.g_stride() = _g_stride ;
}

void setTriplet(Range& rng, const int lb, const int ub, const int stride,
                Range parent) {

  int extent  = (ub - lb) / stride + 1 ;
  int base    = (lb - parent.g_lb()) / parent.g_stride() ;
  int stride_ = stride / parent.g_stride() ;
  rng = parent.subrng(extent, base, stride_) ;

  rng.g_lb()     = lb ;
  rng.g_stride() = stride ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Range creation.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_range_distribute(int* g_lb, int* g_ub,
                              int* distribution, int* grp_grid, int* dim) {

  Range* rng = new Range ;
  Group* grpPtr = (Group*) ad_ToPointer(*grp_grid) ;
  setDistribute(*rng, *g_lb, *g_ub, *distribution, grpPtr, *dim) ;

  return ad_FromPointer(rng) ;
}

//wjp add 1999.4.8
int pcrc_gen_range_distribute(int* g_lb, int* g_ub, int* rank, int* index,
                              int* distribution, int* grp_grid, int* dim) {

  Range* rng = new Range ;
  Group* grpPtr = (Group*) ad_ToPointer(*grp_grid) ;
  setGenDistribute(*rng, *g_lb, *g_ub, *rank, index, *distribution, grpPtr, *dim) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_collapse(int* g_lb, int* g_ub) {

  Range* rng = new Range ;
  setCollapse(*rng, *g_lb, *g_ub) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_align(int* g_lb, int* g_ub,
                          int* a, int* b, int* rng_parent) {
  Range* rng = new Range ;
  setAlign(*rng, *g_lb, *g_ub, *a, *b, 0, 0,
           *((Range*) ad_ToPointer(*rng_parent))) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_loop(int* g_lb, int* g_ub, int* g_stride,
                         int* a, int* b, int* rng_parent) {
  Range* rng = new Range ;
  setLoop(*rng, *g_lb, *g_ub, *g_stride, *a, *b,
          *((Range*) ad_ToPointer(*rng_parent))) ;

  return ad_FromPointer(rng) ;
}

int pcrc_new_range_copy(int* rng_parent) {
  Range* rng = new Range ;

  *rng = *((Range*) ad_ToPointer(*rng_parent)) ;

  return ad_FromPointer(rng) ;
}

void pcrc_delete_range(int* handle) {
  Range* rng = (Range*) ad_ToPointer(*handle) ;

  ad_RmPointer(*handle) ;
  delete rng ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class pcrcfDAD.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

pcrcfDAD :: pcrcfDAD(void* _data,
                     const int _elementType, const int _elementLen,
                     const int _rank, const Group& _group, int _major) :
                DAD(_rank, _group),
                elementLen(_elementLen), elementType(_elementType),
                size(1), data(_data)  {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;

    // for OUT-OF-CORE
  memdata = _data;
  memHandle = -1;
  data = _data;
  pfile = NULL;
  icla_size = 1;
  major = _major;

}


pcrcfDAD :: pcrcfDAD(const int _rank, pcrcfDAD& parent)  :
                DAD(_rank, parent.grp()),
                elementLen(parent.elementLen), elementType(parent.elementType),
                size(parent.size), data(parent.data)  {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
  {
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
// added by  zheng.gb
    rngs[r]=parent.rngs[r];
    strs[r]=parent.strs[r];
  }

    // add by Zheng Gengbin
    // for OUT-OF-CORE
  status = parent.status;
  major = parent.major;

  memdata = parent.memdata;
  memHandle = parent.memHandle;
  if (status == OUT_OF_CORE) {
	// we may change the headerSize & ocla array.
      pfile = CopyPfile(parent.pfile);
      pfile->share->refcount ++;
      data = (void *)pfile;
  }
  else {
      data = memdata;
  }
  icla_size=parent.icla_size;

}

pcrcfDAD :: pcrcfDAD(void* _data, pcrcfDAD& parent)  :
                DAD(parent),
                elementLen(parent.elementLen), elementType(parent.elementType),
                size(parent.size), memdata(_data)  {
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;
  grpHandle   = ad_FromPointer(&(group)) ;
  for(int r = 0 ; r < rank ; r++)
  {
    rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
// added by Zheng Gengbin
    rngs[r]=parent.rngs[r];
    strs[r]=parent.strs[r];
  }
  DAD :: ranges  = rngs ;
  DAD :: strides = strs ;

    // add by Zheng Gengbin
  if ( parent.rank == 0 ) status = IN_CORE;
  else status = parent.status;
  major = parent.major;

  memdata = _data;
  memHandle = -1;
  if (status == OUT_OF_CORE) {
      pfile = parent.pfile;
      pfile->share->refcount ++;
      data = (void *)parent.pfile;
  }
  else {
      data = memdata;
      pfile = NULL;
  }
  icla_size=parent.icla_size;

}

pcrcfDAD :: ~pcrcfDAD() {
  ad_RmPointer(grpHandle) ;
  for(int r = 0 ; r < rank ; r++)
    ad_RmPointer(rngHandles [r]) ;

    // add by Zheng Gengbin
    // clean temp file
  if (status == OUT_OF_CORE) {
    pfile->share->refcount --;
    if (pfile->share->refcount == 0)
       LAF_close(pfile);
  }
}

void pcrcfDAD :: addRange(const int r, Range x) {
  ranges [r] = x ;

  Stride g(x.stride(size)) ;
  strides [r] = g ;
// Zheng gengbin
  rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
  rngs[r] =  x;
  strs[r] = g;

   // for OUT-OF-CORE
//  if (status == IN_CORE) icla_size = size;

//wjp update for gen_block
  if (status == IN_CORE) icla_size = size<(STATIC_BUFFER_SIZE/4)?(STATIC_BUFFER_SIZE/4):size;
}

void pcrcfDAD :: addTriplet(const int r,
                            const int lb, const int ub, const int stride,
                            pcrcfDAD& parent, const int s) {
  Range x ;
  setTriplet(x, lb, ub, stride, parent.rng(s)) ;

  ranges  [r] = x ;
  strides [r] = parent.str(s) ;
// Zheng gengbin
//  rngHandles [r] = ad_FromPointer(&(ranges [r])) ;
  rngs[r] = x;
  strs[r] = parent.str(s);

  if (status == OUT_OF_CORE)
  {
	pfile->ocla_size[r] = parent.pfile->ocla_size[s];
  }
}

void pcrcfDAD :: addScalar(const int u, pcrcfDAD& parent, const int s) {
  Range y = parent.rng(s) ;
//AdlibPrintf("pcrcfDAD::addScalar:y()=%d\n",(u - y.g_lb()) / y.g_stride());
  Location i(y((u - y.g_lb()) / y.g_stride())) ;

//AdlibPrintf("pcrcfDAD::addScalar: u=%d,y.glb()=%d,y.g_stride()=%d\n",u,y.g_lb(),y.g_stride());
//AdlibPrintf("pcrcfDAD::addScalar: elementLen=%d,offset=%d\n",elementLen,parent.str(s).offset(i));
  if (status == IN_CORE) {
    (char*&) memdata += elementLen * parent.str(s).offset(i) ;
    data = memdata;
  }
  else {		//  for OUT_OF_CORE
    pfile->HeaderSize += elementLen * parent.str(s).offset(i) ;
  }
  group /= i ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Array creation.                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_data(void* data,
                        int* element_type, int* element_size,
                        int* rank, int* grp, int *major) {
  pcrcfDAD* ptr = new pcrcfDAD(data, *element_type, *element_size,
                     *rank, *((Group*) ad_ToPointer(*grp)), *major) ;

  return ad_FromPointer(ptr) ;
}

void pcrc_set_array_distribute(int* dad, int* dim, int* g_lb, int* g_ub,
                               int* distribution,
                               int* grp_grid, int *grid_dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setDistribute(x, *g_lb, *g_ub, *distribution,
                (Group*) ad_ToPointer(*grp_grid), *grid_dim) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_collapse(int* dad, int* dim, int* g_lb, int* g_ub) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setCollapse(x, *g_lb, *g_ub) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_align(int* dad, int* dim,
                          int* g_lb, int* g_ub,
                          int* a, int* b,
                          int* gl, int* gu,
                          int* rng_parent) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  Range x ;
  setAlign(x, *g_lb, *g_ub, *a, *b, *gl, *gu,
           *((Range*) ad_ToPointer(*rng_parent))) ;

  ptr->addRange(*dim - 1, x) ;
}

void pcrc_set_array_copy(int* dad, int* dim, int* rng_parent) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addRange(*dim - 1, *((Range*) ad_ToPointer(*rng_parent))) ;

}

void pcrc_delete_array(int* dad) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ad_RmPointer(*dad) ;

  delete ptr ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Section creation.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_section(int* rank, int* dad_parent) {
  pcrcfDAD* ptr = new pcrcfDAD(*rank, *((pcrcfDAD*) ad_ToPointer(*dad_parent))) ;

  return ad_FromPointer(ptr) ;
}


void pcrc_set_array_triplet(int* dad, int* dim,
			    int *lb, int *ub, int *stride,
			    int* dad_parent, int* dim_parent) {

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addTriplet(*dim - 1, *lb, *ub, *stride,
                  *((pcrcfDAD*) ad_ToPointer(*dad_parent)), *dim_parent - 1) ;
}

void pcrc_set_array_scalar(int* dad,
		           int *u,
		           int* dad_parent, int* dim_parent) {

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->addScalar(*u, *((pcrcfDAD*) ad_ToPointer(*dad_parent)), *dim_parent - 1) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Array copying.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_new_array_copy(void* data, int* dad_parent) {
  pcrcfDAD* ptr = new pcrcfDAD(data, *((pcrcfDAD*) ad_ToPointer(*dad_parent))) ;

  return ad_FromPointer(ptr) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Misc operations on arrays and query functions on DAD.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_group(int* dad) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->grpHandle ;
}

int pcrc_range(int* dad, int* dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->rngHandles [*dim - 1] ;
}

int pcrc_stride(int* dad, int* dim) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->str(*dim - 1).len() ;
}

//wjp add 1999-6-24
int pcrc_mem_stride(int* dad, int* dim) {

  const int r = *dim - 1;
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x = ptr->rng(r) ;

  int sub_stp ;
//  int count, sub, sub_stp ;
//  x.block(x.dim().crd(), &sub, &count) ;
////AdlibPrintf("coef:y.block(): %d,%d,%d\n",x.dim().crd(),sub,count);
  x.block_step(&sub_stp) ;

  Stride u = ptr->str(r) ;

  return u.disp_step(sub_stp) ;
}

  // set type        add by Zheng Gengbin
void pcrc_set_type(int *dad, int *type, int *len) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->elementType = *type;
  ptr->elementLen  = *len;
}

int  pcrc_local_size(int* dad) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->size ;
}

// return mem handle
int pcrc_array_memdata(int *dad) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  return ptr->memHandle ;
}

// called by new_array_section
void pcrc_reset_array_base(int* dad, void* data, int *mh) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

//  modified by Zheng Gengbin
//  ptr->data = data ;

  ptr->memdata = data ;
  ptr->memHandle = *mh;

  if (ptr->status == OUT_OF_CORE) {
      pcrc_oc_reset(dad);
  }
  else
      ptr->data = data ;
}

// called by new_array_data when used in temporary DAD with pcrc_oc_init.
void pcrc_reset_in_core_base(int* dad, void* data, int *mh) {
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;

  ptr->memdata = data ;
  ptr->memHandle = *mh;

  if (ptr->status == IN_CORE)
	ptr->data = data ;
}

void pcrc_reset_push_bounds(int* dad, int* dim, int* lb,
                            int* lb_parent, int* ub_parent,
                            int* stride_parent ) {

  // (Last argument now ignored.)

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x  = ptr->rng(*dim - 1) ;

  *lb_parent     = x.g_lb() ;
  *stride_parent = x.g_stride() ;

  x.g_lb()     = *lb ;
  x.g_stride() = 1 ;
}

void pcrc_reset_pop_bounds(int* dad, int* dim,
                           int* lb_parent, int* ub_parent, int* stride_parent) {

  // (Last argument now ignored.)

  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x  = ptr->rng(*dim - 1) ;

  x.g_lb()     = *lb_parent ;
  x.g_stride() = *stride_parent ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Address translation.                                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int pcrc_glob(int* rng, int* l_i) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;

//  printf("pcrc_glob: x.g_lb=%d,x.glb()=%d,x.g_stride=%d\n",x.g_lb(),x.glb(*l_i,x.dim().crd()),x.g_stride());
  return x.g_lb() + x.glb(*l_i, x.dim().crd()) * x.g_stride() ;
}

int pcrc_loc(int* rng, int* g_i) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;
  int l_i ;

  const int glb = (*g_i - x.g_lb()) / x.g_stride() ;
  x.ker_glb(glb, &l_i) ;
  return l_i ;
}

void pcrc_loop_bounds(int* rng, int* l_lb, int* l_ub, int* l_stride) {
  Range x = *((Range*) ad_ToPointer(*rng)) ;

  // Parameters of local block of `x'.

  int count ;

  x.block(x.dim().crd(), l_lb, &count) ;
//  AdlibPrintf("loop bound:x.block(): %d,%d,%d\n",x.dim().crd(),*l_lb,count);

  x.block_step(l_stride) ;
  *l_ub = *l_lb + *l_stride * (count - 1) ;

//  wjp addd
//  AdlibPrintf("loop bound:%d,%d,%d\n",*l_lb,*l_ub,*l_stride);

}

void pcrc_coef(int *dad, int *dim, int *l, int *k, int *s, int *a, int *b,
               int *amount, int *m, int *c) {

  const int r = *dim - 1;
  pcrcfDAD* ptr = (pcrcfDAD*) ad_ToPointer(*dad) ;
  Range x = ptr->rng(r) ;

  const int t_stride = x.str() / x.g_stride() ;

  // Assume an imaginary, shifted version of the data, moved `amount'
  // places to the left in the template.

  // The subrange `y' defines the section of the imaginary array
  // selected by the loop bounds.

  const int extent = (*k - *l) / *s + 1 ;
  const int base   = x.bas() - *amount + t_stride * (*a * *l + *b - x.g_lb()) ;
  const int stride = t_stride * *a * *s ;

  Range y = x.tem().subrng(extent, base, stride) ;

  // Parameters of local block of `y'.

  int count, sub, sub_stp ;
  y.block(y.dim().crd(), &sub, &count) ;
//AdlibPrintf("coef:y.block(): %d,%d,%d\n",y.dim().crd(),sub,count);
  y.block_step(&sub_stp) ;

  // Convert to local subscript.  Actual data comes from `amount'
  // places to the right of where it would be in the imaginary array
  // (ie, go into ghost regions of actual data).

  Stride u = ptr->str(r) ;

  *m = u.disp_step(sub_stp) ;
  *c = u.disp(sub) + u.len() * *amount ;
}


void pcrc_offset(int *dad, int *dim, int *subscript, int *v)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  const int r = *dim-1;
  Range x = ptr->rng(r);
  Location i(x((*subscript - x.g_lb()) / x.g_stride())) ;

  *v = ptr->str(r).offset(i);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Communication.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void pcrc_shift(int* dad, int* dad_source, int* dim, int* amount) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  Shift schedule(ptr, srcPtr, ptr->elementLen, *amount, *dim - 1, EDGE) ;
  schedule.execute(ptr->data, srcPtr->data) ;
}


void pcrc_tshift(void* data, int* dad_source, int* dim, int* amount) {

  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;
  int d = *dim - 1 ;

  // Create `DAD' record containing a primitive range in
  // place of the actual source array range.

  const int rank = srcPtr->rnk() ;
  Range  rngs [8] ;
  Stride strs [8] ;
  for(int r = 0 ; r < rank ; r++) {
    if(r == d)
      rngs [r] = srcPtr->rng(r).tem() ;
    else
      rngs [r] = srcPtr->rng(r) ;
    strs [r] = srcPtr->str(r) ;
  }
  DAD tmp(rank, srcPtr->grp(), rngs, strs) ;

  // Do the shift

  Shift schedule(&tmp, &tmp, srcPtr->elementLen, *amount, d, EDGE) ;
  schedule.execute(data, srcPtr->data) ;
}


void pcrc_remap(int* dad, int* dad_source) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

  #ifdef PROFILELIB
  {
        ProfBeginRemap();
  }
  #endif

  ProfBeginRemap();

  int defaultSave = -1;
  if (ptr->major == srcPtr->major && defaultStorageMode != ptr->major) {
	defaultSave = defaultStorageMode;
	defaultStorageMode = ptr->major;
  }

  Remap schedule(ptr, srcPtr, srcPtr->elementLen) ;
  schedule.execute(ptr->data, srcPtr->data) ;

  if (defaultSave != -1) defaultStorageMode = defaultSave;

  ProfEndRemap();

  #ifdef PROFILELIB
  {
        ProfEndRemap();
  }
  #endif
}

void pcrc_remap_transpose(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

/*
  if (ptr->elementType != srcPtr->elementType) {
    fprintf(stderr, "** RUNTIME ERROR **  Remap type mismatched. \n");
    return;
  }
*/

  int no=0;
  int dim[7];
  if (*d1!=-1) dim[no++]=*d1;
  if (no == 1 && *d2!=-1) dim[no++]=*d2;
  if (no == 2 && *d3!=-1) dim[no++]=*d3;
  if (no == 3 && *d4!=-1) dim[no++]=*d4;
  if (no == 4 && *d5!=-1) dim[no++]=*d5;
  if (no == 5 && *d6!=-1) dim[no++]=*d6;
  if (no == 6 && *d7!=-1) dim[no++]=*d7;

  pcrcfDAD *trans= new pcrcfDAD(srcPtr->memdata,*srcPtr);  // for out of core
  for (int i=0; i<no; i++)
  {
        trans->rngs[i]=srcPtr->rngs[dim[i]];
        trans->strs[i]=srcPtr->strs[dim[i]];
  }

  int defaultSave = -1;
  if (ptr->major == srcPtr->major && defaultStorageMode != ptr->major) {
	defaultSave = defaultStorageMode;
	defaultStorageMode = ptr->major;
  }

  Remap schedule(ptr, trans, srcPtr->elementLen) ;
  schedule.execute(ptr->data, srcPtr->data) ;

  if (defaultSave != -1) defaultStorageMode = defaultSave;

  delete trans;
}

void pcrc_scatter_remap(int* dad, int* dad_source) {
  pcrc_remap(dad, dad_source);
}

void pcrc_gather_remap(int* dad, int* dad_source) {
extern int InverseRemp;
  InverseRemp = 1;
  pcrc_remap(dad, dad_source);
  InverseRemp = 0;
}

// add by Zheng Gengbin
// interface for Gather
void pcrc_gather(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7,int *d8) 
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

//	gather(ptr,srcPtr,ptr1);
  switch (srcPtr->rank) {
  case 1:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
	Gather schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq1<const DAD*>(ptr1), Seq1<int*>((int*)ptr1->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  case 2:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
  	pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*d2) ;
	Gather schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq2<const DAD*>(ptr1,ptr2), Seq2<int*>((int*)ptr1->data,(int*)ptr2->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  case 3:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
  	pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*d2) ;
  	pcrcfDAD* ptr3    = (pcrcfDAD*) ad_ToPointer(*d3) ;
	Gather schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq3<const DAD*>(ptr1,ptr2,ptr3), Seq3<int*>((int*)ptr1->data,(int*)ptr2->data,(int*)ptr3->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  }
}

// add by Zheng Gengbin
// interface for Scatter
void pcrc_scatter(int* dad, int* dad_source,int *d1,int *d2,int *d3,int *d4,int *d5,int *d6,int *d7,int *d8) 
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  pcrcfDAD* srcPtr = (pcrcfDAD*) ad_ToPointer(*dad_source) ;

//	scatter(ptr,srcPtr,ptr1);
  switch (srcPtr->rank) {
  case 1:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
	Scatter schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq1<const DAD*>(ptr1), Seq1<int*>((int*)ptr1->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  case 2:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
  	pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*d2) ;
	Scatter schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq2<const DAD*>(ptr1,ptr2), Seq2<int*>((int*)ptr1->data,(int*)ptr2->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  case 3:  {
  	pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*d1) ;
  	pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*d2) ;
  	pcrcfDAD* ptr3    = (pcrcfDAD*) ad_ToPointer(*d3) ;
	Scatter schedule(ptr, srcPtr, srcPtr->elementLen,
                  Seq3<const DAD*>(ptr1,ptr2,ptr3), Seq3<int*>((int*)ptr1->data,(int*)ptr2->data,(int*)ptr3->data),
                  0, 0) ;
  	schedule.execute(ptr->data, srcPtr->data) ;
	break;
	}
  }
}

void pcrc_write_halo(int* dad, int* wlo, int* whi) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  const int rank = ptr->rnk() ;
//AdlibPrintf("pcrc_write_halo: rank=%d\n",rank);
  int* w = new int [2 * rank] ;

  int* wloAdj = w ; 
  int* whiAdj = w + rank ; 

  for(int r = 0 ; r < rank ; r++) {
    const int str = ptr->rng(r).str() ;
    wloAdj [r] = (wlo [r] + str - 1) / str ;
    whiAdj [r] = (whi [r] + str - 1) / str ;
//AdlibPrintf("pcrc_write_halo: wloAdj[,%d,]=%d\n",r,wloAdj[r]);
//AdlibPrintf("pcrc_write_halo: whiAdj[,%d,]=%d\n",r,whiAdj[r]);
  }

  WriteHalo schedule(ptr, ptr->elementLen, wloAdj, whiAdj) ;

  schedule.execute(ptr->data) ;

  delete [] w ;
}

//wjp add 1999.6.8 for write back shadows to their original locations
void pcrc_write_back_halo(int* dad, int* wlo, int* whi) {
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

  const int rank = ptr->rnk() ;
//AdlibPrintf("pcrc_write_back_halo: rank=%d\n",rank);
  int* w = new int [2 * rank] ;

  int* wloAdj = w ; 
  int* whiAdj = w + rank ; 

  for(int r = 0 ; r < rank ; r++) {
    const int str = ptr->rng(r).str() ;
    wloAdj [r] = (wlo [r] + str - 1) / str ;
    whiAdj [r] = (whi [r] + str - 1) / str ;
//AdlibPrintf("pcrc_write_back_halo: wloAdj[,%d,]=%d\n",r,wloAdj[r]);
//AdlibPrintf("pcrc_write_back_halo: whiAdj[,%d,]=%d\n",r,whiAdj[r]);
  }

  WriteBackHalo schedule(ptr, ptr->elementLen, wloAdj, whiAdj) ;

  schedule.execute(ptr->data) ;

  delete [] w ;
}

int pcrc_distcode(int *dad,int *dim)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  if (*dim > ptr->rank || *dim < 1) printf("pcrc_rng_distcode:Out of rank.\n");
  //Range x= ptr->rngs[*dim-1];
  switch (ptr->rngs[*dim-1].tem().format()) {
  case DIST_COLLAPSED:
	return 0;
  case DIST_BLOCK:
	return 1;
  case DIST_CYCLIC:
	return 2;
  }
  return -1;
}

// calculate the address distance of dad to addr2
int pcrc_byte_space_index(int *dad,int *addr2)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  unsigned char *data = (unsigned char*)(ptr->memdata);
  return (int)(data-(unsigned char *)addr2+1);
}

// calculate the address distance of dad to addr2
int pcrc_word_space_index(int *dad,int *addr2)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  short *data = (short *)(ptr->memdata);
  return (int)(data-(short *)addr2+1);
}

// calculate the address distance of dad to addr2
int pcrc_int_space_index(int *dad,int *addr2)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  int *data = (int*)(ptr->memdata);
  return (int)(data-addr2+1);
}

// calculate the address distance of dad to addr2
int pcrc_double_space_index(int *dad,int *addr2)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  double *data = (double*)(ptr->memdata);
  double *data1 = (double*)(addr2);
  return (int)(data-data1+1);
}

/*****************************************************************************/
/*	Runtime support for creating new derived type.                       */
/*****************************************************************************/
#if 0
int pcrc_new_type_struct(int *len, int *blocklen, int *type)
{
}
#endif

/*****************************************************************************/

// add by Zheng Gengbin
// compare if two array is same data mapping.
int pcrc_same_data_mapping(int *dad1, int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1) ;
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2) ;

  const int debug = 0;

  if (ptr1->elementType != ptr2->elementType) 
  { 
    if (debug && AdlibLocId() == 0) printf("different data type.\n"); return 0; 
  }
  if (ptr1->rank != ptr2->rank) 
  { 
    if (debug && AdlibLocId() == 0) printf("different rank.\n"); return 0;
  }

  // same processors group
//  if (ptr1->grpHandle != ptr2->grpHandle) return 0;

  // same ranges
  for (int i=0; i<ptr1->rank ; i++) {
  	if (ptr1->rngs[i].size() != ptr2->rngs[i].size()) 
	{
	  if (debug && AdlibLocId() == 0) printf("different size.\n");return 0;
	}
  	if (ptr1->rngs[i].tem().format() != ptr2->rngs[i].tem().format()) {
	  if (debug) printf("different dist code.\n");return 0;
	  return 0;
	}
	if (ptr1->rngs[i].tem().format() != DIST_COLLAPSED) 
	{
  	  if (ptr1->rngs[i].tem().size() != ptr2->rngs[i].tem().size()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different tem size:%d!=%d.\n",ptr1->rngs[i].tem().size(),ptr2->rngs[i].tem().size());return 0;
	  }
  	  if (ptr1->rngs[i].str() != ptr2->rngs[i].str()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different str.\n");return 0;
	  }
  	  if (ptr1->rngs[i].bas() != ptr2->rngs[i].bas()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different bas:%d!=%d.\n",ptr1->rngs[i].bas(),ptr2->rngs[i].bas());return 0;
	  }
	}
  	if (ptr1->rngs[i].dim() != ptr2->rngs[i].dim()) return 0;
  }
  if (debug && AdlibLocId() == 0) printf("same data mapping .\n");
  return 1;
}

//add wjp
int pcrc_same_dad_check(int *dad1, int *dad2)
{
  pcrcfDAD* ptr1    = (pcrcfDAD*) ad_ToPointer(*dad1) ;
  pcrcfDAD* ptr2    = (pcrcfDAD*) ad_ToPointer(*dad2) ;

  const int debug = 0;

  if (ptr1->status != ptr2->status) 
  { 
    if (debug && AdlibLocId() == 0) printf("different status.\n"); return 0; 
  }
  if (ptr1->rank != ptr2->rank) 
  { 
    if (debug && AdlibLocId() == 0) printf("different rank.\n"); return 0;
  }

  // same processors group
//  if (ptr1->grpHandle != ptr2->grpHandle) return 0;

  // same ranges
  for (int i=0; i<ptr1->rank ; i++) {
  	if (ptr1->rngs[i].size() != ptr2->rngs[i].size()) 
	{
	  if (debug && AdlibLocId() == 0) printf("different size.\n");return 0;
	}
  	if (ptr1->rngs[i].tem().format() != ptr2->rngs[i].tem().format()) {
	  if (debug) printf("different dist code.\n");return 0;
	  return 0;
	}
	if (ptr1->rngs[i].tem().format() != DIST_COLLAPSED) 
	{
  	  if (ptr1->rngs[i].wlo() != 0 || ptr1->rngs[i].whi() != 0 || ptr2->rngs[i].wlo() != 0 || ptr2->rngs[i].whi() != 0) 
	  {
		if (debug && AdlibLocId() == 0) printf(":existed shadow.\n");return 0;
	  }
  	  if (ptr1->rngs[i].tem().size() != ptr2->rngs[i].tem().size()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different tem size:%d!=%d.\n",ptr1->rngs[i].tem().size(),ptr2->rngs[i].tem().size());return 0;
	  }
  	  if (ptr1->rngs[i].str() != ptr2->rngs[i].str()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different str.\n");return 0;
	  }
  	  if (ptr1->rngs[i].bas() != ptr2->rngs[i].bas()) 
	  {
		if (debug && AdlibLocId() == 0) printf("different bas:%d!=%d.\n",ptr1->rngs[i].bas(),ptr2->rngs[i].bas());return 0;
	  }
	}
  	if (ptr1->rngs[i].dim() != ptr2->rngs[i].dim()) return 0;
  }
  if (debug && AdlibLocId() == 0) printf("same data mapping .\n");
  return 1;
}


//#include "detect_comm.h"
//inline int max(int x, int y) {return x > y ? x : y ; }
//inline int min(int x, int y) {return x < y ? x : y ; }

// 0: NONE 1: SHIFT 2: REMAP
int pcrc_communication_detection(int* dad, int* dad_src,int *dim, 
                            int *l0, int *u0, int *s0, int *index0,
			    int *l1, int *u1, int *s1, int *index1, int *amount) 
{
  int pcrc_detect_communication(detect_DIM* dad_x, detect_DIM* dad_y,
                              int* l0, int* u0, int* s0, 
			      int* l1, int* u1, int* s1, int *amount);
  int i,j;
  int detect_result=-1;
  pcrcfDAD* lhsDad    = (pcrcfDAD*)ad_ToPointer(*dad);
  pcrcfDAD* rhsDad    = (pcrcfDAD*)ad_ToPointer(*dad_src);
  int rank = rhsDad->rank;
  detect_DIM dad_x,dad_y;
  if (lhsDad->grpHandle != lhsDad->grpHandle) return 2;
/*
  if (lhsDad->status == IN_CORE && rhsDad->status == OUT_OF_CORE) return 2;
  if (rhsDad->status == IN_CORE && lhsDad->status == OUT_OF_CORE) return 2;
*/
  for (i=0; i<rank; i++) amount[i]=0;
  for (i=0;  i<lhsDad->group.prc()->rnk(); i++ )
  {
	Dimension  adim = lhsDad->group.prc()->dim(i);
	int lpd=-1, rpd=-1;
	for (j=0; j<lhsDad->rank ; j++)
		if (lhsDad->rngs[j].dim() == adim) { lpd = j; break;}
	for (j=0; j<rhsDad->rank ; j++)
		if (rhsDad->rngs[j].dim() == adim) { rpd = j; break;}
//if (debug)
//printf("dim:%d, lpd:%d,rpd:%d\n",i,lpd,rpd);
	if (rpd == -1) { detect_result = 0; goto result;}
	if (lpd == -1) { detect_result = 2; goto result;}

	if (index0[lpd]!=index1[rpd]) { detect_result=2; break; }

	// fill DIM lhs
  	dad_x.numProcs = lhsDad->rngs[lpd].dim().size();
  	switch (lhsDad->rngs[lpd].format()) 
  	{
  	case DIST_COLLAPSED: dad_x.distCode=0; break;
  	case DIST_BLOCK:     dad_x.distCode=1; break;
  	case DIST_CYCLIC:    dad_x.distCode=2; break;
  	}
  	dad_x.t_lb = lhsDad->rngs[lpd].tem().g_lb();
  	dad_x.t_size = lhsDad->rngs[lpd].tem().size() ;
  	dad_x.t_stride = lhsDad->rngs[lpd].str()/lhsDad->rngs[lpd].g_stride();
  	dad_x.t_offset = dad_x.t_lb+lhsDad->rngs[lpd].bas()-dad_x.t_stride*lhsDad->rngs[lpd].g_lb();
//if (AdlibLocId() == 0)
//printf("lhs:dim:%d,distCode:%d,t_stride:%d,t_offset:%d,t_lb:%d,t_size:%d\n",lpd,dad_x.distCode,dad_x.t_stride,dad_x.t_offset,dad_x.t_lb,dad_x.t_size);
	// fill DIM rhs
  	dad_y.numProcs = rhsDad->rngs[rpd].dim().size();
  	switch (rhsDad->rngs[rpd].format()) 
  	{
  	case DIST_COLLAPSED: dad_y.distCode=0; break;
  	case DIST_BLOCK:     dad_y.distCode=1; break;
  	case DIST_CYCLIC:    dad_y.distCode=2; break;
  	}
  	dad_y.t_lb = rhsDad->rngs[rpd].tem().g_lb();
  	dad_y.t_size = rhsDad->rngs[rpd].tem().size();
  	dad_y.t_stride = rhsDad->rngs[rpd].str()/rhsDad->rngs[rpd].g_stride();
  	dad_y.t_offset = dad_y.t_lb+rhsDad->rngs[rpd].bas()-dad_y.t_stride*rhsDad->rngs[rpd].g_lb();
//if (AdlibLocId() == 0)
//printf("rhs:dim:%d,distCode:%d,t_stride:%d,t_offset:%d,t_lb:%d,t_size:%d\n",rpd,dad_y.distCode,dad_y.t_stride,dad_y.t_offset,dad_y.t_lb,dad_y.t_size);

//if (AdlibLocId() == 0)
//printf("index:%d,l0:%d,u0:%d,s0:%d,l1:%d,u1:%d,s1:%d\n",index0[lpd],l0[lpd],u0[lpd],s0[lpd],l1[rpd],u1[rpd],s1[rpd]);
	int result = pcrc_detect_communication(&dad_x,&dad_y,
                             &l0[lpd],&u0[lpd],&s0[lpd],
			     &l1[rpd],&u1[rpd],&s1[rpd],&amount[rpd]);
	switch (result) {
        case 0:
		if (detect_result != 1) detect_result=0;
		break;
        case 1:
		detect_result=1;
		break;
        case 2:
		detect_result=2;
		goto result;
	}
  }

    // check if overflow
  if (detect_result == 1) {
    for ( i =0; i < rank; i++) {
      if (amount[i] > 0) {
	if (amount[i] > rhsDad->rngs[i].whi())
		detect_result = 2;
      }
      if (amount[i] < 0) {
	if (-amount[i] > rhsDad->rngs[i].wlo())
		detect_result = 2;
      }
      if ( detect_result == 2 ) {
	if (runtimeDebug == 1) 
	   AdlibPrintf("Shift reset as Remap, due to the overflow of GhostArea(%d needed).\n",amount[i]);
        break;
      }
    }
  }

    // remap cyclic  ???
  if (detect_result == 1) {
	for (i=0; i<rank; i++)
	  if (amount[i]!=0 && rhsDad->rngs[i].format() == DIST_CYCLIC) { 
		detect_result = 2; 
	        if (runtimeDebug == 1) AdlibPrintf("Shift reset as Remap, due to CYCLIC data distribution.\n");
		break; 
	  }
  }
    // data dependence checking
  if ( lhsDad == rhsDad && detect_result != 2 ) {
    int depend = 0;
    for (i=0; i<lhsDad->rank; i++ )
    {
	if (index0[i] != index1[i]) printf("Why???\n");

	int ll0 ,uu0, ll1, uu1;
	if (s0[i] >= 0) { ll0 = l0[i]; uu0 = u0[i]; }
	else { ll0 = u0[i]; uu0 = l0[i]; }
	if (s1[i] >= 0) { ll1 = l1[i]; uu1 = u1[i]; }
	else { ll1 = u1[i]; uu1 = l1[i]; }

	  // begin to check if not verlapped
	if ( min(uu0,uu1) <= max(ll0,ll1) ) 
	   { depend = 0; break;}
	if (l0[i] == l1[i] && u0[i] == u1[i] && s0[i] * s1[i] > 0) { continue; }
	if ( s0[i] * s1[i] < 0 ) { depend =1; continue;}
	if (s0[i]>0 && l1[i] > l0[i]) { continue;}   // anti-dependence ??
	if (s0[i]<0 && l0[i] > l1[i]) { continue;}   // anti-dependence ??
	depend = 1; 
    } // end of for
    if (depend == 1) {
	detect_result=2;
if (AdlibLocId() == 0)
printf("pcrc_communication_detection:Maybe some data dependence.!\n");
    }
  }

  
result:
/*
printf("amount:");
for (i=0; i<rank; i++ ) printf("%d ",amount[i]);
printf("\n");
*/
/*
if (AdlibLocId() == 0)
printf ("detect_result:%d\n",detect_result);
*/
  return detect_result;
}


int pcrc_in_core_size(int *dad)
{
  pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
  return (ptr->status == OUT_OF_CORE)?  ptr->icla_size: ptr->size;
}

// allow node program change default storageMode temporarily
void pcrc_set_global_major(int *major)
{
  static int oldStorageMode = -1;
  if (oldStorageMode == -1) {	// set to new major mode temporarily
     oldStorageMode = defaultStorageMode;
     defaultStorageMode = *major;
  }
  else {	// recover to old major mode
     defaultStorageMode = oldStorageMode;
     oldStorageMode = -1;
  }
}

void pcrc_set_io_process(int *io_grp)
{
  Group* grpPtr = (Group*) ad_ToPointer(*io_grp) ;
  PtrArray[3].ptr = grpPtr;
}

//        For Intrinsic functions
//#include "intrinsic.C"

// -------------------------------------------------------------
// Things to do
//
//   Delete `tshift_', and remove the `amount' argument from `pcrc_coef_'.
//
//   Negotiate to lose interim `g_lb', `g_stride' members on `Range',
//       thus lose `pcrc_reset_push_bounds_', `pcrc_reset_pop_bounds_'.
//
//   Layer this as a Fortran-callable wrapper around a C-callable interface?
//
//   Finally (?) sort out the argc, argv business.
//
//   Move `extern' function prototypes into `pcrc.h'.

