#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

extern "C" {
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
}

extern "C" {
int close(...);
int read(...);
int write(...);
long int lseek(...);
}

#include "adlib.h"

#include "extname.h"

#define  MAXVALUE  10
#define  LEN  sizeof(struct device)
 
extern int defaultStorageMode;

extern "C" {
   void pcrc_open(int *unit, char *pt, int *len, int *status,int *access, int *form, int *recl);
   int pcrc_close(int *unit);
   void pcrc_read(int *unit, char *buffer,int *dim,int *major,int *order, int *count, int *size, int *index);
   void pcrc_write(int *unit, char *buffer,int *dim,int *major,int *order, int *count,int *size, int *index);
   void pcrc_seek(int *unit, int *rec);
}

typedef struct device {
   int unit;       /* Fortran 77 */
   char *fileName;
   int fd;         /* C */
   int status;
   int access;
   int form;
   int recl;
   struct device *next;
}DEVICE, *PDEVICE;

static PDEVICE devHeadPtr=NULL;

void pcrc_open(int *unit, char *pt, int *len, int *status,int *access, int *form, int *recl)
{
 PDEVICE p1,p2;
 char *ptCopy=(char *)malloc(sizeof(char)*(*len+1));
 int Fd;
 int i;

 PDEVICE tp = devHeadPtr;
 while (tp) {
   if (tp->unit == *unit)
        AdlibError("file has been opened, but still not closed\n");
   tp = tp->next;
 }

 ptCopy=strncpy(ptCopy,pt,*len);
 for (i=*len-1; i>=0 ; i--) if ( ptCopy[i] != ' ' ) break;
 ptCopy[i+1]='\0';
 
 if(*status==-1)
    Fd=open(ptCopy,O_RDWR|O_CREAT,0644);  
 else  
  {
    Fd=open(ptCopy,O_RDWR,0644);    
    if(*status==0)    //'old'
   {
    //file not exist, but its status is 'old'
    if(Fd==-1)
      AdlibError("open: No such file or directory\n");
   }
   else              //'new'
   {
    //file exist,but its status is 'new'
    if(Fd!=-1)
      AdlibError("open:'new' file exists\n");
    close(Fd);
    Fd=open(ptCopy,O_RDWR|O_CREAT,0644);  
   }
  }

  if(Fd==-1)
     AdlibError("open: failure\n");

 p1=(PDEVICE)malloc(LEN);
 p1->unit=*unit;
 p1->fileName=ptCopy;
 p1->fd=Fd;
 p1->status=*status;
 p1->access=*access;
 p1->form=*form;
 p1->recl=*recl;
//AdlibPrintf("fd:%d\n",p1->fd);
 p1->next=NULL;
 if(devHeadPtr==NULL)
     devHeadPtr=p1;
 else    //device list's head !=NULL
    { 
     p2=devHeadPtr;
     while( p2->next!=NULL)
           p2=p2->next;
     p2->next=p1;
    }   //device list's head !=NULL
}


int pcrc_close(int *unit)
{
 PDEVICE pre,current;

 if(devHeadPtr==NULL) 
       return(0);
//     AdlibError("close not existing file\n");
 else
    { 
      current=devHeadPtr;
      pre=NULL;
      while(current && (current->unit!=*unit))
      {
       pre=current;
       current=current->next;
      }  	     
      if (!current)
         return(0);
//          AdlibError("close not existing file\n");
      close(current->fd);
      if (pre == NULL) 
	{devHeadPtr=devHeadPtr->next;}
      else 
      	{pre->next=current->next;}
      free(current->fileName);
      free(current);
    }
   return(1);  
}

void pcrc_seek(int *unit,int *rec)
{
  PDEVICE current;
   current=devHeadPtr;
   while( current && (current->unit!=*unit))
    {
     current=current->next;
    }

    if(current==NULL) 
          AdlibError("seek not existing file\n");

    lseek(current->fd,(current->recl)*(*rec-1),0);

}

//major=0: COLUMN_MAJOR
//major=1: ROW_MAJOR
//count's order: a(i,j,k)---count[0]=width of dim i;count[1]=j's;count[2]=k's  
//index's meaning: (((a(i,j,k),j=...),k=...),i=...)
//index[0]=1,index[1]=2,index[2]=0
/*      
           [0]  [1]    [2]         [1]    [2]      [0]
          a(i,   j,     k)        (j=...,  k=...,  i=...)
            |    |      |            |       |      | 
            |    |      |            |       |      |
     count[0] count[1] count[2]   index[0] index[1] index[2]
*/
         

void pcrc_read(int *unit, char *buffer,int *dim,int *major,int *order, int *count,int *size, int *index)
{
  PDEVICE current;
  int flag;
  int i,j,k;
  int storage=defaultStorageMode;
  
   current=devHeadPtr;
   while( current && (current->unit!=*unit))
    {
     current=current->next;
    }

    if(current==NULL) 
          AdlibError("read: not opened file\n");

   if(current->status==1)
          AdlibError("start: end of file\n"); 

    ProfBeginRead()

    switch(*dim)
    {
      case 1:
          flag=read(current->fd,buffer,count[0]*(*size));  
          if(flag==-1)
              AdlibError("file read error\n");
          if(flag==0)
              printf("file end\n");
          break;
      case 2:
          if(*order==0) 
            {
             flag=read(current->fd,buffer,count[0] * count[1] * (*size));  
             if(flag==-1)
                AdlibError("file read error\n");
             if(flag==0)
               {
                printf("file end\n");
                exit(0);
               }
            }
          else   //*order=1 
            {
              int a[2];
              char *tempBuffer=(char *)malloc((*size)*count[0]*count[1]);
              flag=read(current->fd,tempBuffer,count[0] * count[1] * (*size));  
              if(flag==-1)
                  AdlibError("file read error\n");
              if(flag==0)
                  printf("file end\n");  //??? exit()

              i=0;
              if(storage==COLUMN_MAJOR) {
               for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
	        {
		     memcpy(buffer+(a[0]+count[0]*a[1])*(*size), tempBuffer+i,
			*size);	
		     i += *size;
		}
	      }
	      else {
               for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
		{
		     memcpy(buffer+(a[1]+count[1]*a[0])*(*size), tempBuffer+i,
			*size);	
		     i += *size;
		}
	      }
              free(tempBuffer);
            }
          break;
      case 3:
          if(*order==0)
            {
             flag=read(current->fd,buffer,count[0] * count[1] * count[2] * (*size));  
             if(flag==-1)
                AdlibError("file read error\n");
             if(flag==0)
                printf("file end\n");
            }
          else //*order=1
          { 
              int a[3];
              char *tempBuffer=(char *)malloc((*size)*count[0]*count[1]*count[2]);
              flag=read(current->fd,tempBuffer,count[0] * count[1] * count[2] * (*size));  
              if(flag==-1)
                AdlibError("file read error\n");
              if(flag==0)
                printf("file end\n");

              i=0; 
              if(storage==COLUMN_MAJOR) {
               for(a[index[2]]=0;a[index[2]]<count[index[2]];a[index[2]]++)
                for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                  for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
		  {
		     memcpy(buffer+ ((a[2]*count[1]+a[1])*count[0]+a[0])*(*size), tempBuffer+i,
			*size);
		     i += *size;
		  }
	      }
	      else {
               for(a[index[2]]=0;a[index[2]]<count[index[2]];a[index[2]]++)
                for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                  for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
		  {
		     memcpy(buffer+ ((a[0]*count[1]+a[1])*count[2]+a[2])*(*size), tempBuffer+i,
			*size);
		     i += *size;
		  }
	      }
              free(tempBuffer);
          }
          break;
    } 

    ProfEndRead()
#ifdef RUNTIME_PROFILE
    read_ops++;
#endif
}

void pcrc_write(int *unit,char *buffer,int *dim,int *major, int *order,int *count,int *size, int *index)
{
  PDEVICE current;
  int flag;
  int i,j,k;
  int storage=defaultStorageMode;
  
   current=devHeadPtr;
   while( current && (current->unit!=*unit))
    {
     current=current->next;
    }

    if(current==NULL) 
          AdlibError("write: not opened file\n");

//    if(current->status==0)
//          AdlibError("cannot write old file\n");  

    ProfBeginWrite()

    switch(*dim)
    {
      case 1:
          flag=write(current->fd,buffer,count[0]*(*size));  
          if(flag==-1)
              AdlibError("file write error\n");
          if(flag==0)
              AdlibError("nothing was written \n");
          break;
      case 2:
          if(*order==0)
            {
             flag=write(current->fd,buffer,count[0] * count[1] * (*size));  
             if(flag==-1)
                AdlibError("file write error\n");
             if(flag==0)
                AdlibError("nothing was written \n");
            }
          else        // (*order==1)) 
            {
              int a[2];
              char *tempBuffer=(char *)malloc((*size)*count[0]*count[1]);
           
              i=0;
              if(storage==COLUMN_MAJOR) {
               for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
	        {
		     memcpy(tempBuffer+i, buffer + (a[0]+count[0]*a[1])*(*size),
				*size);
		     i+=*size;
                }
              }
	      else {
               for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
                {
		     memcpy(tempBuffer+i, buffer + (a[1]+count[1]*a[0])*(*size),
				*size);
		     i+=*size;
                }
             }
             flag=write(current->fd,tempBuffer,count[0] * count[1] * (*size));  
             if(flag==-1)
                AdlibError("file write error\n");
             if(flag==0)
                AdlibError("nothing was written \n");
             free(tempBuffer);
            }
          break;
      case 3:
          if(*order==0)
            {
             flag=write(current->fd,buffer,count[0] * count[1] * count[2] * (*size));  
             if(flag==-1)
                AdlibError("file write error\n");
             if(flag==0)
                AdlibError("nothing was written \n");
            }
          else   //(*order==1)
           { 
              int a[3];
              char *tempBuffer=(char *)malloc((*size)*count[0]*count[1]*count[2]);
              i=0;
              if(storage==COLUMN_MAJOR) {
               for(a[index[2]]=0;a[index[2]]<count[index[2]];a[index[2]]++)
                for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                  for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
		  {
		     memcpy(tempBuffer+i, buffer+((a[2]*count[1]+a[1])*count[0]+a[0])*(*size),
				*size);
		     i+=*size;
		  }
              }
	      else {
              for(a[index[2]]=0;a[index[2]]<count[index[2]];a[index[2]]++)
                for(a[index[1]]=0;a[index[1]]<count[index[1]];a[index[1]]++)
                  for(a[index[0]]=0;a[index[0]]<count[index[0]];a[index[0]]++)
		  {
		     memcpy(tempBuffer+i, buffer+((a[0]*count[1]+a[1])*count[2]+a[2])*(*size),
				*size);
		     i+=*size;
		  }
	      }
              flag=write(current->fd,tempBuffer,count[0] * count[1] * count[2] * (*size));  
              if(flag==-1)
                    AdlibError("file write error\n");
              if(flag==0)
                    AdlibError("nothing was written \n");
              free(tempBuffer);
           }
          break;
    } 

    ProfEndWrite()
#ifdef RUNTIME_PROFILE
    write_ops++;
#endif
}
