      subroutine find_rhs_psi(alpha, h2, psi, zeta, rhs, N)
      implicit none
      integer , intent(in):: N 
      real , intent(in):: alpha,h2 
      real ,  dimension(:,:),intent(in):: psi,zeta 
      real ,  dimension(:,:),intent(out):: rhs 
!HPF$      distribute * (*,block) :: psi,zeta,rhs
      integer :: i,k,j1,jN
      real :: b
      j1 = 2
      jN = N-1
      b = alpha-2.0
      do  k=j1,jN
         rhs(k,j1:jN) = (psi(k-1,j1:jN)+b*psi(k,j1:jN)+psi(k+1,j1:jN))-(&
     &h2*zeta(k,j1:jN))
      enddo
      rhs(1,:) = psi(1,:)
      rhs(N,:) = psi(N,:)
      do  k=j1,jN
         rhs(k,1) = psi(k,1)
         rhs(k,N) = psi(k,N)
      enddo
      return
      end subroutine

      program adi
      implicit none
      integer ,  parameter:: NN=8 
      real ,  parameter:: DT=0.1 
      real ,  parameter:: H=0.2 
      integer ,  parameter:: STDIN=5 
      integer ,  parameter:: STDOUT=6 
      integer ,  parameter:: FILEIN=8 
      integer ,  parameter:: FILEOUT=9 
      interface 
         subroutine tridiag_psi(alpha, beta, rhs, u, temp, N)
         integer , intent(in):: N 
         real , intent(in):: alpha 
         real ,  dimension(:,:),intent(in):: rhs 
         real ,  dimension(:,:),intent(out):: u,temp,beta 
!HPF$      distribute * (*,block) :: beta,rhs,u,temp
         end subroutine

         subroutine find_rhs_psi(alpha, h2, psi, zeta, rhs, N)
         integer , intent(in):: N 
         real , intent(in):: alpha,h2 
         real ,  dimension(:,:),intent(in):: psi,zeta 
         real ,  dimension(:,:),intent(out):: rhs 
!HPF$      distribute * (*,block) :: psi,zeta,rhs
         end subroutine

      end interface
      integer :: i,j
      integer :: iTime
      integer :: NTime
      real :: alpha,h2
      real ,  dimension(NN,NN):: beta,psi,zeta,rhs,wspace 
!HPF$      processors :: procs(NUMBER_OF_PROCESSORS ())
!HPF$      template, dimension(NN) :: tplate
!HPF$      distribute  (block) onto  procs :: tplate
!HPF$      align beta (*,:) with tplate(:)
!HPF$      align psi (*,:) with tplate(:)
!HPF$      align zeta (*,:) with tplate(:)
!HPF$      align rhs (*,:) with tplate(:)
!HPF$      align wspace (*,:) with tplate(:)
      h2 = H*H
      alpha = 2.0*h2/DT
      NTime = 10
      rhs = 1.0
      wspace = 0.0
      psi = 5.0
      psi(2:NN-1,2:NN-1) = 2.0
      zeta = 0.0
      do  iTime=1,NTime
         call tridiag_psi (alpha,beta,rhs,psi,wspace,NN)
         call find_rhs_psi (alpha,h2,psi,zeta,rhs,NN)
         psi = TRANSPOSE (psi)
         zeta = TRANSPOSE (zeta)
         rhs = TRANSPOSE (rhs)
         call tridiag_psi (alpha,beta,rhs,psi,wspace,NN)
         call find_rhs_psi (alpha,h2,psi,zeta,rhs,NN)
         psi = TRANSPOSE (psi)
         zeta = TRANSPOSE (zeta)
         rhs = TRANSPOSE (rhs)
         write (unit=STDOUT,fmt=1001) iTime
         do  i=1,NN
            write (unit=STDOUT,fmt=1002) (psi(i,j), j=1,NN)
         enddo
      enddo
      write (unit=STDOUT,fmt=*) 'Final Result: for psi()'
      do  i=1,NN
         write (unit=STDOUT,fmt=1002) (psi(i,j), j=1,NN)
      enddo
      stop
1001      format ('Status: iTime=',I6)
1002      format (12F9.3)
      end program


      subroutine tridiag_psi(alpha, beta, rhs, u, temp, N)
      implicit none
      integer , intent(in):: N 
      real , intent(in):: alpha 
      real ,  dimension(:,:),intent(in):: rhs 
      real ,  dimension(:,:),intent(out):: u,temp,beta 
!HPF$      distribute * (*,block) :: beta,rhs,u,temp
      integer :: i,k,j1,jN
      real :: a,b,c
      j1 = 2
      jN = N-1
      a = (-1.0)
      b = alpha+2.0
      c = (-1.0)
      u(j1,j1:jN) = (rhs(j1,j1:jN)-(a*u(1,j1:jN)))/b
      beta(j1,j1:jN) = b
      do  k=j1+1,jN
         temp(k,j1:jN) = c/beta(k-1,j1:jN)
         beta(k,j1:jN) = b-(a*temp(k,j1:jN))
         u(k,j1:jN) = (rhs(k,j1:jN)-(a*u(k-1,j1:jN)))/beta(k,j1:jN)
      enddo
      temp(N,j1:jN) = c/beta(jN,j1:jN)
      do  k=jN,j1,(-1)
         u(k,j1:jN) = u(k,j1:jN)-(temp(k+1,j1:jN)*u(k+1,j1:jN))
      enddo
      return
      end subroutine

