#include "compilerApi.h"
#include "../../depInterface.h"

SgProject       *pTheProject = (SgProject *) NULL;
SgStatement	*pSgGlobalCopy = (SgStatement *) NULL;
SgFile          *pTheFile = (SgFile *) NULL;
SgStatement     **ppSgForall = (SgStatement **) NULL,**ppSgForallConstruct = (SgStatement **) NULL;
int             nForallCount=0,nForallConstructCount=0; /** how many forall and
forallconstruct stmt**/


extern int ForallToFor(SgStatement **pSgForall);
extern int ForallConstructToFor(SgStatement **pSgForallConstruct);
extern doDependenceAnalysisForAFunction(SgFile *file, SgStatement *func,int printdep, int printannotation, int verbose);

int modifyForallList(void);
int modifyForallConstructList(void);

/**reset dep tool kit**/

void resetDepToolKit(void)
{
    pTheFile= (SgFile *)NULL;

    if(ppSgForall) delete [] ppSgForall;
    ppSgForall=(SgStatement **)NULL;

    if(ppSgForallConstruct) delete [] ppSgForallConstruct;
    ppSgForallConstruct=(SgStatement **)NULL;


    nForallCount=0;
    nForallConstructCount=0;
}

/**establish the relations between the two trees.So that we know which statement in the new tree correspondes to which statement in the original tree**/
/** which means they have the same Id**/

void createLinksBetweenTrees()
{
   SgStatement * pSgOld, *pSgNew;
   SgExprListExp *pSgList;
   SgExpression *pMask;   
   int *pAtt;
   int nestedStack[20];//forall construct maximum nested level is 20
   int nP=-1;		
   int nNest=0;
   int i=0;
   
   pSgOld=pTheFile->firstStatement();
   pSgNew=pSgGlobalCopy;

   while(pSgNew && pSgOld)
   {
	if(pSgOld->variant()!=FORALL_STMT && pSgOld->variant()!=FORALL_CONSTRUCT_STMT)
	{
		if(pSgOld->variant()!=CONTROL_END)	//commonplace stmt
		{
#if 0
			printf("----common statements:\n");
			pSgOld->unparsestdout();
#endif

			pAtt=new int(pSgOld->id());
			pSgNew->addAttribute(FOR_FORALL_ATTRIBUTE,(void *)pAtt,0);

			//pSgNew->setId(pSgOld->id());
			pSgNew=pSgNew->lexNext();
			pSgOld=pSgOld->lexNext();
		}
		else	//control end
		{
			pSgOld->unparsestdout();
			pSgOld=pSgOld->lexNext();
		
			i=0;
			while(i<nestedStack[nP])
			{
				if(pSgNew->variant()==CONTROL_END) i++; 
				pSgNew=pSgNew->lexNext();
			}
			nP--;
		
		}
	}
	else	//forall stmt and forall construct stmt
	{
#if 0
		printf("----forall stmt or forall construct \n");
		pSgOld->unparsestdout();
#endif
		pAtt=new int(pSgOld->id());
		pSgNew->addAttribute(FOR_FORALL_ATTRIBUTE,(void *)pAtt,0);
		//pSgNew->setId(pSgOld->id());
		pSgList=(SgExprListExp*) (pSgOld->expr(0))->lhs();
   		nNest=pSgList->length();
		pMask=pSgNew->expr(1);

		for(i=0;i<nNest;i++)
			pSgNew=pSgNew->lexNext();

		if(pMask)
		{ 
			pSgNew=pSgNew->lexNext();
			nNest++;
		}
		nestedStack[++nP]=nNest;

		pSgOld=pSgOld->lexNext();
	}	
   }
	
}

int initializeDepToolKit(SgProject *pSgProj)
{
	SgStatement *pSgStmt,*pSgTemp;
	int *pAtt;
	int i,nForall=0,nForallConstruct=0;

	resetDepToolKit();

	pTheProject=pSgProj;
        pTheFile=&(pTheProject->file(0));

	pSgGlobalCopy=BfndMapping(duplicateStmts(pTheFile->firstStatement()->thebif));

	pSgStmt=pSgGlobalCopy;
   	while(pSgStmt)
   	{
     	  if((pSgStmt->variant())==FORALL_STMT) nForallCount++;
    	  else
     	  if((pSgStmt->variant())==FORALL_CONSTRUCT_STMT) nForallConstructCount++;    

	 pSgStmt=pSgStmt->lexNext();
        }

  	ppSgForall =new SgStatement *[nForallCount];
   	ppSgForallConstruct =new SgStatement *[nForallConstructCount];

	pSgStmt=pSgGlobalCopy;
   	while(pSgStmt)
   	{
     	  if((pSgStmt->variant())==FORALL_STMT)
     	  {
        	ppSgForall[nForall]=pSgStmt;
        	nForall++;
     	  }
     	  else
          if((pSgStmt->variant())==FORALL_CONSTRUCT_STMT)
           {
                ppSgForallConstruct[nForallConstruct]=pSgStmt;
                nForallConstruct++;
           }
     	  pSgStmt=pSgStmt->lexNext();
        }

	modifyForallList();
	modifyForallConstructList();
	printf("running just below modifying forall construct list \n");
	pSgGlobalCopy->unparsestdout();
	

//set up links between the original tree and the new tree to show which do loop is transformed from which forall stmt or forall construct
	createLinksBetweenTrees();

// get each function or subroutine and do dependence analysis to them
        i=0; 
	for(i=0;i<pSgGlobalCopy->numberOfChildrenList1();i++)
	{
		pSgStmt=pSgGlobalCopy->childList1(i);
		while(pSgStmt && pSgStmt->variant()!=PROG_HEDR && pSgStmt->variant()!=PROC_HEDR && pSgStmt->variant()!=FUNC_HEDR)
			pSgStmt=pSgStmt->lexNext();
		ASSERT(pSgStmt);
		doDependenceAnalysisForAFunction(pTheFile,pSgStmt,1,1,0);
	}

	return 0;
	
}

void finalizeDepToolKit(void)
{
	resetDepToolKit();
}


/** modify all the forall statements in the ppSgForall[] into for statement **/

int modifyForallList(void)
{
    int i;

    ASSERT(ppSgForall) ;
    for(i=0;i<nForallCount;i++)
    {
       ForallToFor(&(ppSgForall[i]));
    }
}


/** modify all the forall construct statements in the ppSgForallConstruct[] into for statement **/
/* this is proceeding from first transforming the deepest nested forall construct then to the less nested forall construct**/

int modifyForallConstructList(void)
{
    int i,j,n=0,nNestedLevelMax=0;
    SgStatement *pTemp;
    int *pnLevel;

    ASSERT(ppSgForallConstruct) ;
    pnLevel= new int [nForallConstructCount];
    for(i=0;i<nForallConstructCount;i++)
    {
	n=0;
        pTemp=ppSgForallConstruct[i];
        while(pTemp->controlParent()!=NULL)
        {
                pTemp=pTemp->controlParent();
                n++;
        }
        pnLevel[i]=n;
        if(n>nNestedLevelMax) nNestedLevelMax=n;
    }

    for(j=nNestedLevelMax;j>0;j--)
    {
        for(i=0;i<nForallConstructCount;i++)
                if(pnLevel[i]==j)
                        ForallConstructToFor(&(ppSgForallConstruct[i]));
    }

    delete pnLevel;
}

//judge whether any dependence exists between the two statement
depNode* isDependentBetweenStmts(SgStatement *pOldF, SgStatement *pOldT) 
{
	SgStatement *pStmt;
	depGraph *pGraph;
	depNode *pNode;
	int fromID,toID,*pAttF,*pAttT;

	fromID=pOldF->id();
	toID=pOldT->id();

	pStmt=pSgGlobalCopy;
	while(pStmt)
	{
		if(pStmt->variant()==FOR_NODE)
		{
			pGraph=(depGraph *) (pStmt->attributeValue(0,DEPGRAPH_ATTRIBUTE));
			ASSERT(pGraph);
			pNode=pGraph->first;
			while(pNode)
			{
				pAttF=(int *)(pNode->stmtin->attributeValue(0,FOR_FORALL_ATTRIBUTE));
				pAttT=(int *)(pNode->stmtout->attributeValue(0,FOR_FORALL_ATTRIBUTE));
				if(*pAttF== fromID && *pAttT==toID) return pNode;
				else	pNode=pNode->next;
						
			}
			pStmt=pStmt->lastNodeOfStmt()->lexNext();
		}
		else
			pStmt=pStmt->lexNext();
	}	
	return NULL;
}

//judge whether any dependence exists between the two statement
depNode* isDependentBetweenStmts(int fromID,int toID) 
{
	SgStatement *pStmt;
	depGraph *pGraph;
	depNode *pNode;
	int *pAttF,*pAttT;

	printf("In checking dependence:\n");
        
	pStmt=pSgGlobalCopy;
	while(pStmt)
	{
		if(pStmt->variant()==FOR_NODE)
		{
			pGraph=(depGraph *) (pStmt->attributeValue(0,DEPGRAPH_ATTRIBUTE));
			ASSERT(pGraph);
			pNode=pGraph->first;
			while(pNode)
			{       
				pNode->displayDep();
                                
				pNode->stmtin->unparsestdout();
				pNode->stmtout->unparsestdout();
				pAttF=(int *)(pNode->stmtin->attributeValue(0,FOR_FORALL_ATTRIBUTE));
				pAttT=(int *)(pNode->stmtout->attributeValue(0,FOR_FORALL_ATTRIBUTE));
				printf("fromId == %d toID==%d\n",fromID,toID);
				printf("from id=%d to id= %d\n",*pAttF,*pAttT);
				printf("this is here\n");
                                if((*pAttF== fromID) && (*pAttT==toID))
                                    return pNode;
				else
                        	    pNode=pNode->next;
			}
                        pStmt=pStmt->lastNodeOfStmt()->lexNext();
		}
		else
			pStmt=pStmt->lexNext();
	}	
	return NULL;
}
