#include <stdio.h>
#include <malloc.h>
#include "compilerApi.h"
#include "definesValues.h"
#include "set.h"
#include "definitionSet.h"
#include "inducVar.h"
#include "depGraph.h"

#define MAXVALUE 1000
///////////////////////////////////////////////////////////////////////////////////////
// some global declarations
///////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////
// Loop Transformation Sets. (the validity checkings for the transformations
// are in the depGraph method.
///////////////////////////////////////////////////////////////////////////////////////

//
// b is the outer loopm 
// permut specify the loop permutations
// the n outermost loops must be perfectly nested
// permut start at 0; Example
// do i = 0
//    do j = 8
//      do k = 1
//       ....
// permut = [2,1,0]
// gives
// do  k = 1
//    do j = 8
//      do i = 0
//       ....
//

// modified by slx 97/12/20

int loopInterchange(SgStatement *loop, int *permut, int n)
{
  SgSymbol *tab_symb[MAXNESTEDLOOP];  
  SgExpression *tab_bf[MAXNESTEDLOOP];
  SgExpression * tab_bi[MAXNESTEDLOOP];
  SgExpression * tab_step[MAXNESTEDLOOP];
  SgForStmt *doloop, *tloop;
  int i;

  if (!(doloop = isSgForStmt(loop)))
     return 0;
 
  // loop number exchanged should < or = perfectloopnest number 
  if (n > doloop->isPerfectLoopNest())   
    {
      Message("Interchange loop error, loops are not perfectly nested",doloop->lineNumber());
      return 0;
    }

  tloop = doloop;
  for (i = 0; (i < n); i++)
    {
      tab_symb[i] = tloop->symbol();
      tab_bf[i] = tloop->start()->copyPtr();        // loop lower bound
      tab_bi[i] = tloop->end()->copyPtr();          // loop upper bound
      tab_step[i] = tloop->step()?tloop->step()->copyPtr():0;    // loop step 
      if (tloop = isSgForStmt(tloop->getNextLoop()))        
         printf("%s\n",tloop->symbol()->identifier());
    }

  tloop = doloop;
  for (i = 0; (i < n); i++)
    {
      if (tab_symb[permut[i]])
        tloop->setSymbol(*tab_symb[permut[i]]);
      if (tab_bf[permut[i]])
        tloop->setStart(*tab_bf[permut[i]]);
      if (tab_bi[permut[i]])
        tloop->setEnd(*tab_bi[permut[i]]);
//      if (tab_step[permut[i]])
      tloop->setStep(tab_step[permut[i]]);
      tloop = isSgForStmt(tloop->getNextLoop());        
    }
  return 1;
}

//**************************************************************************
// modified by slx 98/1/12
// Tile a loop assumed to be perfectly nested
// and assume that each loop in nested loop has step=1
// user can give (tileSize) and (offset)

// original loop :
// do i=lo, hi
//      a(i)=b(i)
// enddo

// is tiled into:
//(1) do it=floor((lo-offset)/tileSize)*tileSize+offste, floor((hi-offste)/tileSize)*tilesize+offset, tileSize
//(2)  do i=max(lo,it), min(hi,it+tileSize-1)
//        a(i)=b(i)
//     enddo
//    enddo
//(1) is called tile loop, (2) is called element loop

//To be simple, assume offset=0
// So we have:
//(11) do it=floor(lo/tileSize)*tileSize, floor(hi/tileSize)*tilesize, tileSize
//(22)  do i=max(lo,it), min(hi,it+tileSize-1)

//In C or Fortran language, if lo and tileSize are integer,
//then lo/tileSize=floor(lo/tileSize) 
//So we again have:
//(111) do it=(lo/tileSize)*tileSize, (hi/tileSize)*tilesize, tileSize
//(222)  do i=max(lo,it), min(hi,it+tileSize-1)

//****************************************************************************************
// argument "func" is convient to declare new index variable within "func"
// number of pointer array(**size) element should be equal to "nb"

// detect whether outer index variable appears in inner loop
// return 0, means "no appear"
// return 1, means "appear in lb"
// return 2, means "appear in ub"

int  detect_index_variable(SgForStmt *doloop, SgSymbol **tab_symb, int nb, int *outerL, int *innerL, int *outerU, int *innerU)
{
  int i,j, flagLB=0,flagUB=0;
  SgForStmt *temloo;

  for(j=0;j<nb-1;j++)
  {
    if (temloo=isSgForStmt(doloop->getNextLoop()))
	 for (i=1; i<nb; i++)
	{
           //search lower bound
	   if(temloo->start()->IsSymbolInExpression(*tab_symb[j]))
	    {printf("appear in lb\n");
	     outerL[i-1]=j;
	     innerL[i-1]=i;
             flagLB=1; 
	    } 
           //search upper bound
	   if(temloo->end()->IsSymbolInExpression(*tab_symb[j]))
	       {printf("appear in ub\n");
	        outerU[i-1]=j;
	        innerU[i-1]=i;
                flagUB=1;
	       }
	   temloo=isSgForStmt(temloo->getNextLoop());
         }
   }

  if((!flagLB) && (!flagUB)) return (0);
  if((flagLB) && (!flagUB))  return (1);
  if((!flagLB) && (flagUB))  return (2);
  if((flagLB) && (flagUB)) return (3);
 
}

SgForStmt * simple_construct_new_doloop(SgForStmt *doloop, SgSymbol **tab_symb, SgExpression **lb, SgExpression **ub, SgExpression **step, int nb)
{
  SgForStmt *loopCopy=(SgForStmt *)doloop->copyPtr();
  SgStatement *enddo=new SgStatement(ENDDO_STMT);
  SgForStmt *newloop,*temloo;
  int i;

  // set original loop's lb and ub into max() and min() respectively
  temloo=loopCopy;
  for(i=0;i<nb;i++)
   {
     temloo->setStart(*lb[i]);
     temloo->setEnd(*ub[i]);
     temloo = isSgForStmt(temloo->getNextLoop());  
   }

  newloop = loopCopy;
  for(i=nb-1; i>=0; i--)
  {
/*
    newloop=new SgForStmt(tab_symb[nb+i],lb[nb+i],ub[nb+i],step[nb+i],NULL);
    newloop->insertStmtAfter(*loopCopy, *newloop);
    loopCopy=newloop; 
*/
    newloop = new SgForStmt(tab_symb[nb+i],lb[nb+i],ub[nb+i],step[nb+i],newloop);
    }

  return newloop;
}

int loopTiling(SgStatement *func, SgStatement *loop, SgExpression **size, int nb)
{
#define MAX_LEVEL  30
  SgForStmt *doloop,*last_loop,*temloo,*newloop;
  SgSymbol *tab_symb[MAX_LEVEL];
  SgExpression *lb[MAX_LEVEL];
  SgExpression *ub[MAX_LEVEL];
  SgExpression *step[MAX_LEVEL];
  int i,innerL[MAX_LEVEL-1],outerL[MAX_LEVEL-1],innerU[MAX_LEVEL-1],outerU[MAX_LEVEL-1];
  char strfoname[256];

  if(!( doloop = isSgForStmt(loop)))  return 0;

  SgSymbol *minSymb=doloop->getSymbolInCurrentScope("min");
  SgSymbol *maxSymb=doloop->getSymbolInCurrentScope("max");
/*
  if (nb > doloop->isPerfectLoopNest())
    {
     Message("loop Tilling error, loops are not perfectly nested",doloop->lineNumber());
     return 0;
   }
*/
  if (nb == 0) return 0;

  last_loop = doloop;
  for ( i = 0; i < nb; i++)
    {
      int result;
      if (last_loop->step())
       { 
         if (last_loop->step()->valueInteger(result)==-1)
            {
              Message("Cannot tile loop, there is already a step",last_loop->lineNumber());
              return 0;
            }
         else 
            if (result!=1)
            {
              Message("Cannot tile loop, there is already a step",last_loop->lineNumber());
              return 0;
            }
       }     
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  temloo = doloop;
  for (i = 0; i < nb; i++)
    {
      tab_symb[i] = temloo->symbol();
      lb[i]=temloo->start()->copyPtr();
      ub[i]=temloo->end()->copyPtr();
      temloo = isSgForStmt(temloo->getNextLoop());  
    }
  // generate the new symbol and declare them
  // generate new loop's lower_bound,upper_bound, step
  for ( i = 0; i < nb; i++)
    {
/*
      sprintf(strfoname,"%s%s%d",tab_symb[i]->identifier(),"T",doloop->id());
      tab_symb[i+nb] =  new SgVariableSymb(strfoname);
      tab_symb[i+nb]->declareTheSymbol(*func);
*/
      //in order to reduce variable declare, we adopt the name beginning with pcrc_T
      tab_symb[i+nb]=loop->getSymbolInCurrentScope("pcrc","T",tab_symb[i]->identifier());
      lb[i+nb]=&(*lb[i]->copyPtr() / *size[i]->copyPtr() * *size[i]->copyPtr());
      lb[i+nb]=calculateExp(lb[i+nb]);
      ub[i+nb]=&(*ub[i]->copyPtr() / *size[i]->copyPtr() * *size[i]->copyPtr());
      ub[i+nb]=calculateExp(ub[i+nb]);
      step[i+nb]=size[i]->copyPtr();
    }
 
  // modify original loop's lower_bound, upper_bound
  for (i=0; i<nb; i++)
  {
    SgFunctionCallExp *maxExp= new SgFunctionCallExp(*maxSymb);
    SgFunctionCallExp *minExp= new SgFunctionCallExp(*minSymb);
    SgVarRefExp *varref=new SgVarRefExp(tab_symb[i+nb]);
    maxExp->addArg(*lb[i]->copyPtr());
    maxExp->addArg(*varref->copyPtr());
    minExp->addArg(*ub[i]->copyPtr());
    minExp->addArg(*varref->copyPtr()+*size[i]->copyPtr()-*(new SgValueExp(1)));
    lb[i]=maxExp;
    ub[i]=minExp;  
  }

  for(i=0;i<nb-1;i++)
  {
     outerL[i]=MAXVALUE;
     innerL[i]=MAXVALUE;
     outerU[i]=MAXVALUE;
     innerU[i]=MAXVALUE;
  }
 
  switch (detect_index_variable(doloop,tab_symb,nb,outerL,innerL,outerU,innerU))
  {
   case 0:
          break;
   case 1:
      {
        for(i=0;i<nb-1;i++)
        {
           if(outerL[i]!=MAXVALUE)  
             {
 	        lb[innerL[i]+nb]->replaceSymbolByExpression(*tab_symb[outerL[i]],*lb[outerL[i]]->copyPtr()); 
                for(int j=i;j>0; j--)
                {
                  if((outerL[j]==innerL[j-1]) && (outerL[j]!=MAXVALUE))
 	             lb[innerL[i]+nb]->replaceSymbolByExpression(*tab_symb[outerL[j-1]],*lb[outerL[j-1]]->copyPtr()); 
                }
             }
        } 
        break;
      }

   case 2:
      {
        for(i=0;i<nb-1;i++)
        {
           if(outerU[i]!=MAXVALUE)  
            {
               ub[innerU[i]+nb]->replaceSymbolByExpression(*tab_symb[outerU[i]],*ub[outerU[i]]->copyPtr()); 
                for(int j=i;j>0; j--)
                {
                  if((outerU[j]==innerU[j-1]) && (outerU[j]!=MAXVALUE)) 
 	             ub[innerU[i]+nb]->replaceSymbolByExpression(*tab_symb[outerU[j-1]],*ub[outerU[j-1]]->copyPtr()); 
                }
            }
        }
        break;
      }

   case 3:

      {
        for(i=0;i<nb-1;i++)
        {
           if(outerL[i]!=MAXVALUE)  
             {
 	        lb[innerL[i]+nb]->replaceSymbolByExpression(*tab_symb[outerL[i]],*lb[outerL[i]]->copyPtr()); 
                for(int j=i;j>0; j--)
                {
                  if((outerL[j]==innerL[j-1]) && (outerL[j]!=MAXVALUE)) 
 	             lb[innerL[i]+nb]->replaceSymbolByExpression(*tab_symb[outerL[j-1]],*lb[outerL[j-1]]->copyPtr()); 
                }   
             }
        } 

        for(i=0;i<nb-1;i++)
        {
           if(outerU[i]!=MAXVALUE)  
            {
               ub[innerU[i]+nb]->replaceSymbolByExpression(*tab_symb[outerU[i]],*ub[outerU[i]]->copyPtr()); 
                for(int j=i;j>0; j--)
                {
                  if((outerU[j]==innerU[j-1]) && (outerU[j]!=MAXVALUE)) 
 	             ub[innerU[i]+nb]->replaceSymbolByExpression(*tab_symb[outerU[j-1]],*ub[outerU[j-1]]->copyPtr()); 
                }
            }
        }

       break; 
      }
   default:  
      {
        printf("Sorry !!! Now can't deal with this\n"); 
        return 0;
      }
  }

  newloop=simple_construct_new_doloop(doloop,tab_symb,lb,ub,step,nb);
  doloop->insertStmtAfter(*newloop,*doloop->controlParent());
  doloop->deleteStmt();  
  return (1);
}

/*
int loopTiling(SgStatement *func, SgStatement *b, int *size, int nb)
{
  int i,j;
  SgSymbol *tab_symb[MAXNESTEDLOOP];
  SgStatement *body, *ptstmt; 
  SgForStmt *internal_loop, *temp, *last_loop;
  SgForStmt *loop, *temloo;
  SgExpression *borne_inf[MAXNESTEDLOOP];
  SgExpression *borne_sup[MAXNESTEDLOOP];
  SgExpression *step[MAXNESTEDLOOP];
  SgType *typint;
  SgSymbol  *min_fonc;
  char strfoname[256];

  if (!(loop = isSgForStmt(b)) || !func)
     return 0;
  if (nb > loop->isPerfectLoopNest())
    {
     printf("Perfect Loop Nest :%d\n",loop->isPerfectLoopNest());
     Message("loop Tilling error, loops are not perfectly nested",loop->lineNumber());
     return 0;
   }
  if (nb == 0) return 0;

  if (nb >= MAXNESTEDLOOP)
    {
      Message("Too many nested level",b->lineNumber());
      return 0;
    }
  for (i = 0 ; i<MAXNESTEDLOOP; i++)
    {
      tab_symb[i] = NULL;
      borne_inf[i] = NULL;
      borne_sup[i] = NULL;
      step[i] = NULL;
    }
  last_loop = loop;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        {
          if (last_loop->step())
            {
              Message("Cannot tile loop, there is already a step",last_loop->lineNumber());
              return 0;
            }
        }
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  // Get integer type;
  typint = SgTypeInt();
  // get the indexes of the loops;
  temloo = loop;
  for (i = 0; i < nb; i++)
    {
      tab_symb[i] = temloo->symbol();
      internal_loop = temloo;
      temloo = isSgForStmt(temloo->getNextLoop());  
    }
  if (!internal_loop)
    {
      Message("Internal loop not found",loop->lineNumber());
      return 0;
    }
  // generate the new symbol and declare them
  for ( i = 0; i < nb; i++)
    {
      sprintf(strfoname,"iT%d_%d",loop->id(),i);
      tab_symb[i+nb] =  new SgVariableSymb(strfoname);
      tab_symb[i+nb]->declareTheSymbol(*func);
    }
  last_loop = loop;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        last_loop->setStep((new SgValueExp(size[i]+1)));
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  
  // create a min tion; 
  min_fonc = new SgSymbol(FUNCTION_NAME,"min",*typint,*(loop->controlParent()));
  // creation of the bound expressions;
  last_loop = loop;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] != NO_STEP)
        {
          SgFunctionCallExp *cexp;
          borne_inf[i+nb] = new SgValueExp(0);
          cexp = new SgFunctionCallExp(*min_fonc);
          borne_sup[i+nb] =  cexp;
          cexp->addArg(((last_loop->end())->copy()) - *(new SgVarRefExp(*tab_symb[i])));
          cexp->addArg(*(new SgValueExp(size[i])));
        }
      else
        {          
          borne_inf[i+nb] = new SgValueExp(1);
          borne_sup[i+nb] = &((last_loop->end())->copy());
        }      
      step[i+nb] =  NULL;
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  // remove the loop body;
  // update the steps;
  body = internal_loop->extractStmtBody();
  printf("*************%d\n",body->variant());
  // upper bound for NO_STEP loops
  last_loop = loop;
  for ( i = 0; i < nb; i++)
    {
      if (size[i] == NO_STEP)
        last_loop->setEnd(*(new SgValueExp(1)));
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  // create the new loop;
  last_loop = internal_loop;
  for ( i = 0; i < nb; i++)
    {
      loop = new SgForStmt(tab_symb[i+nb], borne_inf[i+nb], borne_sup[i+nb],
                                step[i+nb], NULL);
      last_loop->insertStmtAfter(*loop,*last_loop);
      last_loop = loop; 
    }

  // update the loop body;
  if (!body || !last_loop)
    {
      Message("The loop may not have a body or internal error",0);
      return 0;
    }
  last_loop->insertStmtAfter(*body,*last_loop);
  
  for (j=0; j < last_loop->numberOfChildrenList1(); j++)
    {
      ptstmt =  last_loop->childList1(j);
      if (ptstmt)
        {
          for ( i = 0; i < nb; i++)
            {
              if (size[i] != NO_STEP)
                ptstmt->replaceSymbByExp(*tab_symb[i], *(new SgVarRefExp(*tab_symb[i]))
                                   +  *(new SgVarRefExp(*tab_symb[i+nb])));
              else
                ptstmt->replaceSymbByExp(*tab_symb[i], *(new SgVarRefExp(*tab_symb[i])));
            }
        }
    }
  return 1;
}
*/
//
// distribute the loops according to the 
// table of SCC.
// for instance consider the following loop
//  
//      do i= 1,100
//[1]      x  =  2
//[2]      do j= 1,100
//           D(j+i) = A(i,j)
//           D(j+i) = A(i,j)
//         enddo
//[3]      x  =  2 
//      enddo
// and the following content of the 
// [[1,3],2] scc table index
//

int
distributeLoopSCC(SgStatement *b, int *sccTable, int leadingdim, int numSCC)
{
  int i,j;
  SgStatement *last, **sccloop, **sccextracted, *ptstm; 
  SgStatement *previous, *cp, *copl;
  SgForStmt *loop;
  int nbstat;

  if (!(loop = isSgForStmt(b)) || !sccTable)
     return 0;

  nbstat = loop->numberOfChildrenList1() -1; // does not remove the controlend;
  sccloop = new SgStatement *[numSCC];
  sccextracted = new SgStatement *[nbstat];
   for (j=0; j < nbstat; j++)
    {
      ptstm =  loop->childList1(j);
      sccextracted[j] = ptstm;
    }
   for (j=0; j < nbstat; j++)
     {
       if (sccextracted[j])
         sccextracted[j]->extractStmt();
     }
  // now creates the loop;
  previous = loop->nodeBefore();
  cp = loop->controlParent();
  loop->extractStmt();
  copl = &(loop->copy());
  previous->insertStmtAfter(*loop,*cp);
  sccloop[0] = loop;
  last =  loop->lastNodeOfStmt();
  // need to extractloop before doing the copy;
  
  for (j=1; j < numSCC; j++)
    {
      if (j ==  numSCC-1)
        sccloop[j] = copl;
      else
        sccloop[j] = &(copl->copy());
      last->insertStmtAfter(*(sccloop[j]),*cp);
      last = sccloop[j]->lastNodeOfStmt();
    }

  for (j=0; j < numSCC; j++)
    {
      for (i=leadingdim-1; i>= 0; i--)
        {
          // insert in the right body the statement;
          // go revserse order; simpler;
          if (sccTable[j*leadingdim + i])
            {
              if (sccextracted[sccTable[j*leadingdim + i]-1])
                {
                  sccloop[j]->insertStmtAfter(*sccextracted[sccTable[j*leadingdim + i]-1],
                                              *sccloop[j]);
                }
            }
        }       
    }

  return 1;
}


//
// Takes two loops and fusion the bodies
//

//modified by slx 97/12/19
int
loopFusion(SgStatement *loop1,SgStatement *loop2)
{
  SgStatement *body,*last;
  int i;

  if (!loop1->symbol() || !loop2->symbol())
    return 0;

// find last effective statement of loop1
  last = loop1->lastNodeOfStmt();                    // 104 CONTROL_END
  last = (last->nodeBefore())->nodeBefore();                         // 194 ENDDO_STMT--> 111 

  int org_numberOfChildren=loop2->numberOfChildrenList1();
  SgStatement * * children = new SgStatement * [org_numberOfChildren];


  for(i=0; i<org_numberOfChildren; i++)
    children[i]=loop2->childList1(i)->copyPtr();

// insert loop2 body after loop1 body
// if two loops's induction variable are not same, then replace second loop symbol by first loop symbol
 
  if (strcmp(loop1->symbol()->identifier(),loop2->symbol()->identifier()))
   {
     SgExpression *exp = new SgExpression(VAR_REF,NULL,NULL,loop1->symbol()); 
     for (i=0; i<org_numberOfChildren-2; i++)
      {
       switch (children[i]->variant())
       {
        case ASSIGN_STAT:
        case PRINT_STAT: 
             children[i]->expr(0)->replaceSymbolByExpression(*(loop2->symbol()),*exp);           
             children[i]->expr(1)->replaceSymbolByExpression(*(loop2->symbol()),*exp);           
             break;
        case LOGIF_STMT:
             children[i]->expr(0)->replaceSymbolByExpression(*(loop2->symbol()),*exp);           
             break;
        }
      }  
    }

  for (i=0; i<org_numberOfChildren-2; i++)
    {
      last->insertStmtAfter(*children[i], *loop1);
      last = children[i];  
    }
  
  delete [] children; 


  loop2->deleteStmt();
  return 1;
}

// added by slx 97/12/27
// To strip-mine the following loop with a strip size of s:
// (1) do i = 1,N 
// (2)    a(i) = b(i) + c(i)
// (3) end do
// the loop is decomposed into two nested loops; the outer strip loop will 
// have a step size of s (the strip size), and the inner loop limits depend 
// on the outer loop index:
// (1) do Is = 1, N , s             -------strip loop
// (1)   do I = Is, min(N,Is+s-1)   -------element loop
// (2)        a(I) = b(I) + c(I)
// (3)   end do
// (3) end do
 
// argument "func" is convient to declare new index variable within "func"
int stripMining(SgStatement *func, SgStatement *loop, SgExpression *stripSize)
{
  SgForStmt *doloop;
  SgSymbol *symb;
  int org_numberOfChildren=loop->numberOfChildrenList1();
  SgStatement * * children = new SgStatement * [org_numberOfChildren];
  SgStatement * * childrenCopy = new SgStatement * [org_numberOfChildren-2];
  int i; 

  if(!( doloop = isSgForStmt(loop)))  return 0;

  for (i=0; i<org_numberOfChildren; i++)
     children[i]=loop->childList1(i);

  for (i=0; i<org_numberOfChildren-2; i++)
     childrenCopy[i]=loop->childList1(i)->copyPtr() ;

// add inner loop, its induction variable is original outer loop's.

  symb= new SgVariableSymb(loop->symbol()->identifier());
  SgForStmt  *innerloop=new SgForStmt(symb,NULL,NULL,NULL,NULL);
  loop->insertStmtAfter(*innerloop, *loop);
  SgStatement *enddo=new SgStatement(ENDDO_STMT);
  innerloop->insertStmtAfter(*enddo,*innerloop);

// modify outer loop's induction variable name, add a suffix "S"
  
  char strfoname[256];
  sprintf(strfoname,"%s%s%d",loop->symbol()->identifier(),"S",loop->id());
  symb= new SgVariableSymb(strfoname);
  symb->declareTheSymbol(*func);
  loop->setSymbol(*symb);
  SgExpression *step= stripSize->copyPtr();
  doloop->setStep(step);

// set inner loop's lbound,ubound
  
  SgVarRefExp *lbound= new SgVarRefExp(symb);
  innerloop->setStart(*lbound);
  SgSymbol *minSymb=innerloop->getSymbolInCurrentScope("min");
  SgFunctionCallExp *minExp= new SgFunctionCallExp(*minSymb);
  minExp->addArg(*(doloop->end()->copyPtr()));
  minExp->addArg(*lbound->copyPtr()+*step->copyPtr()-*(new SgValueExp(1)));
  innerloop->setEnd(*minExp);
 
// add outer loop body to inner loop body

  for (i=org_numberOfChildren-3; i>=0; i--)
     innerloop->insertStmtAfter(*childrenCopy[i], *innerloop);

// delete outer loop body

  for (i=0; i<org_numberOfChildren-2; i++)
     children[i]->deleteStmt();
 
  delete  [] children;
  delete  [] childrenCopy;

  return 1;

}

SgStatement * forall_loopTiling(SgStatement *func, SgStatement *loop, SgExpression **size, int nb)
{
  SgForStmt *doloop,*last_loop,*temloo,*newloop;
  SgSymbol *tab_symb[MAX_LEVEL];
  SgExpression *lb[MAX_LEVEL];
  SgExpression *ub[MAX_LEVEL];
  SgExpression *step[MAX_LEVEL];
  int i;

  if(!( doloop = isSgForStmt(loop)))  return (SgStatement *)NULL;
  SgSymbol *minSymb=doloop->getSymbolInCurrentScope("min");

  if (nb == 0) return (SgStatement *)NULL;

  last_loop = doloop;
  for ( i = 0; i < nb; i++)
    {
      int result;
      if (last_loop->step())
       { 
         if (last_loop->step()->valueInteger(result)==-1)
            {
              Message("Cannot tile loop, there is already a step",last_loop->lineNumber());
              return (SgStatement *)NULL;
            }
         else 
            if (result!=1)
            {
              Message("Cannot tile loop, there is already a step",last_loop->lineNumber());
              return (SgStatement *)NULL;
            }
       }     
      last_loop = isSgForStmt(last_loop->getNextLoop()); 
    }
  temloo = doloop;
  for (i = 0; i < nb; i++)
    {
      tab_symb[i] = temloo->symbol();
      lb[i]=temloo->start()->copyPtr();
      ub[i]=temloo->end()->copyPtr();
      temloo = isSgForStmt(temloo->getNextLoop());  
    }
  // generate the new symbol and declare them
  // generate new loop's lower_bound,upper_bound, step
  for ( i = 0; i < nb; i++)
    {
      //in order to reduce variable declare, we adopt the name beginning with pcrc_T
      tab_symb[i+nb]=loop->getSymbolInCurrentScope("pcrc","T",tab_symb[i]->identifier());
      lb[i+nb]=lb[i]->copyPtr();
      lb[i+nb]=calculateExp(lb[i+nb]);
      ub[i+nb]=ub[i]->copyPtr(); 
      ub[i+nb]=calculateExp(ub[i+nb]);
      step[i+nb]=size[i]->copyPtr();
    }

  // modify original loop's lower_bound, upper_bound
  for (i=0; i<nb; i++)
  {
    SgFunctionCallExp *minExp= new SgFunctionCallExp(*minSymb);
    SgVarRefExp *varref=new SgVarRefExp(tab_symb[i+nb]);
    minExp->addArg(*ub[i]->copyPtr());
    minExp->addArg(*varref->copyPtr()+*size[i]->copyPtr()-*(new SgValueExp(1)));
    lb[i]=varref;
    ub[i]=minExp;  
  }

  newloop=simple_construct_new_doloop(doloop,tab_symb,lb,ub,step,nb);
  doloop->insertStmtAfter(*newloop,*doloop->controlParent());
  doloop->deleteStmt();  
  return (newloop);
}
