/**@name Strategy for the compiler
We define some macro which can affect the compiler's action during 
compile-time. If you change any of these macros, you have to re-compile the
whole compiler and runtime to let the new strategies take effect.
This give much flexiblity for the compiler and users.

{\small implemented by \URL[Zheng Gengbin]{http://162.105.179.137/users/zhenggb}.}

[\URL[Back]{http://162.105.179.137/pact/index.html}]
 */

//@{

/**@name Intrinsic function call
The macro STRATEGY_ELEMENT_INTRINSIC_CALL defined how to deal the intrinsic
function call which will return a scalar type result in an expression.
@memo intrinsic function call in expression process strategy
 */
//@{

///  declare a temporary variable and add an assignment to assign the function to it.
#define EXTRACT_ELEMENT_INTRINSIC_CALL 1
///  create stack to store temporary function result for every function call
#define STACK_ELEMENT_INTRINSIC_CALL   2
///  define which strategy compiler will use, default is the second
#define STRATEGY_ELEMENT_INTRINSIC_CALL       STACK_ELEMENT_INTRINSIC_CALL

//@}

/**@name Optimization
In order to improve the speedup, we must assume some optimize strategy.
These optimization include:
\begin{itemize}
\item caculation of array subscripts.( in node program, array subscripts are linerized.
\item local index to global index conversion.
\item compare dummy and actual distribute pattern to decide if remap is needed.
\item for elementary forall statement, for example forall(i=0:0) a(i)=b(i), create a less procgroup for broadcast to scalar communication. 
\end{itemize}
@memo optimization strategy
*/
//@{
/// define ENABLE_OPTIMIZE
#define DISABLE_OPTIMIZE 0
/// define ENABLE_OPTIMIZE
#define ENABLE_OPTIMIZE 1
/// optimize the local array subscripts calculation
#define INDEX_OPTIMIZE	0x0001
/// optimize the local index to global index conversion.
#define LOCAL2GLOBAL	0x0002
/// optimize the runtime communication detection
#define RUNTIMECOMM	0x0004
/// other optimizaton
#define OTHER_OPTIMIZE	0x8000
//@}

/**@name Default data mapping strategy
We can define the macro to decide how to distribute the array across processors.
If we undefine this macro , compiler will take another strategy that it will 
assume BLOCK mapping mode by default
@memo default data mapping strategy
 */
//@{
/// if an array is not distributed , distributed it as collapsed by default
#define DEFAULT_COLLPASE
//@}

/**@name Processors Shape Definition
We can specify the default processor number and dimension
@memo define the default processor's shape
*/
//@{
// used in trans1
/// define the default number of processors, default number_of_processors()
#define NUMBER_OF_PROC  -1                           

/// define the default dimension of processors
#define DIM_OF_PROC	1
//@}

/**@name Memory Management Strategy
We can choose the memory management strategy by define the macro or not.
Currently, there are two strategies:
\begin{itemize}
\item Built-in stack management.
\item \URL[Memory Allocation Package]{http://www.emsl.pnl.gov:2080/docs/parsoft/ma/MA.html} 
\end{itemize}
@memo choose memory management strategy
*/
//@{
/// use the memory allocation(MA) strategy
#define MEM_ALLOC       1
//@}

/**@name Dynamic Memory Size
@memo define maximun stack and heap size.
*/
//@{
/// define maximun stack size
#define MAX_STACK_SIZE 16000000 	//for mem init  16M
#define MIN_STACK_SIZE 8000000 		//for mem init  8M
/// define maximun heap size
#define MAX_HEAP_SIZE  6000		//for mem init
#define MIN_HEAP_SIZE  6000		//for mem init
//@}

//@}
