// written by Zheng Gengbin

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "tof77.h"

#include "compilerApi.h"

#define RECORD_ALIAS 1000

/***************************************************************************/

int is_pointer(SgSymbol *symb)
{
   return symb->attributes() &  POINTER_BIT ? 1:0;
}

void trans_record(SgStatement *s, SgSymbol *recordName)
{
    SgClassType *recordType = isSgClassType(recordName->type());
    ASSERT(recordType);
    char *record_str = recordName->identifier();
    
    char *prefix = (char *)malloc(strlen(record_str) + 2);
    prefix = strcat(prefix, record_str);
    prefix = strcat(prefix, "_");

    SgSymbol *first = recordType->firstFieldSymb();
    while (first)
    {
	if (isSgClassType(first->type()) && !is_pointer(first)) {
	  SgSymbol *typeName = isSgClassType(first->type())->recordName();
	  for (int i=0; i<typeName->numberOfAttributes(RECORD_ALIAS); i++)
	  {
	   SgSymbol *field = (SgSymbol *)typeName->attributeValue(i, RECORD_ALIAS);
	   char *str = field->identifier();
	   char *alias = (char *)malloc( strlen(str) + strlen(prefix) );
	   strcpy(alias, prefix); 
	   strcat(alias, str);
	   SgSymbol * aliasSymb = s->getSymbolInCurrentScope(alias);
	   aliasSymb->setType(*field->type());
	   aliasSymb->setAttribute(field->attributes());
	   recordName->addAttribute(RECORD_ALIAS, aliasSymb, sizeof(aliasSymb));
printf("RECORD FIELDS:%s: ALIAS:%s\n",first->identifier(),alias);
	  }
	}
	else {
	   char *name =  first->identifier();
	   char *alias = (char *)malloc( strlen(name) + strlen(prefix) );
	   strcpy(alias, prefix); 
	   strcat(alias, name);
	   SgSymbol * aliasSymb = s->getSymbolInCurrentScope(alias);
	   aliasSymb->setType(*first->type());
	   aliasSymb->setAttribute(first->attributes());
	   recordName->addAttribute(RECORD_ALIAS, aliasSymb, sizeof(aliasSymb));
printf("FIELDS:%s: ALIAS:%s\n",first->identifier(),alias);
	}
	first = first->nextInScope();
    }
}

void trans_record_decl(SgStatement *s)
{
    SgSymbol *recordName = s->symbol();
    trans_record(s, recordName);
    s->deleteStmt();
}

/***************************************************************************
	trans_record_in_vardecl()

	change VAR declaration. 
	type (node) :: a
	=> a_node ...
***************************************************************************/

void trans_record_in_vardecl(SgStatement *s)
{
    SgExpression *typeop = s->expr(1);
    SgClassType *classType = isSgClassType(typeop->type());
    if (!classType) return;

    SgSymbol *recordName = classType->recordName();
    for (SgExpression *varlist = s->expr(0) ; varlist; varlist=varlist->rhs())
    {
	SgExpression *expr = varlist->lhs();
	SgSymbol *symb = expr->symbol();
	  // ???????
    	//if (is_pointer(symb)) continue;
	for (int i=0; i<recordName->numberOfAttributes(RECORD_ALIAS); i++)
	{
	   SgSymbol * aliasSymb = (SgSymbol *)recordName->attributeValue(i, RECORD_ALIAS);
	   SgSymbol * newSymb = s->getSymbolInCurrentScope(symb->identifier(),aliasSymb->identifier(),NULL);
	   newSymb->setType(*aliasSymb->type());
	   newSymb->setAttribute(aliasSymb->attributes());
	   newSymb->setAttribute(symb->attributes());

	   SgStatement *stmt = newSymb->makeVarDeclStmt();
	     // set attribute op
	   if (newSymb->haveAttribute(POINTER_BIT)) {
		SgExpression *op = new SgExpression(POINTER_OP, NULL,NULL,NULL);
		isSgVarDeclStmt(stmt)->addAttributeOp(*op);
  	   }
	   if (newSymb->haveAttribute(TARGET_BIT)) {
		SgExpression *op = new SgExpression(TARGET_OP, NULL,NULL,NULL);
		isSgVarDeclStmt(stmt)->addAttributeOp(*op);
  	   }
	   s->insertStmtBefore(*stmt, *s->controlParent());
	}
    }
    s->deleteStmt();
}

/***************************************************************************/

SgExpression *recordRef2var(SgStatement *s, SgRecordRefExp *recordExp)
{
    SgSymbol *varName = recordExp->recordName();
    SgSymbol *typeName = isSgClassType(varName->type())->recordName();
    char *name = (char *)malloc(strlen(varName->identifier()) + strlen(typeName->identifier()) +2);
    strcpy(name,varName->identifier());
    strcat(name,"_");
    strcat(name,typeName->identifier());
    SgRecordRefExp *recordList = (SgRecordRefExp *)recordExp->rhs();
    while (recordList) {
	SgSymbol *field = recordList->recordName();
	ASSERT(field);
	  // use type name instead.
      	if (isSgClassType(field->type())) 
	     field=isSgClassType(field->type())->recordName();
	char *str = (char *)malloc(strlen(name)+strlen(field->identifier())+2);
	strcpy(str,name);
	strcat(str,"_");
	strcat(str,field->identifier());
	free(name);
	name=str;
	recordList = (SgRecordRefExp *)recordList->rhs();
    }
    SgSymbol *symb = s->getSymbolInCurrentScope(name);
    SgExpression *exp = new SgVarRefExp(symb);
    recordExp->replace(*exp);
    return exp;
}

void parse_record_in_expression(SgStatement *s, SgExpression *expr)
{
    if (expr)
    switch (expr->variant())
    {
    case RECORD_REF:
      {
	recordRef2var(s, isSgRecordRefExp(expr));
	break;
      }
    case VAR_REF:
    case ARRAY_REF:
	break;
    default: 
      {
	SgExpression *lhs = expr->lhs();
	if (lhs) parse_record_in_expression(s, lhs);
	SgExpression *rhs = expr->rhs();
	if (rhs) parse_record_in_expression(s, rhs);
	break;
      }
    }
}

/***************************************************************************
	trans_allocate()

***************************************************************************/

// after trans1, allocate is splitted as one dummy a time.
void trans_allocate(SgStatement *s)
{
   SgExpression *allocee = s->expr(0)->lhs();
   SgClassType *type = isSgClassType(allocee->type());
   if (!type) return;
   SgSymbol *symb = s->expr(0)->lhs()->symbol();
   
   SgSymbol *recordName = type->recordName();
   ASSERT(recordName);
   for (int i=0; i<recordName->numberOfAttributes(RECORD_ALIAS); i++)
   {
	SgSymbol * aliasSymb = (SgSymbol *)recordName->attributeValue(i, RECORD_ALIAS);
	SgSymbol * newSymb = s->getSymbolInCurrentScope(symb->identifier(),aliasSymb->identifier(),NULL);
	SgExpression *newExp = new SgVarRefExp(newSymb);
	SgStatement *newAlloc = s->copyPtr();
	newAlloc->expr(0)->setLhs(*newExp);
	s->insertStmtBefore(*newAlloc, *s->controlParent());
   }
   s->deleteStmt();
}

/***************************************************************************/

void analyse7(SgStatement * s)
{
  switch(s->variant())
  {
    case GLOBAL:
      {
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
             children[i]=s->childList1(i);
         for(i=0;i<org_numberOfChildren;i++)
             analyse7(children[i]);

        delete [] children;

        break;
      };

   case PROG_HEDR:
      { 
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
             children[i]=s->childList1(i); 

         in_main_program=on;

         for(i=0;i<org_numberOfChildren;i++)
             analyse7(children[i]);

         delete [] children;

         in_main_program=off;
  
         break;
       };

   case PROC_HEDR:
      {
        SgProcHedrStmt * prochedr=(SgProcHedrStmt *)s;
        int org_numberOfChildren=s->numberOfChildrenList1();
        SgStatement * * children=new SgStatement * [org_numberOfChildren];

        int i;

        for(i=0;i<org_numberOfChildren;i++)
            children[i]=s->childList1(i);

        in_procedure=on;

        for(i=0;i<org_numberOfChildren;i++)
             analyse7(children[i]);
     
        delete [] children;

        in_procedure=off;

        break;
      };

    case FUNC_HEDR:
      {
         SgFuncHedrStmt * prochedr=(SgFuncHedrStmt *)s;
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
              children[i]=s->childList1(i);

          in_function=on;

         for(i=0;i<org_numberOfChildren;i++)
             analyse7(children[i]);

         delete [] children;

         in_function=off;

         break;
      };  
  case FOR_NODE:
  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse7(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);			

      for (i=0; i<org_numberOfChildren2; i++)
	analyse7(children2[i]);
      
      delete [] children2;

      break;

    }
   // statements of interest
  case STRUCT_DECL:
    {
      trans_record_decl(s);
      break;
    }
  case VAR_DECL:
    {
      trans_record_in_vardecl(s);
      break;
    }
  case FORALL_STMT:
    {
	s=s->childList1(0);
    }
  case ASSIGN_STAT:
    {
      parse_record_in_expression(s, s->expr(0));
      parse_record_in_expression(s, s->expr(1));
      break;
    }
  case ALLOCATE_STMT:
    {
      trans_allocate(s);
      break;
    }
  case READ_STAT:
  case PRINT_STAT:
  case WRITE_STAT:
    {
      parse_record_in_expression(s, s->expr(1));
      break;
    }
  }
 
}
