//
//	DataTable.cc
//		Implement CDataTable, defined DataTable.h
//

#define _CPP
#include "headers.h"
#include "Xheaders.h"

#include "ProfUtil.h"
#include "DataTable.h"

/////////////////////////////////////////////////////////////////
//	Implemention
static int g_nSortCol=0;
static int g_bSortAsc=TRUE;

static int SortProc(const void *p1, const void *p2)
{
	const int *a1=(const int *)p1;
	const int *a2=(const int *)p2;
	return (g_bSortAsc? a1[g_nSortCol]-a2[g_nSortCol] :
						a2[g_nSortCol]-a1[g_nSortCol] );
}

CDataTable::CDataTable(int nColumn)
{
	assert(nColumn>0);
	m_nColumn=nColumn;
	for (int i=0; i<nColumn; i++)
	{
		m_aWidth.insert(m_aWidth.end(), 10);
	}
}

BOOL CDataTable::SetShape(int nColumn)
{
	RemoveAll();
	m_nColumn=nColumn;
	while (!m_aWidth.empty())
	{
		m_aWidth.pop_back();
	}
	for (int i=0; i<nColumn; i++)
	{
		m_aWidth.insert(m_aWidth.end(), 10);
	}
	return TRUE;
}

BOOL CDataTable::SetWidth(int nScale, int *aWidth)
{
	for (int i=0; i<m_nColumn; i++)
		m_aWidth[i]=aWidth[i]/nScale;
/*
		if (aWidth[i]>=90)
			m_aWidth[i]=aWidth[i]/(nScale-2);
		else if (aWidth[i]>=70)
			m_aWidth[i]=aWidth[i]/(nScale-2);
		else
			m_aWidth[i]=aWidth[i]/(nScale-1);
*/
	return TRUE;
}

BOOL CDataTable::SetWidth(int nColumn, int nWidth)
{
	assert(nColumn>=0 && nColumn<m_nColumn);
	assert(nWidth<32);
	m_aWidth[nColumn]=nWidth;
	return TRUE;
}

BOOL CDataTable::AddRow(int *aData)
{
	assert(NULL!=aData);
	for (int i=0; i<m_nColumn; i++)
	{
		m_aData.insert(m_aData.end(), aData[i]);
	}
	return TRUE;
}

BOOL CDataTable::RemoveAll()
{
	while (!m_aData.empty())
	{
		m_aData.pop_back();
	}
	return TRUE;
}

BOOL CDataTable::DeleteRows(int nStart, int nCount=1)
{
	return FALSE;
}

BOOL CDataTable::SetData(int nRow, int nCol, int nVal)
{
	int nOffset=nRow*m_nColumn+nCol;
	assert(nOffset<m_aData.size());
	m_aData[nOffset]=nVal;
	return TRUE;
}

BOOL CDataTable::Sort(int nCol, BOOL bAsc) //, CDataTable &destTable)
{
	//destTable.SetShape(m_nColumn, m_aWidth);
	assert(nCol>=0 && nCol<m_nColumn);
	//	setup sort parameters
	g_nSortCol=nCol;
	g_bSortAsc=bAsc;
	//	sort
	qsort((void *)m_aData.begin(), RowCount(),
		m_nColumn*sizeof(int), SortProc);
	return TRUE;
}

BOOL CDataTable::DisplayMerge(Widget w, int nStart, int nStride)
{
	nStride=2;	//	force stride 2 now

	XmListDeleteAllItems(w);

	if (m_aData.size()<=0)
		return TRUE;

	int nOffset=0;
	for (int nRow=0; nRow<RowCount(); nRow++)
	{
		char szLine[1024];
		szLine[0]='\0';

		for (int nCol=0; nCol<m_nColumn; nCol++)
		{
			char szBuf[80];
			char szBuf1[80];
			char szFormat[10];
			int nDataCol=
				nCol>=nStart? nStart+(nCol-nStart)*nStride : nCol;

			if (m_aWidth[nCol]>0)
			{
				if (nCol<nStart)
				{
					sprintf(szFormat, "%%%dd", m_aWidth[nCol]);
					sprintf(szBuf, szFormat, m_aData[nOffset]);
				}
				else
				{
					sprintf(szBuf1, " %d/%d", 
						GetData(nRow, nDataCol), GetData(nRow, nDataCol+1));
					int len=strlen(szBuf1);
					for (int n=0; n<m_aWidth[nCol]-len; n++)
					{
						szBuf[n]=' ';
					}
					szBuf[m_aWidth[nCol]-len]='\0';
					strcat(szBuf, szBuf1);
				}
				if (strlen(szLine)+strlen(szBuf)<sizeof(szLine))
					strcat(szLine, szBuf);
			}
			nOffset++;
			if (nDataCol>=m_nColumn-2)
				break;
		}
		strcat(szLine, " ");
		CXmString xstr(szLine);
		XmListAddItem(w, xstr.m_str, 0);
	}

	return TRUE;
}

BOOL CDataTable::Display(Widget w, BOOL bSummary=TRUE)
{
	XmListDeleteAllItems(w);

	if (m_aData.size()<=0)
		return TRUE;

	int nOffset=0;
	for (int nRow=0; nRow<RowCount(); nRow++)
	{
		char szLine[1024];
		szLine[0]='\0';

		for (int nCol=0; nCol<m_nColumn; nCol++)
		{
			char szBuf[80];
			char szFormat[10];
			if (m_aWidth[nCol]>0)
			{
				sprintf(szFormat, "%%%dd ", m_aWidth[nCol]-1);
				sprintf(szBuf, szFormat, m_aData[nOffset]);
				if (strlen(szLine)+strlen(szBuf)<sizeof(szLine))
					strcat(szLine, szBuf);
			}
			nOffset++;
		}
		CXmString xstr(szLine);
		XmListAddItem(w, xstr.m_str, 0);
	}

	//	add summary rows
	if (bSummary)
	{
		CXmString xstr("---------");
		XmListAddItem(w, xstr.m_str, 0);

		char szMaxLine[1024];
		char szMinLine[1024];
		char szAvgLine[1024];
		char szTotalLine[1024];
		char szFormat[10];

		sprintf(szFormat, "%%%ds", m_aWidth[0]);
		sprintf(szMaxLine, szFormat, "Max");
		sprintf(szMinLine, szFormat, "Min");
		sprintf(szAvgLine, szFormat, "Average");
		sprintf(szTotalLine, szFormat, "Total");
		for (int nCol=1; nCol<m_nColumn; nCol++)
		{
			char szBuf[80];
			int nMax, nMin, nAvg, nTotal;

			if (m_aWidth[nCol]>0)
			{
				GetSummary(m_aData, m_nColumn, nCol, nMax, nMin, nAvg, nTotal);
				sprintf(szFormat, "%%%dd ", m_aWidth[nCol]-1);
				sprintf(szBuf, szFormat, nMax);
				if (strlen(szMaxLine)+strlen(szBuf)<sizeof(szMaxLine))
					strcat(szMaxLine, szBuf);
				sprintf(szBuf, szFormat, nMin);
				if (strlen(szMinLine)+strlen(szBuf)<sizeof(szMinLine))
					strcat(szMinLine, szBuf);
				sprintf(szBuf, szFormat, nAvg);
				if (strlen(szAvgLine)+strlen(szBuf)<sizeof(szAvgLine))
					strcat(szAvgLine, szBuf);
				sprintf(szBuf, szFormat, nTotal);
				if (strlen(szTotalLine)+strlen(szBuf)<sizeof(szTotalLine))
					strcat(szTotalLine, szBuf);
			}
		}
		CXmString xMaxStr(szMaxLine);
		CXmString xMinStr(szMinLine);
		CXmString xAvgStr(szAvgLine);
		CXmString xTotalStr(szTotalLine);
		XmListAddItem(w, xMaxStr.m_str, 0);
		XmListAddItem(w, xMinStr.m_str, 0);
		XmListAddItem(w, xAvgStr.m_str, 0);
		XmListAddItem(w, xTotalStr.m_str, 0);
	}
	return TRUE;
}

CharBuf aColorMap[]=
{
	"Red",
	"Blue",
	"Green",
	"Yellow",
	"Cyan",
	"Gray",
	"White",
	"Black"
};

int CDataTable::GetColColor(CDC *pDC, int nCol)
{
	return pDC->GetColor(aColorMap[nCol % numberof(aColorMap)]);
}

BOOL CDataTable::DisplayChart(Widget drawArea,
    vector <int> &aDispCols, int nAnnoCol, int nType, CRect &rcChart)
{
	assert(aDispCols.size()>0);
	if (RowCount()<=0) return FALSE;

	int nColCount=0;
	double dfUp=INT_MIN, dfLow=INT_MAX, dfTotal=0, dfAvg=0;

	for (int nRow=0; nRow<RowCount(); nRow++)
	{
		for (int nCol=0; nCol<aDispCols.size(); nCol++)
		{
			int nVal=GetData(nRow, aDispCols[nCol]);
			dfUp=max((double)nVal, dfUp);
			dfLow=min((double)nVal, dfLow);
			dfTotal+=nVal;
			dfAvg=dfTotal/RowCount();
		}
	}

	if (nType==chart_bar)
	{
		DrawBarChart(drawArea, "", nAnnoCol, aDispCols, dfUp, dfLow, rcChart);
	}
	else if (nType==chart_pie)
	{
		DrawPieChart(drawArea, "", nAnnoCol, aDispCols[0], dfTotal, rcChart);
	}
	else if (nType==chart_radar)
	{
		DrawRadarChart(drawArea, "", nAnnoCol, 
			aDispCols[0], dfAvg, dfUp, rcChart);
	}
	return TRUE;
}

BOOL CDataTable::DrawBarChart(Widget drawArea, char *lpszAxis, int nAnnoCol,
	vector <int> &aDispCols, double dfUp, double dfLow, CRect &rcChart)
{
	CDC dc(drawArea);
	CDC *pDC=&dc;

	if(dfLow>=dfUp)
		return FALSE;

	double dfInterval, dfIntMax, dfIntMin;
	double dfFactor=1;
	int nLevel;

	dfInterval=dfUp-dfLow;
	dfIntMax=dfInterval/5;
	dfIntMin=dfInterval/10;
	while (dfIntMin>1)
	{
		if (dfIntMin/2<1)
		{
			dfFactor*=2;
			dfIntMin/=2;
			dfIntMax/=2;
		}
		else if (dfIntMin/5<1)
		{
			dfFactor*=5;
			dfIntMin/=5;
			dfIntMax/=5;
		}
		else
		{
			dfFactor*=10;
			dfIntMin/=10;
			dfIntMax/=10;
		}
	}
	while (dfIntMax<1)
	{
		if (dfIntMax*2>1)
		{
			dfFactor/=2;
			dfIntMin*=2;
			dfIntMax*=2;
		}
		else if (dfIntMax*5>1)
		{
			dfFactor/=5;
			dfIntMin*=5;
			dfIntMax*=5;
		}
		else
		{
			dfFactor/=10;
			dfIntMin*=10;
			dfIntMax*=10;
		}
	}
	dfLow=((int)((dfLow-dfFactor/2)/dfFactor))*dfFactor;
	dfUp=((int)((dfUp+dfFactor)/dfFactor))*dfFactor;
	nLevel=(int)((dfUp-dfLow)/dfFactor);

	BOOL bSingleCol;
	bSingleCol=(aDispCols.size()==1);

	int nRowCount=RowCount();
	int nColCount=aDispCols.size();

	//pDC->SelectFont("10x20");
	pDC->SelectFont("9x15");
	pDC->SelectColor("Black");

	int nLeftMargin=rcChart.Width()/10;
	int nBottomMargin=rcChart.Height()/10;

	CharBuf strScaleWidth;
	sprintf(strScaleWidth, "%0.0lf", dfUp);
	CSize size(strlen(strScaleWidth)*12, 20);

	size.cx+=15;
	size.cy+=10;
	if (nLeftMargin<size.cx) nLeftMargin=size.cx;
	if (nBottomMargin<size.cy) nBottomMargin=size.cy;

	//	draw bars and annotations
	CRect rcBound;
	rcBound.left=rcChart.left+nLeftMargin;
	rcBound.right=rcChart.right-rcChart.Width()/20;
	rcBound.top=rcChart.top+rcChart.Height()/20;
	rcBound.bottom=rcChart.bottom-nBottomMargin;

	// Draw Scale
	CRect rcScaleBound;
	rcScaleBound.left=rcChart.left;
	rcScaleBound.right=rcChart.left+nLeftMargin;
	rcScaleBound.top=rcBound.top;
	rcScaleBound.bottom=rcBound.bottom;

	for (int i=0; i<=nLevel; i++)
	{
		double dfScale=dfLow+i*dfFactor;
		CharBuf strScale;

		int y=(long)(rcBound.bottom-rcScaleBound.Height()*
			((dfScale-dfLow)/(dfUp-dfLow)));
		
		sprintf(strScale, "%0.0lf", dfScale);

		pDC->TextOut(rcScaleBound.left+5, 
			y+5, strScale);
		pDC->MoveTo(rcBound.left, y);
		pDC->LineTo(rcBound.right+1, y);
	}

	CRect rcValue(rcBound.left,0,0,rcBound.bottom-1);
	int nBarWidth=
		bSingleCol? rcBound.Width()/nRowCount :
			rcBound.Width()/(aDispCols.size()*nRowCount + (nRowCount+1)/2);

	for (int nRow=0; nRow<nRowCount; nRow++)
	{
		CharBuf strAnno;

		if (!bSingleCol)
		{
			rcValue.left=rcBound.left+rcBound.Width()*nRow/nRowCount;
			rcValue.left+=nBarWidth/2;
		}

		if (nAnnoCol>=0)
		{
			pDC->SelectColor("Black");
			sprintf(strAnno, "%d", GetData(nRow, nAnnoCol));
			pDC->TextOut(rcValue.left, rcValue.bottom+20,
				strAnno);
		}

		for (int nCol=0; nCol<nColCount; nCol++)
		{	//	draw a bar
			int nColor;
			if (nColCount>1)
				nColor=GetColColor(pDC, nCol);
			else nColor=GetColColor(pDC, nRow);

			pDC->SelectColor(nColor);
			pDC->SelectBrush(TRUE);
			rcValue.right=rcValue.left+nBarWidth-1;
			rcValue.top=(long)(rcValue.bottom-
				rcBound.Height()*
					((GetData(nRow, aDispCols[nCol])-dfLow)/(dfUp-dfLow)));
			pDC->Rectangle(rcValue);
			pDC->SelectColor("Black");
			pDC->SelectBrush(FALSE);
			pDC->Rectangle(rcValue);

			rcValue.left=rcValue.right+1;
		}
	}
	
	// Draw Axis
	pDC->SelectColor("Black");
	pDC->MoveTo(rcBound.left, rcBound.top);
	pDC->LineTo(rcBound.left, rcBound.bottom);
	pDC->LineTo(rcBound.right+1, rcBound.bottom);
	pDC->LineTo(rcBound.right+1, rcBound.top);

	return TRUE;
}

extern "C" double cos(double);
extern "C" double sin(double);
#define PI 3.1415926

BOOL CDataTable::DrawPieChart(Widget drawArea, char *lpszAxis, 
	int nAnnoCol, int nDataCol, double dfTotal, CRect &rcChart)
{
	CDC dc(drawArea);
	CDC *pDC=&dc;

	//pDC->SelectFont("10x20");
	pDC->SelectFont("9x15");
	pDC->SelectBrush(TRUE);

	CRect rcBound;
	int nRadius=min(rcChart.Width(), rcChart.Height())*7/20;
	rcBound.left=(rcChart.left+rcChart.right)/2-nRadius;
	rcBound.right=(rcChart.left+rcChart.right)/2+nRadius;
	rcBound.top=(rcChart.top+rcChart.bottom)/2-nRadius;
	rcBound.bottom=(rcChart.top+rcChart.bottom)/2+nRadius;

	double anStart,anTemp,anPrev;
	CPoint ptStart,ptEnd,ptText;
	long lCenterX,lCenterY;
	int nRowCount=RowCount();

	if(nRowCount<=1) // Draw ellipse
	{
		pDC->SelectColor(GetColColor(pDC, 0));
		pDC->Ellipse(rcBound);
		return TRUE;
	}
	lCenterX=(long)(((double)rcBound.left+rcBound.right)/2);
	lCenterY=(long)(((double)rcBound.top+rcBound.bottom)/2);

	CharBuf strAnno;

	double dfCurTotal=0;
	anStart=0.0;

	ptStart.x=(long)(((double)rcBound.left+rcBound.right)/2);
	ptStart.y=rcBound.bottom;

	for(int i=0;i<nRowCount;i++)
	{
		// Get end point and draw the pie
		anPrev=dfCurTotal/dfTotal*2*PI;
		dfCurTotal+=GetData(i, nDataCol);
		anTemp=dfCurTotal/dfTotal*2*PI;
		double anTemp1=anTemp+0.01;

		ptText.x=(long)(1.2*nRadius*cos(-(anPrev+anTemp)/2)+lCenterX);
		ptText.y=(long)(1.2*nRadius*sin(-(anPrev+anTemp)/2)+lCenterY);

		pDC->SelectColor(GetColColor(pDC, i));
		pDC->Arc(rcBound, (int)(anPrev*64*180/PI), 
			(int)((anTemp-anPrev)*64*180/PI));
		if (nAnnoCol>=0)
		{
			sprintf(strAnno, "%d", GetData(i, nAnnoCol));
			CSize sizeText(strlen(strAnno)*12, 100);

			CRect rcText;
/*
			switch ( (int)((anPrev+anTemp-PI)*2/PI+0.5) )
			{
			case 0:
			case 8:
				ptText=CPoint(ptText.x-sizeText.cx, ptText.y);
				break;
			case 1:
				ptText=CPoint(ptText.x-sizeText.cx, ptText.y);
			case 2:
				ptText=CPoint(ptText.x-sizeText.cx, ptText.y);
				break;
			case 3:
				ptText=CPoint(ptText.x-sizeText.cx, ptText.y);
				break;
			case 4:
				ptText=CPoint(ptText.x, ptText.y+sizeText.cy);
				break;
			case 5:
				ptText=CPoint(ptText.x, ptText.y+sizeText.cy);
				break;
			case 6:
				ptText=CPoint(ptText.x, ptText.y);
				break;
			case 7:
				ptText=CPoint(ptText.x, ptText.y);
				break;
			}
*/
			pDC->TextOut(ptText.x, ptText.y, strAnno);
		}
		ptEnd.x=(long)(nRadius*cos(anTemp)+lCenterX);
		ptEnd.y=(long)(nRadius*sin(anTemp)+lCenterY);
		ptStart=ptEnd;
	}
	pDC->SelectBrush(FALSE);
	pDC->SelectColor("Black");
	pDC->Ellipse(rcBound);

	return TRUE;
}

BOOL CDataTable::DrawRadarChart(Widget drawArea, char *lpszAxis, 
	int nAnnoCol, int nDataCol, 
	double dfAvg, double dfMax, CRect &rcChart)
{
	if (dfMax<=0) return FALSE;
	CDC dc(drawArea);
	CDC *pDC=&dc;

	//pDC->SelectFont("10x20");
	pDC->SelectFont("9x15");
	pDC->SelectBrush(TRUE);

	CRect rcBound;
	int nRadius=min(rcChart.Width(), rcChart.Height())*7/20;
	rcBound.left=(rcChart.left+rcChart.right)/2-nRadius;
	rcBound.right=(rcChart.left+rcChart.right)/2+nRadius;
	rcBound.top=(rcChart.top+rcChart.bottom)/2-nRadius;
	rcBound.bottom=(rcChart.top+rcChart.bottom)/2+nRadius;

	long lCenterX=(long)(((double)rcBound.left+rcBound.right)/2);
	long lCenterY=(long)(((double)rcBound.top+rcBound.bottom)/2);

	pDC->SelectBrush(TRUE);
	pDC->SelectColor("Red");
	pDC->Ellipse(rcBound);

	int nAvgRadius=(int)(nRadius*dfAvg/dfMax); 
	CRect rcAvg(lCenterX-nAvgRadius, lCenterY-nAvgRadius,
		lCenterX+nAvgRadius, lCenterY+nAvgRadius);

	double angle;
	XPoint aPt[MAX_PROC+1];
	XPoint aPtRadius[MAX_PROC];
	CPoint ptText;
	int nRowCount=RowCount();

	if(nRowCount<=2 || nRowCount>MAX_PROC) // Draw ellipse
	{
		return TRUE;
	}

	CharBuf strAnno;

	pDC->SelectColor("Black");
	for(int i=0;i<nRowCount;i++)
	{
		// Get end point and draw the pie
		angle=((double)i)/nRowCount*2*PI;
		double dfRatio=GetData(i, nDataCol)/dfMax;

		ptText.x=(long)(1.2*nRadius*cos(angle)+lCenterX);
		ptText.y=(long)(1.2*nRadius*sin(angle)+lCenterY);

		aPt[i].x=(short)(dfRatio*nRadius*cos(angle)+lCenterX);
		aPt[i].y=(short)(dfRatio*nRadius*sin(angle)+lCenterY);
	
		aPtRadius[i].x=(short)(nRadius*cos(angle)+lCenterX);
		aPtRadius[i].y=(short)(nRadius*sin(angle)+lCenterY);
	
		if (nAnnoCol>=0)
		{
			sprintf(strAnno, "%d", GetData(i, nAnnoCol));
			CSize sizeText(strlen(strAnno)*12, 100);
			pDC->TextOut(ptText.x, ptText.y, strAnno);
		}
	}
	aPt[nRowCount]=aPt[0];
	pDC->SelectColor("Green");
	pDC->SelectBrush(TRUE);
	pDC->Polygon(aPt, nRowCount+1);
	pDC->SelectColor("Black");
	pDC->SelectBrush(FALSE);
	pDC->Polygon(aPt, nRowCount+1);

	for (int i=0; i<nRowCount; i++)
	{
		pDC->MoveTo(lCenterX, lCenterY);
		pDC->LineTo(aPtRadius[i].x, aPtRadius[i].y);
	}

	pDC->SelectBrush(FALSE);
	pDC->SelectColor("Black");
	pDC->Ellipse(rcBound);
	pDC->Ellipse(rcAvg);

	return TRUE;
}

static CharBuf aSzCommColor[9]=
{
	"green",
	"chartreuse",
	"GreenYellow",
	"yellow",
	"gold",
	"orange",
	"DarkOrange1",
	"red",
	"red",
};

static void GetProcPos(int nProcNum, int nProc, int &x, int &y)
{
	double angle=((double)nProc)/nProcNum*2*PI;
	x=(long)(180*cos(angle)+225);
	y=(long)(180*sin(angle)+190);
}

BOOL CDataTable::DrawCommChart(Widget drawArea, int nRow, CRect &rcChart)
{
	CDC dc(drawArea);

	//	classify comm volumn
	int nMaxVal=0;
	int nMaxRow, nMinRow, nRowCount=RowCount();
	
	nMaxRow=nRowCount;
	nMinRow=0;

	for (int i=nMinRow; i<nMaxRow; i++)
	{
		for (int j=2; j<=nRowCount*2; j+=2)
		{
			if (nMaxVal<GetData(i, j))
				nMaxVal=GetData(i, j);
		}
	}
	int nScale=nMaxVal % 8? nMaxVal/8+1 : nMaxVal/8;
	if (nScale==0)
	{
		dc.SelectColor("black");
		dc.SelectFont("8x13");
		dc.TextOut(160, 200, "No Communication");
		return TRUE;
	}
	
	if (nRow==-1)
	{
		nMaxRow=nRowCount;
		nMinRow=0;
	}
	else
	{
		nMaxRow=nRow+1;
		nMinRow=nRow;
	}

	//	draw comm lines
	for (int i=nMinRow; i<nMaxRow; i++)
	{
		for (int j=0; j<nRowCount; j++)
		{
			if (i==j) continue;
			int nSend=GetData(i, j*2+2);
			int nRecv= nRow!=-1? GetData(j, i*2+2) : 0;
			int x1, y1, x2, y2;

			GetProcPos(nRowCount, i, x1, y1);
			GetProcPos(nRowCount, j, x2, y2);
			if (nSend>0)
			{
				dc.SelectColor(aSzCommColor[nSend/nScale]);
				dc.SelectPen(nSend/nScale+1);
				dc.MoveTo((x1+x2)/2, (y1+y2)/2);
				dc.ArrowLineTo(x2, y2);
			}
				
			if (nRecv>0)
			{
				dc.SelectColor(aSzCommColor[nRecv/nScale]);
				dc.SelectPen(nRecv/nScale+1);
				dc.MoveTo((x1+x2)/2, (y1+y2)/2);
				dc.ArrowLineTo(x1, y1);
			}
		}
	}

	//	draw legend table
	dc.SelectFont("8x13");
	for (int i=0; i<8; i++)
	{
		CharBuf szLegend;
		int nMin=i*nScale+1;
		int nMax=nMin+nScale-1;

		sprintf(szLegend, "%d - %d", nMin, nMax);
		dc.SelectColor(aSzCommColor[i]);
		dc.SelectPen((i+1));
		dc.MoveTo(450, i*50+10);
		dc.LineTo(550, i*50+10);
		dc.SelectColor("black");
		dc.TextOut(450, i*50+35, szLegend);
	}
	return TRUE;
}
