//
//	DataUI.cc
//		Implement UI for displaying distributed array information
//		including
//			array distribution information
//			array communication information
//

#define _CPP
#define _PROCINFO
#include "headers.h"
#include "Xheaders.h"

//#include "Prof.h"
#include "ProfP.h"
#include "ProfUtil.h"

////////////////////////////////////////////////////////////////
//	declarations
int PrepareDataLayout();
int ShowDataLayout(int nSymbolID);
int PrepareDataComm();
int ShowDataComm(int nSymbolID, int nStmtID);

////////////////////////////////////////////////////////////////
//	Implementation
extern Widget topLevel;
static Widget wDataDistDlg=NULL;
static Widget wSymbolList=NULL;
static Widget wDataDistArea=NULL;

static int nCurSymbol=0;

int GotoSymbol(int nSymbolID);
void GetSymbol(CharBuf szSymbol, int nSymbolID);

void CbkSelSymbolList(Widget w, void *pData, XmListCallbackStruct *pCallData);
void CbkExposeDataDist(Widget w, void *pData,
	XmDrawingAreaCallbackStruct *pCallData);

int PrepareDataLayout() 
{
	if (wDataDistDlg!=NULL)
	{
		XtDestroyWidget(wDataDistDlg);
		wDataDistDlg=NULL;
	}
	return TRUE;
}

int ShowDataLayout(int nSymbolID)
{
	int    n=0, m;
	Arg    wargs[MAXARGS];
	char   str[100],*s1;
	XmString  xmstr;
	int    n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
	int    lineno,no=0;
	
	if (wDataDistDlg!=NULL)
	{
		XtManageChild(wDataDistDlg);
		return GotoSymbol(nSymbolID);
	}
	//	create UI component
	n=0;
	XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
	XtSetArg(wargs[n],XmNwidth,630);	n++;
	XtSetArg(wargs[n],XmNheight,490);	n++;
	wDataDistDlg=XmCreateBulletinBoardDialog(topLevel,
		"Data Distribution\n",wargs,n);
	
	XtAddCallback(wDataDistDlg, XmNdestroyCallback,
		CbkDestroyWidget, &wDataDistDlg);
	
	XtSetArg(wargs[0],XmNx,150);
	XtSetArg(wargs[1],XmNy,20);
	XtSetArg(wargs[2],XmNwidth,450);
	XtSetArg(wargs[3],XmNheight,450);
	Widget wDataDistArea=XmCreateDrawingArea(wDataDistDlg,
		"DrawArea", wargs, 4);
	XtManageChild(wDataDistArea);

	XtAddCallback(wDataDistArea, XmNexposeCallback,
		CbkExposeDataDist, NULL);
	
	XtSetArg(wargs[0],XmNx,30);
	XtSetArg(wargs[1],XmNy,400);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,40);
	Widget bt=XmCreatePushButton(wDataDistDlg,
		"Close", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkUnmanageDlg, (void*)wDataDistDlg);

	XtSetArg(wargs[0],XmNx,15);
	XtSetArg(wargs[1],XmNy,20);
	XtSetArg(wargs[2],XmNwidth,120);
	XtSetArg(wargs[3],XmNheight,360);
	wSymbolList=XmCreateList(wDataDistDlg,
		"List", wargs, 4);
	XtAddCallback(wSymbolList, XmNbrowseSelectionCallback,
		CbkSelSymbolList, NULL);
	XtAddCallback(wSymbolList, XmNsingleSelectionCallback,
		CbkSelSymbolList, NULL);
	XtManageChild(wSymbolList);

	//	Prepare symbols in list
	CProcInfo *pProc=GetProc(0);
	if (pProc!=NULL)
	{
		POSITION pos=pProc->GetFirstArrayPos();
		while (-1!=pos)
		{
			CDistArrayInfo *pArray=pProc->GetNextArray(pos);
			CharBuf szSymbol;
	
			GetSymbol(szSymbol, pArray->m_nArrayID);
			CXmString strItem(szSymbol);
			XmListAddItem(wSymbolList, strItem.m_str, 0);
		}
	}
	XtManageChild(wDataDistDlg);
	return GotoSymbol(nSymbolID);
}

void GetSymbol(CharBuf szSymbol, int nSymbolID)
{
	sprintf(szSymbol, "%d", nSymbolID);
}

int GotoSymbol(int nSymbol)
{
	if (nSymbol!=0)
	{
		CProcInfo *pProc=GetProc(0);
		if (pProc==NULL) return FALSE;
		CDistArrayInfo *pArray=pProc->FindDistArray(nSymbol);
		if (pArray==NULL)
			return FALSE;
		else nCurSymbol=nSymbol;
	    XClearWindow(XtDisplay(wDataDistArea), XtWindow(wDataDistArea));
    	CbkExposeDataDist(wDataDistArea, NULL, NULL);
	}
	return TRUE;
}

void CbkSelSymbolList(Widget w, void *pData, XmListCallbackStruct *pCallData)
{
	CProcInfo *pProc=GetProc(0);
	if (pProc==NULL) return;
	CDistArrayInfo *pArray=pProc->GetArray(pCallData->item_position-1);
	if (pArray==NULL)
		return;
	else nCurSymbol=pArray->m_nArrayID;
    XClearWindow(XtDisplay(wDataDistArea), XtWindow(wDataDistArea));
   	CbkExposeDataDist(wDataDistArea, NULL, NULL);

	return;
}

void CbkExposeDataDist(Widget w, void *pData,
	XmDrawingAreaCallbackStruct *pCallData)
{
    if ( pCallData!=NULL && ((XExposeEvent*)(pCallData->event))->count>0)
        return;
    
	CProcInfo *pProc=GetProc(0);
	if (pProc==NULL) return;
	CDistArrayInfo *pArray=pProc->FindDistArray(nCurSymbol);
	if (pArray!=NULL)
	{
	}
}

int PrepareDataComm() { return FALSE; };
int ShowDataComm(int nSymbolID, int nStmtID) { return FALSE; };

