//
//	EventUI.cc
//		Implement UI for displaying event information
//		including
//			event static graph
//			event animation
//			event time seeking
//

#define _CPP
#define _PROCINFO
#define _RUNTIME
#include "headers.h"
#include "Xheaders.h"

//#include "Prof.h"
#include "ProfP.h"
#include "ProfUtil.h"

#define PI 3.1415926

////////////////////////////////////////////////////////////////
//	declarations
int PrepareEventGraph();
int ShowEventGraph(int bAnnimate);

////////////////////////////////////////////////////////////////
//	Implementation
int RoundUp(int nVal)
{
	if (nVal<1000) return 1000;
	else
	{
		int i;
		for (i=10; nVal / i >=10; i*=10);
		if (nVal / i==1)
			nVal=15*i/10;
		else nVal=(nVal/i + 1)*i;
	}
	return nVal;
}

int Swap(int &x1, int &x2)
{
	int t=x1;
	x1=x2;
	x2=t;
	return t;
}

int RoundUp(int nVal1, int nVal2)
{
	int t1=nVal1, t2=nVal2;
	int scale=1;
	int i=0;
	int aScale[3]={ 10, 5, 2 };

	if ((t1<0 && t2>=0) || (t1>=0 && t2<0))
		return 0;
	if (scale==1 && i<3)
	{
		while (t1/aScale[i]!=t2/aScale[i])
		{
			if (t1/aScale[i]==0)
				return t1*scale;
			if (t2/aScale[i]==0)
				return t2*scale;
			t1/=aScale[i];
			t2/=aScale[i];
			scale*=aScale[i];
		}
		i++;
	}
	return (max(t1,t2)*scale);
}

#define G_SIZE 500
class CEventGraphInfo
{
public:
	int m_nElapse;	//	total time elapse
	int m_nStartTime, m_nEndTime;	//	visible time segment
	int m_aStatus[MAX_PROC][G_SIZE];
	int m_aWidth[MAX_PROC][G_SIZE];

public:
	CEventGraphInfo();
	int Zoom(double dfRatio);
	int ZoomTo(int nStart, int nEnd=0);
	int LineUp();
	int LineDown();
	int Reset();

	int GetX(int nTime);

	int Draw(Widget w);

protected:
	int Recalc();
};

CEventGraphInfo::CEventGraphInfo()
{
}

int CEventGraphInfo::LineUp()
{
	if (m_nStartTime<=0)
		return FALSE;

	int nSpan=m_nEndTime-m_nStartTime+1;

	//int nGap=RoundUp(nSpan*5/6, nSpan/12);
	int nGap1=nSpan*5/6;
	int nGap2=nSpan*6/7;

	if (nGap1==0) return FALSE;

	int nPoint=RoundUp(m_nStartTime-nGap1, m_nStartTime-nGap2);

	return ZoomTo(nPoint);
}

int CEventGraphInfo::LineDown()
{
	if (m_nEndTime>=m_nElapse-1)
		return FALSE;

	int nSpan=m_nEndTime-m_nStartTime+1;

	int nGap1=nSpan*5/6;
	int nGap2=nSpan*6/7;

	if (nGap1==0) return FALSE;

	int nPoint=RoundUp(m_nStartTime+nGap1, m_nStartTime+nGap2);

	return ZoomTo(nPoint);
}

int CEventGraphInfo::ZoomTo(int nStart, int nEnd)
{
	if (nEnd==0)
	{
		if (nStart==m_nStartTime)
			return FALSE;
		if (nStart<=m_nStartTime && m_nStartTime<=0)
			return FALSE;
		if (nStart>=m_nStartTime && m_nEndTime>=m_nElapse-1)
			return FALSE;
	}

	int nSpan=m_nEndTime-m_nStartTime;

	m_nStartTime=nStart;
	if (nEnd!=0) m_nEndTime=nEnd;
	else m_nEndTime=nStart+nSpan;
	return Recalc();
}

int CEventGraphInfo::Zoom(double dfRatio)
{
	int nCenter=(m_nStartTime+m_nEndTime)/2;
	int nSpan=m_nEndTime-m_nStartTime+1;

	if (nSpan<=1000 && dfRatio<=1)
		return FALSE;
	if (nSpan>=m_nElapse && dfRatio>=1)
		return FALSE;

	nSpan=(int)(nSpan*dfRatio);
	if (nSpan<1000) nSpan=1000;
//	m_nStartTime=nCenter-nSpan/2-1;
	m_nEndTime=m_nStartTime+nSpan-1;
	return Recalc();
}

int CEventGraphInfo::Reset()
{
	m_nElapse=RoundUp(GetMaxTime());
	m_nStartTime=0;
	m_nEndTime=m_nElapse-1;
	return Recalc();
}

int CEventGraphInfo::Recalc()
{
	if (m_nStartTime<0)
	{
		m_nEndTime=min(m_nElapse-1, m_nEndTime-m_nStartTime);
		m_nStartTime=0;
	}
	if (m_nEndTime>=m_nElapse)
	{
		m_nStartTime=max(0, m_nStartTime-m_nEndTime+m_nElapse-1);
		m_nEndTime=m_nElapse-1;
	}

	int nSpan=m_nEndTime-m_nStartTime+1;
	int nEventID;

	for (int nProc=0; nProc<nMaxProc+1; nProc++)
	{	//	calc each processor event graph
		int y=480/(nMaxProc+1)+60;
		int nHeight=min(50, 360/(nMaxProc+1));

		CProcInfo *pProc=GetProc(nProc);
		
		if (pProc==NULL)
			continue;

		for (int i=0; i<G_SIZE; i++)
		{
			int nTime=m_nStartTime+(nSpan/G_SIZE*i);
			pProc->GetStatusBySysTime(nTime, nSpan/G_SIZE, 
				m_aStatus[nProc][i], nEventID);
			m_aWidth[nProc][i]=0;
		}
		for (int i=0, nPos=0; i<G_SIZE; i++)
		{
			if (m_aStatus[nProc][i]==m_aStatus[nProc][nPos])
			{
				m_aWidth[nProc][nPos]++;
			}
			else
			{
				nPos=i;
				m_aWidth[nProc][i]=1;
			}
		}
	}
	return TRUE;
}

int CEventGraphInfo::GetX(int nTime)
{
	if (m_nStartTime==m_nEndTime) return 0;
	double x=((double)(nTime-m_nStartTime))/(m_nEndTime-m_nStartTime+1)
		*500+70;
	return (int)x;
}

CharBuf aColors[]=
	{	"green3",
		"blue3",
		"blue3",
		"red3",
		"yellow2"
	};

int CEventGraphInfo::Draw(Widget w)
{
	CDC dc(w);

	int nSpan=m_nEndTime-m_nStartTime+1;

	int nGap=RoundUp(nSpan/10, nSpan/12);
	if (nGap==0) return FALSE;

	CharBuf szPoint;
	int y=20, y1=40;

	CDC sdc[5];

	for (int i=0; i<5; i++)
	{
		sdc[i].Attach(w);
		sdc[i].SelectColor(aColors[i]);
		sdc[i].SelectBrush(TRUE);
	}

	dc.SelectColor("black");
	dc.SelectFont("9x15");

	for (int nPoint=RoundUp(m_nStartTime, m_nStartTime+nGap);
		nPoint<=m_nEndTime+1; nPoint+=nGap)
	{
		int x=GetX(nPoint);
		dc.MoveTo(x, 50);
		dc.LineTo(x, 550);
		if (nGap>100000)
			sprintf(szPoint, "%1.1lfs", ((double)nPoint)/1000000 );
		else if (nGap>1000)
			sprintf(szPoint, "%3.1lfms", ((double)(nPoint % 1000000))/1000 );
		else
			sprintf(szPoint, "%1.3lfms", ((double)(nPoint % 1000000))/1000 );
		if (nPoint!=m_nStartTime)
			dc.TextOut(x-20, Swap(y, y1), szPoint);
	}
	dc.MoveTo(GetX(m_nStartTime), 50);
	dc.LineTo(GetX(m_nStartTime), 550);
	dc.LineTo(GetX(m_nEndTime+1), 550);
	dc.LineTo(GetX(m_nEndTime+1), 50);
	dc.MoveTo(GetX(m_nEndTime+1), 55);
	dc.LineTo(GetX(m_nStartTime), 55);

	sprintf(szPoint, "%1.4lfs", ((double)m_nStartTime)/1000000 );
	dc.TextOut(5, 50, szPoint);

	CharBuf szProc;

	for (int nProc=0; nProc<nMaxProc+1; nProc++)
	{	//	calc each processor event graph
		int y=480*nProc/(nMaxProc+1)+70;
		int nHeight=min(50, 360/(nMaxProc+1));

		sprintf(szProc, "Proc %d", nProc);
		dc.TextOut(5, y+30, szProc);

		CProcInfo *pProc=GetProc(nProc);
		
		if (pProc==NULL)
			continue;
		for (int i=0; i<G_SIZE; i+=m_aWidth[nProc][i])
		{
			if (m_aStatus[nProc][i]>=0)
			{
				sdc[m_aStatus[nProc][i] % numberof(aColors)].Rectangle(i+70, y, 
					min(570, i+70+m_aWidth[nProc][i]-1), y+nHeight-1);
			}
		}
	}
	return TRUE;
}

////////////////////////////
//	CEventAnimate
class CDataPacket
{
public:
	int m_nSource;
	int m_nTarget;
	int m_nSize;
	int m_nStep;
	int m_nID;

public:
	CDataPacket() { m_nSource=-1; m_nStep=0; };
	CDataPacket(int nSource, int nTarget, int nSize, int nID)
	{ m_nSource=nSource; m_nTarget=nTarget; m_nSize=nSize; 
	  m_nID=nID; m_nStep=0; };

	void Set(int nSource, int nTarget, int nSize, int nID)
	{ m_nSource=nSource; m_nTarget=nTarget; m_nSize=nSize; 
	  m_nID=nID; m_nStep=0; };

	BOOL Step() 
	{ m_nStep++; if (m_nStep>=3) return FALSE; else return TRUE; };
	BOOL IsDead() 
	{ return (m_nSource==-1 || m_nStep>=3 || m_nSource==m_nTarget); };
	BOOL Kill() { return m_nSource==-1; };

	void GetPosition(int &x, int &y);

	void Clear(CDC *pDC);
	void Draw(CDC *pDC, CDC *pTextDC);
};

void CDataPacket::Clear(CDC *pDC)
{
	if (IsDead()) return;

	int x, y;
	
	GetPosition(x, y);
	pDC->ClearArea(x-15, y-15, x+15, y+15);
}

void CDataPacket::Draw(CDC *pDC, CDC *pTextDC)
{
	if (IsDead()) return;

	int x, y;
	int size;
	
	GetPosition(x, y);
	int x1, y1;

	double angle=((double)m_nSource)/(nMaxProc+1)*2*PI;

	x1=(long)(200*cos(angle)+290);
	y1=(long)(170*sin(angle)+265);

	if (m_nSize<100)
		size=6;
	else if (m_nSize<1000)
		size=7;
	else if (m_nSize<10000)
		size=8;
	else size=9;

	pDC->MoveTo(x1, y1);
	pDC->ArrowTo(x, y, size);
}

void CDataPacket::GetPosition(int &x, int &y)
{
	int x1, y1;
	int x2, y2;

	double angle=((double)m_nSource)/(nMaxProc+1)*2*PI;

	x1=(long)(200*cos(angle)+290);
	y1=(long)(170*sin(angle)+265);

	angle=((double)m_nTarget)/(nMaxProc+1)*2*PI;

	x2=(long)(200*cos(angle)+290);
	y2=(long)(170*sin(angle)+265);

	x=x2*(m_nStep+1)/4+x1*(3-m_nStep)/4;
	y=y2*(m_nStep+1)/4+y1*(3-m_nStep)/4;
/*
	x=x2*(m_nStep+2)/6+x1*(4-m_nStep)/6;
	y=y2*(m_nStep+2)/6+y1*(4-m_nStep)/6;
	x=x2*((m_nStep+2)*3+(m_nID%3))/18+x1*((5-m_nStep)*3-(m_nID%3))/18;
	y=y2*((m_nStep+2)*3+(m_nID%3))/18+y1*((5-m_nStep)*3-(m_nID%3))/18;
*/
}
	
class CEventAnimate
{
public:
	int m_nElapse;
	int m_nCurTime;
	int m_nSpeed;

	BOOL m_bPause;
	
protected:
	CDataPacket m_aPackets[100];

	BOOL NewDataPacket(int nSource, int nTarget, int nSize, int nID);

public:
	CEventAnimate() {};

	BOOL Reset();

	BOOL SetSpeed(double dblScale);
	BOOL SetTime(int nTime);

	BOOL Step() 
	{ 	
		if (m_nCurTime>=m_nElapse-1)
		{
			m_nCurTime=0;
			Pause();
		}
		else m_nCurTime=min(m_nElapse-1, m_nCurTime+m_nSpeed);
		return TRUE;
	};
	BOOL Continue(){ m_bPause=FALSE; return TRUE; };
	BOOL Pause(){ m_bPause=TRUE; return TRUE; };
	BOOL IsPause() { return m_bPause; };

	BOOL DrawAnimate(Widget w);
	BOOL DrawFixed(Widget w);
};

BOOL CEventAnimate::Reset()
{
	m_bPause=TRUE;
	m_nElapse=GetMaxTime();
	m_nCurTime=0;

	if (m_nElapse<10000) SetSpeed(0.00001);
	else if (m_nElapse<100000) SetSpeed(0.0001);
	else if (m_nElapse<1000000) SetSpeed(0.001);
	else if (m_nElapse<10000000) SetSpeed(0.01);
	else if (m_nElapse<100000000) SetSpeed(0.1);
	else SetSpeed(1);
		
	return TRUE;
}

BOOL CEventAnimate::SetSpeed(double dblScale)
{
	if (dblScale<=0.000001)
		dblScale=0.000001;
	m_nSpeed=(int)(1000000*dblScale);
	return TRUE;
}

BOOL CEventAnimate::SetTime(int nTime)
{
	m_nCurTime=nTime;
	return TRUE;
}

BOOL CEventAnimate::NewDataPacket(int nSource, 
	int nTarget, int nSize, int nID)
{
	for (int i=0; i<numberof(m_aPackets); i++)
	{
		if (m_aPackets[i].IsDead())
		{
			m_aPackets[i].Set(nSource, nTarget, nSize, nID);
			return TRUE;
		}
	}
	return FALSE;
}

CharBuf aBrightColors[]=
	{	"green",
		"blue",
		"blue",
		"red",
		"yellow"
	};

CharBuf aPacketColor[]=
{
	"green",
	"yellow",
	"gold",
	"red",
};

CharBuf aLightColor[]=
{
	"red",
	"green",
	"gray",
};

int aPrevEvent[MAX_PROC];

BOOL CEventAnimate::DrawAnimate(Widget w)
{	//	draw animate part of animation
	static int nStep=0;
	int aSend[MAX_PROC], aSendAmount[MAX_PROC];

	nStep=(nStep+1) % 3;

	CharBuf szBuf;

	CDC dc(w);
	CDC blackDC(w);

	dc.SelectBrush(TRUE);

	blackDC.SelectColor("black");
	blackDC.SelectFont("9x15");
	blackDC.SelectBrush(TRUE);

	CDC sdc[5];
	CDC ddc[4];

	for (int i=0; i<5; i++)
	{
		sdc[i].Attach(w);
		sdc[i].SelectColor(aBrightColors[i]);
		sdc[i].SelectBrush(TRUE);
		sdc[i].SelectFont("9x15");
		sdc[i].SelectPen(2);
	}

	for (int i=0; i<4; i++)
	{
		ddc[i].Attach(w);
		ddc[i].SelectColor(aPacketColor[i]);
		ddc[i].SelectBrush(TRUE);
	}

	sprintf(szBuf, "%0.6lfs", ((double)m_nCurTime)/1000000);
	blackDC.ClearArea(550, 120, 650, 145);
	blackDC.TextOut(550, 140, szBuf);

	// draw active data packet
	if (m_nCurTime>0)
	{
		for (int i=0; i<numberof(m_aPackets); i++)
		{
			m_aPackets[i].Clear(&blackDC);
			m_aPackets[i].Step();
		}
	}
	//	draw processor states
	for (int nProc=0; nProc<=nMaxProc; nProc++)
	{
		int x, y, x2, y2;

		double angle=((double)nProc)/(nMaxProc+1)*2*PI;

		x=(long)(250*cos(angle)+285);
		y=(long)(230*sin(angle)+255);

		x2=(long)(200*cos(angle)+290);
		y2=(long)(170*sin(angle)+265);

		for (int j=nProc+1; j<=nMaxProc; j++)
		{
			int x1, y1;
			double angle=((double)j)/(nMaxProc+1)*2*PI;

			x1=(long)(200*cos(angle)+290);
			y1=(long)(170*sin(angle)+265);

			blackDC.Line(x2, y2, x1, y1);
		}
		blackDC.Rectangle(x-14, y-7, x+34, y+32);
		//	Query processor states
		CProcInfo *pProc=GetProc(nProc);
		int nStatus, nEventID;
		
		if (pProc==NULL)
			continue;

		pProc->GetCommBySysTime(m_nCurTime, m_nSpeed, aSend, aSendAmount);
		for (int l=0; l<MAX_PROC; l++)
		{
			if (aSendAmount[l]>0)
				NewDataPacket(nProc, l, aSendAmount[l], 0);
		}
			
		pProc->GetStatusBySysTime(m_nCurTime, m_nSpeed,
			nStatus, nEventID);
		if (m_nCurTime==0) nStatus=-1;

		dc.SelectColor(aLightColor[nStep]);

		switch (nStatus)
		{
		case stat_serial:
			sdc[nStatus].Line(x+10, y-5, x+10, y+2+13*nStep);
			break;
		case stat_para:
			sdc[nStatus].Line(x-7, y-5, x-7, y+4+13*nStep);
			sdc[nStatus].Line(x+4, y-5, x+4, y+4+13*nStep);
			sdc[nStatus].Line(x+15, y-5, x+15, y+4+13*nStep);
			sdc[nStatus].Line(x+26, y-5, x+26, y+4+13*nStep);
			break;
		case stat_send:
			{
				CExecEvent *pEvent=pProc->GetEvent(nEventID);
				sdc[nStatus].TextOut(x+5, y+19, "S");
/*
				if (pEvent!=NULL && nEventID!=aPrevEvent[nProc])
					NewDataPacket(nProc, pEvent->m_nComTarget, 
						pEvent->m_nComSize, nEventID);
*/
			}
			break;
		case stat_recv:
			sdc[nStatus].TextOut(x+5, y+19, "R");
			break;
		default:
			sdc[0].TextOut(x+5, y+19, "E");
			dc.SelectColor(aLightColor[2]);
		}
		dc.Rectangle(x+25, y+40, x+30, y+45);
		aPrevEvent[nProc]=nEventID;

	}
	// draw active data packet
	if (m_nCurTime>0)
	{
		for (int i=0; i<numberof(m_aPackets); i++)
		{
			CDC *pDC;
	
			if (m_aPackets[i].m_nSize<100)
				pDC=&ddc[0];
			else if (m_aPackets[i].m_nSize<1000)
				pDC=&ddc[1];
			else if (m_aPackets[i].m_nSize<10000)
				pDC=&ddc[2];
			else pDC=&ddc[3];
	
			m_aPackets[i].Draw(pDC, &blackDC);
		}
	}
	return TRUE;
}

BOOL CEventAnimate::DrawFixed(Widget w)
{	//	draw fixed part of animation
	CharBuf szBuf;

	CDC dc(w);
	CDC blackDC(w);

	blackDC.SelectColor("black");
	blackDC.SelectFont("9x15");
	blackDC.SelectBrush(FALSE);
	dc.SelectColor("honeydew");
	dc.SelectBrush(TRUE);
	for (int i=0; i<=nMaxProc; i++)
	{
		int x, y;

		double angle=((double)i)/(nMaxProc+1)*2*PI;

		x=(long)(250*cos(angle)+285);
		y=(long)(230*sin(angle)+255);

		dc.Rectangle(x-20, y-10, x+40, y+50);
		blackDC.Rectangle(x-20, y-10, x+40, y+50);

		sprintf(szBuf, "%d", i);
		blackDC.TextOut(x-10, y+48, szBuf);
	}
	sprintf(szBuf, "%0.6lf", ((double)m_nSpeed)/1000000);
	blackDC.TextOut(550, 40, "Speed:");
	blackDC.ClearArea(550, 50, 650, 75);
	blackDC.TextOut(550, 70, szBuf);
	blackDC.TextOut(550, 110, "Time:");

	CDC sdc[5];
	CDC ddc[4];

	for (int i=0; i<5; i++)
	{
		sdc[i].Attach(w);
		sdc[i].SelectColor(aBrightColors[i]);
		sdc[i].SelectBrush(TRUE);
		sdc[i].SelectFont("9x15");
		sdc[i].SelectPen(2);
	}

	blackDC.TextOut(550, 400, "Status:");
	blackDC.TextOut(580, 430, "Stop");
	sdc[0].TextOut(550, 430, "E");
	blackDC.TextOut(580, 460, "Serial");
	sdc[0].Line(560, 445, 560, 465);
	blackDC.TextOut(580, 490, "Para");
	sdc[1].Line(550, 475, 550, 495);
	sdc[1].Line(560, 475, 560, 495);
	sdc[1].Line(570, 475, 570, 495);
	blackDC.TextOut(580, 520, "Send");
	sdc[3].TextOut(550, 520, "S");
	blackDC.TextOut(580, 550, "Recv");
	sdc[4].TextOut(550, 550, "R");
	
	for (int i=0; i<4; i++)
	{
		ddc[i].Attach(w);
		ddc[i].SelectColor(aPacketColor[i]);
		ddc[i].SelectBrush(TRUE);
	}

	blackDC.TextOut(10, 430, "Send:");
	blackDC.TextOut(25, 460, "1-100");
	ddc[0].MoveTo(10, 450);
	ddc[0].ArrowTo(20, 450, 6);
	blackDC.TextOut(25, 490, "101-1000");
	ddc[1].MoveTo(10, 480);
	ddc[1].ArrowTo(20, 480, 7);
	blackDC.TextOut(25, 520, "1001-10000");
	ddc[2].MoveTo(10, 510);
	ddc[2].ArrowTo(20, 510, 8);
	blackDC.TextOut(25, 550, ">10000");
	ddc[3].MoveTo(10, 540);
	ddc[3].ArrowTo(20, 540, 9);


	return TRUE;
}

//////////////////////////
//	UI Implementation
extern Widget topLevel;

Widget wEventDlg=NULL;	
Widget wEventGraphForm, wEventGraph;
Widget wEventAnimateForm, wEventAnimate;
Widget wBtGraphTime, wBtAnimateTime, wBtAnimateSpeed;

static CEventGraphInfo eventInfo;
static CEventAnimate eventAnimate;

static int aBtEventWidth[7]=
	{ 80, 80, 80, 80, 80, 80, 80 };

static CharBuf aEventGraphBt[7]=
	{ "ZoomIn", "ZoomOut", "<-", "->", "Time", "Animate", "Close" };

static CharBuf aEventAnimBt[6]=
	{ "Speed", "Continue", "Pause", "Time", "Chart", "Close" };

void CbkChangeEventRatio(Widget w, void * /* int (multiply 10)  */ nScale,
	XmAnyCallbackStruct *pCallData);
void CbkPaneEventGraph(Widget w, void * /* BOOL */ bLineUp,
	XmAnyCallbackStruct *pCallData);
void CbkSetEventGraphTime(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData);

void CbkExposeEventGraph(Widget w, caddr_t data,
	XmDrawingAreaCallbackStruct *pCallData);

void CbkChangeAnimateSpeed(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData);
void CbkContinueAnimate(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData);
void CbkPauseAnimate(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData);

void CbkExposeEventAnimate(Widget w, caddr_t data,
	XmDrawingAreaCallbackStruct *pCallData);

void CbkOpenTimeDlg(Widget w, void * /* BOOL */ bAnimate,
	XmAnyCallbackStruct *pCallData);
void CbkChangeEventStyle(Widget w, void * /* int */ nStatus,
	XmAnyCallbackStruct *pCallData);

int PrepareEventGraph()
{
	eventInfo.Reset();
	eventAnimate.Reset();

	if (NULL!=wEventDlg)
	{
		XtDestroyWidget(wEventDlg);
		wEventDlg=NULL;
	}

	return TRUE;
}

int ShowEventGraph(int bAnimate)
{
	Widget wChartDlg;
	int n=0;
	Arg wargs[MAXARGS];
	Widget awButton[20];

	if (NULL==wEventDlg)
	{
		n=0;
		XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
		XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
		XtSetArg(wargs[n],XmNmarginWidth,0); n++;    
		XtSetArg(wargs[n],XmNmarginHeight,0); n++;    
		XtSetArg(wargs[n],XmNwidth,650);	n++;
		XtSetArg(wargs[n],XmNheight,630);	n++;
		wEventDlg=XmCreateBulletinBoardDialog(topLevel,
			"Event Graph\n",wargs,n);
	
		XtAddCallback(wEventDlg, XmNdestroyCallback,
			CbkDestroyWidget, &wEventDlg);

		//	create UI for event graph
		n=0;
		XtSetArg(wargs[n],XmNmarginWidth,0); n++;    
		XtSetArg(wargs[n],XmNmarginHeight,0); n++;    
		XtSetArg(wargs[n],XmNx,0);	n++;
		XtSetArg(wargs[n],XmNy,0);	n++;
		XtSetArg(wargs[n],XmNwidth,650);	n++;
		XtSetArg(wargs[n],XmNheight,630);	n++;
		wEventGraphForm=XmCreateBulletinBoard(wEventDlg,
			"Graph\n",wargs,n);

		XtSetArg(wargs[0],XmNx,0);
		XtSetArg(wargs[1],XmNy,0);
		XtSetArg(wargs[2],XmNwidth,650);
		XtSetArg(wargs[3],XmNheight,555);
		wEventGraph=XmCreateDrawingArea(wEventGraphForm,
			"DrawArea", wargs, 4);
		XtManageChild(wEventGraph);
	
		XtAddCallback(wEventGraph, XmNexposeCallback,
			CbkExposeEventGraph, NULL);
		
		n=0;
		XtSetArg(wargs[n],XmNy,560); n++;
		XtSetArg(wargs[n],XmNheight,40); n++;
		CreateButtonArray(numberof(aEventGraphBt), wEventGraphForm,
			awButton, aEventGraphBt, aBtEventWidth, 15, wargs, n);
	
		XtAddCallback(awButton[0], XmNactivateCallback,
			CbkChangeEventRatio, (void*)5);
		XtAddCallback(awButton[1], XmNactivateCallback,
			CbkChangeEventRatio, (void*)20);

		XtAddCallback(awButton[2], XmNactivateCallback,
			CbkPaneEventGraph, (void*)1);
		XtAddCallback(awButton[3], XmNactivateCallback,
			CbkPaneEventGraph, (void*)0);

		XtAddCallback(awButton[4], XmNactivateCallback,
			CbkOpenTimeDlg, (void*)FALSE);

		XtAddCallback(awButton[5], XmNactivateCallback,
			CbkChangeEventStyle, (void*)1);
		XtAddCallback(awButton[6], XmNactivateCallback,
			CbkDestroyDlg, (void*)wEventDlg);

		wBtGraphTime=awButton[4];

		for (int i=0; i<numberof(aEventGraphBt); i++)
			XtManageChild(awButton[i]);
	
		if (!bAnimate)
			XtManageChild(wEventGraphForm);	//	show graph by default

		//	create UI for event animation
		n=0;
		XtSetArg(wargs[n],XmNmarginWidth,0); n++;    
		XtSetArg(wargs[n],XmNmarginHeight,0); n++;    
		XtSetArg(wargs[n],XmNx,0);	n++;
		XtSetArg(wargs[n],XmNy,0);	n++;
		XtSetArg(wargs[n],XmNwidth,650);	n++;
		XtSetArg(wargs[n],XmNheight,630);	n++;
		wEventAnimateForm=XmCreateBulletinBoard(wEventDlg,
			"Animate\n",wargs,n);

		XtSetArg(wargs[0],XmNx,0);
		XtSetArg(wargs[1],XmNy,0);
		XtSetArg(wargs[2],XmNwidth,650);
		XtSetArg(wargs[3],XmNheight,555);
		wEventAnimate=XmCreateDrawingArea(wEventAnimateForm,
			"DrawArea", wargs, 4);
		XtManageChild(wEventAnimate);
	
		XtAddCallback(wEventAnimate, XmNexposeCallback,
			CbkExposeEventAnimate, NULL);
		
		n=0;
		XtSetArg(wargs[n],XmNy,560); n++;
		XtSetArg(wargs[n],XmNheight,40); n++;
		CreateButtonArray(numberof(aEventAnimBt), wEventAnimateForm,
			awButton, aEventAnimBt, aBtEventWidth, 15, wargs, n);
	
		XtAddCallback(awButton[0], XmNactivateCallback,
			CbkChangeAnimateSpeed, NULL);
		XtAddCallback(awButton[1], XmNactivateCallback,
			CbkContinueAnimate, NULL);
		XtAddCallback(awButton[2], XmNactivateCallback,
			CbkPauseAnimate, NULL);

		XtAddCallback(awButton[3], XmNactivateCallback,
			CbkOpenTimeDlg, (void*)TRUE);
		XtAddCallback(awButton[4], XmNactivateCallback,
			CbkChangeEventStyle, (void*)0);
		XtAddCallback(awButton[5], XmNactivateCallback,
			CbkDestroyDlg, (void*)wEventDlg);

		wBtAnimateSpeed=awButton[0];
		wBtAnimateTime=awButton[1];

		for (int i=0; i<numberof(aEventAnimBt); i++)
			XtManageChild(awButton[i]);
	
		if (bAnimate)
			XtManageChild(wEventGraphForm);	//	show animate by default

		//	manage dialog
		XtManageChild(wEventDlg);

		return TRUE;
	}
	else XtManageChild(wEventDlg);

	return TRUE;
}

/////////////////////////////////////////////////////
//	static graph call backs
BOOL RedrawEventGraph()
{
/*
	XClearWindow(XtDisplay(wEventGraph), XtWindow(wEventGraph));
	CbkExposeEventGraph(wEventGraph, NULL, NULL);
*/
	CDC dc(wEventGraph);
	dc.ClearArea(0, 0, 1000, 1000, TRUE);
	return TRUE;
}

void CbkExposeEventGraph(Widget w, caddr_t data,
	XmDrawingAreaCallbackStruct *pCallData)
{
	if ( pCallData!=NULL && ((XExposeEvent*)(pCallData->event))->count>0)
		return;
		
	//	draw time scale
	eventInfo.Draw(w);

	//	draw legend
	{
		CDC dc(w);

		dc.SelectBrush(TRUE);
		dc.SelectColor(aColors[0]);
		dc.Rectangle(600, 120, 630, 150);
		dc.SelectColor(aColors[1]);
		dc.Rectangle(600, 200, 630, 230);
		dc.SelectColor(aColors[3]);
		dc.Rectangle(600, 280, 630, 310);
		dc.SelectColor(aColors[4]);
		dc.Rectangle(600, 360, 630, 390);
		dc.SelectFont("9x15");
		dc.SelectColor("black");
		dc.TextOut(585, 180, "Serial");
		dc.TextOut(600, 260, "Para");
		dc.TextOut(600, 340, "Send");
		dc.TextOut(600, 420, "Recv");
	}
}

void CbkChangeEventRatio(Widget w, void * /* int (multiply 10)  */ nScale,
	XmAnyCallbackStruct *pCallData)
{
	BOOL bNeedDraw;
	bNeedDraw=eventInfo.Zoom( ((double)(int)nScale)/10 );

	if (bNeedDraw) RedrawEventGraph();
}

void CbkPaneEventGraph(Widget w, void * /* BOOL */ bLineUp,
	XmAnyCallbackStruct *pCallData)
{
	BOOL bNeedDraw;

	if (bLineUp)
		bNeedDraw=eventInfo.LineUp();
	else bNeedDraw=eventInfo.LineDown();

	if (bNeedDraw) RedrawEventGraph();
}

void CbkSetEventGraphTime(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData)
{
	//	ask for a time value
	int nTime=0;

	//	and set it
	BOOL bNeedDraw;
	bNeedDraw=eventInfo.ZoomTo(nTime);

	if (bNeedDraw) RedrawEventGraph();
}

/////////////////////////////////////////////////////
//	animate call backs
void ShowTimeDialog(BOOL bAnimation);
void ShowSpeedDialog(double dblSpeed);

void TimerMajorAnim(caddr_t data, XtIntervalId *id);
void TimerMinorAnim(caddr_t data, XtIntervalId *id);

BOOL RedrawEventAnimate()
{
/*
	XClearWindow(XtDisplay(wEventGraph), XtWindow(wEventGraph));
	CbkExposeEventGraph(wEventGraph, NULL, NULL);
*/
	CDC dc(wEventAnimate);
	dc.ClearArea(0, 0, 1000, 1000, TRUE);
	return TRUE;
}

void CbkExposeEventAnimate(Widget w, caddr_t data,
	XmDrawingAreaCallbackStruct *pCallData)
{
	if ( pCallData!=NULL && ((XExposeEvent*)(pCallData->event))->count>0)
		return;
	
	eventAnimate.DrawFixed(w);	
	eventAnimate.DrawAnimate(w);
}

void CbkChangeAnimateSpeed(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData)
{
	eventAnimate.Pause();
	ShowSpeedDialog(((double)eventAnimate.m_nSpeed)/1000000);
}

void CbkContinueAnimate(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData)
{
	if (eventAnimate.IsPause())
	{
		eventAnimate.Continue();
		XtAddTimeOut(100, TimerMajorAnim, NULL);
	}
}

void CbkPauseAnimate(Widget w, caddr_t data,
	XmAnyCallbackStruct *pCallData)
{
	if (!eventAnimate.IsPause())
	{
		eventAnimate.Pause();
	}
}

void TimerMajorAnim(caddr_t data, XtIntervalId *id)
{
	if (wEventDlg==NULL)
		return;

	if (eventAnimate.IsPause())
	{
		//	do nothing
		return;
	}
	//	calculate animation time
	eventAnimate.Step();

	//	calculate procesor status and communication events

	//	draw
	eventAnimate.DrawAnimate(wEventAnimate);

	//	generate next time outs
	XtAddTimeOut(100, TimerMajorAnim, NULL);
//	XtAddTimeOut(25, TimerMinorAnim, NULL);
}

void TimerMinorAnim(caddr_t data, XtIntervalId *id)
{
	if (eventAnimate.IsPause())
	{
		//	do nothing
		return;
	}
/*
	//	draw animation
	XtAddTimeOut(25, TimerMinorAnim, NULL);
*/
}

///////////////////////////////////////////////////////
//	common callbacks
void CbkChangeEventStyle(Widget w, void * /* int */ nStatus,
	XmAnyCallbackStruct *pCallData)
{
	eventAnimate.Pause();
	if (nStatus)
	{
		XtManageChild(wEventAnimateForm);
		XtUnmanageChild(wEventGraphForm);
	}
	else
	{
		XtManageChild(wEventGraphForm);
		XtUnmanageChild(wEventAnimateForm);
	}
}

void CbkOpenTimeDlg(Widget w, void * /* BOOL */ bAnimate,
	XmAnyCallbackStruct *pCallData)
{
	ShowTimeDialog((BOOL)bAnimate);
}

/////////////////////////////////////////////////
//	dialog to change animation speed
Widget wAnimSpeedDlg;
Widget wSpeedText;

void CbkSpeedOK(Widget w, void * data, XmAnyCallbackStruct *pCallData);

void ShowSpeedDialog(double dblSpeed)
{
	int n=0;
	Arg wargs[MAXARGS];
	Widget awButton[20];
	char szSpeed[80];

	n=0;
	XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
	XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
	XtSetArg(wargs[n],XmNmarginWidth,0); n++;    
	XtSetArg(wargs[n],XmNmarginHeight,0); n++;    
	XtSetArg(wargs[n],XmNwidth,300);	n++;
	XtSetArg(wargs[n],XmNheight,200);	n++;
	wAnimSpeedDlg=XmCreateBulletinBoardDialog(topLevel,
		"Animation Speed\n",wargs,n);

	XtSetArg(wargs[0],XmNx,20);
	XtSetArg(wargs[1],XmNy,80);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	Widget label=XmCreateLabel(wAnimSpeedDlg, "Speed:", wargs, 4);
	XtManageChild(label);

	XtSetArg(wargs[0],XmNx,105);
	XtSetArg(wargs[1],XmNy,80);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	wSpeedText=XmCreateText(wAnimSpeedDlg, "Speed", wargs, 4);
	sprintf(szSpeed, "%1.4lf", dblSpeed);
	XmTextSetString(wSpeedText, szSpeed);
	XtManageChild(wSpeedText);

	XtSetArg(wargs[0],XmNx,200);
	XtSetArg(wargs[1],XmNy,30);
	XtSetArg(wargs[2],XmNwidth,70);
	XtSetArg(wargs[3],XmNheight,40);
	Widget bt=XmCreatePushButton(wAnimSpeedDlg,
		"OK", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkSpeedOK, NULL);

	XtSetArg(wargs[0],XmNx,200);
	XtSetArg(wargs[1],XmNy,100);
	XtSetArg(wargs[2],XmNwidth,70);
	XtSetArg(wargs[3],XmNheight,40);
	bt=XmCreatePushButton(wAnimSpeedDlg,
		"Cancel", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkDestroyDlg, (void*)wAnimSpeedDlg);

	XtManageChild(wAnimSpeedDlg);
}

void CbkSpeedOK(Widget w, void * data,
	XmAnyCallbackStruct *pCallData)
{
	char *lpszSpeed=XmTextGetString(wSpeedText);

	double dblSpeed;

	dblSpeed=atof(lpszSpeed);
	XtFree(lpszSpeed);

	XtDestroyWidget(wAnimSpeedDlg);
	eventAnimate.SetSpeed(dblSpeed);
}

/////////////////////////////////////////////////
//	dialog to zoom to a time quickly
Widget wEventTimeDlg;
Widget wFromText, wToText;

void CbkZoomTimeOK(Widget w, void * /* BOOL */ bAnimate,
	XmAnyCallbackStruct *pCallData);

void ShowTimeDialog(BOOL bAnimation)
{
	int n=0;
	Arg wargs[MAXARGS];
	Widget awButton[20];
	char szLabel[80];

	n=0;
	XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
	XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
	XtSetArg(wargs[n],XmNmarginWidth,0); n++;    
	XtSetArg(wargs[n],XmNmarginHeight,0); n++;    
	XtSetArg(wargs[n],XmNwidth,300);	n++;
	XtSetArg(wargs[n],XmNheight,200);	n++;
	wEventTimeDlg=XmCreateBulletinBoardDialog(topLevel,
		"Zoom To Time\n",wargs,n);

	XtSetArg(wargs[0],XmNx,20);
	XtSetArg(wargs[1],XmNy,30);
	XtSetArg(wargs[2],XmNwidth,180);
	XtSetArg(wargs[3],XmNheight,60);
	sprintf(szLabel, "Min: 0.000s\nMax: %1.3lfs", 
		((double)eventInfo.m_nElapse)/1000000);
	Widget label=XmCreateLabel(wEventTimeDlg,
		szLabel, wargs, 4);
	XtManageChild(label);

	XtSetArg(wargs[0],XmNx,20);
	XtSetArg(wargs[1],XmNy,90);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	label=XmCreateLabel(wEventTimeDlg, "From(s):", wargs, 4);
	XtManageChild(label);

	XtSetArg(wargs[0],XmNx,105);
	XtSetArg(wargs[1],XmNy,90);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	wFromText=XmCreateText(wEventTimeDlg, "FromTime", wargs, 4);
	XtManageChild(wFromText);

	XtSetArg(wargs[0],XmNx,20);
	XtSetArg(wargs[1],XmNy,140);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	label=XmCreateLabel(wEventTimeDlg, "To  (s):", wargs, 4);
	XtManageChild(label);

	XtSetArg(wargs[0],XmNx,105);
	XtSetArg(wargs[1],XmNy,140);
	XtSetArg(wargs[2],XmNwidth,80);
	XtSetArg(wargs[3],XmNheight,35);
	wToText=XmCreateText(wEventTimeDlg, "ToTime", wargs, 4);
	XtManageChild(wToText);

	XtSetArg(wargs[0],XmNx,200);
	XtSetArg(wargs[1],XmNy,30);
	XtSetArg(wargs[2],XmNwidth,70);
	XtSetArg(wargs[3],XmNheight,40);
	Widget bt=XmCreatePushButton(wEventTimeDlg,
		"OK", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkZoomTimeOK, (void*)bAnimation);

	XtSetArg(wargs[0],XmNx,200);
	XtSetArg(wargs[1],XmNy,100);
	XtSetArg(wargs[2],XmNwidth,70);
	XtSetArg(wargs[3],XmNheight,40);
	bt=XmCreatePushButton(wEventTimeDlg,
		"Cancel", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkDestroyDlg, (void*)wEventTimeDlg);

	XtManageChild(wEventTimeDlg);
}

void CbkZoomTimeOK(Widget w, void * /* BOOL */ bAnimate,
	XmAnyCallbackStruct *pCallData)
{
	char *lpszFrom=XmTextGetString(wFromText);
	char *lpszTo=XmTextGetString(wToText);

	double dblFrom, dblTo;

	dblFrom=atof(lpszFrom);
	dblTo=atof(lpszTo);
	XtFree(lpszFrom);
	XtFree(lpszTo);

	if (!bAnimate)
	{
		eventInfo.ZoomTo((int)(dblFrom*1000000), (int)(dblTo*1000000));
		XtDestroyWidget(wEventTimeDlg);

		RedrawEventGraph();
	}
	else
	{	//	restart animation at a given point
		eventAnimate.Pause();
		eventAnimate.SetTime((int)(dblFrom*1000000));
		XtDestroyWidget(wEventTimeDlg);

		RedrawEventAnimate();
	}
}
