//
//	ProfUtil.cc
//		Implement misc util classes and functions for Prof Visualization
//		part
//

#define _CPP
#define _PROCINFO
#include "headers.h"

#include "Xheaders.h"

#include "ProfUtil.h"

#include "ProfP.h"

CProcInfo proc;
CParaStmtInfo stmt;

/////////////////////////////////////////////////////////////
//	Data summary functions

int GetSummary(vector <int> aData, int nColCount, int nCol,
	int &nMax, int &nMin, int &nAvg, int &nTotal)
{
	assert(nCol<nColCount);
	int nCount=0;

	nMax=INT_MIN;
	nMin=INT_MAX;
	nAvg=0;
	nTotal=0;
	for (int nRow=0; nRow<aData.size()/nColCount; nRow++)
	{
		nCount++;
		int nVal=aData[nRow*nColCount+nCol];
		nMax=max( nVal, nMax );
		nMin=min( nVal, nMin );
		nTotal+=nVal;
		nAvg=nTotal / nCount;
	}
	return TRUE;
}

int GetProcStatictics(int nOffset, 
	int &nMax, int &nMin, int &nAvg, int &nTotal)
{
	int nCount=0;

	nMax=INT_MIN;
	nMin=INT_MAX;
	nAvg=0;
	nTotal=0;
	for (int i=0; i<nMaxProc+1; i++)
	{
		CProcInfo *pProc=GetProc(i);
		if (pProc!=NULL)
		{
			nCount++;
			int nVal=*(((int *)pProc)+nOffset);
			nMax=max( nVal, nMax );
			nMin=min( nVal, nMin );
			nTotal+=nVal;
			nAvg=nTotal / nCount;
		}
	}
	return TRUE;
}

int GetStmtStatistics(int nStmtID, int nOffset,
	int &nMax, int &nMin, int &nAvg, int &nTotal)
{
	int nCount=0;

	nMax=INT_MIN;
	nMin=INT_MAX;
	nAvg=0;
	nTotal=0;
	for (int i=0; i<nMaxProc+1; i++)
	{
		CProcInfo *pProc=GetProc(i);
		if (pProc!=NULL)
		{
			CParaStmtInfo *pStmt=pProc->FindStmt(nStmtID);
			
			if (pStmt!=NULL)
			{
				nCount++;
				int nVal=*(((int *)pStmt)+nOffset);
				nMax=max( nVal, nMax );
				nMin=min( nVal, nMin );
				nTotal+=nVal;
				nAvg=nTotal / nCount;
			}
		}
	}
	return TRUE;
}


/////////////////////////////////////////////////////////////
//	Button Array creation
BOOL CreateButtonArray(int nCount, Widget wParent, Widget *aWidget, 
	CharBuf *aSzName, int *aSize, int nStartPos, 
	Arg *wargs, int nArgc, BOOL bHorz) 
{
	int x, y;
	int n;
	x=y=nStartPos;

	for (int i=0; i<nCount; i++)
	{
		n=nArgc;
		if (bHorz)
		{
			XtSetArg(wargs[n],XmNx,x); n++;
			XtSetArg(wargs[n],XmNwidth,aSize[i]); n++;
			x+=aSize[i];
		}
		else
		{
			XtSetArg(wargs[n],XmNy,y); n++;
			XtSetArg(wargs[n],XmNheight,aSize[i]); n++;
			y+=aSize[i];
		}
		aWidget[i]=XmCreatePushButton(wParent,
				aSzName[i], wargs, n);
		XtManageChild(aWidget[i]);
	}
	return TRUE;
}

/////////////////////////////////////////////////////////////
//	XmString package
CXmString::CXmString(char *lpszText)
{
	m_str=XmStringCreateSimple(lpszText);
}

CXmString::~CXmString()
{
	XmStringFree(m_str);
}

/////////////////////////////////////////////////////////////
//	simple functions
void SetWidgetText(Widget w, char *lpszText)
{
	Arg arg;

	CXmString str(lpszText);
	XtSetArg(arg, XmNlabelString, str.m_str);
	XtSetValues(w, &arg, 1);
}

/////////////////////////////////////////////////////////////
//	common callbacks
void CbkUnmanageWidget(Widget w, void *pData, XmAnyCallbackStruct *pCallData)
{
	XtUnmanageChild(w);
}

void CbkUnmanageDlg(Widget w, void *dlg, XmAnyCallbackStruct *pCallData)
{
    XtUnmanageChild((Widget)dlg);
}

void CbkDestroyWidget(Widget w, Widget *pWidget,
	XmAnyCallbackStruct *pCallData)
{
	if (w==*pWidget)
		*pWidget=NULL;
}

void CbkDestroyDlg(Widget w, void *dlg, XmAnyCallbackStruct *pCallData)
{
	XtDestroyWidget((Widget)dlg);
}

///////////////////////////////////////////////////////////////
//	CDC, wrapper class for X drawing functions
CDC::~CDC()
{
	XFreeGC(m_pDisplay, m_gc);
}

BOOL CDC::Attach(Widget w)
{
	m_pDisplay=XtDisplay(w);
	m_window=XtWindow(w);
	m_w=w;
	m_gc=XCreateGC(m_pDisplay, m_window, 0, &m_gcValue);
	m_bFill=FALSE;
	m_x=m_y=0;
	return TRUE;
}

extern "C" double sqrt(double);

void CDC::ArrowLineTo(int x, int y, int nSize)
{
	Line(m_x, m_y, x, y);
	int d=(int)(sqrt((x-m_x)*(x-m_x)+(y-m_y)*(y-m_y)));
	//	draw an arrow
	int ax1=x*3/4+m_x/4, ay1=y*3/4+m_y/4;
	int dx=nSize*(x-m_x)/d, dy=nSize*(y-m_y)/d;
	int ax2=ax1-dx-dy, ay2=ay1-dy+dx;
	int ax3=ax1-dx+dy, ay3=ay1-dy-dx;
	Line(ax1, ay1, ax2, ay2);
	Line(ax1, ay1, ax3, ay3);

	MoveTo(x, y);
}

void CDC::ArrowTo(int x, int y, int nSize)
{
	int d=(int)(sqrt((x-m_x)*(x-m_x)+(y-m_y)*(y-m_y)));

	//	draw an arrow
	XPoint aPt[4];
	aPt[0].x=aPt[3].x=(short)x; aPt[0].y=aPt[3].y=(short)y;
	int dx=nSize*(x-m_x)/d, dy=nSize*(y-m_y)/d;
	aPt[1].x=(short)(aPt[0].x-dx-dy); aPt[1].y=(short)(aPt[0].y-dy+dx);
	aPt[2].x=(short)(aPt[0].x-dx+dy); aPt[2].y=(short)(aPt[0].y-dy-dx);

	Polygon(aPt, 4);

	MoveTo(x, y);
}
	
