//
//  ProfUtil.h
//      define misc util classes and functions for Prof Visualization
//      part
//

#ifndef _PROFUTIL_H
#define _PROFUTIL_H

typedef char CharBuf[30];

#ifdef _PROCINFO
extern CProcInfo proc;
extern CParaStmtInfo stmt;
#endif

////////////////////////////
//	summary
#define PROCOFFSET( field ) ( ((int *)(&(proc.field))) - ((int*)(&proc)) )
#define STMTOFFSET( field ) ( ((int *)(&(stmt.field))) - ((int*)(&stmt)) )

int GetSummary(vector <int> aData, int nColCount, int nCol,
    int &nMax, int &nMin, int &nAvg, int &nTotal);
int GetProcStatictics(int nOffset,
    int &nMax, int &nMin, int &nAvg, int &nTotal);
int GetStmtStatistics(int nStmtID, int nOffset,
    int &nMax, int &nMin, int &nAvg, int &nTotal);


/////////////////////////////////
//	button array creation
BOOL CreateButtonArray(int nCount, Widget wParent, Widget *aWidget,
    CharBuf *aSzName, int *aSize, int nStartPos,
    Arg *wargs, int nArgc, BOOL bHorz=TRUE);

/////////////////////////////////
//	wrapper of XmString
class CXmString
{
public:
	CXmString(char *lpszText);
	~CXmString();

public:
	XmString m_str;
};

/////////////////////////////////
//	simple functions
void SetWidgetText(Widget w, char *lpszText);

/////////////////////////////////
//	common callback
void CbkDestroyWidget(Widget w, Widget *pWidget,
        XmAnyCallbackStruct *pCallData);
void CbkUnmanageDlg(Widget w, void *dlg, XmAnyCallbackStruct *pCallData);
void CbkUnmanageWidget(Widget w, void *pData, XmAnyCallbackStruct *pCallData);
void CbkDestroyDlg(Widget w, void *dlg, XmAnyCallbackStruct *pCallData);

////////////////////////////////////////////////////
//	CDC, wrapper class for X drawing functions
extern "C" int XuGetNamedColor(Display *pDisp, char *lpszName);
extern "C" int XuGetRGBColor(Display *pDisp, int r, int g, int b);

class CPoint
{
public:
	int x, y;

	CPoint() { x=y=0; };
	CPoint(int nX, int nY) { x=nX; y=nY; };
	CPoint &operator =(const CPoint &pt) { x=pt.x; y=pt.y; return *this; };
};

class CSize
{
public:
	int cx, cy;

	CSize() { cx=cy=0; };
	CSize(int wid, int ht) { cx=wid; cy=ht; };
};

class CRect
{
public:
	int top, bottom, left, right;
	CRect() { top=bottom=left=right=0; };
	CRect(int nLeft, int nTop, int nRight, int nBottom) 
		{ left=nLeft; top=nTop; right=nRight; bottom=nBottom; };

	int Width() { return right-left+1; };
	int Height() { return bottom-top+1; };
	CSize Size() { return CSize(Width(), Height()); };
};

class CDC
{
public:
	Display *m_pDisplay;
	Window m_window;
	Widget m_w;
	GC m_gc;
	XGCValues m_gcValue;
	BOOL m_bFill;
	int m_x, m_y;
	
public:
	CDC() {};
	CDC(Widget w){ Attach(w); };
	~CDC();
	BOOL Attach(Widget w);

public:
	//	drawing functions
	void Point(int x, int y) 
	{ XDrawPoint(m_pDisplay, m_window, m_gc, x, y); };

	void MoveTo(int x, int y) { m_x=x; m_y=y; }
	void LineTo(int x, int y) { Line(m_x, m_y, x, y); MoveTo(x, y); };
	void ArrowLineTo(int x, int y, int nSize=12);
	void ArrowTo(int x, int y, int nSize=12);

	void Line(CRect &rect) 
	{ Line(rect.left, rect.top, rect.right, rect.bottom); };

	void Line(int x1, int y1, int x2, int y2)
	{ XDrawLine(m_pDisplay, m_window, m_gc, x1, y1, x2, y2); };

	void Rectangle(CRect &rect)
	{ Rectangle(rect.left, rect.top, rect.right, rect.bottom); };

	void Rectangle(int x1, int y1, int x2, int y2)
	{ 
		if (!m_bFill)
			XDrawRectangle(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1); 
		else
			XFillRectangle(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1); 
	};

	void TextOut(int x, int y, char *lpszText)
	{	XDrawString(m_pDisplay, m_window, m_gc, x, y,
			lpszText, strlen(lpszText)); };

	void Ellipse(CRect rect)
	{ Ellipse(rect.left, rect.top, rect.right, rect.bottom); };
	void Ellipse(int x1, int y1, int x2, int y2)
	{
		if (!m_bFill)
			XDrawArc(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1, 0, 360*64); 
		else
			XFillArc(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1, 0, 360*64); 
	};

	void Arc(CRect rect, int nStart, int nEnd)
	{ Arc(rect.left, rect.top, rect.right, rect.bottom, nStart, nEnd); };

	void Arc(int x1, int y1, int x2, int y2, int nStart, int nEnd)
	{
		if (!m_bFill)
			XDrawArc(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1, nStart, nEnd); 
		else
			XFillArc(m_pDisplay, m_window, m_gc, 
				x1, y1, x2-x1+1, y2-y1+1, nStart, nEnd);
	};

	void Polygon(XPoint *aPt, int nPoint)
	{
		if (!m_bFill)
			XDrawLines(m_pDisplay, m_window, m_gc,
				aPt, nPoint, CoordModeOrigin);
		else
			XFillPolygon(m_pDisplay, m_window, m_gc,
				aPt, nPoint, Nonconvex, CoordModeOrigin);
	}

	void ClearArea(int x1, int y1, int x2, int y2, BOOL bUpdate=FALSE)
	{
		XClearArea(m_pDisplay, m_window, x1, y1, x2-x1+1, y2-y1+1, bUpdate);
	}
	void ClearArea(CRect rect, BOOL bUpdate=FALSE)
	{
		ClearArea(rect.left, rect.top, rect.right, rect.bottom, bUpdate);
	}

	//	attribute setting
	void SelectColor(char *lpszFore, char *lpszBack=NULL)
	{	if (NULL!=lpszFore) 
			XSetForeground(m_pDisplay, m_gc, GetColor(lpszFore));
		if (NULL!=lpszBack) 
			XSetBackground(m_pDisplay, m_gc, GetColor(lpszBack));
	};
	
	void SelectColor(int nFore, int nBack=-1)
	{	XSetForeground(m_pDisplay, m_gc, nFore);
		if (nBack!=-1)
			XSetBackground(m_pDisplay, m_gc, nBack); };

	void SelectBrush(BOOL bFill=TRUE, int nFillStyle=FillSolid) 
	{	XSetFillStyle(m_pDisplay, m_gc, nFillStyle);
		m_bFill=bFill; }

	void SelectPen(int nWidth=1, int nLineStyle=LineSolid, 
		int nCapStyle=CapNotLast, int nJoinStyle=JoinMiter)
	{	XSetLineAttributes(m_pDisplay, m_gc, 
			nWidth, nLineStyle, nCapStyle, nJoinStyle); };

	void SelectFont(Font font)
	{	XSetFont(m_pDisplay, m_gc, font); };

	void SelectFont(char *lpszName)
	{	XSetFont(m_pDisplay, m_gc, GetFont(lpszName)); };
	
	//	attribute querying

	//	flush display
	void Flush() { XFlush(m_pDisplay); };

	//	get resources
	int GetColor(int r, int g, int b)
	{ 	return XuGetRGBColor(m_pDisplay, r, g, b); };
	int GetColor(char *lpszName)
	{	return XuGetNamedColor(m_pDisplay, lpszName); };
	int GetFont(char *lpszName)
	{	return XLoadFont(m_pDisplay, lpszName); };
};

#endif
