//
//	StmtUI.cc
//		Implement UI for displaying statement information
//		including
//			General information
//			comm graph for the program
//			statement summary list
//			statement detail list
//			statement comm graph
//

#define _CPP
#define _PROCINFO
#include "headers.h"
#include "Xheaders.h"

//#include "Prof.h"
#include "ProfP.h"
#include "ProfUtil.h"
#include "DataTable.h"

////////////////////////////////////////////////////////////////
//	declarations
int PrepareGeneralInfo();
int ShowGeneralInfo();
int ShowCommGraph(int nStmtID);

int PrepareStmtList();
int ShowStmtList();
int ShowStmtDetail(int nStmtID);

////////////////////////////////////////////////////////////////
//	Implementation

static CharBuf aChartFields[MAX_PROC]=
	{ 
		"chart", "chart", "chart", "chart", "chart", "chart", "chart", "chart", 
		"chart", "chart", "chart", "chart", "chart", "chart", "chart", "chart", 
		"chart", "chart", "chart", "chart", "chart", "chart", "chart", "chart", 
		"chart", "chart", "chart", "chart", "chart", "chart", "chart", "chart"
	};
 
typedef struct stChartContext
{
	CDataTable *pTable;
	int nCol;
	int nAnnoCol;
	int nStyle;
	BOOL bRadar;
	char szTitle[80];
	char szSubTitle[80];
/*
	CharBuf szTitle;
	CharBuf szSubTitle;
*/
	CharBuf szDesc;
	Widget wArea;
} ChartContext;

BOOL CreateChartDlg(ChartContext *pContext);

//	General Information
static CDataTable g_tbGenInfo(7);
static Widget wGInfoDlg=NULL, wGInfoList=NULL;

extern Widget topLevel;

//	call backs
void CbkShowCommGraph(Widget w, void *nCol, XmAnyCallbackStruct *pCallData);
void CbkSortGenInfo(Widget w, /* int */ void *nCol, 
	XmAnyCallbackStruct *pCallData);
void CbkChartGenInfo(Widget w, /* int */ void *nCol, 
	XmAnyCallbackStruct *pCallData);

int PrepareGeneralInfo()
{
	g_tbGenInfo.RemoveAll();
	for (int i=0; i<=nMaxProc; i++)
	{
		int aData[7];
		CProcInfo *pProc=GetProc(i);
		if (pProc!=NULL)
		{
			aData[0]=i;
			aData[1]=pProc->m_nElapseCPU;
			aData[2]=pProc->m_nElapseSystem;
			pProc->GetSendTotal(aData[3], aData[4]);
			pProc->GetRecvTotal(aData[5], aData[6]);
			g_tbGenInfo.AddRow((int*)aData);
		}
	}
	if (NULL!=wGInfoDlg)
	{
		XtDestroyWidget(wGInfoDlg);
		wGInfoDlg=NULL;
	}
 
	return TRUE;
}

static int aGenInfoWidth[7]=
	{ 80, 100, 100, 60, 80, 60, 80 };

static int aGenInfoWidth1[7]=
	{ 9, 13, 12, 8, 9, 8, 10 };

static CharBuf aGenInfoFields[7]=
	{ "CPU", "CPU\n(us)", "System\n(us)", "Send",
		"Send\n(Byte)", "Recv", "Recv\n(Byte)" };

int ShowGeneralInfo()
{
	Widget awButtons[20];
	int    n=0, m;
	Arg    wargs[MAXARGS];
	char   str[100],*s1;
	XmString  xmstr;
	int    n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
	int    lineno,no=0;
	
	if (wGInfoDlg==NULL)
	{
		XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	//	XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
		XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
		XtSetArg(wargs[n],XmNwidth,600);	n++;
		XtSetArg(wargs[n],XmNheight,350);	n++;
		wGInfoDlg=XmCreateBulletinBoardDialog(topLevel,
			"General Information\n",wargs,n);
	
		XtAddCallback(wGInfoDlg, XmNdestroyCallback,
			CbkDestroyWidget, &wGInfoDlg);

		n=0;
		XtSetArg(wargs[n],XmNy,10);	n++;
		XtSetArg(wargs[n],XmNheight,70); n++;
		CreateButtonArray(numberof(aGenInfoFields), wGInfoDlg,
			awButtons, aGenInfoFields, aGenInfoWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aGenInfoFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkSortGenInfo, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		n=0;
		XtSetArg(wargs[n],XmNy,80);	n++;
		XtSetArg(wargs[n],XmNheight,30); n++;
		CreateButtonArray(numberof(aGenInfoFields), wGInfoDlg,
			awButtons, aChartFields, aGenInfoWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aGenInfoFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkChartGenInfo, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,110);
		XtSetArg(wargs[2],XmNwidth,550);
		XtSetArg(wargs[3],XmNheight,190);
		wGInfoList=XmCreateList(wGInfoDlg,
			"List", wargs, 4);
		XtManageChild(wGInfoList);
		g_tbGenInfo.SetWidth(1, aGenInfoWidth1);
		g_tbGenInfo.Display(wGInfoList);

		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,300);
		XtSetArg(wargs[2],XmNwidth,150);
		XtSetArg(wargs[3],XmNheight,40);
		Widget bt=XmCreatePushButton(wGInfoDlg,
			"Communication", wargs, 4);
		XtManageChild(bt);
		XtAddCallback(bt, XmNactivateCallback,
			CbkShowCommGraph, (void*)-1);

		XtSetArg(wargs[0],XmNx,165);
		XtSetArg(wargs[1],XmNy,300);
		XtSetArg(wargs[2],XmNwidth,155);
		XtSetArg(wargs[3],XmNheight,40);
		bt=XmCreatePushButton(wGInfoDlg,
			"Close", wargs, 4);
		XtManageChild(bt);
		XtAddCallback(bt, XmNactivateCallback,
			CbkUnmanageDlg, (void*)wGInfoDlg);
	}
	XtManageChild(wGInfoDlg);
	
	return TRUE;
}

void CbkSortGenInfo(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol==0)
		g_tbGenInfo.Sort((int)nCol);
	else
		g_tbGenInfo.Sort((int)nCol, FALSE);

	g_tbGenInfo.Display(wGInfoList);
}

void CbkChartGenInfo(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol>0)
	{
		ChartContext *pChartContext=new ChartContext;

		pChartContext->pTable=&g_tbGenInfo;
		pChartContext->nCol=(int)nCol;
		pChartContext->nAnnoCol=0;
		pChartContext->nStyle=chart_bar;
		pChartContext->bRadar=TRUE;
		strcpy(pChartContext->szTitle, "General Info");
		strcpy(pChartContext->szSubTitle, aGenInfoFields[(int)nCol]);
		strcpy(pChartContext->szDesc, "");
		CreateChartDlg(pChartContext);
	}
}

///////////////////////////////////////////////////////
//	statement list display
static CDataTable g_tbStmtList(12);
static Widget wStmtListDlg=NULL, wStmtListList=NULL;

//	call backs
void CbkSelStmtList(Widget w, void *pData, XmListCallbackStruct *pCallData);
void CbkDblClkStmtList(Widget w, void *pData, XmListCallbackStruct *pCallData);
void CbkSortStmtList(Widget w, /* int */ void *nCol, 
	XmAnyCallbackStruct *pCallData);
void CbkChartStmtList(Widget w, void *nCol, XmAnyCallbackStruct *pCallData);

extern "C" int GetIdBFND(int);
extern "C" int GetIdLine(int);
extern "C" void SetCurBFND(
		Widget  w, caddr_t ind, XmAnyCallbackStruct *call_data);

static int aStmtListWidth[12]=
	{ 60, 60, 80, 80, 100, 100, 100, 100,
		60, 60, 80, 80 };
static int aStmtListWidth1[12]=
	{ 7, 8, 10, 9, 13, 12, 13, 12,
		8, 8, 9, 9 };

static CharBuf aStmtListFields[12]=
	{ "ID", "Line", "Loop\nmax", "Loop\ntotal",
		"Time\nmax\n(us)", "Time\ntotal\n(us)",
		"Com Time\nmax\n(us)", "Com Time\ntotal\n(us)",
		"Send\nmax", "Send\ntotal",
		"Send\nmax\n(byte)", "Send\ntotal\n(byte)" };

int PrepareStmtList()
{
	g_tbStmtList.RemoveAll();

	CProcInfo *pProc;

	for (int i=0; i<=nMaxProc; i++)
	{
		int aData[12];
		pProc=GetProc(i);

		if (pProc!=NULL)
		{
			POSITION pos=pProc->GetFirstStmtPos();
			int nMin, nAvg;

			while (pos>=0)
			{
				CParaStmtInfo *pStmt=pProc->GetNextStmt(pos);
				if (pStmt!=NULL)
				{
					aData[0]=pStmt->m_nStmtID;
					aData[1]=GetIdLine(aData[0]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nLoopCounter),
						aData[2], nMin, nAvg, aData[3]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nElapseCPU),
						aData[4], nMin, nAvg, aData[5]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nComElapseCPU),
						aData[6], nMin, nAvg, aData[7]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nSend),
						aData[8], nMin, nAvg, aData[9]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nSendAmount),
						aData[10], nMin, nAvg, aData[11]);
/*
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nRecv),
						aData[12], nMin, nAvg, aData[13]);
					GetStmtStatistics(pStmt->m_nStmtID, 
						STMTOFFSET(m_nRecvAmount),
						aData[14], nMin, nAvg, aData[15]);
*/

					g_tbStmtList.AddRow((int*)aData);
				}
				else break;
			}
			break;		//	break after one proc is not null
		}
	}
	if (NULL!=wStmtListDlg)
	{
		XtDestroyWidget(wStmtListDlg);
		wStmtListDlg=NULL;
		PrepareStmtDetail(-1);
	}
 
	return TRUE;
}

int ShowStmtList()
{
	Widget awButtons[20];
	int    n=0, m;
	Arg    wargs[MAXARGS];
	char   str[100],*s1;
	XmString  xmstr;
	int    n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
	int    lineno,no=0;
	
	if (wStmtListDlg==NULL)
	{
		XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
		XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
		XtSetArg(wargs[n],XmNwidth,990);	n++;
		XtSetArg(wargs[n],XmNheight,450);	n++;
		wStmtListDlg=XmCreateBulletinBoardDialog(topLevel,
			"Statement List\n",wargs,n);
	
		XtAddCallback(wStmtListDlg, XmNdestroyCallback,
			CbkDestroyWidget, &wStmtListDlg);
		
		n=0;
		XtSetArg(wargs[n],XmNy,10);	n++;
		XtSetArg(wargs[n],XmNheight,90); n++;
		CreateButtonArray(numberof(aStmtListFields), wStmtListDlg,
			awButtons, aStmtListFields, aStmtListWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aStmtListFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkSortStmtList, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		n=0;
		XtSetArg(wargs[n],XmNy,100);	n++;
		XtSetArg(wargs[n],XmNheight,30); n++;
		CreateButtonArray(numberof(aStmtListFields), wStmtListDlg,
			awButtons, aChartFields, aStmtListWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aStmtListFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkChartStmtList, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,130);
		XtSetArg(wargs[2],XmNwidth,950);
		XtSetArg(wargs[3],XmNheight,260);
		wStmtListList=XmCreateList(wStmtListDlg,
			"List", wargs, 4);
		XtAddCallback(wStmtListList, XmNbrowseSelectionCallback,
			CbkSelStmtList, NULL);
		XtAddCallback(wStmtListList, XmNsingleSelectionCallback,
			CbkSelStmtList, NULL);
		XtAddCallback(wStmtListList, XmNdefaultActionCallback,
			CbkDblClkStmtList, NULL);
		XtManageChild(wStmtListList);
		g_tbStmtList.SetWidth(1, aStmtListWidth1);
		g_tbStmtList.Display(wStmtListList, FALSE);

		XtSetArg(wargs[0],XmNx,30);
		XtSetArg(wargs[1],XmNy,390);
		XtSetArg(wargs[2],XmNwidth,150);
		XtSetArg(wargs[3],XmNheight,40);
		Widget bt=XmCreatePushButton(wStmtListDlg,
			"Close", wargs, 4);
		XtManageChild(bt);
		XtAddCallback(bt, XmNactivateCallback,
			CbkUnmanageDlg, (void*)wStmtListDlg);
	}
	XtManageChild(wStmtListDlg);
	
	return TRUE;
}

int ProfSelBFND(int nStmtID)
{
	if (wStmtListDlg==NULL) return FALSE;
	for (int i=0; i<g_tbStmtList.RowCount(); i++)
	{
		if (g_tbStmtList.GetData(i, 0)==nStmtID)
		{
			XmListSelectPos(wStmtListList, i+1, FALSE);
			return TRUE;
		}
	}
	XmListDeselectAllItems(wStmtListList);
	return FALSE;
}

void CbkSelStmtList(Widget w, void *pData, XmListCallbackStruct *pCallData)
{
	int nRow=pCallData->item_position-1;

	int nStmtID;

	if (nRow>=g_tbStmtList.RowCount() || nRow<0)
		return;

	nStmtID=g_tbStmtList.GetData(nRow, 0);
	SetCurBFND(w, (caddr_t)GetIdBFND(nStmtID), NULL);
}

void CbkDblClkStmtList(Widget w, void *pData, XmListCallbackStruct *pCallData)
{
	int nRow=pCallData->item_position-1;

	int nStmtID;

	if (nRow>=g_tbStmtList.RowCount() || nRow<0)
		return;

	nStmtID=g_tbStmtList.GetData(nRow, 0);
	ShowStmtDetail(nStmtID);
}

void CbkSortStmtList(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol<=1)
		g_tbStmtList.Sort((int)nCol);
	else
		g_tbStmtList.Sort((int)nCol, FALSE);

	g_tbStmtList.Display(wStmtListList, FALSE);
}

void CbkChartStmtList(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol>1)
	{
		ChartContext *pChartContext=new ChartContext;

		pChartContext->pTable=&g_tbStmtList;
		pChartContext->nCol=(int)nCol;
		pChartContext->nAnnoCol=0;
		pChartContext->nStyle=chart_bar;
		pChartContext->bRadar=FALSE;
		strcpy(pChartContext->szTitle, "Statement List");
		strcpy(pChartContext->szSubTitle, aStmtListFields[(int)nCol]);
		strcpy(pChartContext->szDesc, "");
		CreateChartDlg(pChartContext);
	}
}

/////////////////////////////////////////////////
//	Statement Detail Dialog
static int nCurStmtID=-1;
static CDataTable g_tbStmtDetail(13);
static Widget wStmtDetailDlg=NULL, wStmtDetailList=NULL;
static char szStmtTitle[80];

//	call backs
void CbkSortStmtDetail(Widget w, /* int */ void *nCol, 
	XmAnyCallbackStruct *pCallData);
void CbkChartStmtDetail(Widget w, void *nCol, XmAnyCallbackStruct *pCallData);

static int aStmtDetailWidth[13]=
	{ 60, 80, 100, 100, 100, 
		60, 80, 60, 80, 60, 80, 60, 80 };
static int aStmtDetailWidth1[13]=
	{ 7, 9, 13, 13, 12, 
		8, 10, 8, 9, 8, 10, 8, 9 };

static CharBuf aStmtDetailFields[13]=
	{ "Proc", "Loop", 
		"Time(us)", "Time\n(inc sub)\n(us)", "Com Time\n(us)", 
		"Shift", "Shift\n(byte)", "Remap", "Remap\n(byte)",
		"Send", "Send\n(byte)", "Recv", "Recv\n(byte)"};

int PrepareStmtDetail(int nStmtID)
{
	g_tbStmtDetail.RemoveAll();
	sprintf(szStmtTitle, "Detail: statement\n");
	nCurStmtID=nStmtID;

	if (nStmtID>=0)
	{
		CProcInfo *pProc;
	
		for (int i=0; i<=nMaxProc; i++)
		{
			int aData[13];
			pProc=GetProc(i);
	
			if (pProc!=NULL)
			{
				CParaStmtInfo *pStmt=pProc->FindStmt(nStmtID);
				if (pStmt!=NULL)
				{
					sprintf(szStmtTitle, "Detail: statement %d, line %d\n",
						pStmt->m_nStmtID, GetIdLine(pStmt->m_nStmtID));
					aData[0]=i;
					aData[1]=pStmt->m_nLoopCounter;
					aData[2]=pStmt->m_nElapseCPU;
					aData[3]=pStmt->m_nTotalElapseCPU;
					aData[4]=pStmt->m_nComElapseSystem;
					aData[5]=pStmt->m_nShift;
					aData[6]=pStmt->m_nShiftAmount;
					aData[7]=pStmt->m_nRemap;
					aData[8]=pStmt->m_nRemapAmount;
					aData[9]=pStmt->m_nSend;
					aData[10]=pStmt->m_nSendAmount;
					aData[11]=pStmt->m_nRecv;
					aData[12]=pStmt->m_nRecvAmount;
	
					g_tbStmtDetail.AddRow((int*)aData);
				}
			}
		}
	}
	if (NULL!=wStmtDetailDlg)
	{
		XtDestroyWidget(wStmtDetailDlg);
		wStmtDetailDlg=NULL;
	}
	
	return TRUE;
}

int ShowStmtDetail(int nStmtID)
{
	Widget awButtons[20];
	int    n=0, m;
	Arg    wargs[MAXARGS];
	char   str[100],*s1;
	XmString  xmstr;
	int    n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
	int    lineno,no=0;
	
	PrepareStmtDetail(nStmtID);
	if (wStmtDetailDlg==NULL)
	{
		XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
		XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
		XtSetArg(wargs[n],XmNwidth,1030);	n++;
		XtSetArg(wargs[n],XmNheight,350);	n++;
		wStmtDetailDlg=XmCreateBulletinBoardDialog(topLevel,
			szStmtTitle,wargs,n);
	
		XtAddCallback(wStmtDetailDlg, XmNdestroyCallback,
			CbkDestroyWidget, &wStmtDetailDlg);
		
		n=0;
		XtSetArg(wargs[n],XmNy,10);	n++;
		XtSetArg(wargs[n],XmNheight,90); n++;
		CreateButtonArray(numberof(aStmtDetailFields), wStmtDetailDlg,
			awButtons, aStmtDetailFields, aStmtDetailWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aStmtDetailFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkSortStmtDetail, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		n=0;
		XtSetArg(wargs[n],XmNy,100);	n++;
		XtSetArg(wargs[n],XmNheight,30); n++;
		CreateButtonArray(numberof(aStmtDetailFields), wStmtDetailDlg,
			awButtons, aChartFields, aStmtDetailWidth, 15, wargs, n);
			
		for (int i=0; i<numberof(aStmtDetailFields); i++)
		{
			XtAddCallback(awButtons[i], XmNactivateCallback,
				CbkChartStmtDetail, (void*)i);
			XtManageChild(awButtons[i]);
	    }
	
		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,130);
		XtSetArg(wargs[2],XmNwidth,1000);
		XtSetArg(wargs[3],XmNheight,170);
		wStmtDetailList=XmCreateList(wStmtDetailDlg,
			"List", wargs, 4);
		XtManageChild(wStmtDetailList);
		g_tbStmtDetail.SetWidth(1, aStmtDetailWidth1);
		g_tbStmtDetail.Display(wStmtDetailList, FALSE);

		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,300);
		XtSetArg(wargs[2],XmNwidth,150);
		XtSetArg(wargs[3],XmNheight,40);
		Widget bt=XmCreatePushButton(wStmtDetailDlg,
			"Communication", wargs, 4);
		XtManageChild(bt);
		XtAddCallback(bt, XmNactivateCallback,
			CbkShowCommGraph, (void*)nStmtID);

		XtSetArg(wargs[0],XmNx,165);
		XtSetArg(wargs[1],XmNy,300);
		XtSetArg(wargs[2],XmNwidth,150);
		XtSetArg(wargs[3],XmNheight,40);
		bt=XmCreatePushButton(wStmtDetailDlg,
			"Close", wargs, 4);
		XtManageChild(bt);
		XtAddCallback(bt, XmNactivateCallback,
			CbkUnmanageDlg, (void*)wStmtDetailDlg);
	}
	XtManageChild(wStmtDetailDlg);
	
	return TRUE;
}

void CbkShowCommGraph(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	ShowCommGraph((int)nCol);
}

void CbkSortStmtDetail(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol<1)
		g_tbStmtDetail.Sort((int)nCol);
	else
		g_tbStmtDetail.Sort((int)nCol, FALSE);

	g_tbStmtDetail.Display(wStmtDetailList, FALSE);
}

void CbkChartStmtDetail(Widget w, void *nCol, XmAnyCallbackStruct *pCallData)
{
	if ((int)nCol>0)
	{
		ChartContext *pChartContext=new ChartContext;

		pChartContext->pTable=&g_tbStmtDetail;
		pChartContext->nCol=(int)nCol;
		pChartContext->nAnnoCol=0;
		pChartContext->nStyle=chart_bar;
		pChartContext->bRadar=FALSE;
		strcpy(pChartContext->szTitle, szStmtTitle);
		strcpy(pChartContext->szSubTitle, aStmtDetailFields[(int)nCol]);
		strcpy(pChartContext->szDesc, "");
		CreateChartDlg(pChartContext);
	}
}

///////////////////////////////////////////////////////
//	Comm graph processing
static CDataTable g_tbStmtComm(MAX_PROC*2+1);
static Widget wStmtCommDlg=NULL, wStmtCommList=NULL;
static Widget wCommDrawArea;
static char szStmtCommTitle[80];
static int g_nDispProc;

void EvhExposeCommGraph(Widget w, void *pData, XEvent *pEvent, char *pCh);
void CbkExposeCommGraph(Widget w, void *pData,
	XmDrawingAreaCallbackStruct *pCallData);
void CbkFilterCommGraph(Widget w, void */* int */nProcID,
	XmAnyCallbackStruct *pCallData);

static int aStmtCommWidth1[MAX_PROC+1]=
	{ 10, 
		15, 15, 15, 15, 15, 15, 15, 15, 
		15, 15, 15, 15, 15, 15, 15, 15, 
		15, 15, 15, 15, 15, 15, 15, 15, 
		15, 15, 15, 15, 15, 15, 15, 15
	}; 

static int aStmtCommWidth[MAX_PROC+1]=
	{ 100, 
		120, 120, 120, 120, 120, 120, 120, 120, 
		120, 120, 120, 120, 120, 120, 120, 120, 
		120, 120, 120, 120, 120, 120, 120, 120, 
		120, 120, 120, 120, 120, 120, 120, 120
	}; 

static CharBuf aStmtCommFields[MAX_PROC+1]=
	{ "   To\n   \\\nFrom", 
		"Proc 0", "Proc 1", "Proc 2", "Proc 3",
		"Proc 4", "Proc 5", "Proc 6", "Proc 7", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
		"Proc 1", "Proc 2", "Proc 3", "Proc 4", 
	};

int PrepareCommGraph(int nStmtID)
{
	int aData[MAX_PROC*2+1];
	CProcInfo *pProc;

	g_tbStmtComm.SetShape((nMaxProc+1)*2+1);
	g_nDispProc=-1;

	sprintf(szStmtCommTitle, "Communication: statement\n");
	nCurStmtID=nStmtID;

	if (nStmtID>=0)
	{
		for (int i=0; i<=nMaxProc; i++)
		{
			pProc=GetProc(i);
	
			if (pProc!=NULL)
			{
				CParaStmtInfo *pStmt=pProc->FindStmt(nStmtID);
				if (pStmt!=NULL)
				{
					sprintf(szStmtCommTitle, 
						"Communication: statement %d, line %d\n",
						pStmt->m_nStmtID, 0/*line number*/);
					aData[0]=i;
					for (int j=0; j<MAX_PROC*2; j+=2)
					{
						aData[j+1]=pStmt->m_aSend[j/2];
						aData[j+2]=pStmt->m_aSendAmount[j/2];
					}
	
					g_tbStmtComm.AddRow((int*)aData);
				}
			}
			else
			{
				aData[0]=i;
				for (int j=0; j<MAX_PROC*2+1; j++)
				{
					aData[j]=0;
				}
				g_tbStmtComm.AddRow((int*)aData);
			}
		}
	}
	else if (nStmtID==-1)
	{
		sprintf(szStmtCommTitle, "Communication Summary\n");
		for (int i=0; i<=nMaxProc; i++)
		{
			pProc=GetProc(i);
	
			if (pProc!=NULL)
			{
				aData[0]=i;
				for (int j=0; j<MAX_PROC*2; j+=2)
				{
					aData[j+1]=pProc->m_aSend[j/2];
					aData[j+2]=pProc->m_aSendAmount[j/2];
				}
	
				g_tbStmtComm.AddRow((int*)aData);
			}
			else
			{
				aData[0]=i;
				for (int j=0; j<MAX_PROC*2+1; j++)
				{
					aData[j]=0;
				}
				g_tbStmtComm.AddRow((int*)aData);
			}
		}
		
	}
	if (NULL!=wStmtCommDlg)
	{
		XtDestroyWidget(wStmtCommDlg);
		wStmtCommDlg=NULL;
	}
	
	return TRUE;
}

#define PI 3.1415926

int ShowCommGraph(int nStmtID)
{
	Widget awButtons[20];
	int nWidth=(nMaxProc+1)*120+130;
	int    n=0, m;
	Arg    wargs[MAXARGS];
	char   str[100],*s1;
	XmString  xmstr;
	int    n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
	int    lineno,no=0;
	
	PrepareCommGraph(nStmtID);
	if (wStmtCommDlg==NULL)
	{
		XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
		XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    
		XtSetArg(wargs[n],XmNwidth,nWidth);	n++;
		XtSetArg(wargs[n],XmNheight,750);	n++;
		wStmtCommDlg=XmCreateBulletinBoardDialog(topLevel,
			szStmtCommTitle,wargs,n);
	
		XtAddCallback(wStmtCommDlg, XmNdestroyCallback,
			CbkDestroyWidget, &wStmtCommDlg);
		
		n=0;
		XtSetArg(wargs[n],XmNy,10);	n++;
		XtSetArg(wargs[n],XmNheight,60); n++;
		CreateButtonArray(nMaxProc+2, wStmtCommDlg,
			awButtons, aStmtCommFields, aStmtCommWidth, 15, wargs, n);
			
		for (int i=0; i<nMaxProc+2; i++)
		{
			XtManageChild(awButtons[i]);
		}
	
		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,70);
		XtSetArg(wargs[2],XmNwidth,nWidth-30);
		XtSetArg(wargs[3],XmNheight,200);
		wStmtCommList=XmCreateList(wStmtCommDlg,
			"List", wargs, 4);
		XtManageChild(wStmtCommList);
		g_tbStmtComm.SetWidth(1, aStmtCommWidth1);
		g_tbStmtComm.DisplayMerge(wStmtCommList);
/*
*/
		//	create buttons to represent processors
		for (int i=0; i<=nMaxProc; i++)
		{
			int x, y;
			CharBuf szTitle;

			sprintf(szTitle, "%d", i);

			double angle=((double)i)/(nMaxProc+1)*2*PI;

			x=(long)(180*cos(angle)+220);
			y=(long)(180*sin(angle)+460);

			XtSetArg(wargs[0],XmNx,x);
			XtSetArg(wargs[1],XmNy,y);
			XtSetArg(wargs[2],XmNwidth,40);
			XtSetArg(wargs[3],XmNheight,40);
			Widget wProcBt=XmCreatePushButton(wStmtCommDlg,
				szTitle, wargs, 4);
			XtManageChild(wProcBt);
		
			XtAddCallback(wProcBt, XmNactivateCallback,
				CbkFilterCommGraph, (void*)i);
		}
		//	create a button to show all communication
		XtSetArg(wargs[0],XmNx,30);
		XtSetArg(wargs[1],XmNy,690);
		XtSetArg(wargs[2],XmNwidth,80);
		XtSetArg(wargs[3],XmNheight,40);
		Widget wAllBt=XmCreatePushButton(wStmtCommDlg,
			"All", wargs, 4);
		XtManageChild(wAllBt);
		
		XtAddCallback(wAllBt, XmNactivateCallback,
			CbkFilterCommGraph, (void*)-1);
		
		XtSetArg(wargs[0],XmNx,110);
		XtSetArg(wargs[1],XmNy,690);
		XtSetArg(wargs[2],XmNwidth,80);
		XtSetArg(wargs[3],XmNheight,40);
		Widget bt=XmCreatePushButton(wStmtCommDlg,
			"Close", wargs, 4);
		XtManageChild(bt);
		
		XtAddCallback(bt, XmNactivateCallback,
			CbkUnmanageDlg, (void*)wStmtCommDlg);

		XtSetArg(wargs[0],XmNx,15);
		XtSetArg(wargs[1],XmNy,300);
		XtSetArg(wargs[2],XmNwidth,580);
		XtSetArg(wargs[3],XmNheight,400);
		wCommDrawArea=XmCreateDrawingArea(wStmtCommDlg,
			"DrawArea", wargs, 4);
	
		XtAddCallback(wCommDrawArea, XmNexposeCallback,
			CbkExposeCommGraph, NULL);
		XtManageChild(wCommDrawArea);
	}
	XtManageChild(wStmtCommDlg);

	return TRUE;
}

void CbkExposeCommGraph(Widget w, void *pData, 
	XmDrawingAreaCallbackStruct *pCallData)
{
	if ( pCallData!=NULL && ((XExposeEvent*)(pCallData->event))->count>0)
		return;
		
	CRect rect(15, 300, 515, 700);
	g_tbStmtComm.DrawCommChart(w, g_nDispProc, rect);
}

void EvhExposeCommGraph(Widget w, void *pData, XEvent *pEvent, char *pCh)
{
	if (pEvent!=NULL && ((XExposeEvent*)pEvent)->count>0)
		return;
	CRect rect(15, 300, 515, 700);
	g_tbStmtComm.DrawCommChart(w, g_nDispProc, rect);
}

void CbkFilterCommGraph(Widget w, void */* int */nProcID,
	XmAnyCallbackStruct *pCallData)
{
	g_nDispProc=(int)nProcID;
	XClearWindow(XtDisplay(wCommDrawArea), XtWindow(wCommDrawArea));
	CbkExposeCommGraph(wCommDrawArea, NULL, NULL);
}

/////////////////////////////////////////////////////////////////
//	Charting

void CbkDestroyChartDlg(Widget w, ChartContext *pContext, 
	XmAnyCallbackStruct *pCallData);
void CbkExposeChartDlg(Widget w, ChartContext *pContext, 
	XmDrawingAreaCallbackStruct *pCallData);
void CbkChangeChartStyle(Widget w, ChartContext *pContext, 
	XmAnyCallbackStruct *pCallData);

BOOL CreateChartDlg(ChartContext *pContext)
{
	Widget wChartDlg;
	int	n=0;
	Arg	wargs[MAXARGS];
	char   szTitle[256];
	
	XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	XtSetArg(wargs[n],XmNnoResize,TRUE); n++;	
	XtSetArg(wargs[n],XmNwidth,430);	n++;
	XtSetArg(wargs[n],XmNheight,400);	n++;
	sprintf(szTitle, "%s Chart: %s\n",
		pContext->szTitle, pContext->szSubTitle);
	for (int j=0; j<strlen(szTitle)-1; j++)
		if (szTitle[j]=='\n') szTitle[j]=' ';
	wChartDlg=XmCreateBulletinBoardDialog(topLevel,
		szTitle,wargs,n);

	XtAddCallback(wChartDlg, XmNdestroyCallback,
		CbkDestroyChartDlg, pContext);
	
	XtSetArg(wargs[0],XmNx,15);
	XtSetArg(wargs[1],XmNy,0);
	XtSetArg(wargs[2],XmNwidth,395);
	XtSetArg(wargs[3],XmNheight,320);
	Widget wDrawArea=XmCreateDrawingArea(wChartDlg,
		"DrawArea", wargs, 4);
	XtManageChild(wDrawArea);
	pContext->wArea=wDrawArea;

	XtAddCallback(wDrawArea, XmNexposeCallback,
		CbkExposeChartDlg, pContext);
	
	XtSetArg(wargs[0],XmNx,5);
	XtSetArg(wargs[1],XmNy,350);
	XtSetArg(wargs[2],XmNwidth,150);
	XtSetArg(wargs[3],XmNheight,40);
	Widget bt=XmCreatePushButton(wChartDlg,
		"Change Style", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkChangeChartStyle, pContext);

	XtSetArg(wargs[0],XmNx,165);
	XtSetArg(wargs[1],XmNy,350);
	XtSetArg(wargs[2],XmNwidth,150);
	XtSetArg(wargs[3],XmNheight,40);
	bt=XmCreatePushButton(wChartDlg,
		"Close", wargs, 4);
	XtManageChild(bt);
	XtAddCallback(bt, XmNactivateCallback,
		CbkDestroyDlg, (void*)wChartDlg);

	XtManageChild(wChartDlg);
	return TRUE;
}

void CbkDestroyChartDlg(Widget w, ChartContext *pContext, 
	XmAnyCallbackStruct *pCallData)
{
	delete pContext;	//	destroy when widget is destroyed
}

void CbkExposeChartDlg(Widget w, ChartContext *pContext, 
	XmDrawingAreaCallbackStruct *pCallData)
{
	if ( pCallData!=NULL && ((XExposeEvent*)(pCallData->event))->count>0)
		return;
		
	if ( ((XExposeEvent*)(pCallData->event))->width<390 ||
		((XExposeEvent*)(pCallData->event))->height<310 )
	{
		XClearWindow(XtDisplay(w), XtWindow(w));
	}
		
	vector <int> aCol;
	aCol.insert(aCol.end(), pContext->nCol);
	CRect rect(0, 0, 394, 319);
	pContext->pTable->DisplayChart(w, aCol, 0, pContext->nStyle, rect);
}

void CbkChangeChartStyle(Widget w, ChartContext *pContext, 
	XmAnyCallbackStruct *pCallData)
{
	if (pContext->nStyle==chart_bar)
		pContext->nStyle=chart_pie;
	else if (pContext->nStyle==chart_pie && pContext->bRadar)
		pContext->nStyle=chart_radar;
	else pContext->nStyle=chart_bar;

	XtUnmanageChild(pContext->wArea);
	XtManageChild(pContext->wArea);
}
