/*****************************************************************************
        Visualization tool of AST 

    author:  Zheng Gengbin
                            Peking Univ. P.R.C
*****************************************************************************/

#include <Xm/BulletinB.h>
#include <Xm/Label.h>

#include "Xapp.h"
#include "XSimUtil.h"

#include "process.h"

#define FOLLOW_BIF_POINTER_TO_ID(VAR) \
    (bf_ptr->entry.Template.VAR? bf_ptr-> entry.Template.VAR->id: -1)

#define FOLLOW_SYMB_POINTER_2_TO_ID(VAR) \
    (sy_ptr->entry.VAR? sy_ptr->entry.VAR->id: -1)

#define FOLLOW_TYPE_POINTER_TO_ID(VAR) \
    (ty_ptr->entry.VAR? ty_ptr->entry.VAR->id: -1)

#define MAX_TAG  900

void Ok_callback();
void Ok_callback2();
void Ok_callback3();                    /*    Multi_LLND_dlg OK callback  */
void show_BFND_dlg();
void show_SYM_dlg();
void show_LLND_dlg();
void show_LLND_dlg2();
void show_multi_LLND_dlg3();
void show_TYPE_dlg();
void show_TYPE_dlg1();
void show_LABEL_dlg();
void show_CMNT_dlg();
void show_CMNT_dlg1();


/*                  TOOLS                                                  */
BOOL  is_expr_node(var)
int var;
{
	switch (var) {
	case INT_VAL:
	case FLOAT_VAL:
	case DOUBLE_VAL:
	case BOOL_VAL:
	case CHAR_VAL:
	case STRING_VAL:
	case CONST_REF:
	case VAR_REF:
	case ARRAY_REF:
	case RECORD_REF:
	case ENUM_REF:
	case VAR_LIST:
	case COMPLEX_VAL:
	case EQ_OP:
	case LT_OP:
	case GT_OP:
	case NOTEQL_OP:
	case LTEQL_OP:
	case GTEQL_OP:
	case ADD_OP:
	case SUBT_OP:
	case OR_OP:
	case MULT_OP:
	case DIV_OP:
	case MOD_OP:
	case AND_OP:
	case EXP_OP:
	case ARRAY_MULT:
	case CONCAT_OP:
	case XOR_OP:
	case EQV_OP:
	case NEQV_OP:
	case MINUS_OP:
	case NOT_OP:
	case ASSGN_OP:
	case FUNCTION_OP:
	case BITAND_OP:
	case BITOR_OP:
	case FUNC_CALL:
	case NAMELIST_LIST:
	case BIT_COMPLEMENT_OP:
	case FUNCTION_REF:
	case LSHIFT_OP:
	case RSHIFT_OP:
	case UNARY_ADD_OP:
	case SIZE_OP:
	case INTEGER_DIV_OP:
	case SUB_OP:
	case LE_OP:
	case GE_OP:
	case NE_OP:
			return TRUE;
	default:
			return FALSE;
	}
}

BOOL  ll_has_sym(var)
int var;
{
	switch (var) {
	case (INT_VAL) :
	case (DOUBLE_VAL):
	case (FLOAT_VAL):
	case (BOOL_VAL):
	case (CHAR_VAL):
	case (STMT_STR):
	case (KEYWORD_VAL):
	case (STRING_VAL):
			return FALSE;
	default :       return TRUE;
	}
}

BOOL  ll_has_ll(var)
int var;
{
	switch (var) {
	case (BINARY_VAL):
	case (OCTN_VAL):
	case (HEX_VAL):
	case (INT_VAL) :
	case (DOUBLE_VAL):
	case (FLOAT_VAL):
	case (BOOL_VAL):
	case (CHAR_VAL):
	case (STMT_STR):
	case (KEYWORD_VAL):
	case (STRING_VAL):
/*	case (RANGE_OP):                  modify by HIT  */
	case (UPPER_OP):
	case (LOWER_OP):
	case (LABEL_REF):
			return FALSE;
	default :       return TRUE;
	}
}


Widget create_label1(parent,x,lineno,str)
Widget parent;
int    x,
	lineno;
char   *str;
{
	Widget w1,w2;
	Arg    args[5];
	
	XtSetArg(args[0],XmNx,x);
	XtSetArg(args[1],XmNy,GETY(lineno));
	XtSetArg(args[2],XmNshadowType,XmSHADOW_IN);
	w1=XtCreateManagedWidget("frame",xmFrameWidgetClass,parent,args,3);
	XtSetArg(args[0],XmNwidth,LABEL_WIDTH);
	XtSetArg(args[1],XmNheight,LABEL_HEIGHT);
	w2=XtCreateManagedWidget(str,xmLabelWidgetClass,w1,args,2);
	return w1;
}
	

Widget create_label2(parent,x,lineno,str)
Widget parent;
int    x,
	lineno;
char   *str;
{
	Widget w;
	Arg    args[5];
	
	XtSetArg(args[0],XmNx,x);
	XtSetArg(args[1],XmNy,GETY(lineno));
	XtSetArg(args[2],XmNwidth,LABEL_WIDTH+60);
	XtSetArg(args[3],XmNheight,LABEL_HEIGHT);
	w=XtCreateManagedWidget(str,xmLabelWidgetClass,parent,args,4);
	return w;
}
	
Widget create_button(parent,x,lineno,str)
Widget parent;
int    x,
	lineno;
char   *str;
{
	Widget w;
	Arg    args[5];
	
	XtSetArg(args[0],XmNx,x);
	XtSetArg(args[1],XmNy,GETY(lineno));
	XtSetArg(args[2],XmNwidth,BUTTON_WIDTH);
	XtSetArg(args[3],XmNheight,LABEL_HEIGHT);
	w=XtCreateManagedWidget(str,xmPushButtonWidgetClass,parent,args,4);
	return w;
}

Widget create_button2(parent,x,lineno,str)
Widget parent;
int    x,
	lineno;
char   *str;
{
	Widget w;
	Arg    args[5];
	
	XtSetArg(args[0],XmNx,x==COL_X1?COL_X1:COL_X2+30);
	XtSetArg(args[1],XmNy,GETY(lineno));
	XtSetArg(args[2],XmNwidth,LABEL_WIDTH+32);
	XtSetArg(args[3],XmNheight,LABEL_HEIGHT);
	w=XtCreateManagedWidget(str,xmPushButtonWidgetClass,parent,args,4);
	return w;
}


/*****************************************************************************
                   SYMBOL  Dialog  Block :


	Widget create_BIF_dlg(int index);

	void show_BIF_dlg(Widget w,int *d,call_data);
****************************************************************************/

#define  S_ID      4
#define  S_VAR     5
#define  S_ATTR    6
#define  S_NAME    7

char  *SYM_buttons[]  ={"Ok","Type"};
char  *SYM_labels[]   ={"ID", "Variant", "Attrib","Name",
                        "f1", "f2","f3","f4"};

Widget create_SYM_dlg(sy_ptr)
PTR_SYMB    sy_ptr;
{
 Widget SYM_info_dlg;
 Widget buttons[20],labels[20];
 Widget w[20];
 int    n=0,i,m;
 Arg    wargs[MAXARGS];
 char   str[100],*s1;
 XmString  xmstr;
 int	n1, n2, n3, n4, n5, n6, n7, n8,n9, n10;
 int    lineno,no=0;

	XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
	XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
/*	XtSetArg(wargs[n],XmNnoResize,TRUE); n++;    */
	SYM_info_dlg=XmCreateBulletinBoardDialog(topLevel,
                           "SYM_Node_Info",wargs,n);


	lineno=0;
	XtSetArg(wargs[0],XmNx,COL_X2+100);
	XtSetArg(wargs[1],XmNy,GETY(lineno));
	XtSetArg(wargs[2],XmNwidth,100);
	XtSetArg(wargs[3],XmNheight,LABEL_HEIGHT);
	XtSetArg(wargs[4],XmNuserData,(caddr_t)SYM_info_dlg);
        w[no++]=XtCreateWidget("Ok",xmPushButtonWidgetClass,SYM_info_dlg,
                              wargs,5);
push(w[no-1]);
        XtAddCallback(w[no-1],XmNactivateCallback,Ok_callback,SYM_info_dlg);
	lineno++;


	w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"ID");
        n1 = sy_ptr->id;
        sprintf(str,"%5d  ",n1);
	if (sy_ptr->decl==1) strcat(str," declared");
	else strcat(str," notdeclared");
	w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
	lineno++;

	w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Variant");
	w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,tag[sy_ptr->variant]);
	lineno++;

	w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Attrib");
	str[0]='\0';
	if (sy_ptr->attr & ALLOCATABLE_BIT)
		strcat(str,"ALLOC ");	
	if (sy_ptr->attr & DIMENSION_BIT)
		strcat(str,"DIM ");	
	if (sy_ptr->attr & EXTERNAL_BIT)
		strcat(str,"EXTERN ");	
	if (sy_ptr->attr & IN_BIT)
		strcat(str,"IN ");	
	if (sy_ptr->attr & INOUT_BIT)
		strcat(str,"INOUT ");	
	if (sy_ptr->attr & INTRINSIC_BIT)
		strcat(str,"INTRIN ");	
	if (sy_ptr->attr & OPTIONAL_BIT)
		strcat(str,"OPT ");	
	if (sy_ptr->attr & OUT_BIT)
		strcat(str,"OUT ");	
	if (sy_ptr->attr & PARAMETER_BIT)
		strcat(str,"PARAM ");	
	if (sy_ptr->attr & POINTER_BIT)
		strcat(str,"PTR ");	
	if (sy_ptr->attr & PRIVATE_BIT)
		strcat(str,"PRIV ");	
	if (sy_ptr->attr & PUBLIC_BIT)
		strcat(str,"PUB ");	
	if (sy_ptr->attr & SAVE_BIT)
		strcat(str,"SAVE ");	
	if (sy_ptr->attr & SEQUENCE_BIT)
		strcat(str,"SEQ ");	
	if (sy_ptr->attr & RECURSIVE_BIT)
		strcat(str,"RECURS ");	
	if (sy_ptr->attr & TARGET_BIT)
		strcat(str,"TARGET ");	
	if (sy_ptr->attr & PROCESSORS_BIT)
		strcat(str,"PROCESSORS ");	
	if (sy_ptr->attr & INHERIT_BIT)
		strcat(str,"IHRT ");	
	if (sy_ptr->attr & DYNAMIC_BIT)
		strcat(str,"DYM ");	
	if (sy_ptr->attr & ALIGN_BIT)
		strcat(str,"ALN ");	
	if (sy_ptr->attr & DISTRIBUTE_BIT)
		strcat(str,"DIST ");	
	if (sy_ptr->attr & INTERFACE_BIT)
		strcat(str,"INTER ");	
	if (sy_ptr->attr & PURE_BIT)
		strcat(str,"PURE ");	
	if (sy_ptr->attr & EXTRINSIC_HPF_BIT)
		strcat(str,"EXT_HPF ");	
	if (sy_ptr->attr & EXTRINSIC_HPFLOCAL_BIT)
		strcat(str,"EXT_LOCAL ");	
	if (sy_ptr->attr & ENTRY_BIT)
		strcat(str,"ENTRY ");	
	if (sy_ptr->attr & DEFINED_OPERATOR_BIT)
		strcat(str,"DEF_OP ");	
	if (sy_ptr->attr & DEFINED_ASSIGNMENT_BIT)
		strcat(str,"DEF_AGN ");	
	if (sy_ptr->attr & INTRINSIC_FUNC_BIT)
		strcat(str,"INTR_FUN ");	
	if (strlen(str)==0) strcat(str,"None");
	w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
	lineno++;

	w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Identifer");
	w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,sy_ptr->ident);
	lineno++;

	w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Type");
	XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,sy_ptr->type);
	if (sy_ptr->type==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;

	w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Scope");
	XtAddCallback(w[no-1],XmNactivateCallback,show_BFND_dlg,sy_ptr->scope);
	if (sy_ptr->scope==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;

	w[no++]=create_button2(SYM_info_dlg,COL_X1,lineno,"Next Symbol");
	XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->next_symb);
	if (sy_ptr->next_symb==NULL) XtSetSensitive(w[no-1],FALSE);
/*	lineno++; */

	w[no++]=create_button2(SYM_info_dlg,COL_X2,lineno,"Thread");
	XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->thread);
	if (sy_ptr->thread==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;

	n6 = sy_ptr->scope? sy_ptr->scope->id: -1;

	switch (sy_ptr->variant) {
	    case DEFAULT:
	    case TYPE_NAME:
		break;
	    case CONST_NAME:
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Constant");
		XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.
							Template.const_value);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,show_multi_LLND_dlg3,
				NULL);
		if (sy_ptr->entry.Template.const_value==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	    case ENUM_NAME:
	    case FIELD_NAME:
		n1 = sy_ptr->entry.field.tag;

		w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Tag");
	        sprintf(str,"%5d",sy_ptr->entry.field.tag);
		w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Offset");
	        sprintf(str,"%5d",sy_ptr->entry.field.offset);
		w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Next Field");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.field.next);
		if (sy_ptr->entry.field.next==NULL) XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Base Name");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.field.base_name);
		if (sy_ptr->entry.field.base_name==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
/*
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
                                           "Declared  Name");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.field.declared_name);
		if (sy_ptr->entry.field.declared_name==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
*/

		break;
	    case VARIABLE_NAME:
/*   add by HIT   */
		w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,"Local");
		if (sy_ptr->entry.var_decl.local > MAX_TAG) {
			sprintf(str,"Error: field Local: %d exceed max tag ! ",sy_ptr->entry.var_decl.local);
			BlockReportError(str);
			strcpy(str,"Abnomal !!!");
		}
		else
	        	sprintf(str,"%s",tag[sy_ptr->entry.var_decl.local]);
		
		w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,
							"Offset:Dovar");
	        sprintf(str,"%5d:%5d",sy_ptr->entry.var_decl.offset,
				sy_ptr->entry.var_decl.dovar);
		w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Next In");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.var_decl.next_in);
		if (sy_ptr->entry.var_decl.next_in==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Next Out");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.var_decl.next_out);
		if (sy_ptr->entry.var_decl.next_out==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		if (sy_ptr->entry.Template.const_value) {
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Initial");
		XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.
							Template.const_value);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,show_multi_LLND_dlg3,
				NULL);
		lineno++;
		}
		if (sy_ptr->entry.var_decl.hpfattr) {
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"HPF Attr");
                XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.var_decl.hpfattr);
                XtSetValues(w[no-1],wargs,1);
                XtAddCallback(w[no-1],XmNactivateCallback,show_multi_LLND_dlg3,
                                NULL);
                lineno++;
		}
		break;
	    case PROGRAM_NAME:
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Symbol List");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.prog_decl.symb_list);
		if (sy_ptr->entry.prog_decl.symb_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Proc Header");
		XtAddCallback(w[no-1],XmNactivateCallback,show_BFND_dlg,
                                  sy_ptr->entry.prog_decl.prog_hedr);
		if (sy_ptr->entry.prog_decl.prog_hedr==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Use List");
		XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.prog_decl.use_list);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_multi_LLND_dlg3, NULL);
		if (sy_ptr->entry.prog_decl.use_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		break;
	    case MODULE_NAME:
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Use List");
		XtSetArg(wargs[0],XmNuserData,
				(XtPointer)sy_ptr->entry.prog_decl.use_list);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_multi_LLND_dlg3, NULL);
		if (sy_ptr->entry.prog_decl.use_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	    case FUNCTION_NAME:
            case PROCESS_NAME:
	    case PROCEDURE_NAME:
	    case ROUTINE_NAME:
		w[no++]=create_label1(SYM_info_dlg,COL_X1,lineno,
                           "Number IO:I:O");
	        sprintf(str,"%d:%d:%d",sy_ptr->entry.proc_decl.num_io,
					sy_ptr->entry.proc_decl.num_input,
					sy_ptr->entry.proc_decl.num_output);
		w[no++]=create_label2(SYM_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button2(SYM_info_dlg,COL_X1,lineno,
 			"In List");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.proc_decl.in_list);
		if (sy_ptr->entry.proc_decl.in_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
/*		lineno++;*/

		w[no++]=create_button2(SYM_info_dlg,COL_X2,lineno,
 			"Out List");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.proc_decl.out_list);
		if (sy_ptr->entry.proc_decl.out_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Symbol List");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.proc_decl.symb_list);
		if (sy_ptr->entry.proc_decl.symb_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Use List");
		XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.proc_decl.use_list);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_multi_LLND_dlg3, NULL);
		if (sy_ptr->entry.proc_decl.use_list==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Proc Header");
		XtAddCallback(w[no-1],XmNactivateCallback,show_BFND_dlg,
                                  sy_ptr->entry.proc_decl.proc_hedr);
		if (sy_ptr->entry.proc_decl.proc_hedr==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Next function");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.Template.next);
		if (sy_ptr->entry.Template.next==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;

		break;
	case INTERFACE_NAME:

		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Next function");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.Template.next);
		if (sy_ptr->entry.Template.next==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
/*
		if (sy_ptr->entry.Template.base_name) {
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,"Matched Function");
                XtSetArg(wargs[0],XmNuserData,sy_ptr->entry.Template.base_name);
                XtSetValues(w[no-1],wargs,1);
                XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg, sy_ptr->entry.Template.base_name);
                lineno++;
		}
*/
		break;
	default:
		break;
	}

	if (sy_ptr->entry.Template.declared_name) 
	{
		w[no++]=create_button(SYM_info_dlg,COL_X1,lineno,
 			"Decalred name  (symbol)");
		XtAddCallback(w[no-1],XmNactivateCallback,show_SYM_dlg,
                                  sy_ptr->entry.Template.declared_name);
		lineno++;
	}


	for (i=0;i<no;i++)
		XtAddEventHandler(w[i],KeyPressMask,FALSE,
					(XtEventHandler)closed,w[0]);


/*               Separate    */
	if (no<20) {
	XtSetArg(wargs[0],XmNx,COL_X1);
	XtSetArg(wargs[1],XmNy,GETY(lineno));
	XtSetArg(wargs[2],XmNwidth,LABEL_WIDTH+40);
	XtSetArg(wargs[3],XmNheight,0);
	w[no++]=XtCreateManagedWidget("",xmLabelWidgetClass,
 			SYM_info_dlg,wargs,4);
	}

	XtManageChildren(w,no);

        return SYM_info_dlg;
}


void Ok_callback(w,d,call_data)
Widget  w;
Widget  d;
XmAnyCallbackStruct     *call_data;
{
        XtDestroyWidget(XtParent(d));
}

void show_SYM_dlg(w,d,call_data)
Widget     w;
PTR_SYMB       d;
XmAnyCallbackStruct     *call_data;
{
	char str[100];

	Widget sym_dlg;
	if (d==NULL) {
		BlockReportError("Error: Null Symbol found ! ");
		return;
	}
	if (d->id > ptr_file->num_symbs || d->id<1 )  {
		sprintf(str,"Error: Invalid SYMBOL id: %d !",d->id);
		BlockReportError(str);
		return;
	}
	sym_dlg=create_SYM_dlg(d);
        if (sym_dlg==NULL) {
             fprintf(stderr,"No more dlg !\n");
             return;
	}
	XtManageChild(sym_dlg);
}


void show_SYM_dlg1(w,d,call_data)
Widget     w;
int        *d;
XmAnyCallbackStruct     *call_data;
{
	show_SYM_dlg(w,b_tree[*d].bfnd->entry.Template.symbol,call_data);
}

/*                    CALL BY  Multi_LLND_Dialog   */
void show_SYM_dlg2(w,d,call_data)
Widget     w;
caddr_t        d;
XmAnyCallbackStruct     *call_data;
{
	Arg  wargs[3];
	NODE_PTR    node_ptr;

/*	if (d  && *d<0) return; */
	XtSetArg(wargs[0],XmNuserData,&node_ptr);
	XtGetValues(w,wargs,1);
	show_SYM_dlg(w,node_ptr->llnd->entry.Template.symbol,call_data);
}



/**************************************************************************
                   SINGLE Low_Level Node Dialog Block

***************************************************************************/

Widget create_LLND_dlg(ll_ptr)
PTR_LLND    ll_ptr;
{
 Widget LLND_info_dlg;
 Widget buttons[20],labels[20];
 Widget w[20];
 int    n=0,i,m;
 Arg    wargs[MAXARGS];
 char   str[100],*s1;
 XmString  xmstr;
 int n1, n2, n3, n4, n5, n6, n7, n8, n9, n10, n11, n12, n13, n14, n15, n16, n17;
 int    lineno,no=0;

        XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
        XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
        LLND_info_dlg=XmCreateBulletinBoardDialog(topLevel,
                           "LLND_Node_Info",wargs,n);

        lineno=0;
        XtSetArg(wargs[0],XmNx,COL_X2+100);
        XtSetArg(wargs[1],XmNy,GETY(lineno));
        XtSetArg(wargs[2],XmNwidth,100);
        XtSetArg(wargs[3],XmNheight,LABEL_HEIGHT);
        w[no++]=XtCreateWidget("Ok",xmPushButtonWidgetClass,LLND_info_dlg,
                              wargs,4);
push(w[no-1]);
        XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)Ok_callback,LLND_info_dlg);
        lineno++;

	w[no++]=create_label1(LLND_info_dlg,COL_X1,lineno,"ID");
	n1 = ll_ptr->id;
        sprintf(str,"%5d",n1);
	w[no++]=create_label2(LLND_info_dlg,COL_X2,lineno,str);
	lineno++;


	w[no++]=create_label1(LLND_info_dlg,COL_X1,lineno,"Variant");
	w[no++]=create_label2(LLND_info_dlg,COL_X2,lineno,tag[ll_ptr->variant]);
	lineno++;

	w[no++]=create_label1(LLND_info_dlg,COL_X1,lineno,"Attrib");
	str[0]='\0';
	if (ll_ptr->attr & ALLOCATABLE_BIT)
		strcat(str,"ALLOC ");	
	if (ll_ptr->attr & DIMENSION_BIT)
		strcat(str,"DIM ");	
	if (ll_ptr->attr & EXTERNAL_BIT)
		strcat(str,"EXTERN ");	
	if (ll_ptr->attr & IN_BIT)
		strcat(str,"IN ");	
	if (ll_ptr->attr & INOUT_BIT)
		strcat(str,"INOUT ");	
	if (ll_ptr->attr & INTRINSIC_BIT)
		strcat(str,"INTRINSIC ");	
	if (ll_ptr->attr & OPTIONAL_BIT)
		strcat(str,"OPTIONAL ");	
	if (ll_ptr->attr & OUT_BIT)
		strcat(str,"OUT ");	
	if (ll_ptr->attr & PARAMETER_BIT)
		strcat(str,"PARAM ");	
	if (ll_ptr->attr & POINTER_BIT)
		strcat(str,"PTR ");	
	if (ll_ptr->attr & PRIVATE_BIT)
		strcat(str,"PRIVATE ");	
	if (ll_ptr->attr & PUBLIC_BIT)
		strcat(str,"PUB ");	
	if (ll_ptr->attr & SAVE_BIT)
		strcat(str,"SAVE ");	
	if (ll_ptr->attr & SEQUENCE_BIT)
		strcat(str,"SEQUENCE ");	
	if (ll_ptr->attr & RECURSIVE_BIT)
		strcat(str,"RECURSIVE ");	
	if (ll_ptr->attr & TARGET_BIT)
		strcat(str,"TARGET ");	
	if (ll_ptr->attr & PROCESSORS_BIT)
		strcat(str,"PROCESSORS ");	
	if (ll_ptr->attr & INHERIT_BIT)
		strcat(str,"IHRT ");	
	if (ll_ptr->attr & DYNAMIC_BIT)
		strcat(str,"DYM ");	
	if (ll_ptr->attr & ALIGN_BIT)
		strcat(str,"ALN ");	
	if (ll_ptr->attr & DISTRIBUTE_BIT)
		strcat(str,"DIST ");	
	if (ll_ptr->attr & INTERFACE_BIT)
                strcat(str,"INTER ");
        if (ll_ptr->attr & PURE_BIT)
                strcat(str,"PURE ");
        if (ll_ptr->attr & EXTRINSIC_HPF_BIT)
                strcat(str,"EXT_HPF ");
        if (ll_ptr->attr & EXTRINSIC_HPFLOCAL_BIT)
                strcat(str,"EXT_LOCAL ");  
        if (ll_ptr->attr & ENTRY_BIT)
                strcat(str,"ENTRY ");
        if (ll_ptr->attr & DEFINED_OPERATOR_BIT)
                strcat(str,"DEF_OP ");
        if (ll_ptr->attr & DEFINED_ASSIGNMENT_BIT)
                strcat(str,"DEF_AGN ");
	if (strlen(str)==0) strcat(str,"None");
	w[no++]=create_label2(LLND_info_dlg,COL_X2,lineno,str);
	lineno++;

	w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Type");
	XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                             ll_ptr->type);
	if (ll_ptr->type==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;

	switch (ll_ptr->variant) {
	case INT_VAL:
		w[no++]=create_label1(LLND_info_dlg,COL_X1,lineno,"Int Value");
		sprintf(str,"%d",ll_ptr->entry.ival);
        	w[no++]=create_label2(LLND_info_dlg,COL_X2,lineno,str);
        	lineno++;
		break;
	case EXPR_LIST:
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Item (1)");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                   ll_ptr->entry.list_s.item);
		if (ll_ptr->entry.list_s.item==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
		
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Next (2)");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                   ll_ptr->entry.list_s.next);
		if (ll_ptr->entry.list_s.next==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	case VAR_REF:
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Symbol");
		XtAddCallback(w[no-1],XmNactivateCallback,
					(XtCallbackProc)show_SYM_dlg,
	                                ll_ptr->entry.var_ref_s.symbol);
		if (ll_ptr->entry.var_ref_s.symbol==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	case ARRAY_REF:
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Symbol");
		XtAddCallback(w[no-1],XmNactivateCallback,
					(XtCallbackProc)show_SYM_dlg,
	                                ll_ptr->entry.array_ref_s.symbol);
		if (ll_ptr->entry.array_ref_s.symbol==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,
							"Subscripts (1)");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                  ll_ptr->entry.array_ref_s.subscripts);
		if (ll_ptr->entry.array_ref_s.subscripts==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	case TYPE_OP:
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,
							"Selector (1)");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                  ll_ptr->entry.type_op.selector);
		if (ll_ptr->entry.type_op.selector==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
		break;
	default:
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Symbol");
		XtAddCallback(w[no-1],XmNactivateCallback,
					(XtCallbackProc)show_SYM_dlg,
	                                ll_ptr->entry.Template.symbol);
		if (ll_ptr->entry.Template.symbol==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
	
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Left Ptr");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                   ll_ptr->entry.Template.ll_ptr1);
		if (ll_ptr->entry.Template.ll_ptr1==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
	
		w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Right Ptr");
		XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
	                                   ll_ptr->entry.Template.ll_ptr2);
		if (ll_ptr->entry.Template.ll_ptr2==NULL) 
				XtSetSensitive(w[no-1],FALSE);
		lineno++;
	
	}
	
	w[no++]=create_button(LLND_info_dlg,COL_X1,lineno,"Next (thread)");
	XtAddCallback(w[no-1],XmNactivateCallback,show_LLND_dlg,
                                             ll_ptr->thread);
	if (ll_ptr->thread==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;
	

	for (i=0;i<no;i++)
		XtAddEventHandler(w[i],KeyPressMask,FALSE,
				(XtEventHandler)closed,w[0]);

	XtSetArg(wargs[0],XmNx,COL_X1);
	XtSetArg(wargs[1],XmNy,GETY(lineno));
	XtSetArg(wargs[2],XmNwidth,LABEL_WIDTH+40);
	XtSetArg(wargs[3],XmNheight,0);
	w[no++]=XtCreateManagedWidget("",xmLabelWidgetClass,
 			LLND_info_dlg,wargs,4);

	XtManageChildren(w,no);

        return LLND_info_dlg;
}


char  *LLND_buttons[]  ={"Ok","Type"};
char  *LLND_labels[]   ={"ID", "variant",
                        "f1", "f2"};
Widget   *l_btn,
	 *l_lines;

void show_LLND_tree(dlg)
Widget dlg;
{
 Arg     wargs_b[10],wargs_l[10];
 int     i,j,n,n0;
 int     num_b,num_l;
 int     x,y,xp,yp;
 XmString  xmstr;
 int     parent;
 char    ed_str[100];
 int     n1, n2, n3, n4, n5, n6, n7, n8, n9, n10;

	l_btn=(Widget *)calloc(nLLND,sizeof(Widget*));    /*   Alloc buttons */
	l_lines=(Widget *)calloc(nLLND*2,sizeof(Widget*)); 

	n=0;
	XtSetArg(wargs_b[n],XmNwidth,BT_WIDTH);n++;
	XtSetArg(wargs_b[n],XmNheight,BT_HEIGHT);n++;
	n0=n;
	num_l=0;  num_b=0;
	for (i=0;i<nLLND;i++){
         n=n0;
         x=GetXByDepth(l_tree[i].depth);
         y=GetYByHeight(l_tree[i].height);
         XtSetArg(wargs_b[n],XmNx,x); n++;
         XtSetArg(wargs_b[n],XmNy,y); n++;
         l_btn[num_b]=XtCreateWidget(tag[l_tree[i].llnd->variant],
                                 xmPushButtonGadgetClass,dlg, wargs_b,n);
/*
         XtAddCallback(l_btn[num_b],XmNactivateCallback,curBFND,(caddr_t)i);
*/
         num_b++;
      parent=l_tree[i].parent;
      if (parent!=-1) {
           x=GetXByDepth(l_tree[i].depth);
           y= GetYByHeight(l_tree[i].height)+BT_HEIGHT/2;
           xp=GetXByDepth(l_tree[parent].depth)+LINEMARGIN_X;
           yp=GetYByHeight(l_tree[parent].height)+BT_HEIGHT-1;

             j=0;
             XtSetArg(wargs_l[j],XmNx,xp);j++;
             XtSetArg(wargs_l[j],XmNy,yp);j++;
             XtSetArg(wargs_l[j],XmNheight,y-yp+1);j++;
             XtSetArg(wargs_l[j],XmNorientation,XmVERTICAL);j++;
             l_lines[num_l]=XtCreateWidget("",xmSeparatorGadgetClass,dlg,
                                         wargs_l,j);
             num_l++;
             j=0;
           XtSetArg(wargs_l[j],XmNx,xp);j++;
           XtSetArg(wargs_l[j],XmNy,y);j++;
           XtSetArg(wargs_l[j],XmNwidth,x-xp+1);j++;
           XtSetArg(wargs_l[j],XmNorientation,XmHORIZONTAL);j++;
           l_lines[num_l]=XtCreateWidget("",xmSeparatorGadgetClass,dlg,
                                         wargs_l,j);
           num_l++;
      }
 }
 XtManageChildren(l_lines,num_l);
 XtManageChildren(l_btn,nLLND);
 free(l_btn);
 free(l_lines);
}



Widget create_LLND_dlg1(ll_ptr)
PTR_LLND ll_ptr;
{
 Widget LLND_info_dlg,mainW,topB,bottomT;
 Widget scrollW,draw,buttons[20],labels[20],
                        top_sym,top_type,top_label,top_ok,top_data,root;
 int    n=0,i;
 Arg    wargs[MAXARGS];
 char   str[20];
 XmString  xmstr;
 int MARGIN=3, BUTTON_WID=100, EDIT_WID=300, BUTTON_HEIGHT=40,x,y;

        XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
        XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
        LLND_info_dlg=XmCreateFormDialog(topLevel,
                           "LLND_Multi_Node_Info",wargs,n);
        n=0;
        XtSetArg(wargs[n],XmNtopAttachment,XmATTACH_FORM);  n++;
        XtSetArg(wargs[n],XmNbottomAttachment,XmATTACH_FORM);  n++;
        XtSetArg(wargs[n],XmNrightAttachment,XmATTACH_FORM);  n++;
        XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_FORM);  n++;
        XtSetArg(wargs[n],XmNwidth,620);  n++;
        XtSetArg(wargs[n],XmNheight,420);  n++;
        mainW=XmCreateMainWindow(LLND_info_dlg,"main",wargs,n);
        XtManageChild(mainW);

        topB=XtCreateManagedWidget("TopForm",xmFormWidgetClass,mainW,NULL,0);
        n=0;
        XtSetArg(wargs[n], XmNwidth, BUTTON_WID);        n++;
        XtSetArg(wargs[n], XmNheight, BUTTON_HEIGHT);        n++;
        XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM);     n++;
        XtSetArg(wargs[n], XmNtopOffset, MARGIN);     n++;
        XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM);  n++;
        XtSetArg(wargs[n], XmNbottomOffset, MARGIN);  n++;

        x=MARGIN;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
        top_sym=XtCreateManagedWidget("Symbol",
                    xmPushButtonWidgetClass,topB, wargs, n);

        XtAddCallback(top_sym,XmNactivateCallback,(XtCallbackProc)show_SYM_dlg2
						,NULL);

	x+=BUTTON_WID+5;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
        XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET);  n++;
        XtSetArg(wargs[n], XmNleftWidget, top_sym);  n++;
        top_type=XtCreateManagedWidget("Type",
                    xmPushButtonWidgetClass,topB, wargs, n);
        XtAddCallback(top_type,XmNactivateCallback,
			(XtCallbackProc)show_TYPE_dlg1, NULL);
        x+=BUTTON_WID+5;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
        XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET);  n++;
        XtSetArg(wargs[n], XmNleftWidget, top_type);  n++;
        top_label=XtCreateManagedWidget("Label",
                    xmPushButtonWidgetClass,topB, wargs, n);
        XtAddCallback(top_label,XmNactivateCallback,show_LABEL_dlg1,
                        NULL);

        XtSetSensitive(top_sym,FALSE);
        XtSetSensitive(top_type,FALSE);
        XtSetSensitive(top_label,FALSE);
	
	n=0;
        XtSetArg(wargs[n], XmNeditable, False);  n++;
        bottomT=XtCreateManagedWidget("bottom_Text",xmTextWidgetClass,
                mainW, wargs, n);

        n=0;
        XtSetArg(wargs[n],XmNscrollingPolicy,XmAUTOMATIC);n++;
        scrollW=XtCreateManagedWidget("scrollW",
                xmScrolledWindowWidgetClass, mainW, wargs,n);
        draw=XtCreateManagedWidget("l_draw",xmDrawingAreaWidgetClass,
                scrollW,NULL,0);

        XmMainWindowSetAreas(mainW,topB,NULL,bottomT, NULL, scrollW);


	buttons[0]=XtCreateManagedWidget("Ok",xmPushButtonWidgetClass,
                                      LLND_info_dlg,NULL,0);
push(buttons[0]);
	XtAddCallback(buttons[0],XmNactivateCallback,Ok_callback2,
                               LLND_info_dlg);
	show_LLND_tree(draw);

/*
        for (i=0;i<XtNumber(LLND_buttons);i++)
             buttons[i]=XtCreateWidget(LLND_buttons[i],xmPushButtonWidgetClass,
                                      LLND_info_dlg,NULL,0);
	if (ll_ptr->type==NULL)
                    XtSetSensitive(buttons[1],FALSE);

        for (i=0;i<XtNumber(LLND_labels);i++)
             labels[i]=XtCreateWidget(LLND_labels[i],xmLabelWidgetClass,
                                      LLND_info_dlg,NULL,0);
        sprintf(str,"%d",ll_ptr->id);
        xmstr=XmStringCreate(str,XmSTRING_DEFAULT_CHARSET);
        XtSetArg(wargs[0],XmNlabelString,xmstr);
        XtSetValues(labels[0],wargs,1);                  
        sprintf(str,"%d",ll_ptr->variant);
        xmstr=XmStringCreate(str,XmSTRING_DEFAULT_CHARSET);
        XtSetArg(wargs[0],XmNlabelString,xmstr);
        XtSetValues(labels[1],wargs,1);                

        XtAddCallback(buttons[0],XmNactivateCallback,Ok_callback2,
                               LLND_info_dlg);
        XtManageChildren(buttons,XtNumber(LLND_buttons));
        XtManageChildren(labels,XtNumber(LLND_labels));
*/

        return LLND_info_dlg;
}

 
void Ok_callback2(w,d,call_data)
Widget  w;
Widget  d;
XmAnyCallbackStruct     *call_data;
{
	Arg args[2];
	LLND_TREE  tree;

        XtUnmanageChild(d);
	XtSetArg(args[0],XmNuserData,&tree);
	XtGetValues(d,args,1);
	if (tree==l_tree)  fprintf(stderr,"Ok\n");
	else  fprintf(stderr,"Error !\n");
        XtDestroyWidget(XtParent(d));
	free(tree);
}


void show_LLND_dlg1(w,d,call_data)
Widget     w;
int        *d;
XmAnyCallbackStruct     *call_data;
{
	show_LLND_dlg(w,b_tree[*d].bfnd->entry.Template.ll_ptr1);
}



void show_LLND_dlg(w,d,call_data)
Widget     w;
PTR_LLND      d;
XmAnyCallbackStruct     *call_data;
{
Widget LLND_info_dlg;
char    str[40];

/*	proc_llnd_tree(d);   */

	if (d==NULL) {
                BlockReportError("Error: Null LLND found ! ");
                return;
        }
	if (d->id>ptr_file->num_llnds || d->id<1) {
                sprintf(str,"Error: Invalid LLND Id :%s! ",d->id);
                BlockReportError("Error: Invalid LLND Id ! ");
                return;
        }
        LLND_info_dlg=create_LLND_dlg(d);
        XtManageChild(LLND_info_dlg);

}

/*               CALL BY  TYPE Dialog   */
void show_LLND_dlg2(w,d,call_data)
Widget     w;
caddr_t        d;
XmAnyCallbackStruct     *call_data;
{
        Arg  wargs[3];
        NODE_PTR    node_ptr;
	PTR_LLND  LLND_cb_ptr[3];
	char      str[40];

        XtSetArg(wargs[0],XmNuserData,&node_ptr);
        XtGetValues(w,wargs,1);
	if (node_ptr == NULL)
		BlockReportError("Error LLND found !");
	if (node_ptr->llnd->id > ptr_file->num_llnds || node_ptr->llnd->id<1)  {
		sprintf(str,"Error: Invalid LLND id: %d !",node_ptr->llnd->id);
		BlockReportError(str);
		return;
	}
        show_LLND_dlg(w,node_ptr->llnd,call_data);
	
}

/**************************************************************************
                   MULTIPLE  Low_Level Node Dialog Block

***************************************************************************/

#define L_PANEL_X  150
#define L_PANEL_Y   90
#define L_MARGIN_X  27
#define L_MARGIN_Y  25
#define L_BT_WIDTH  L_PANEL_X-2*L_MARGIN_X
#define L_BT_HEIGHT  L_PANEL_Y-2*L_MARGIN_Y
#define LLND_GetXByDepth(depth)     ( (depth*L_PANEL_X)+L_MARGIN_X )
#define LLND_GetYByHeight(height)   ( ((height+1)*L_PANEL_Y)+L_MARGIN_Y )
#define L_GetMidX(depth)   ((depth*L_PANEL_X)+L_PANEL_X/2)
#define L_GetMidY(height)  (((height+1)*L_PANEL_Y)+L_PANEL_Y/2)

void free_multi_llnd_tree(head)
LIST_PTR head;
{
	LIST_PTR p,old_p;
	NODE_PTR q,old_q;
	for (p=head;p;)  {
	  for (q=p->right;q;) {
		old_q=q;
		q=q->right;
		free(old_q);
	  }
	old_p=p;
	p=p->next;
	free(old_p);
	}
}

void Ok_callback3(w,d,call_data)
Widget  w;
Widget  d;
XmAnyCallbackStruct     *call_data;
{
        Arg args[2];
        LIST_PTR  list_head;

        XtSetArg(args[0],XmNuserData,&list_head);
        XtGetValues(d,args,1);
        XtDestroyWidget(XtParent(d)); 
        free_multi_llnd_tree(list_head);
	XuPressButton(NULL,1);                  /*  resume button background */
}


void set_cur_llnd(w,node_ptr,call_data)
Widget   w;
NODE_PTR  node_ptr;
XmAnyCallbackStruct     *call_data;
{
	Widget sym_w,type_w,label_w,data_w,detail_w,dlg,notify,old_w;
	Arg    wargs[5];
	int     n;
	char    str[100],str1[100];
	int	n1, n2, n3;
	NODE_PTR  old_ptr;

	XtSetArg(wargs[0],XmNuserData,(caddr_t)&dlg); 
	XtGetValues(w,wargs,1);
	sym_w=XtNameToWidget(dlg,"*Symbol");
	type_w=XtNameToWidget(dlg,"*Type");
	label_w=XtNameToWidget(dlg,"*Label");
	XtSetSensitive(label_w,FALSE);
	detail_w=XtNameToWidget(dlg,"*Detail");
	XtSetSensitive(detail_w,TRUE);

	if (ll_has_sym(node_ptr->llnd->variant)&&
                       node_ptr->llnd->entry.Template.symbol)
                    XtSetSensitive(sym_w,TRUE);
        else XtSetSensitive(sym_w,FALSE);
        if (node_ptr->llnd->type) XtSetSensitive(type_w,TRUE);
        else XtSetSensitive(type_w,FALSE);

	switch (node_ptr->llnd->variant) {
/* add by HIT  */
	    case (INT_VAL):
	        sprintf(str,"ID:%-4d Var:%s INT:%-4d",node_ptr->llnd->id,
                                tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.ival);
		break;
	    case (DOUBLE_VAL):
	    case (FLOAT_VAL):
	        sprintf(str,"ID:%4d Var:%s FLOAT:%4s",node_ptr->llnd->id,
                                tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
	    case (BOOL_VAL):
	        sprintf(str,"ID:%-4d Var:%s BOOL:%-4s",node_ptr->llnd->id,
                                tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.bval? ".T" : ".F");
		break;
	    case (CHAR_VAL):
	        sprintf(str,"ID:%-4d Var:%s CHAR:%-4c",node_ptr->llnd->id,
                                tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.cval);
		break;
	    case (STMT_STR):
	        sprintf(str,"ID:%-4d Var:%s Statement:%-4s",
                              node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
	    case (KEYWORD_VAL):
	        sprintf(str,"ID:%-4d Var:%s Keyword:%-4s",
                               node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
	    case (STRING_VAL):
	        sprintf(str,"ID:%-4d Var:%s String:%-4s",
                              node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
/* HIT */
	    case (BINARY_VAL):
	        sprintf(str,"ID:%-4d Var:%s Binary:%-4s",
                              node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
	    case (OCTN_VAL):
	        sprintf(str,"ID:%-4d Var:%s Octn:%-4s",
                              node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
	    case (HEX_VAL):
	        sprintf(str,"ID:%-4d Var:%s Hex:%-4s",
                              node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.string_val);
		break;
/*   motify by HIT  */
/*	    case (RANGE_OP): */
	    case (UPPER_OP):
	    case (LOWER_OP):
	        sprintf(str,"ID:%-4d Var:%s Dimmension:%-4d",
                             node_ptr->llnd->id,tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.array_op.dim);
   
		break;

	    case (LABEL_REF):
	    case (END_LABEL):
	    case (ERR_LABEL):
		XtSetArg(wargs[0],XmNuserData,(caddr_t)node_ptr); 
		XtSetValues(label_w,wargs,1);
		XtSetSensitive(sym_w,FALSE);
	        sprintf(str,"ID:%-4d Var:%s ",node_ptr->llnd->id,
                                   tag[node_ptr->llnd->variant]);
		if (node_ptr->llnd->entry.label_list.lab_ptr)
			XtSetSensitive(label_w,TRUE);
		else{
			XtSetSensitive(label_w,FALSE);
			BlockReportError("Error: NULL Label found ! ");
		}
		break;
/*   add by HIT   */	
	    case (LABEL_LIST):
		XtSetArg(wargs[0],XmNuserData,(caddr_t)node_ptr); 
		XtSetValues(label_w,wargs,1);
	        sprintf(str,"ID:%-4d Var:%s ",node_ptr->llnd->id,
                                   tag[node_ptr->llnd->variant]);
		XtSetSensitive(sym_w,FALSE);
		if (node_ptr->llnd->entry.label_list.lab_ptr)
			XtSetSensitive(label_w,TRUE);
		else{
			XtSetSensitive(label_w,FALSE);
			BlockReportError("Error: NULL Label found ! ");
		}
		break;
	    case (VAR_REF):
	    case (CONST_REF):
	    case (ARRAY_REF):
	        sprintf(str,"ID:%-4d Var:%s   Ident:%s",node_ptr->llnd->id,
                                   tag[node_ptr->llnd->variant],
				node_ptr->llnd->entry.var_ref_s.symbol->ident);
		if (node_ptr->llnd->entry.var_ref_s.symbol == NULL )
			BlockReportError("Error: Symbol NULL ! ");
		break;
	
	    case (TYPE_OP):		
	        sprintf(str,"ID:%-4d Var:%s ",node_ptr->llnd->id,
                                   tag[node_ptr->llnd->variant]);
		break;
	   default:
	        sprintf(str,"ID:%-4d Var:%s ",node_ptr->llnd->id,
                                   tag[node_ptr->llnd->variant]);
/*
		if (node_ptr->llnd->type) {
		  switch (node_ptr->llnd->type->variant) {
			case T_INT: 
				sprintf(str1,"%s Value:%d",str,node_ptr->llnd->entry.ival);
fprintf("I.m here \n");
				strcpy(str,str1);
				break;
			case T_FLOAT: sprintf(str1,"%s Value:%f",str,node_ptr->llnd->entry.dval);
				strcpy(str,str1);
				break;
			case T_STRING: sprintf(str1,"%s Value:%s",str,node_ptr->llnd->entry.string_val);
				strcpy(str,str1);
				break;
			case T_CHAR: sprintf(str1,"%s Value:%s",str,node_ptr->llnd->entry.cval);
				strcpy(str,str1);
				break;
			case T_BOOL: sprintf(str1,"%s Value:%s",str,node_ptr->llnd->entry.bval);
				strcpy(str,str1);
				break;
		  }
		}
*/


	}    /* end of switch */
/*      add by zheng.gb   */
	if (node_ptr->llnd->attr & PARAMETER_BIT) 
		sprintf(str,"%s Value:%d",str,node_ptr->llnd->const_val);


	notify=XtNameToWidget(dlg,"*bottom_Text");
	XmTextSetString(notify,str);

	data_w=XtNameToWidget(dlg,"*data");
	XtSetArg(wargs[0],XmNuserData,(caddr_t)&old_w);
        XtGetValues(data_w,wargs,1);	
	if (old_w) XuPressButton(old_w,1);
	XtSetArg(wargs[0],XmNuserData,w);
	XtSetValues(data_w,wargs,1);
	XuPressButton(w,0);

	XtSetArg(wargs[0],XmNuserData,(caddr_t)node_ptr); 
	XtSetValues(sym_w,wargs,1);
	XtSetArg(wargs[0],XmNuserData,(caddr_t)node_ptr); 
	XtSetValues(type_w,wargs,1);
	XtSetArg(wargs[0],XmNuserData,(caddr_t)node_ptr); 
	XtSetValues(detail_w,wargs,1);
/*
	if (ll_has_sym(node_ptr->llnd->variant)&&
                       node_ptr->llnd->entry.Template.symbol)  
                    XtSetSensitive(sym_w,TRUE);
	else XtSetSensitive(sym_w,FALSE);
	if (node_ptr->llnd->type) XtSetSensitive(type_w,TRUE);
	else XtSetSensitive(type_w,FALSE);
*/
	if (is_expr_node(node_ptr->llnd->variant)&& !node_ptr->llnd->type )
		BlockReportError("Error: Type is NULL ! ");
}

/*                 draw area  EXPOSURE call back   */
void show_multi_lines(w,list_head,call_data)
Widget        w;
LIST_PTR      list_head;
XmAnyCallbackStruct     *call_data;
{
 LIST_PTR  p;
 NODE_PTR  q;
 GC        gc1,gc2,gc3;
 XGCValues values,values2;
 Arg       wargs[10];
 int       x,y,x0,y0;
 unsigned long fg,bg;
 int       lcolor,rcolor,col1;
	
	XtSetArg(wargs[0],XtNforeground,&values.foreground);
	XtSetArg(wargs[1],XtNbackground,&values.background);
	XtGetValues(w,wargs,2);
	fg=values.foreground;
	bg=values.background;

	lcolor=XuGetNamedColor(XtDisplay(w),"red");
	values.foreground=lcolor;  
	gc1=XtGetGC(w,GCForeground | GCBackground, &values);
	XSetLineAttributes(XtDisplay(w),gc1,2,LineSolid,CapNotLast,JoinMiter);

	rcolor=XuGetNamedColor(XtDisplay(w),"green");
	values.foreground=rcolor;
	gc2=XtGetGC(w,GCForeground | GCBackground, &values);
	XSetLineAttributes(XtDisplay(w),gc2,2,LineSolid,CapNotLast,JoinMiter);

	col1=XuGetNamedColor(XtDisplay(w),"blue");
	values.foreground=col1;
	gc3=XtGetGC(w,GCForeground | GCBackground, &values);
	XSetLineAttributes(XtDisplay(w),gc3,2,LineSolid,CapNotLast,JoinMiter);

	x0=30 ;  x=30+60;
	y0=35 ;  y=35+20;
        XDrawLine(XtDisplay(w),XtWindow(w),gc1,x0,y0,x,y0);
        XDrawLine(XtDisplay(w),XtWindow(w),gc2,x0,y,x,y);
	XDrawString(XtDisplay(w), XtWindow(w), gc1, x0+70, y0, "Left",4);
	XDrawString(XtDisplay(w), XtWindow(w), gc2, x0+70, y, "Right",5);

	x0=L_GetMidX(1);
	q=list_head->right;
	x0=L_GetMidX((q->depth+q->right->right->depth)/2);
	y0=L_GetMidY(-1);
	x=L_GetMidX(q->depth);
        y=L_GetMidY(q->height);
        XDrawLine(XtDisplay(w),XtWindow(w),gc3,x0,y0,x,y);
	q=q->right;
	x=L_GetMidX(q->depth);
        y=L_GetMidY(q->height);
        XDrawLine(XtDisplay(w),XtWindow(w),gc3,x0,y0,x,y);
	q=q->right;
	x=L_GetMidX(q->depth);
        y=L_GetMidY(q->height);
        XDrawLine(XtDisplay(w),XtWindow(w),gc3,x0,y0,x,y);
	for (p=list_head;p;p=p->next)
	 for (q=p->right;q;q=q->right)
	   if (q->llnd) {
		x0=L_GetMidX(q->depth);
		y0=L_GetMidY(q->height);
		if (q->son1) {
		   x=L_GetMidX(q->son1->depth);
                   y=L_GetMidY(q->son1->height);
		   XDrawLine(XtDisplay(w),XtWindow(w),gc1,x0,y0,x,y);
		}
		if (q->son2) {
		   x=L_GetMidX(q->son2->depth);
                   y=L_GetMidY(q->son2->height);
		   XDrawLine(XtDisplay(w),XtWindow(w),gc2,x0,y0,x,y);
                }
	   }
	XFlush(XtDisplay(w));
}


Widget create_multi_LLND_dlg(list_head)
LIST_PTR  list_head;
{
 Widget LLND_info_dlg,mainW,topB,bottomT;
 Widget scrollW,draw,buttons[20],labels[20],
		top_sym,top_type,top_label,top_ok,top_data,top_detail,root;
 Widget *button;
 int    n=0,n0,i,sum;
 Arg    wargs[MAXARGS];
 LIST_PTR   p;
 NODE_PTR   q;
 PTR_SYMB   pt;
 int MARGIN=3, BUTTON_WID=100, EDIT_WID=300, BUTTON_HEIGHT=40,x,y;
 int DIALOG_WID=620;

        XtSetArg(wargs[n],XmNuserData,(caddr_t)list_head); n++;
        XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
        XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
        LLND_info_dlg=XmCreateFormDialog(topLevel,
                           "LLND_Multi_Node_Info",wargs,n);
	n=0;
	XtSetArg(wargs[n],XmNtopAttachment,XmATTACH_FORM);  n++;
	XtSetArg(wargs[n],XmNbottomAttachment,XmATTACH_FORM);  n++;
	XtSetArg(wargs[n],XmNrightAttachment,XmATTACH_FORM);  n++;
	XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_FORM);  n++;
	XtSetArg(wargs[n],XmNwidth,DIALOG_WID);  n++;
	XtSetArg(wargs[n],XmNheight,420);  n++;
	mainW=XmCreateMainWindow(LLND_info_dlg,"main",wargs,n);
	XtManageChild(mainW);

	n=0;
        XtSetArg(wargs[n], XmNheight, BUTTON_HEIGHT);        n++;
	topB=XtCreateManagedWidget("TopForm",xmFormWidgetClass,mainW,wargs,n);
	n=0;
        XtSetArg(wargs[n], XmNwidth, BUTTON_WID);        n++;
        XtSetArg(wargs[n], XmNheight, BUTTON_HEIGHT);        n++;
        XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM);     n++;
        XtSetArg(wargs[n], XmNtopOffset, MARGIN);     n++;
        XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM);  n++;
        XtSetArg(wargs[n], XmNbottomOffset, MARGIN);  n++;

        x=MARGIN;
        XtSetArg(wargs[n], XmNx, x);             n++;
	XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_FORM);  n++;
        top_sym=XtCreateManagedWidget("Symbol",
                    xmPushButtonWidgetClass,topB, wargs, n);

        XtAddCallback(top_sym,XmNactivateCallback,(XtCallbackProc)show_SYM_dlg2
					,NULL);

        x+=BUTTON_WID+5;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
	XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_WIDGET);  n++;
	XtSetArg(wargs[n],XmNleftWidget,top_sym);  n++;
        top_type=XtCreateManagedWidget("Type",
                    xmPushButtonWidgetClass,topB, wargs, n);
        XtAddCallback(top_type,XmNactivateCallback,show_TYPE_dlg1,
			NULL);
        x+=BUTTON_WID+5;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
	XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_WIDGET);  n++;
	XtSetArg(wargs[n],XmNleftWidget,top_type);  n++;
        top_label=XtCreateManagedWidget("Label",
                    xmPushButtonWidgetClass,topB, wargs, n);
        XtAddCallback(top_label,XmNactivateCallback,show_LABEL_dlg1,
			NULL);

        x+=BUTTON_WID+5;
        n=6;
        XtSetArg(wargs[n], XmNx, x);             n++;
	XtSetArg(wargs[n],XmNleftAttachment,XmATTACH_WIDGET);  n++;
	XtSetArg(wargs[n],XmNleftWidget,top_label);  n++;
        top_detail=XtCreateManagedWidget("Detail",
                    xmPushButtonWidgetClass,topB, wargs, n);
        XtAddCallback(top_detail,XmNactivateCallback,(XtCallbackProc)show_LLND_dlg2,
			NULL);

        XtSetSensitive(top_sym,FALSE);
        XtSetSensitive(top_type,FALSE);
        XtSetSensitive(top_label,FALSE);
        XtSetSensitive(top_detail,FALSE);

	n=0;
        XtSetArg(wargs[n], XmNwidth, BUTTON_WID);        n++; 
        XtSetArg(wargs[n], XmNheight, BUTTON_HEIGHT);        n++;
	XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM);  n++;
        XtSetArg(wargs[n], XmNtopOffset, MARGIN);     n++;
        XtSetArg(wargs[n], XmNrightOffset, MARGIN);     n++;
	XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM);  n++;
	XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM);  n++;
	XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_NONE);  n++; 
	XtSetArg(wargs[n], XmNuserData,(caddr_t)LLND_info_dlg);  n++;  /*  Escape cb  */
        top_ok=XtCreateManagedWidget("Ok",
                    xmPushButtonWidgetClass,topB, wargs, n);
push(top_ok);
        XtAddCallback(top_ok,XmNactivateCallback,(XtCallbackProc)Ok_callback3,
							LLND_info_dlg);

	XtAddEventHandler(top_type,KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);
	XtAddEventHandler(top_sym,KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);
	XtAddEventHandler(top_ok,KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);
	XtAddEventHandler(top_detail,KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);
	XtAddEventHandler(top_label,KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);

	n=0;
	XtSetArg(wargs[n], XmNuserData,(caddr_t)NULL); n++;    /*last button*/
        top_data=XtCreateWidget("data",
                    xmPushButtonWidgetClass,topB, wargs, n);

	n=0;
	XtSetArg(wargs[n], XmNeditable, False);  n++;
        bottomT=XtCreateManagedWidget("bottom_Text",xmTextWidgetClass,
                mainW, wargs, n);
	
        n=0;
        XtSetArg(wargs[n],XmNscrollingPolicy,XmAUTOMATIC);n++;
        scrollW=XtCreateManagedWidget("scrollW",
                xmScrolledWindowWidgetClass, mainW, wargs,n);
	n=0;
	XtSetArg(wargs[n],XmNwidth,DIALOG_WID); n++;
        draw=XtCreateManagedWidget("l_draw",xmDrawingAreaWidgetClass,
                scrollW,wargs,n);
	XtAddCallback(draw,XmNexposeCallback,
				(XtCallbackProc)show_multi_lines,list_head);

	XmMainWindowSetAreas(mainW,topB,NULL,bottomT, NULL, scrollW);
	
	n=0;
	n0=XuGetNamedColor(XtDisplay(draw),"gray65");
	XtSetArg(wargs[n],XmNbackground,n0);  n++;    
	q=list_head->right;
	XtSetArg(wargs[n],XmNx,LLND_GetXByDepth((q->depth+q->right->right->depth)/2)); n++;
	XtSetArg(wargs[n],XmNy,LLND_GetYByHeight(-1)); n++;
        XtSetArg(wargs[n],XmNwidth,L_BT_WIDTH);n++;
        XtSetArg(wargs[n],XmNheight,L_BT_HEIGHT);n++;
	root=XtCreateManagedWidget("Root", 
                                 xmPushButtonWidgetClass, draw,wargs,n);
/*
	root=XtCreateManagedWidget(tag[b_tree[curBFND].bfnd->variant], 
                                 xmPushButtonWidgetClass, draw,wargs,n);
*/
	sum=0;
	for (p=list_head;p;p=p->next) sum+=p->num;
        button=(Widget *)calloc(sum,sizeof(Widget*));    /*   Alloc buttons */

        n=0;
        XtSetArg(wargs[n],XmNwidth,L_BT_WIDTH);n++;
        XtSetArg(wargs[n],XmNheight,L_BT_HEIGHT);n++;
	i=0;
	n0=n;
	for (p=list_head;p;p=p->next)
  	  for (q=p->right;q;q=q->right) {
            if (q->llnd == NULL) {
	      x=LLND_GetXByDepth(q->depth);
              y=LLND_GetYByHeight(q->height);
	      n=n0;
              XtSetArg(wargs[n],XmNx,x); n++;
              XtSetArg(wargs[n],XmNy,y); n++;
              button[i]=XtCreateWidget("NULL",
                         xmPushButtonWidgetClass, draw,wargs,n);
	      XtSetSensitive(button[i],FALSE);
	      i++;
	    }
	    else {
 	      x=LLND_GetXByDepth(q->depth);
              y=LLND_GetYByHeight(q->height);
	      n=n0;
              XtSetArg(wargs[n],XmNx,x); n++;
              XtSetArg(wargs[n],XmNy,y); n++;
              XtSetArg(wargs[n],XmNuserData,(caddr_t)LLND_info_dlg); n++;
 	      button[i]=XtCreateWidget(tag[q->llnd->variant],
                           xmPushButtonWidgetClass, draw,wargs,n);
	      XtAddEventHandler(button[i],KeyPressMask,FALSE,
					(XtEventHandler)closed,top_ok);
	      XtAddCallback(button[i],XmNactivateCallback,
				(XtCallbackProc)set_cur_llnd, q);
	      i++;
	    }
	  }
	XtManageChildren(button,sum);
	free(button);
	return LLND_info_dlg;
}
 
	
void show_multi_LLND_dlg(w,ptr,call_data)
Widget     w;
PTR_LLND   *ptr;
XmAnyCallbackStruct     *call_data;
{
        LIST_PTR  list_head;
	Widget    dlg;
        list_head=proc_multi_llnd_tree(ptr[0],ptr[1],ptr[2]);
	dlg=create_multi_LLND_dlg(list_head);
	XtManageChild(dlg);
}


/*               CALL BY MAIN  BFND tree   */
void show_multi_LLND_dlg1(w,d,call_data)
Widget     w;
int        *d;
XmAnyCallbackStruct     *call_data;
{
	PTR_LLND  LLND_cb_ptr[3];
	if (*d<0) return;
	LLND_cb_ptr[0]=b_tree[*d].bfnd->entry.Template.ll_ptr1;
	LLND_cb_ptr[1]=b_tree[*d].bfnd->entry.Template.ll_ptr2;
	LLND_cb_ptr[2]=b_tree[*d].bfnd->entry.Template.ll_ptr3;
        show_multi_LLND_dlg(w,LLND_cb_ptr,call_data);
}

/*               CALL BY MULTI BFND tree   */
void show_multi_LLND_dlg2(w,d,call_data)
Widget     w;
caddr_t        d;
XmAnyCallbackStruct     *call_data;
{
        Arg  wargs[3];
        NODE_PTR    node_ptr;
	PTR_LLND  LLND_cb_ptr[3];

        XtSetArg(wargs[0],XmNuserData,&node_ptr);
        XtGetValues(w,wargs,1);
	LLND_cb_ptr[0]=node_ptr->llnd->entry.Template.ll_ptr1;
	LLND_cb_ptr[1]=node_ptr->llnd->entry.Template.ll_ptr2;
	LLND_cb_ptr[2]=NULL;
        show_multi_LLND_dlg(w,LLND_cb_ptr,call_data);
	
}

/*               CALL BY  TYPE Dialog   */
void show_multi_LLND_dlg3(w,d,call_data)
Widget     w;
caddr_t        d;
XmAnyCallbackStruct     *call_data;
{
        Arg  wargs[3];
        PTR_LLND    node_ptr;
	PTR_LLND  LLND_cb_ptr[3];
	char      str[40];

        XtSetArg(wargs[0],XmNuserData,&node_ptr);
        XtGetValues(w,wargs,1);
	if (node_ptr == NULL)
		BlockReportError("Error LLND found !");
	if (node_ptr->id > ptr_file->num_llnds || node_ptr->id<1 )  {
		sprintf(str,"Error: Invalid LLND id: %d !",node_ptr->id);
		BlockReportError(str);
		return;
	}
	LLND_cb_ptr[0]=node_ptr;
	LLND_cb_ptr[1]=NULL;
	LLND_cb_ptr[2]=NULL;
        show_multi_LLND_dlg(w,LLND_cb_ptr,call_data);
	
}

/****************************************************************************
                       TYPE Dialog Block
****************************************************************************/

#define  T_ID      4
#define  T_VAR     5
#define  T_NAME    6
#define  T_LEN     7

char  *TYPE_buttons[]  ={"Ok"};
char  *TYPE_labels[]   ={"ID", "Variant", "Name","Length",
                        "f1", "f2","f3","f4"};

Widget create_TYPE_dlg(ty_ptr)
PTR_TYPE      ty_ptr;
{
 Widget      TYPE_info_dlg;
 Widget buttons[20],labels[20],w[30];
 int    n=0,i;
 Arg    wargs[MAXARGS];
 char   str[20];
 XmString  xmstr;
 int     n1, n2, n3;
 int     n4, n5, n6;   /*  NEW ADDED FOR VPC */
 char buf[33];
 int     no=0,lineno;
 PTR_LLND  LLND_cb_ptr[3];

        XtSetArg(wargs[n],XmNautoUnmanage,FALSE); n++;
        XtSetArg(wargs[n],XmNdialogStyle,XmDIALOG_APPLICATION_MODAL);n++;
        TYPE_info_dlg=XmCreateBulletinBoardDialog(topLevel,
                           "TYPE_Node_Info",wargs,n);

	lineno=0;
	XtSetArg(wargs[0],XmNx,COL_X2+110);
	XtSetArg(wargs[1],XmNy,GETY(lineno));
	XtSetArg(wargs[2],XmNwidth,100);
	XtSetArg(wargs[3],XmNheight,LABEL_HEIGHT);
	XtSetArg(wargs[4],XmNuserData,(caddr_t)TYPE_info_dlg);
        w[no++]=XtCreateWidget("Ok",xmPushButtonWidgetClass,TYPE_info_dlg,
                              wargs,5);
push(w[no-1]);
        XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)Ok_callback,TYPE_info_dlg);
	lineno++;

	w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"ID");
        n1 = ty_ptr->id;
        sprintf(str,"%5d",n1);
	w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
	lineno++;

	w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Variant");
	w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,tag[ty_ptr->variant]);
	lineno++;
	w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Length");
        sprintf(str,"%5d",ty_ptr->length);
	w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
	lineno++;

	w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Name");
	XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_SYM_dlg,ty_ptr->name);
	if (ty_ptr->name==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;
	
	w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Next (thread)");
	XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,ty_ptr->thread);
	if (ty_ptr->thread==NULL) XtSetSensitive(w[no-1],FALSE);
	lineno++;
	
        switch (ty_ptr->variant) {
            case DEFAULT:
            case T_INT:
            case T_FLOAT:
            case T_DOUBLE:
            case T_VOID :  /*  NEW ADDED FOR VPC */
            case T_CHAR:
            case T_BOOL:
            case T_STRING:
            case T_COMPLEX:
            case T_DCOMPLEX:
            case T_ENUM_FIELD:
            case T_BYTE:
                break;

/*   del by HIT  */
/*            case T_SUBRANGE:
		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Base Type");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.subrange.base_type);
		if (ty_ptr->entry.subrange.base_type==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Lower");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.subrange.lower);
		if (ty_ptr->entry.subrange.lower==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Upper");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.subrange.upper);
		if (ty_ptr->entry.subrange.upper==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;
*/
            case T_ARRAY:

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Base Type");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.ar_decl.base_type);
		if (ty_ptr->entry.ar_decl.base_type==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dimension");
                n1 = ty_ptr->entry.ar_decl.num_dimensions;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Ranges");
		XtSetArg(wargs[0],XmNuserData,ty_ptr->entry.ar_decl.ranges);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_multi_LLND_dlg3, NULL);
		if (ty_ptr->entry.ar_decl.ranges==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;
/*  del by HIT */
/*
            case T_LIST:
		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Base Type");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.base_type);
		if (ty_ptr->entry.base_type==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;
*/
            case T_RECORD:
		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Fields");
                n1 = ty_ptr->entry.re_decl.num_fields;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Size");
                n1 = ty_ptr->entry.re_decl.recordize;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Mapped");
                n1 = ty_ptr->entry.re_decl.mapped_flag;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"First");
		XtAddCallback(w[no-1],XmNactivateCallback,
		      (XtCallbackProc)show_SYM_dlg,ty_ptr->entry.re_decl.first);
		if (ty_ptr->entry.re_decl.first==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;

            case T_FUNCTION:  
	    case T_POINTER:
		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dummy1");
                n1 = ty_ptr->entry.Template.dummy1;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dummy2");
                n1 = ty_ptr->entry.Template.dummy2;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dummy3");
                n1 = ty_ptr->entry.Template.dummy3;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dummy4");
                n1 = ty_ptr->entry.Template.dummy4;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_label1(TYPE_info_dlg,COL_X1,lineno,"Dummy5");
                n1 = ty_ptr->entry.Template.dummy5;
        	sprintf(str,"%5d",n1);
		w[no++]=create_label2(TYPE_info_dlg,COL_X2,lineno,str);
		lineno++;

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Base Type");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.Template.base_type);
		if (ty_ptr->entry.Template.base_type==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Ranges");
		XtSetArg(wargs[0],XmNuserData,ty_ptr->entry.Template.ranges);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,
				(XtCallbackProc)show_multi_LLND_dlg3, NULL);
		if (ty_ptr->entry.Template.ranges==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;

/*  del by HIT */
/*            case T_DERIVED_TYPE :  
                n1 = FOLLOW_TYPE_POINTER_TO_ID(derived_type.symbol);
                break;
            case T_MEMBER_POINTER:
                n1 = FOLLOW_TYPE_POINTER_TO_ID(col_decl.collection_name);
                n2 = FOLLOW_TYPE_POINTER_TO_ID(col_decl.base_type);
                break;
            case T_DERIVED_COLLECTION:
                n1 = FOLLOW_TYPE_POINTER_TO_ID(col_decl.collection_name);
                n2 = FOLLOW_TYPE_POINTER_TO_ID(col_decl.base_type);
                break;
*/
/*  add by HIT but not defined in tag by HIT */

            case T_KIND:
		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Base Type");
		XtAddCallback(w[no-1],XmNactivateCallback,show_TYPE_dlg,
                                      ty_ptr->entry.kind_decl.base_type);
		if (ty_ptr->entry.kind_decl.base_type==NULL) 
                             XtSetSensitive(w[no-1],FALSE);
		lineno++;

		w[no++]=create_button(TYPE_info_dlg,COL_X1,lineno,"Kind Exp");
		XtSetArg(wargs[0],XmNuserData,ty_ptr->entry.kind_decl.kindexp);
		XtSetValues(w[no-1],wargs,1);
		XtAddCallback(w[no-1],XmNactivateCallback,(XtCallbackProc)show_multi_LLND_dlg3,
				NULL);
		if (ty_ptr->entry.kind_decl.kindexp==NULL) 
                              XtSetSensitive(w[no-1],FALSE);
		lineno++;
                break;


            default:
/*
                fprintf(fd, "%46s", " ");
*/
                break;
        }
/*
        XtManageChildren(buttons,XtNumber(TYPE_buttons));
        XtManageChildren(labels,XtNumber(TYPE_labels));
*/

	for (i=0;i<no;i++)
		XtAddEventHandler(w[i],KeyPressMask,FALSE,
					(XtEventHandler)closed,w[0]);

/*               Separate   */
	XtSetArg(wargs[0],XmNx,COL_X1);
	XtSetArg(wargs[1],XmNy,GETY(lineno));
	XtSetArg(wargs[2],XmNwidth,LABEL_WIDTH+40);
	XtSetArg(wargs[3],XmNheight,0);
	w[no++]=XtCreateManagedWidget("",xmLabelWidgetClass,
 			TYPE_info_dlg,wargs,4);

	XtManageChildren(w,no);
	return TYPE_info_dlg;

}



void show_TYPE_dlg(w,ty_ptr,call_data)
Widget     w;
PTR_TYPE      ty_ptr;
XmAnyCallbackStruct     *call_data;
{
	char str[50];

	Widget TYPE_info_dlg;
	if (ty_ptr==NULL) {
		BlockReportError("Error: Null Type found ! ");
		return;
	}
	if (ty_ptr->id > ptr_file->num_types || ty_ptr->id<1 )  {
		sprintf(str,"Error: Invalid TYPE id: %d !",ty_ptr->id);
		BlockReportError(str);
		return;
	}
        TYPE_info_dlg=create_TYPE_dlg(ty_ptr);
        XtManageChild(TYPE_info_dlg);

}

/*                    CALL BY  Multi_LLND_Dialog   */
void show_TYPE_dlg1(w,d,call_data)
Widget     w;
caddr_t        d;
XmAnyCallbackStruct     *call_data;
{
        Arg  wargs[3];
        NODE_PTR    node_ptr;

        XtSetArg(wargs[0],XmNuserData,&node_ptr);
        XtGetValues(w,wargs,1);
        show_TYPE_dlg(w,node_ptr->llnd->type,call_data);
}

