/*
 *	SimpleUtil.c
 *		including several simple utilities for X/Motif Widget.
 *	    Most of the source code here are based on xslib.c but are
 *	    slightly modified..
 *
 *	Author : Yu, Li
 *	Date :	 04/26/93
 */

#include <X11/Intrinsic.h>

#include "Xapp.h"
#include "XSimUtil.h"

/*
 *	XuWidgetPrintf, act as printf except it send its output to
 *	any XmLabel Widget.
void XuWidgetPrintf(va_alist)
	va_dcl
{
Widget		w;
char    	*format;
va_list		args;
char		str[500];
Arg		wargs[10];
XmString	xmstr;
   
	va_start(args);

        w=va_arg(args,Widget);

        format=va_arg(args,char *); 

        vsprintf(str,format,args);

        xmstr=XmStringLtoRCreate(str,XmSTRING_DEFAULT_CHARSET);
  
        XtSetArg(wargs[0],XmNlabelString,xmstr);
        XtSetValues(w,wargs,1);

        va_end(args);
	XmStringFree(xmstr);

}
 */

/*
 *	XuMessageBoxPrintf, act as printf except it send its output to
 *	any XmMessageBox Widget as its message string.
void XuMessageBoxPrintf(va_alist)
	va_dcl
{
Widget		w;
char    	*format;
va_list		args;
char		str[500];
Arg		wargs[10];
XmString	xmstr;
   
	va_start(args);

        w=va_arg(args,Widget);

        format=va_arg(args,char *); 

        vsprintf(str,format,args);

        xmstr=XmStringLtoRCreate(str,XmSTRING_DEFAULT_CHARSET);
  
        XtSetArg(wargs[0],XmNmessageString,xmstr);
        XtSetValues(w,wargs,1);

        va_end(args);
	XmStringFree(xmstr);

}
 */

/*
 *	XuTextPrintf, act as printf except that it send the output
 *	into a XmText widget .
void XuTextPrintf(va_alist)
	va_dcl
{
Widget		w;
char    	*format;
va_list		args;
char		str[500];

	va_start(args);
        w=va_arg(args,Widget);
        format=va_arg(args,char *); 
        vsprintf(str,format,args);
	XmTextSetString(w,str);
}
 */


/*
 *	XuWidgetGets, return the titleString of a Widget as a string.
 *	Memory are allocated by XtMalloc, should be freed by callee using
 *	XtFree();
 */
String XuWidgetGets(w)
Widget w;
{
	Arg wargs[1];
	char *retStr;

        XtSetArg(wargs[0],XmNlabelString,NULL);
        XtGetValues(w,wargs,1);
	XmStringGetLtoR((XmString)wargs[0].value,XmSTRING_DEFAULT_CHARSET,&retStr);
	return(retStr);
}

/*----------------------------------------------------------------
   	XuCreateMenuButtons
		Create buttons for any menubar or menupanel.
	Slightly improved by Yu, Li to recorgnize whether the
	passed widget is a menubar or a menupanel, and handle 
	them differently.
		This version include two more feature :
		1. caller can get all buttons' Widget if he need.
		2. caller can specify each button's class as long
	as several conditions are met. 
-----------------------------------------------------------------*/

void XuCreateMenuButtons(title,menu,menulist,nitems)
  char			*title;
  Widget		menu;
  MenuButtonsStruct	*menulist;	
  int			nitems;
{
Arg		wargs[1];
int		i,n;
WidgetList	buttons;
int		seperators=0;
unsigned char 	menu_type;
XmString        xmstr;

 /*
  *	Judge whether the passed widget is a menubar or a menupanel
  */
  XtSetArg(wargs[0],XmNrowColumnType,&menu_type);
  XtGetValues(menu,wargs,1);
/*  menu_type=(unsigned char)wargs[0].value;  */

 /*
  * Allocate a widget list to hold all button widgets
  *
  */

  buttons=(WidgetList)XtMalloc(nitems * sizeof(Widget));

 /*
  * If a title is given, create Label and Seperator 
  *
  */

  if (title && menu_type!=XmMENU_BAR)
  {
      XtCreateManagedWidget(title,xmLabelGadgetClass,menu,NULL,0);
      XtCreateManagedWidget("seperator",xmSeparatorGadgetClass,menu,NULL,0);
  }

 /*
  * Create an entry for each item in the menu 
  *
  */
 
  for (i=0; i<nitems; i++)
  {

 /*
  * A NULL name represents a seperator
  */

   if (menulist[i].name ==NULL)
   {
      if (menu_type!=XmMENU_BAR)
      {
         buttons[i]=XtCreateManagedWidget("seperator",xmSeparatorGadgetClass,
		menu,NULL,0);
      }
      /* else do nothing */
   }

 /*
  * If there is a name and a callback,  or user force 
  * to be a ToggleButton or PushButton and no sub_menu 
  * specified, then create a selectable menu entry and
  * register the callback function.
  */

  else if (menulist[i].func || menu_type==XmMENU_OPTION ||
	(!menulist[i].sub_menu && 
	( menulist[i].widgetClass==xmToggleButtonGadgetClass ||
	  menulist[i].widgetClass==xmPushButtonGadgetClass )))
  {
         WidgetClass widgetClass;

	if (menu_type==XmMENU_BAR) widgetClass=xmCascadeButtonGadgetClass;
         else if (menulist[i].widgetClass==NULL ||
		menulist[i].widgetClass==xmPushButtonGadgetClass) 
		widgetClass=xmPushButtonGadgetClass;
	 else if (menulist[i].widgetClass==xmCascadeButtonGadgetClass) 
		widgetClass=xmCascadeButtonGadgetClass; 
	 else widgetClass=xmToggleButtonGadgetClass;

         n=0; 
         XtSetArg(wargs[n],XmNmnemonic,menulist[i].mnemonic);n++;
/*
	 if (menulist[i].accelerator) {
                xmstr=XmStringCreateSimple(menulist[i].accel_text);
                XtSetArg(wargs[n],XmNbuttonAccelerators,menulist[i].accelerator); n++;
                XtSetArg(wargs[n],XmNbuttonAcceleratorText,xmstr); n++;
		XmStringFree(xmstr);
 	 }
*/
         buttons[i]=XtCreateWidget(menulist[i].name, widgetClass,menu,wargs,n);
         if (menulist[i].func)
	 {
		if (widgetClass!=xmToggleButtonGadgetClass)
             		XtAddCallback(buttons[i],XmNactivateCallback,
				menulist[i].func,menulist[i].data);
             	else XtAddCallback(buttons[i],XmNvalueChangedCallback,
				menulist[i].func,menulist[i].data);
	 }

         if (strcmp(menulist[i].name,"Help")==0)        
         {
            n=0;
            XtSetArg(wargs[n],XmNmenuHelpWidget,buttons[i]); n++;
            XtSetValues(menu,wargs,n);
         }   
  }

 /*
  * If there is a name , but no callback function,
  * the entry must be a label , unless there is a  
  * submenu or user force it to be another kind of
  * widget.
  */

  else if (!menulist[i].sub_menu && 
	(menulist[i].widgetClass==NULL || 
	menulist[i].widgetClass==xmLabelGadgetClass)) 
       {
         WidgetClass widgetClass;

         if (menu_type==XmMENU_BAR) widgetClass=xmCascadeButtonGadgetClass;
         else widgetClass=xmLabelGadgetClass;
         n=0; 
         XtSetArg(wargs[n],XmNmnemonic,menulist[i].mnemonic);n++;
         buttons[i]=XtCreateWidget(menulist[i].name,widgetClass,menu,wargs,n);
         if (strcmp(menulist[i].name,"Help")==0)        
         {
            n=0;
            XtSetArg(wargs[n],XmNmenuHelpWidget,buttons[i]); n++;
            XtSetValues(menu,wargs,n);
         }   
       }
   
 /*
  * else make the submenu 
  */

  else if (menulist[i].sub_menu!=NULL || 
	menulist[i].widgetClass==xmCascadeButtonGadgetClass)
  {
   	n=0; 
 	XtSetArg(wargs[n],XmNmnemonic,menulist[i].mnemonic);n++;
   	buttons[i] = XtCreateWidget(menulist[i].name,
		xmCascadeButtonGadgetClass,menu,wargs,n);
	if (menulist[i].sub_menu!=NULL)
	{
   	  Widget	sub_menu;
	  String        title=menulist[i].sub_menu_title;
  
 	  sub_menu=XmCreatePulldownMenu(menu,
		title?title:" ",NULL,0);      
/*		menulist[i].sub_menu_title,NULL,0);       zheng.gb */
   	  XuCreateMenuButtons(menulist[i].sub_menu_title,
                          sub_menu,menulist[i].sub_menu,
                          menulist[i].n_sub_items);
	  n=0;
 	  XtSetArg(wargs[n],XmNsubMenuId,sub_menu);n++;
          XtSetValues(buttons[i],wargs,n);
	}
   	if (strcmp(menulist[i].name,"Help")==0)        
   	{
       	 	n=0;
        	XtSetArg(wargs[n],XmNmenuHelpWidget,buttons[i]); n++;
        	XtSetValues(menu,wargs,n);
	}
  }
  
  /* keep the returned widget */
  if (menulist[i].ret_widget!=NULL)
	*menulist[i].ret_widget=buttons[i];
 
 } 

 XtManageChildren(buttons,nitems-seperators);

}

/* bits for exclamation point in dialog */
char warningBits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xe0, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc0, 0x03, 0x00,
   0x00, 0xe0, 0x07, 0x00, 0x00, 0xf0, 0x0f, 0x00, 0x00, 0xf0, 0x0f, 0x00,
   0x00, 0xf0, 0x0f, 0x00, 0x00, 0xf0, 0x0f, 0x00, 0x00, 0xe0, 0x07, 0x00,
   0x00, 0xc0, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00};

/************************************************************************
 *
 *  CreateDefaultImage - create a default images for warning symbol.
 *
 **********************************<->***********************************/
static XImage * xs_CreateDefaultImage (bits, width, height)
char *bits;
int   width, height;
{

    XImage *image;

    image = (XImage *) XtMalloc (sizeof (XImage));
    image->width = width;
    image->height = height;
    image->data = bits;
    image->depth = 1;
    image->xoffset = 0;
    image->format = XYBitmap;
    image->byte_order = LSBFirst;
    image->bitmap_unit = 8;
    image->bitmap_bit_order = LSBFirst;
    image->bitmap_pad = 8;
    image->bytes_per_line = (width+7)/8;
    return (image);

}

/************************************************************  
static Widget xs_CreateSpecialWarningDialog
              Create special 4 button message box out of a
              Selection box.
*************************************************************/

Widget xs_CreateSpecialWarningDialog (parent, name, image_string, message, arglist, argcount)
Widget          parent;
char            *name;
char            *image_string;
char            *message;
ArgList         arglist;
int             argcount;
{
        Widget          warning_dialog; /*  special warning selection box */
        Widget          work_area;      /*  rowcolumn for pixmap and text */
        Widget          pixmap_label;   /*  pixmap label                  */
        Widget          text_label;     /*  text label                    */
        Widget          apply_button;   /*  apply button                  */
        Widget          ok_button;      /*  ok button                     */
        Widget          kid[5];         /*  buttons                       */
        XImage          *image;         /*  image for warning pixmap    */
        Pixel           foreground;     /*  dialog foreground             */
        Pixel           background;     /*  dialog background             */
        Pixmap          pixmap;         /*  dialog pixmap                 */
        register int    i;              /*  kid index                     */
        Arg             al[10];         /*  arg list                      */
        register int    ac;             /*  arg count                     */

        image = xs_CreateDefaultImage (warningBits, 32, 32);
        XmInstallImage (image, "warning_image");

        warning_dialog = XmCreatePromptDialog(parent, name, arglist, argcount);

        ac = 0;
        XtSetArg(al[ac], XmNorientation, XmHORIZONTAL); ac++;
        work_area = XmCreateRowColumn(warning_dialog, "workarea", al, ac);
        XtManageChild(work_area);

        ac = 0;
        XtSetArg(al[ac], XmNforeground, &foreground); ac++;
        XtSetArg(al[ac], XmNbackground, &background); ac++;
        XtGetValues(warning_dialog, al, ac);

        ac = 0;
        XtSetArg(al[ac], XmNlabelType, XmPIXMAP); ac++;
        pixmap = XmGetPixmap(XtScreen(warning_dialog), image_string,
                             foreground, background);
        XtSetArg(al[ac], XmNlabelPixmap, pixmap); ac++;
        pixmap_label = XmCreateLabel(work_area, "pixmap_label", al, ac);
        XtManageChild(pixmap_label);

        ac = 0;
        XtSetArg(al[ac], XmNlabelString,
                 XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET)); ac++;
        text_label = XmCreateLabel(work_area, "text_label", al, ac);
        XtManageChild(text_label);

        apply_button = XmSelectionBoxGetChild (warning_dialog, XmDIALOG_APPLY_BUTTON);

        ac = 0;
        XtSetArg(al[ac], XmNlabelString,
                 XmStringCreateLtoR("Discard", XmSTRING_DEFAULT_CHARSET)); ac++;
        XtSetValues(apply_button, al, ac);
        XtManageChild(apply_button);

        ok_button = XmSelectionBoxGetChild (warning_dialog,XmDIALOG_OK_BUTTON);
        ac = 0;
        XtSetArg(al[ac], XmNlabelString,
                 XmStringCreateLtoR("Save", XmSTRING_DEFAULT_CHARSET)); ac++;
        XtSetValues(ok_button, al, ac);


        /*      Unmanage unneeded children.
        */
        i = 0;
        kid[i++] = XmSelectionBoxGetChild (warning_dialog, XmDIALOG_TEXT);
        kid[i++] = XmSelectionBoxGetChild (warning_dialog,
                                                     XmDIALOG_SELECTION_LABEL);
        XtUnmanageChildren (kid, i);

        return(warning_dialog);
}

int XuConvertToAbsCoord(disp,w,x,y,x_root,y_root)
Display *disp;
Window w;
int x,y;
int *x_root,*y_root;
{
	Window root,parent;
	Window *kids;
	unsigned int kidsCount;
	int win_x,win_y;
	unsigned int win_wid,win_ht,win_bd,depth;

	if (w!=DefaultRootWindow(disp))
	{
		XQueryTree(disp, w, &root, &parent, &kids, &kidsCount);
		XFree((char *)kids);
		XGetGeometry(disp,w, &root, &win_x,&win_y,&win_wid,&win_ht,
			&win_bd,&depth);
		x+=win_x;	y+=win_y;
		XuConvertToAbsCoord(disp,parent,x,y,x_root,y_root);
	}
	else
	{
		*x_root=x;
		*y_root=y;
	}
	return(SUCCESS);
}
		
		
int XuConvertToWindowCoord(disp,w,x_root,y_root,x,y)
Display *disp;
Window w;
int x_root,y_root;
int *x,*y;
{
	Window root,parent;
	Window *kids;
	unsigned int kidsCount;
	int win_x,win_y;
	unsigned int win_wid,win_ht,win_bd,depth;

	if (w!=DefaultRootWindow(disp))
	{
		XQueryTree(disp, w, &root, &parent, &kids, &kidsCount);
		XFree((char *)kids);
		XGetGeometry(disp,w, &root, &win_x,&win_y,&win_wid,&win_ht,
			&win_bd,&depth);
		x_root-=win_x;	y_root-=win_y;
		XuConvertToWindowCoord(disp,parent,x_root,y_root,x,y);
	}
	else
	{
		*x=x_root;
		*y=y_root;
	}
	return(SUCCESS);
}
		
		
int XuRevertShadow(w)
Widget w;
{
	/* revert widget's shadow */
        Arg args[5];

        XtSetArg(args[0],XmNtopShadowColor,&args[0].value);
        XtSetArg(args[1],XmNbottomShadowColor,&args[1].value);
        XtGetValues(w,args,2);
        XtSetArg(args[0],XmNbottomShadowColor,args[0].value);
        XtSetArg(args[1],XmNtopShadowColor,args[1].value);
        XtSetValues(w,args,2);
	return(SUCCESS);
}

int XuPressButton(w,time)
Widget w;
int time;
{
	Arg args[5];
	static int bkcolor,armcolor;
	static int status=0,index=0;
	static int   st[30];
/*	
	if (status==0) {
		if (w) {
			XtSetArg(args[0],XmNarmColor,&armcolor);
        		XtSetArg(args[1],XmNbackground,&bkcolor);
        		XtGetValues(w,args,2);
        		XtSetArg(args[0],XmNbackground,armcolor);
        		XtSetValues(w,args,1);
			XuRevertShadow(w);
			XSync(XtDisplay(w),False);
		}
		status=1;
	}
	else {
		if (w) {
        		XtSetArg(args[0],XmNbackground,bkcolor);
        		XtSetValues(w,args,1);
			XuRevertShadow(w);
			XSync(XtDisplay(w),False);
		}
		status=0;
	}
*/
	if (time==0) {
                if (w) {
                        XtSetArg(args[0],XmNarmColor,&armcolor);
                        XtSetArg(args[1],XmNbackground,&st[index]);
                        XtGetValues(w,args,2);
                        XtSetArg(args[0],XmNbackground,armcolor);
                        XtSetValues(w,args,1);
                        XuRevertShadow(w);
                        XSync(XtDisplay(w),False);
                }
		index++;
        }
        else {
                if (w) {
                        XtSetArg(args[0],XmNbackground,st[index-1]);
                        XtSetValues(w,args,1);
                        XuRevertShadow(w);
                        XSync(XtDisplay(w),False);
                }
		index--;
        }

}

 
int XuFontHeight(fontList)
XmFontList fontList;
{
	int height;
	XmString str;

	str=XmStringCreateLtoR("lg",XmSTRING_DEFAULT_CHARSET);
	height=XmStringHeight(fontList,str);
	XmStringFree(str);
	return(height);
}

int XuFontWidth(fontList)
XmFontList fontList;
{
	int width;
	XmString str;

	str=XmStringCreateLtoR("A",XmSTRING_DEFAULT_CHARSET);
	width=XmStringWidth(fontList,str);
	XmStringFree(str);
	return(width);
}

int XuGetRGBColor(display, r, g, b)
Display *display;
int r, g, b;
{
   /* return the pixel value of a named color */
   XColor color;
   Colormap cmap;

   cmap=DefaultColormap(display,DefaultScreen(display));

   color.red=r;
   color.green=g;
   color.blue=b;
   XAllocColor(display,cmap,&color);
   return(color.pixel);
}

int XuGetNamedColor(display, name)
Display *display;
char *name;
{
   /* return the pixel value of a named color */
   XColor color,exact_color;
   Colormap cmap;

   cmap=DefaultColormap(display,DefaultScreen(display));
   XAllocNamedColor(display,cmap,name,&color,&exact_color);
   return(color.pixel);
}

int XuClearEvent(display, mask)
Display *display;
long mask;
{
	XEvent event;

	while (XCheckMaskEvent(display, mask, &event));
	return(SUCCESS);
}

int XuDispatchEvent(display, mask)
Display *display;
long mask;
{
	XEvent event;

	while (XCheckMaskEvent(display, mask,&event))
	{
		XtDispatchEvent(&event);
	}
	return(SUCCESS);
}



/*        block function
	to abtain a answer :
			Yes, No, Cancel
*/

static void     response();

int  XzAskUser(parent,question,ans1,ans2,ans3,default_ans)
Widget  parent;
char    *question,*ans1,*ans2,*ans3;
int     default_ans;
{
	Widget   dialog,shell;
	XmString text,yes,no,cancel;
	int      n,answer=0;
	Arg      args[5];
	Atom     WM_DELETE_WINDOW;

	{
		dialog=XmCreateQuestionDialog(parent,"dialog",NULL,0);
		XtVaSetValues(dialog,
			XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL,
			NULL);
		if (ans3==NULL) XtSetSensitive(XmMessageBoxGetChild(dialog,
					XmDIALOG_HELP_BUTTON),FALSE);
		XtAddCallback(dialog,XmNokCallback,response,&answer);
		XtAddCallback(dialog,XmNcancelCallback,response,&answer);
		if (ans3)XtAddCallback(dialog,XmNhelpCallback,response,&answer);
	}
	text=XmStringCreateLtoR(question,XmSTRING_DEFAULT_CHARSET);
	yes=XmStringCreateSimple(ans1);
	no=XmStringCreateSimple(ans2);
	if (ans3) cancel=XmStringCreateSimple(ans3);
	XtVaSetValues(dialog,
			XmNmessageString,   text,
			XmNokLabelString,   yes,
			XmNcancelLabelString,   no,
			NULL);
	n=0;
	switch (default_ans) {
	case ASK_YES: XtSetArg(args[n],XmNdefaultButtonType,XmDIALOG_OK_BUTTON); 
			break;
	case ASK_NO: XtSetArg(args[n],XmNdefaultButtonType,XmDIALOG_CANCEL_BUTTON);
			break;
	case ASK_CANCEL:XtSetArg(args[n],XmNdefaultButtonType,XmDIALOG_HELP_BUTTON);
			break;
	}
	n++;
	if (ans3)  {
		XtSetArg(args[n],XmNhelpLabelString,cancel);
		n++;
	}
	XtSetValues(dialog,args,n);
	XmStringFree(text);
	XmStringFree(yes);
	XmStringFree(no);
	if (ans3) XmStringFree(cancel);
	XtManageChild(dialog);

	shell=XtParent(dialog);
        WM_DELETE_WINDOW=XmInternAtom(XtDisplay(parent),"WM_DELETE_WINDOW",
                                        False);
        XmAddWMProtocolCallback(shell,WM_DELETE_WINDOW,response,&answer);

	while (answer==0) {
		XtProcessEvent(XtIMAll);
		XSync(XtDisplay(dialog),0);
	}
/*
	XuRevertShadow(XmMessageBoxGetChild(dialog,
                                        XmDIALOG_CANCEL_BUTTON));
*/
	XtDestroyWidget(dialog);
/*
	XSync(XtDisplay(dialog),0);
	XmUpdateDisplay(dialog);
*/

	return answer;
}

static void response(w,answer,cbs)
Widget w;
int    *answer;
XmAnyCallbackStruct  *cbs;
{
	struct   _cbs{
		Widget w;
		int    *answer;
	} *close_cbs;

	if (cbs->reason==XmCR_OK)   
		*answer=ASK_YES;
	else if (cbs->reason==XmCR_CANCEL)   
		*answer=ASK_NO;
	     else if (cbs->reason==XmCR_HELP)   
			*answer=ASK_CANCEL;
		  else {
			*answer=ASK_CANCEL;
			}
}
	
	

