/*****************************************************************************
        Visualization tool of AST 

    author:  Zheng Gengbin
                            Peking Univ. P.R.C
*****************************************************************************/


/*     process.c    */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(STDC_HEADERS) && (STDC_HEADERS==1)
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#include "process.h"


BIF_TREE  b_tree = NULL;
int     nBIF = 0;

#ifdef DEBUG
int display()
{
 int n1,n2,n3,n4,n5,n6,n7;
 int i;
 printf("***BIF NODES***\n\n");
 printf("ID  Parent  Son  elder younger depth height nson\n");
 for (i=0;i<nBIF;i++){
   n1=b_tree[i].parent;
   n2=b_tree[i].son;
   n3=b_tree[i].elder;
   n4=b_tree[i].younger;
   n5=b_tree[i].depth;
   n6=b_tree[i].height;
   n7=b_tree[i].nson;
   printf("%-3d %5d %5d %5d %5d %5d %5d %5d \n",i,n1,n2,n3,n4,n5,n6,n7);
 }
}
#endif
  
/*
                      Preprocess()
   alloc array of BIF TREE
   assign every node's bfnd,parent,son,elder,expaned,nson,
   BUT no height   --------  use another function set_height()
*/

#define NORMAL    1
#define ABNORMAL 2

/* correct BFND tree  thread error by HIT */
correct_tree(f)
PTR_FILE f;
{
	PTR_BFND bf_ptr,last;
	int      modified=FALSE;

	last=NULL;
	for (bf_ptr = f->head_bfnd; bf_ptr; bf_ptr = bf_ptr->thread)
	{
		if (bf_ptr->control_parent == NULL) { last=bf_ptr; continue;}
		if (bf_ptr->id == bf_ptr->control_parent->id-1 ){
			last->thread=bf_ptr->control_parent;
			bf_ptr->thread=bf_ptr->control_parent->thread;
			bf_ptr->control_parent->thread=bf_ptr;
			bf_ptr->id++;
			(bf_ptr->control_parent->id)--;
			if (!modified) modified=TRUE;
		}
		last=bf_ptr;
	}	
	if (modified) ReportError("Warning: Abnormal thread found ! ");
}

int preprocess(f)
PTR_FILE f;
{
 register PTR_BFND bf_ptr;
 int   i,p;
 int   state;

	correct_tree(f);
 state=NORMAL;
 if (b_tree != NULL )  free(b_tree);
 b_tree=(BIF_TREE)(f->num_bfnds?calloc(f->num_bfnds,sizeof(struct tree_node))
                                :NULL);
 nBIF=f->num_bfnds;
 if (nBIF>MAXHEIGHT) state=ABNORMAL;
 i=-1;
 for (bf_ptr = f->head_bfnd; bf_ptr; bf_ptr = bf_ptr->thread)
 {
  i++;
  b_tree[i].bfnd=bf_ptr;
  b_tree[i].nson=0;
  b_tree[i].son=-1;
  b_tree[i].elder=-1;
  b_tree[i].younger=-1;
  b_tree[i].lastson=-1;
  p=bf_ptr->control_parent ? bf_ptr->control_parent->id-1 : -1;
  b_tree[i].parent=p;
  if (p!=-1) {
    if (b_tree[p].son==-1)  b_tree[p].son=i;
    b_tree[p].nson++;
    if (b_tree[p].lastson!=-1) b_tree[b_tree[p].lastson].younger=i;
    b_tree[i].elder=b_tree[p].lastson;
    b_tree[p].lastson=i;
    b_tree[i].depth=b_tree[p].depth+1;           /*   Parent plus 1 */
  }
  else b_tree[i].depth=0;                        /*   ROOT NODE   */
  b_tree[i].show=(state==NORMAL?TRUE:FALSE);
 }     /* end of for */ 
 if (state==ABNORMAL) b_tree[0].show=TRUE;
 set_height();
}

static int l;                             /*    indicate Level */

static int travelson(p)
int p;
{
 int ptr;

 if (!b_tree[p].show) return 0;
 b_tree[p].height=l;
 l++;
 for (ptr=b_tree[p].son; ptr>=0; ptr=b_tree[ptr].younger)
	travelson(ptr);
}

int set_height()
{
  l=0;
  travelson(0);
}

 
/*****************************************************************************
          SINGLE LLND tree
******************************************************************************/


LLND_TREE  l_tree;
int        nLLND;

int count(ll_ptr)
PTR_LLND ll_ptr;
{
	PTR_LLND ptr;

	nLLND++;
	ptr=ll_ptr->entry.Template.ll_ptr1;
	if (ptr) count(ptr);
	ptr=ll_ptr->entry.Template.ll_ptr2;
	if (ptr) count(ptr);
}

static int   _n;

int  set_val(ll_ptr,parent)
PTR_LLND ll_ptr;
int      parent;
{
	PTR_LLND  ptr;
	int       whoami;

	whoami=_n;
        l_tree[_n].llnd=ll_ptr;
	l_tree[_n].height=_n;
	l_tree[_n].parent=parent;
	if (parent!=-1) l_tree[_n].depth=l_tree[parent].depth+1;
	else l_tree[_n].depth=0;
	_n++;
	ptr=ll_ptr->entry.Template.ll_ptr1;
        if (ptr) set_val(ptr,whoami);
        ptr=ll_ptr->entry.Template.ll_ptr2;
        if (ptr) set_val(ptr,whoami);
}



int  proc_llnd_tree(ll_ptr)
PTR_LLND ll_ptr;
{
 int i;
	nLLND=0;
	if (ll_ptr==NULL) {
	  fprintf(stderr,"Error ! Null llnd ");
	  l_tree=NULL;
	  return;
	}
	count(ll_ptr);
fprintf(stderr,"Number LLND :%d\n",nLLND);
	l_tree=(LLND_TREE)calloc(nLLND,sizeof(struct tree_llnd));
	_n=0;
        set_val(ll_ptr,-1);
for (i=0;i<nLLND;i++)
   fprintf(stderr,"ID:%d id:%d Depth:%d Height:%d Parent:%d\n",i,l_tree[i].llnd->id,l_tree[i].depth,l_tree[i].height,l_tree[i].parent);

}


/*****************************************************************************

             MUTI_LLND_PROCESS

****************************************************************************/

static void init_node(node_ptr,llnd_ptr,depth,height)
NODE_PTR node_ptr;
PTR_LLND   llnd_ptr;
int        depth,height;
{
	node_ptr->llnd=llnd_ptr;
	node_ptr->son1=NULL;
	node_ptr->son2=NULL;
	node_ptr->depth=depth;
	node_ptr->height=height;
	node_ptr->right=NULL; 
}

static NODE_PTR  search_node(head,llnd)
LIST_PTR  head;
PTR_LLND   llnd;
{
	LIST_PTR p;
	NODE_PTR q;

	if (llnd==NULL) return NULL;
	for (p=head;p;p=p->next)
		for(q=p->right;q;q=q->right)
	      	 if (q->llnd==llnd) return q;
	return NULL;
}


/*
 *   depth_node() -- get the depth of the node p in root ptr
 *    return -1 if not found p.
 */
static int  depth_node(list,p,ptr,level)   /*  P is the llnd to make depth */
LIST_PTR list;
PTR_LLND p,
         ptr;
int      level;
{
	int m1,m2;
	PTR_LLND   q,q1,q2;
	
	if (p==ptr) return level;
	if (ptr==NULL) return -1;
	if (!ll_has_ll(ptr->variant)) return -1;
	q1=ptr->entry.Template.ll_ptr1;
	q2=ptr->entry.Template.ll_ptr2;
	m1=-1;  m2=-1;
	if (q1==NULL && q2==NULL) return -1;
/*   modified on 1/25 1996   by zheng.gb  */
/*   make sure if the same level 
		19 ----   18
	        |          |
	      |   |       |  |
	      8   18      16 17
*/
	if (q1 && !search_node(list,q1)) m1=depth_node(list,p,q1,level+1);
	if (q2 && !search_node(list,q2)) m2=depth_node(list,p,q2,level+1);
	return m1>m2?m1:m2;
}

static BOOL deepest_node(list,q,no,ptr1,ptr2,ptr3)
LIST_PTR     list;
PTR_LLND     q;
int          no;
PTR_LLND     ptr1,
	     ptr2,
	     ptr3;
{
	int n1,n2,n3;
	n1=depth_node(list,q,ptr1,0);
	n2=depth_node(list,q,ptr2,0);
	n3=depth_node(list,q,ptr3,0);
	if (no>=n1 && no>=n2 && no>=n3) return TRUE;
	else return FALSE;
}

static void construct_son(head)
LIST_PTR  head;
{
	LIST_PTR   p;
	NODE_PTR   q,qson;

	for (p=head;p;p=p->next)
	  for (q=p->right;q;q=q->right)
		if (q->llnd && ll_has_ll(q->llnd->variant)) 
		{
			
			if (q->llnd->entry.Template.ll_ptr1) {
			   qson=search_node(head,q->llnd->entry.Template.ll_ptr1
                                        );
			   q->son1=qson;
			}
			if (q->llnd->entry.Template.ll_ptr2) {
                           qson=search_node(head,q->llnd->entry.Template.ll_ptr2
                                        );
                           q->son2=qson;
                        }
		}
}

int  adjust_tree(head)
LIST_PTR head;
{
	NODE_PTR  np,p0,n1,n2;
	LIST_PTR  nl;
	int       d1,d2;

	nl=head->next;
	if (nl) adjust_tree(nl);
	for (np=head->right;np;np=np->right)  {
		n1=np->son1;  n2=np->son2;
		d2=n2?n2->depth:-1;
		d1=n1?n1->depth:-1;
		if (d1!=-1 && d2 !=-1) np->depth=(d1+d2)/2;
		else  if (d1!=-1 || d2!=-1)  np->depth= (d1==-1) ? d2: d1;
	}
	p0=NULL;
	for (np=head->right;np;np=np->right)  {
		if (p0) {
			if (p0->depth>=np->depth) np->depth=p0->depth+1;
		}
		p0=np;
	}
			
}

void display_multi_llnd (head)
LIST_PTR head;
{
	LIST_PTR p;
	NODE_PTR q;

	fprintf(stderr,"********************************\n");
	for (p=head;p;p=p->next) {
		fprintf(stderr,"Depth:%d\n",p->no);
		for (q=p->right;q;q=q->right)
		 if (q->llnd)
			fprintf(stderr,"ID:%-4dson1:%-4dson2:%-4d  ",
				q->llnd->id,q->son1?q->son1->llnd->id:-1,
				q->son2?q->son2->llnd->id:-1);
		fprintf(stderr,"\n");
	}
}


LIST_PTR proc_multi_llnd_tree(ptr1,ptr2,ptr3)
PTR_LLND ptr1,
	 ptr2,
	 ptr3;
{
	LIST_PTR list_head,p,p1;
	NODE_PTR q,q1,q2,q3;
	PTR_LLND pt,qson;
	int      dp1,dp2,dp3,md;
	int      no;
	int      depth;

	list_head=(LIST_PTR)malloc(sizeof(LIST_LLND));
	list_head->next=NULL;  list_head->right=NULL;
	list_head->num=3;      list_head->no=0;
	q1=(NODE_PTR)malloc(sizeof(LIST_NODE));
	init_node(q1,ptr1,0,0);
	list_head->right=q1;
	q2=(NODE_PTR)malloc(sizeof(LIST_NODE));
	init_node(q2,ptr2,1,0);
	q1->right=q2;
	q1=q2;
	q2=(NODE_PTR)malloc(sizeof(LIST_NODE));
        init_node(q2,ptr3,2,0);
	q1->right=q2;
	no=1;
	for (p1=list_head;p1->right;p1=p) {
	 p=(LIST_PTR)malloc(sizeof(LIST_LLND));
	 p1->next=p;
	 p->next=NULL;    p->right=NULL;
	 p->no=no;        p->num=0;
	 no++;
	 q3=NULL;
	 for (q=p1->right;q;q=q->right) {
		pt=q->llnd;
		if (pt==NULL) continue;
		if (!ll_has_ll(pt->variant))  continue;  /* Has no LLND sons */
		qson=pt->entry.Template.ll_ptr1;
		if (qson && deepest_node(list_head,qson,p->no,ptr1,ptr2,ptr3) &&
			search_node(list_head,qson)==NULL)
		{
		  q2=(NODE_PTR)malloc(sizeof(LIST_NODE));
		  depth=q3?q3->depth+1:0;
		  depth=depth > q->depth-1 ? depth : q->depth;
		  init_node(q2,pt->entry.Template.ll_ptr1,depth,
				p->no);
		  p->num++;
		  if (q3) q3->right=q2;
		  else p->right=q2;
		  q3=q2;
		}
		qson=pt->entry.Template.ll_ptr2;
		if (qson && deepest_node(list_head,qson,p->no,ptr1,ptr2,ptr3) &&
                        search_node(list_head,qson)==NULL)
                {
                  q2=(NODE_PTR)malloc(sizeof(LIST_NODE));
		  depth=q3?q3->depth+1:0;
		  depth=depth > q->depth ? depth : q->depth;
                  init_node(q2,pt->entry.Template.ll_ptr2,depth,
                                p->no);
                  p->num++;
                  if (q3) q3->right=q2;
                  else p->right=q2;
                  q3=q2;
                }
	 }
	}
	
	construct_son(list_head);
	adjust_tree(list_head);
/*
	display_multi_llnd(list_head);
*/

	return list_head;
}
	


