//
//	RunTime.cc
//		Implement run time performance data collection
//		definition in RunTime.h and RunTimeP.h
//

#define _CPP
#define _RUNTIMEP
#include "headers.h"


//////////////////////////////////////////////////////////////////
//	global variable, defined in RunTime.h and RunTimeP.h
stack < vector <PtrParaStmtInfo> > g_stParaStmt;
		//      stack of active parallel statement

CProcInfo g_procInfo;
		//      global CProcInfo object to store all prof data
int g_curStmt=0;

/////////////////////////////////////////////////////////////////
//	extern visible global variable
	//	time registers
int gr_nSTimeCPU=0;	//	start CPU time for current statement
int gr_nElapseCPU=0;	//	CPU time elpase for current statement
int gr_nSTimeSystem=0;	//	start system time for current statement
int gr_nElapseSystem=0;	//	system time elpase for current statement
int gr_nTotalElapseCPU=0;		//	total CPU time elpase for current statement
int gr_nTotalElapseSystem=0;	//	total system time elpase for cur statement
int gr_nComSTimeCPU=0;	//	start CPU time for current communication
int gr_nComElapseCPU=0;	//	CPU time elpase of comm for cur statement
int gr_nComSTimeSystem=0;	//	start system time for current communication
int gr_nComElapseSystem=0;	//	system time elpase of comm for cur statement

	//	loop counters
int gr_nLoopCounter=0;	//	times of loop executed

	//	communication
int gr_nShift=0;			//	times of shift for current statement
int gr_nShiftAmount=0;	//	shift amount for current statement
int gr_nRemap=0;			//	times of remap for current statement
int gr_nRemapAmount=0;	//	remap amount for current statement
int gr_aSend[MAX_PROC];			//	time of send to given proc
int gr_aSendAmount[MAX_PROC];	//	ammount of send to given proc
int gr_aRecv[MAX_PROC];			//	time of recv to given proc
int gr_aRecvAmount[MAX_PROC];	//	ammount of recv to given proc

struct timeval start_tv;
struct timeval tv;
struct timezone tz;

double dfStartClock;

int g_nProcNum=0;

//	var control whether to record event infomation
BOOL g_bNoEvent=FALSE;

double MPI_Wtime();
int GetTimeCPU() { return (int)(1000000*(MPI_Wtime()-dfStartClock)); }

//////////////////////////////////////////////////////////////////
//	public functions, called in HPF compiler and AdLib

//	initialization and finalization
BOOL prof_begin__(int *pnProcID, int *pnProcNum)
{
	gettimeofday(&start_tv,&tz);
	g_nProcNum=*pnProcNum;
	dfStartClock=MPI_Wtime();
	RestoreState(NULL);
	return g_procInfo.Start(*pnProcID);
}

BOOL prof_end__()
{
	SaveCurState(NULL);
	return g_procInfo.Stop();
}

//	Virtual Machine manipulation
BOOL prof_pushstack__(int *pnStmtID)
{
	//	first find a existing statement record
	//	or create a new one
	CParaStmtInfo *pParaStmt=g_procInfo.FindStmt(*pnStmtID);
	if (NULL==pParaStmt)
	{
		pParaStmt=g_procInfo.NewStmt(*pnStmtID);
		if (NULL==pParaStmt)
			return FALSE;
	}

	//	save current state to stack top
	CParaStmtInfo *pEnclosingStmt=
		g_stParaStmt.empty()? (CParaStmtInfo*)NULL : g_stParaStmt.top();
	SaveCurState(pEnclosingStmt);

	//	push stack, and prepare state for new stack top statement
	g_stParaStmt.push(pParaStmt);
	gr_nTotalElapseCPU=0;
	gr_nTotalElapseSystem=0;
	RestoreState(pParaStmt);

	ProfNewEvent(stat_para, 0, 0);
	return TRUE;
}

BOOL prof_popstack__()
{
	//	find and save current statement to stack top
	//	stack top should not be NULL
	CParaStmtInfo *pStmt=
		g_stParaStmt.empty()? (CParaStmtInfo*)NULL : g_stParaStmt.top();

	SaveCurState(pStmt);
	if (pStmt!=NULL)
	{
		ProfNewEvent(stat_endpara, 0, 0);
	}
	else assert(FALSE);

	//	pop stack
	g_stParaStmt.pop();

	//	restore current state for new stack top
	CParaStmtInfo *pLastStmt=
		g_stParaStmt.empty()? (CParaStmtInfo*)NULL : g_stParaStmt.top();

	RestoreState(pLastStmt);

	return TRUE;
}

//	Event Recording
BOOL prof_newevent__(int *pnState, int *pnTarget, int *pnComSize)
{
	if (*pnState==stat_send)
	{
		gr_aSend[*pnTarget]++;
		gr_aSendAmount[*pnTarget]+=*pnComSize;
	}
	else if (*pnState==stat_recv)
	{
		gr_aRecv[*pnTarget]++;
		gr_aRecvAmount[*pnTarget]+=*pnComSize;
	}
	
	if (!g_bNoEvent)
	{
		if (*pnState==stat_para)
		{
			//	Check if we can merge this event with existing events
			POSITION pos=g_procInfo.GetLastEventPos();
			CExecEvent *pLastEvent=g_procInfo.GetPrevEvent(pos);
		
			if (pLastEvent!=NULL && 
				pLastEvent->m_nState==stat_endpara &&
				pLastEvent->m_nStmtID==g_curStmt)
			{
				//	if we can, do merge
				g_procInfo.EraseLastEvent();
				return TRUE;
			}
		}
	
		//	create a new event
		return (NULL!=
			g_procInfo.NewEvent(g_curStmt, *pnState, *pnTarget, *pnComSize));
	}
	else return FALSE;
}
	
BOOL prof_endevent__(int *pnState)
{
	if (g_bNoEvent) return FALSE;

	//	check if last event is atomic
	POSITION pos=g_procInfo.GetLastEventPos();
	CExecEvent *pLastEvent=g_procInfo.GetPrevEvent(pos);
	
	if (pLastEvent!=NULL && pLastEvent->m_nState==*pnState &&
		pLastEvent->m_nStmtID==g_curStmt)
	{
		//	set the time elapse field of an atomic event
		pLastEvent->EndEvent();
		return TRUE;
	}
	else return FALSE;
}

void prof_inccounter__(int *pnCount)
{
	gr_nLoopCounter+=*pnCount;
}

BOOL prof_def_processors__(int *pnProcessorsID, int *pnGroupID)
{
	CProcessors *pProcs=g_procInfo.FindProcDefByGrp(*pnGroupID);
	if (pProcs==NULL) return FALSE;
	pProcs->m_nProcessorsID=*pnProcessorsID;
	return TRUE;
}

BOOL prof_def_range__(int *pnTemplateID, int *pnDim, int *pnRangeID)
{
	CTemplate *pTempl=g_procInfo.FindTemplateByRange(*pnRangeID);
	if (pTempl==NULL) return FALSE;
	pTempl->m_nDim=*pnDim;
	pTempl->m_nTemplateID=*pnTemplateID;
	return TRUE;
}

BOOL prof_def_array__(int *pnArray, int *pnDAD)
{
	CDistArrayInfo *pArray=g_procInfo.FindDistArrayByDAD(*pnDAD);
	if (pArray==NULL) return FALSE;
	pArray->m_nArrayID=*pnArray;
	return TRUE;
}

BOOL ProfNewGroupGrid(int nGroup, int nRank, int *aShape)
{
	CProcessors *pProcessors=g_procInfo.NewProcDef(0, nGroup);
	if (pProcessors==NULL) return FALSE;
	pProcessors->m_nDimCount=nRank;
	for (int i=0; i<nRank; i++)
	{
		pProcessors->m_aSize[i]=aShape[i];
	}
	return TRUE;
}

BOOL ProfNewRangeDistribute(int nRange, int nLB, int nUB, int nGrp, int nDist)
{
	CTemplate *pTempl=g_procInfo.NewTemplate(0, 0, nRange);
	if (pTempl==NULL) return FALSE;
	pTempl->SetDist(nLB, nUB, nGrp, nDist);
	return TRUE;
}

BOOL ProfNewRangeCollapse(int nRange, int nLB, int nUB)
{
	CTemplate *pTempl=g_procInfo.NewTemplate(0, 0, nRange);
	if (pTempl==NULL) return FALSE;
	pTempl->SetDist(nLB, nUB, 0, dist_collapse);
	return TRUE;
}

BOOL ProfNewArrayData(int nDAD, int nRank)
{
	CDistArrayInfo *pArray=g_procInfo.NewDistArray(0, nDAD);
	if (pArray==NULL) return FALSE;
	pArray->m_nDimAmount=nRank;
	return TRUE;
}

BOOL ProfSetArrayAlign(int nDAD, int nDim, int nLB, int nUB)
{
	CDistArrayInfo *pArray=g_procInfo.FindDistArrayByDAD(nDAD);
	if (NULL==pArray) return FALSE;
	pArray->SetArrayAlign(nDim, nLB, nUB, 0, 0, 0, align_collapse);
	return TRUE;
}

BOOL ProfSetArrayCollapse(int nDAD, int nDim, int nRange, 
	int nLB, int nUB, int a, int b)
{
	CDistArrayInfo *pArray=g_procInfo.FindDistArrayByDAD(nDAD);
	if (NULL==pArray) return FALSE;
	pArray->SetArrayAlign(nDim, nLB, nUB, a, b, nRange, align);
	return TRUE;
}

BOOL ProfSetArrayCopy(int nDAD, int nDim, int nRange)
{
	CDistArrayInfo *pArray=g_procInfo.FindDistArrayByDAD(nDAD);
	if (NULL==pArray) return FALSE;
	pArray->SetArrayAlign(nDim, 0, 0, 0, 0, nRange, align_copy);
	return TRUE;
}

//////////////////////////////////////////////////////////////////
//	private functions, for internal use only
BOOL SaveCurState(CParaStmtInfo *pParaStmt)
{
	int nElapseCPU=GetTimeCPU()-gr_nSTimeCPU;
	int nElapseSystem=GetTimeSystem()-gr_nSTimeSystem;

	gr_nElapseCPU+=nElapseCPU;
	gr_nElapseSystem+=nElapseSystem;
	gr_nTotalElapseCPU+=nElapseCPU;
	gr_nTotalElapseSystem+=nElapseSystem;

	if (pParaStmt!=NULL)
	{
		pParaStmt->m_nLoopCounter=gr_nLoopCounter;     
		pParaStmt->m_nElapseCPU=gr_nElapseCPU;       
		pParaStmt->m_nElapseSystem=gr_nElapseSystem;    
		pParaStmt->m_nTotalElapseCPU=gr_nTotalElapseCPU;  
		pParaStmt->m_nTotalElapseSystem=gr_nTotalElapseSystem;   
		pParaStmt->m_nComElapseCPU=gr_nComElapseCPU;    
		pParaStmt->m_nComElapseSystem=gr_nComElapseSystem; 
	
		pParaStmt->m_nShift=gr_nShift;           
		pParaStmt->m_nShiftAmount=gr_nShiftAmount;     
		pParaStmt->m_nRemap=gr_nRemap;           
		pParaStmt->m_nRemapAmount=gr_nRemapAmount;     
	
		for (int i=0; i<g_nProcNum; i++)
		{
			g_procInfo.m_aSend[i]+=gr_aSend[i]-pParaStmt->m_aSend[i];   
			g_procInfo.m_aSendAmount[i]+=
					gr_aSendAmount[i]-pParaStmt->m_aSendAmount[i]; 
			g_procInfo.m_aRecv[i]+=gr_aRecv[i]-pParaStmt->m_aRecv[i];   
			g_procInfo.m_aRecvAmount[i]+=
					gr_aRecvAmount[i]-pParaStmt->m_aRecvAmount[i]; 
	
			pParaStmt->m_aSend[i]=gr_aSend[i];   
			pParaStmt->m_aSendAmount[i]=gr_aSendAmount[i]; 
			pParaStmt->m_aRecv[i]=gr_aRecv[i];   
			pParaStmt->m_aRecvAmount[i]=gr_aRecvAmount[i]; 
		}
	}
	else
	{
		for (int i=0; i<g_nProcNum; i++)
		{
			g_procInfo.m_aSend[i]+=gr_aSend[i];
			g_procInfo.m_aSendAmount[i]+=gr_aSendAmount[i];
			g_procInfo.m_aRecv[i]+=gr_aRecv[i];
			g_procInfo.m_aRecvAmount[i]+=gr_aRecvAmount[i];
		}
	}
	return TRUE;
}

BOOL RestoreState(CParaStmtInfo *pParaStmt)
{
	if (pParaStmt!=NULL)
	{
		g_curStmt=pParaStmt->m_nStmtID;
	
		int nTimeCPU=GetTimeCPU();
		int nTimeSystem=GetTimeSystem();
	
		gr_nTotalElapseCPU=pParaStmt->m_nTotalElapseCPU+nTimeCPU-gr_nSTimeCPU;  
		gr_nTotalElapseSystem=
			pParaStmt->m_nTotalElapseSystem+nTimeSystem-gr_nSTimeSystem;   
	
		gr_nSTimeCPU=nTimeCPU;
		gr_nSTimeSystem=nTimeSystem;
	
		gr_nLoopCounter=pParaStmt->m_nLoopCounter;     
		gr_nElapseCPU=pParaStmt->m_nElapseCPU;       
		gr_nElapseSystem=pParaStmt->m_nElapseSystem;    
		gr_nComElapseCPU=pParaStmt->m_nComElapseCPU;    
		gr_nComElapseSystem=pParaStmt->m_nComElapseSystem; 
	
		gr_nShift=pParaStmt->m_nShift;           
		gr_nShiftAmount=pParaStmt->m_nShiftAmount;     
		gr_nRemap=pParaStmt->m_nRemap;           
		gr_nRemapAmount=pParaStmt->m_nRemapAmount;     
	
		for (int i=0; i<g_nProcNum; i++)
		{
			gr_aSend[i]=pParaStmt->m_aSend[i];   
			gr_aSendAmount[i]=pParaStmt->m_aSendAmount[i]; 
			gr_aRecv[i]=pParaStmt->m_aRecv[i];   
			gr_aRecvAmount[i]=pParaStmt->m_aRecvAmount[i]; 
		}
	}
	else
	{
		g_curStmt=-1;
		for (int i=0; i<g_nProcNum; i++)
		{
			gr_aSend[i]=0;
			gr_aSendAmount[i]=0;
			gr_aRecv[i]=0;
			gr_aRecvAmount[i]=0;
		}
	}
	return TRUE;
}

/////////////////////
//	for test
double MPI_Wtime()
{
	return (((double)clock())/CLOCKS_PER_SEC);
}
