//
//	RunTime.h
//		run time performance data collection classes and data structures
//		including:
//			global variables serve as registers of performance data
//

#ifndef _RUNTIME_H
#define _RUNTIME_H

/////////////////////////////////////////////////////////////////////////////
//	definition of running sate
    enum
    {   //  execution state tags
		stat_none=-1,
        stat_serial,    //  in serial computing
        stat_para,      //  in parallel computing
        stat_endpara,       //  in parallel computing
        stat_send,      //  send data to other processors, atomic
        stat_recv,      //  receiving data from other proc, waiting, atomic
        stat_idle,      //  idle, waiting for sync, no use for now, atomic
        stat_term	//  program termination
    };

////////////////////////////////////////////////////
//	definition of data distribute type
	enum
	{
		dist_copy,
		dist_block,
		dist_cyclic,
		dist_collapse
	};

	enum
	{
		align_copy,
		align,
		align_collapse
	};

/////////////////////////////////////////////////
//	virtual registers, need save and restore

	//	time registers
extern int gr_nSTimeCPU;	//	start CPU time for current statement
extern int gr_nElapseCPU;	//	CPU time elpase for current statement
extern int gr_nSTimeSystem;	//	start system time for current statement
extern int gr_nElapseSystem;//	system time elpase for current statement
extern int gr_nComSTimeCPU;	//	start CPU time for current communication
extern int gr_nComElapseCPU;	//	CPU time elpase of comm for cur statement
extern int gr_nComSTimeSystem;	//	start system time for current communication
extern int gr_nComElapseSystem;	//	system time elpase of comm for cur statement

extern int gr_nTotalElapseCPU;	//	all CPU time elpase for cur statement
extern int gr_nTotalElapseSystem;	//	all system time elpase for cur statement

	//	loop counters
extern int gr_nLoopCounter;	//	times of loop executed

	//	communication
extern int gr_nShift;		//	times of shift for current statement
extern int gr_nShiftAmount;	//	shift amount for current statement
extern int gr_nRemap;		//	times of remap for current statement
extern int gr_nRemapAmount;	//	remap amount for current statement
extern int gr_aSend[MAX_PROC];		//	time of send to given proc
extern int gr_aSendAmount[MAX_PROC];//	ammount of send to given proc
extern int gr_aRecv[MAX_PROC];		//	time of recv to given proc
extern int gr_aRecvAmount[MAX_PROC];//	ammount of recv to given proc


extern int g_nProcNum;		//	processor numbers

///////////////////////////////////////////////////////////////////////
//	function prototype, should be able to call from Fortran, 
//	use pointer as parameter

//	get processor information
int GetMaxProc();

//	Fortran callable functions
//	fortran is case insensitive, and add two dash at the end of func name
extern "C"
{
	BOOL prof_begin__(int *pnProcID, int *pnProcNum);
	BOOL prof_end__();
	BOOL prof_pushstack__(int *pnStmtID);
	BOOL prof_popstack__();
	BOOL prof_newevent__(int *pnState, int *pnTarget, int *pnComSize);
	BOOL prof_endevent__(int *pnState);
	void prof_inccounter__(int *pnCount);

	BOOL prof_def_processors__(int *pnProcessorsID, int *pnGroupID);
	BOOL prof_def_range__(int *pnTemplateID, int *pnDim, int *pnRangeID);
	BOOL prof_def_array__(int *pnArray, int *pnDAD);
}

//	initialization and finalization
BOOL ProfBegin(int nProcID, int nProcNum);
BOOL ProfEnd();

//	Virtual Machine manipulation
BOOL ProfPushStack(int nStmtID);
BOOL ProfPopStack();

BOOL ProfBeginCom();
BOOL ProfEndCom();

//	Event Recording
BOOL ProfNewEvent(int nState, int nTarget, int nComSize);
BOOL ProfEndEvent(int nState);

//	Data Distribution
BOOL ProfDefProcessors(int nProcessorsID, int nGroupID);
BOOL ProfDefRange(int nTemplateID, int nDim, int nRangeID);
BOOL ProfDefArray(int nArray, int nDAD);

BOOL ProfNewGroupGrid(int nGroup, int nRank, int *aShape);
BOOL ProfNewRangeDistribute(int nRange, int nLB, int nUB, int nGrp, int nDist);
BOOL ProfNewRangeCollapse(int nRange, int nLB, int nUB);
BOOL ProfNewArrayData(int nDAD, int nRank);
BOOL ProfSetArrayAlign(int nDAD, int nDim, int nLB, int nUB);
BOOL ProfSetArrayCollapse(int nDAD, int nDim, int nRange, 
	int nLB, int nUB, int a, int b);
BOOL ProfSetArrayCopy(int nDAD, int nDim, int nRange);

//	Time query function
int GetTimeCPU();		//	return CPU time in usec
int GetTimeSystem();	//	return system time in usec

////////////////////////////////////////////////////////////////////////
//	inline implementation
extern struct timeval start_tv;
extern struct timeval tv;
extern struct timezone tz;
extern double dfStartClock;

/*
extern "C" double MPI_Wtime();

inline int GetTimeCPU() { return (int)(1000000*(MPI_Wtime()-dfStartClock)); };
*/

inline int GetTimeSystem() 
{ gettimeofday(&tv, &tz); 
	return (int)((tv.tv_sec-start_tv.tv_sec)*1000000+
		(tv.tv_usec-start_tv.tv_usec)); };

inline int GetMaxProc(){ return g_nProcNum; };

inline BOOL ProfBegin(int nProcID, int nProcNum)
	{ return prof_begin__(&nProcID, &nProcNum); };
inline BOOL ProfEnd() { return prof_end__(); };

inline BOOL ProfPushStack(int nStmtID) 
	{ return prof_pushstack__(&nStmtID); };
inline BOOL ProfPopStack()
	{ return prof_popstack__(); };

inline BOOL ProfNewEvent(int nState, int nTarget, int nComSize)
	{ return prof_newevent__(&nState, &nTarget, &nComSize); };
inline BOOL ProfEndEvent(int nState) { return prof_endevent__(&nState); };

inline BOOL ProfBeginCom()
	{ gr_nComSTimeCPU=GetTimeCPU(); gr_nComSTimeSystem=GetTimeSystem(); 
		return TRUE; };
inline BOOL ProfEndCom()
	{ gr_nComElapseCPU+=GetTimeCPU()-gr_nComSTimeCPU;
	  gr_nComElapseSystem+=GetTimeSystem()-gr_nComSTimeSystem;
	  return TRUE; };

inline BOOL ProfDefProcessors(int nProcessorsID, int nGroupID)
{	return prof_def_processors__(&nProcessorsID, &nGroupID);	};
inline BOOL ProfDefRange(int nTemplateID, int nDim, int nRangeID)
{	return prof_def_range__(&nTemplateID, &nDim, &nRangeID);	};
inline BOOL ProfDefArray(int nArray, int nDAD)
{	return prof_def_array__(&nArray, &nDAD);	};

#endif
