#define Forall_Ref   5
#define Assign_Ref  10
#define Io_Ref      15
#define IoScalar_Ref    16
#define Proccall_Ref    17
#define Funccall_Ref    18
#define Independent_Ref 19

#define Proc_Def      20
#define Align_Def     25
#define Temp_Def      30
#define Dist_Def      35
#define DDist_Def     40                  // Distribute without align
#define Dynamic_Def   45
#define Major_Def     46
#define Allocate_Def  48
#define Filemap_Def   49

#define Shad_Def      50

#define Inherit_Def   60
#define Outofcore_Def 65

#define Resident_Def  68

#define NeedDad_Def   70

#define Common_Def    72

#define INDEP_Def     75
#define INDEPHDER_Def 80
#define INDO_CALL_OPT 81

#define FINDEX_Def  85
#define FHDER_Def   90
#define IHDER_Def   95

#define DUMP_Def   100

#define DUMMY_MAPPING	108

#define DUMMY_ADDR 	110		//  dummy first address
#define DUMMY_NAME	120		// dummy name

#define SHIFTARRAY_Def      200
#define GHOST_Def           300
#define UNKNOWSHIFT_Def     350

// record local index , must separated at least by 10
#define INDEX_OPT	400
#define INDEX_ADD	450
#define INDEX_INIT      500
#define INDEX_SDX	550

// tag specify that this expression is already in NODE program mode
//#define IN_NODE  	600

#define IRREG_ARRAY	610

// tag to specify the return label
#define RETURN_LABEL    620

// tag to speicified array_all_collapsed can not be peroformed.
#define ALL_COLLAPSED     630

// specify the inconstant array expression
#define INCONST_ARRAY	   640
// when an array is unknown boundry, we have to allocate memory
#define UNBOUND_ARRAY      650

// for prefetch_read
#define DO_PREFETCH   660
#define PREFETCH_NUM  670

// for out of core
#define OC_ASSIGN  	700
#define OC_READ    	710
#define OC_READ_NUMBER  720

// intrinsic2sub will turn intrinsic functions to subroutine
#define INTRIN_SUB      750

// optimzie flag
#define SET_INCORE       900

#define NOP    	     0
#define BLOCK	    -1
#define CYCLIC	     1

enum Flag { off, on };

#define BEFORE      0
#define AFTER       1

#define COLUMN_MAJOR ((int)0)
#define ROW_MAJOR    ((int)1)

#define IO_FORTRAN	0
#define IO_C		1

// ----------------;

#define MAX_PROCESSOR_DIM 10

///
class Processor_def {
  ///
  int dim;
  ///
  SgSymbol * symb;
  ///
  SgSymbol * interface_name;
  ///
  SgExpression * s[MAX_PROCESSOR_DIM];
  ///
  SgExpression * e[MAX_PROCESSOR_DIM];
  ///
  SgSymbol * grp;
  ///
  SgSymbol * shp;
public:
  ///
  Processor_def(void);
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void set_interface_symbol(SgSymbol *);
  ///
  SgSymbol * get_interface_symbol(void);
  ///
  void add_dimention(SgExpression *, SgExpression *);
  ///
  int getDim(void);
  ///
  SgExpression * getStart(int);
  ///
  SgExpression * getEnd(int);
  ///
  void setgrp(SgSymbol *);
  ///
  SgSymbol * getgrp(void);
  ///
  void setshp(SgSymbol *);
  ///
  SgSymbol * getshp(void);
  ///
  void show(void);
};

// ---------------

#define MAX_TEMPLATE_DIM 10

///
class Template_def {
  ///
  int dim;
  ///
  SgSymbol * symb;
  ///
  SgType * type;
  ///
  SgExpression *s[MAX_TEMPLATE_DIM];
  ///
  SgExpression *e[MAX_TEMPLATE_DIM];
  ///
  SgExpression *localsize[MAX_TEMPLATE_DIM];
  ///
  int lput; //@ManMemo: largest possible used time;
  ///
  SgSymbol * rng;

public:
  ///
  Template_def(void);
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void settype(SgType *);
  ///
  SgType * gettype(void);
  ///
  void add_dimention(SgExpression *, SgExpression *);
  ///
  int getDim(void);
  ///
  SgExpression * getStart(int);
  ///
  SgExpression * getEnd(int);
  ///
  void setlocalsize(int, SgExpression *);
  ///
  SgExpression * getlocalsize(int);
  ///
  void setlput(int);
  ///
  int getlput();
  ///
  void setrng(SgSymbol *);
  ///
  SgSymbol * getrng(void);
  ///
  void show(void);
};

// ----------------

#define MAX_ALIGN_DIM 10

///
class Alignment_def {
  ///
  int dim;
  ///
  int align_with_dim;
  ///
  SgSymbol * fromsymb;
  ///
  SgSymbol * tosymb;
  ///
  SgExpression * stride[MAX_ALIGN_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM];
  ///
  int align_pair[MAX_ALIGN_DIM];
  ///
  SgSymbol * local;
  ///
  SgSymbol * dad;
  ///
  SgSymbol * tmb;
  ///
  SgSymbol * tmbdad;

public:
  ///
  Alignment_def(void);
  ///
  Alignment_def(int ,SgSymbol *, SgSymbol *);
  ///
  void setsymbol(SgSymbol *, SgSymbol *);
  ///
  SgSymbol * getfromsymbol(void);
  ///
  SgSymbol * gettosymbol(void);
  ///
  void add_dimention_info(SgExpression *, SgExpression *);
  ///
  int getDim(void);
  ///
  void setAlignWithDim(int);     // dim of align with dimension
  ///
  int getAlignWithDim(void);      
  ///
  SgExpression * getStride(int);
  ///
  SgExpression * getOffset(int);
  ///
  void setAlignPair(int);
  ///
  int getAlignPair(int);
  ///
  void setlocal(SgSymbol *);
  ///
  SgSymbol * getlocal(void);
  ///
  void setdad(SgSymbol *);
  ///
  SgSymbol * getdad(void);
  ///
  void settmb(SgSymbol *);
  ///
  SgSymbol * gettmb(void);
  ///
  void settmbdad(SgSymbol *);
  ///
  SgSymbol * gettmbdad(void);
  
  ///
  void show(void);
};

// ------------;

#define MAX_DISTRIBUTE_DIM 10

///
class Distribute_def {
  ///
  int dim;
  ///
  SgSymbol * fromsymb;
  ///
  SgArrayRefExp *Procs;
  ///
  SgSymbol * tosymb;
  ///
  int distblock[MAX_DISTRIBUTE_DIM];
  ///
  int descriptive;
  ///
  SgSymbol * grp;
public:
  ///
  Distribute_def(void);
  ///
  Distribute_def(int,SgSymbol *,SgSymbol * );
  ///
  void setProcs(SgExpression *);
  ///
  SgArrayRefExp * getProcs(void);
  ///
  void setsymbol(SgSymbol *, SgSymbol *);
  ///
  SgSymbol * getfromsymbol(void);
  ///
  SgSymbol * gettosymbol(void);
  ///
  int  getDim(void);
  ///
  void add_dimention_info(int);
  ///
  int distCode(int);
  ///
  SgExpression *getProcSize(int);
  ///
  inline void set_descriptive(int des) { descriptive = des; };
  ///
  int is_descriptive() { return descriptive; };
  ///
  void setgrp(SgSymbol *);
  ///
  SgSymbol * getgrp(void);
  ///
  void show(void);
};

// -------------------;

/// 

class Forall_index_def {
  ///
  SgSymbol * indexvar;
  ///
  SgSymbol * lb;
  ///
  SgSymbol * ub;
  ///
  SgSymbol * step;
public:
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void setlb(SgSymbol *);
  ///
  SgSymbol * getlb(void);
  ///
  void setub(SgSymbol *);
  ///
  SgSymbol * getub(void);
  ///
  void setstep(SgSymbol *);
  ///
  SgSymbol * getstep(void);
};

// ------------;

#define MAX_INDEX_DIM 10

///
class Forall_header_def {
  ///
  int number;
  
  struct {
    SgSymbol * indexvar;
    SgExpression * indextriple[3];
  } index[MAX_INDEX_DIM];

  ///
  SgExpression * condition;

public:
  /// The constructor. 
  Forall_header_def(void);

  ///
  int getindexnumber(void);

  ///
  void setindexsymb(SgSymbol *);
  ///
  SgSymbol * getindexsymb(int);
   
  ///
  void setindextriple(int, SgExpression *);
  ///
  SgExpression * getindextriple(int, int);   

  /// 
  void setcondition(SgExpression *);
  ///
  SgExpression * getcondition(void);

  ///
  int evaluatable_index(int);

   ///
   void show(void);
};

// -------------------------;

enum Position {head, left, right};
enum Communication {NONE, SHIFT, REMAP, BROADCAST_SCALAR, BROADCAST_DIM, IRREG, UNKNOW, BROADCAST};

class Shadow_def;

///
class forall_reference {
  ///
  Position side;

  ///
  SgExpression * expr;
  ///
  SgSymbol * arraysymb;
  ///
  int index_level[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * stride[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM][MAX_INDEX_DIM];

  ///
  Communication status;
  /// hold some special information on communication.
  /// 1: shift immediately.
  int	comm_tag;
  ///
  Shadow_def * sdef;
  ///
  int amount[MAX_ALIGN_DIM];
public:
  ///  the number of forall_reference that just same as me
  int sameAs;
  ///
  forall_reference (Position);
  ///
  int getposition(void);
  
  ///
  void setExpr(SgExpression *);
  ///
  SgExpression * getexpression(void);
  ///
  SgSymbol * getarraysymbol(void);
  
  ///
  void setIndex(int, int, int);
  ///
  int getIndex(int, int);

  ///
  void setstride(int, int, SgExpression *);
  ///
  SgExpression * getstride(int, int);

  ///
  void setoffset(int, int, SgExpression *);
  ///
  SgExpression * getoffset(int, int);

  ///
  int scalarDim(int);
  ///
  int tripletDim(int);
  
  ///
  int hasScalarDim(void);

  ///
  int isScalar();  // return 1 if the reference is scalar like a(1) , else return 0

  ///
  int indexNumberOnDim(int);	// return the index number of the dim
 
  ///
  int dimNumberOnIndex(int);	// return the dimension of the index

  ///
  int numberOfIndex();		// return the total number of index refered

  ///
  int numberOfDiffIndex();	// return the total number of different indexes refered

  ///
  int refered(int);		// return 1 , if index is used.
  ///
  int evaluatable(void);

  ///
  int checkRegular(void);

  ///
  void setCommunication(Communication);
  ///
  Communication getCommunication(void);

  ///
  void setIShift();
  ///
  int  isIShift();
  ///
  void setShadow(Shadow_def * );
  ///
  Shadow_def * getShadow();
  ///
  void setAmount(int, int);
  ///
  int getAmount(int);

  ///
  void show(void);
};

// -----------------------;

///
class assign_reference {
  ///
  Position side;
  ///
  SgExpression * expr;
  
  ///
  Communication status;

public:
  ///
  int sameAs;
  ///
  assign_reference (Position);
  ///
  Position position(void);

  ///
  void setExpression(SgExpression *);
  ///
  SgExpression * expression(void);

  ///
  void setCommunication(Communication);
  ///
  Communication getCommunication(void);

  ///
  void show(void);
};

// -----------------------;

#define MAX_INDEX_DIM 10

///
class Impdo_header_def {
  ///
  int number;
  
  struct { 
    SgSymbol * indexvar;    
    SgExpression * indextriple[3];
  } index[MAX_INDEX_DIM];

public:
  ///
  Impdo_header_def(void);

  ///
  int numberOfIndex(void);
   
  ///
  void setIndexSymbol(SgSymbol *);
  ///
  SgSymbol * indexSymbol(int);
   
  ///
  void setIndexTriplet(int, SgExpression *);
  ///
  SgExpression * getIndexTriplet(int, int);
};

// -----------------------;

///
class io_reference {
  ///
  SgExpression * expr;
  ///
  SgSymbol * symb;
  ///
  int index_level[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * stride[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression *func;
public:
  ///
  int sameAs;
  ///
  Impdo_header_def * ihdef;

public:

  ///
  io_reference (void);

  ///
  void setExpression(SgExpression *);
  ///
  SgExpression * expression(void);
  
  ///
  void setFunction(SgExpression *);
  ///
  SgExpression * getFunction(void);
  
  ///
  SgSymbol  *symbol();
  ///
  void setIndex(int, int, int);
  ///
  int getIndex(int, int);

  ///
  void setStride(int, int, SgExpression *);

  ///
  SgExpression * getStride(int, int);

  ///
  void setOffset(int, int, SgExpression *);
  ///
  SgExpression * getOffset(int, int);

  ///
  int hasIndexOnDim(int);
  ///
  int indexSequenceOnDim(int);

  ///
  int evaluatable(void);

  ///
  int checkRegular(void);

  ///
  int reusable(void);
  ///
  int useNo(void);
  ///
  void show(void);
};

#define MAX_PROC_DIM 10

class proc_reference {
  ///
  int  number;
  ///
  SgExpression *actuals[MAX_PROC_DIM];
  ///
  SgSymbol *dummys[MAX_PROC_DIM];
  ///
  SgFunctionSymb *header;
  /// real body header symbol
  SgFunctionSymb *real_header;

  public:
  ///
  proc_reference(void);
  ///
  int  get_number(void);
  ///
  void add_pair(SgExpression *, SgSymbol *);
  ///
  void setHeader(SgFunctionSymb *, SgFunctionSymb *);
  ///
  SgFunctionSymb *getHeader(void);
  ///
  SgFunctionSymb *getRealHeader(void);
  ///
  SgSymbol *get_dummy(int);
  ///
  SgExpression *get_actual(int);
};


// ---------------

#define MAX_SHADOW_DIM 10

enum Org {DEFINED, DETECTED, GNEEDED, LNEEDED, ACTUAL};

///
class Shadow_def {
  ///
  Org tag;
  ///
  int l[MAX_SHADOW_DIM];
  ///
  int u[MAX_SHADOW_DIM];
  ///
  SgStatement * scope;  


public:
  ///
  Shadow_def(Org);
  ///
  Org getTag(void);
  ///
  void setLower(int, int);  
  ///
  void setUpper(int, int);
  ///
  int getLower(int);
  ///
  int getUpper(int);
  ///
  void setScope(SgStatement *s);
  ///
  SgStatement * getScope(void);
  ///
  void show(void);
};

//----------------------------------
#define MAX_NEW_VAR     10
///
class Independent_def {
  ///
  int number;
  ///
  SgExpression *newvar[MAX_NEW_VAR];

public:
  ///
  Independent_def(void);
  ///
  int numberOfNewVar(void);
  ///
  void setNewVar(int, SgExpression *);
  ///
  SgExpression *getNewVar(int);
};

//-------------------------------

class Independent_header_def {
  ///
  int number;

  struct {
    SgSymbol * indexvar;
    SgExpression * indextriple[3];
  } index[MAX_INDEX_DIM];

public:
  /// The constructor.
  Independent_header_def(void);

  ///
  int getindexnumber(void);

  ///
  void setindexsymb(SgSymbol *);
  ///
  SgSymbol * getindexsymb(int);

  ///
  void setindextriple(int, SgExpression *);
  ///
  SgExpression * getindextriple(int, int);

};

//-------------------------------
#define MAX_ARRAY_DIM 10
enum Usedfor {as_lhs, as_rhs};

class independent_reference{    //one dimension, no communication
  ///                           //comm detection, so multi-array is needed
  Usedfor intent;	//used as lhs or rhs
  ///
  int left_count;	//if used as_rhs, it's the corresponding lhs cnt
  ///
  int right_count;
  ///
  int index_level[MAX_ARRAY_DIM][MAX_INDEX_DIM];       
	//now to show which dim is used
  ///
  SgExpression * expr;  //VAR_REF or ARRAY_REF
  ///
  SgStatement * used;
  ///
  SgExpression * stride[MAX_ARRAY_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * offset[MAX_ARRAY_DIM][MAX_INDEX_DIM];
        //stride and offset of every dimension of the array
  ///
  Communication status;
  ///
public:
  ///
  independent_reference(void);
  ///
  int sameAs;
  ///
  void setIntent(Usedfor);
  ///
  int getIntent(void);
  ///
  void setLeftCount(int);
  ///
  int getLeftCount(void);
  ///
  void setRightCount(int);
  ///
  int getRightCount(void);

  ///
  void setExpr(SgExpression *);
  ///
  SgExpression * getexpression(void);
  ///
  void setStmt(SgStatement *);
  ///
  SgStatement * getStmt(void);
  ///
  void setIndex(int, int, int);
  ///
  int getIndex(int, int);

  ///
  void setstride(int, int, SgExpression *);
  ///
  SgExpression * getstride(int, int);

  ///
  void setoffset(int, int, SgExpression *);
  ///
  SgExpression * getoffset(int, int);

  ///
  int hasScalarDim(void);

  ///
  int refered(int);        //retrun 1 if index used

  ///
  int isScalar(void);  // return 1 if the reference is scalar like a(1) , else return 0

  ///
  int indexNumberOnDim(int);    // return the index number of the dim

  ///
  int dimNumberOnIndex(int);    // return the dimension of the index

  ///
  void setCommunication(Communication);
  ///
  Communication getCommunication(void);

};


///
class expression_reference {
  ///
  SgExpression *exp;
  ///
  SgExpression *same_exp;

public:
  ///
  expression_reference(SgExpression *);
  ///
  expression_reference(SgExpression *,SgExpression *);
  ///
  void insert(SgStatement*,int);
  ///
  void set_same_expression(int,SgExpression*,SgExpression *);
  ///
  SgExpression *get_same_expression();
  ///
  SgExpression * get_expression();
};

// extern data
extern Flag in_main_program;
extern Flag in_procedure;
extern Flag in_interface;
extern Flag in_module;
extern Flag in_independent;

extern int opt_strategy;
extern int debug_trans0;
extern int silentFlag;
extern int runtimeDebug;
extern int prefetchFlag;

extern int storageMode;
extern int ioMode;

extern char * mixstring;

extern int stack_size, heap_size;
extern int IN_CORE_SIZE;
extern int RECORD_SIZE;

extern int have_main_program;

// record the level of forall .
extern char forall_level[2] ;
// record the level of expression .
extern char exp_level[1024] ;
// record the level of if_stmt/select-case
extern int if_level;

extern char * nameOfType(SgType *);
extern char * stackNameOfType(SgType *);
extern char * sizeNameOfType(SgType *);
extern char * allocNameOfType(SgType *);
extern char * allocHeapNameOfType(SgType *);
extern char * freeNameOfType(SgType *);

extern SgExpression *get_type_size(SgStatement *s, SgType *type);

// extern   defined in util.cc
void replace_intrinsic_function_name(SgExpression *proc_call,int flag);
extern void intrinsic2sub(SgStatement *s);

extern SgExpression *get_in_core_size(SgStatement *s);
extern SgExpression *get_record_size(SgStatement *s);

extern SgStatement *make_init(SgStatement *s, int store, int debugFlag);
extern SgStatement *make_new_array_section(SgStatement *, SgSymbol *,int, SgSymbol *);
extern SgStatement *make_set_array_triplet(SgStatement *,SgSymbol *,int ,SgExpression *,SgExpression *,SgExpression *,SgSymbol *,int );
extern SgStatement *make_set_array_scalar(SgStatement *,SgSymbol *,SgExpression *,SgSymbol *,int );
extern SgStatement *make_new_array_data(SgStatement *,SgSymbol *,SgType *,int store=-1,int rank=-1,SgExpression *grpExp=NULL);
extern SgStatement *make_new_array_data1(SgStatement *,SgSymbol *,SgExpression *,SgType *,int store=-1,int rank=-1,SgExpression *grpExp=NULL);
extern SgStatement *make_set_type(SgStatement *,SgSymbol *, SgType *);
extern SgStatement * make_new_array_copy(SgStatement *,SgSymbol *,SgSymbol *);
extern SgStatement * make_set_array_copy(SgStatement *, SgSymbol *, int , SgSymbol *, int );
extern SgStatement * make_set_array_collapse(SgStatement *, SgSymbol *, int , SgExpression *, SgExpression *);
extern SgStatement * make_set_array_distribute(SgStatement *, SgSymbol *, int , SgExpression *, SgExpression *, int , SgSymbol *, int );
extern SgStatement * make_set_array_align(SgStatement *, SgSymbol *, int , SgExpression *, SgExpression *, SgExpression * ,SgExpression *, SgExpression *, SgExpression *,SgExpression *);
extern SgStatement * make_remap(SgStatement *,SgSymbol *,SgSymbol *);
extern SgStatement * make_write_halo(SgStatement *, SgSymbol *,SgSymbol *, SgSymbol *);
extern SgStatement * make_write_back_halo(SgStatement *, SgSymbol *,SgSymbol *, SgSymbol *);
extern SgStatement * make_delete_array(SgStatement * , SgSymbol *);
extern SgStatement * make_array_free(SgStatement *, SgSymbol *);
extern SgExpression *make_new_group_grid(SgStatement *,int , SgSymbol *);
extern SgStatement * make_delete_group(SgStatement * , SgSymbol *);
extern SgExpression * make_new_range_distribute(SgStatement *,SgExpression *,SgExpression *,int ,SgSymbol *,int );
extern SgExpression *make_new_range_collapse(SgStatement *s,SgExpression *start,SgExpression *end);
extern SgStatement * make_delete_range(SgStatement * , SgExpression *);
extern SgStatement * make_alloc(SgStatement *,SgType *,SgSymbol *,SgSymbol *);
extern SgStatement * make_stack_alloc(SgStatement *,SgType *,SgSymbol *,SgSymbol *,SgSymbol *);
extern SgStatement * make_stack_alloc_size(SgStatement *,SgType *,SgSymbol *,SgExpression *,SgSymbol *);
extern SgStatement * make_heap_alloc(SgStatement *,SgType *,SgSymbol *,SgSymbol *,SgSymbol *);
extern SgStatement * make_heap_alloc_size(SgStatement *,SgType *,SgSymbol *,SgExpression *,SgSymbol *);
extern SgStatement * make_free(SgStatement *,SgType *,SgSymbol *);
extern SgStatement * make_reset_array_base(SgStatement *,SgSymbol *,SgType *, SgSymbol *, SgSymbol *);
extern SgStatement * make_reset_array_base1(SgStatement *,SgSymbol *,SgExpression *, SgSymbol *);
extern SgStatement * make_reset_in_core_base(SgStatement *,SgSymbol *,SgType *,SgSymbol *, SgExpression *mhExp);
extern SgExpression *make_glob(SgStatement *,SgSymbol *,SgExpression *);
extern SgExpression *make_group(SgStatement *s,SgSymbol *dad);
extern SgExpression *make_range(SgStatement *,SgSymbol *,SgExpression *);
extern SgExpression *make_stride(SgStatement *,SgSymbol *,SgExpression *);
extern SgExpression *make_loc(SgStatement *,SgExpression *,SgExpression *);
extern SgExpression *make_on(SgStatement *s, SgExpression *grpexp);
extern SgExpression *make_in_core_size(SgStatement *, SgSymbol *);
extern SgStatement * make_coef(SgStatement *,SgSymbol *,int ,SgExpression *,SgExpression *,SgExpression *,SgExpression *,SgExpression *,SgExpression * ,int ,int );
extern SgStatement * make_offset(SgStatement *,SgSymbol *,int ,SgExpression *, SgSymbol *);
extern SgStatement * make_same_data_mapping(SgStatement *,SgSymbol *,SgSymbol *,SgSymbol *);
extern SgStatement * make_index(SgStatement *s,SgSymbol *lhs,SgSymbol *dad, SgType *type);
extern SgExpression * make_distcode(SgStatement *,SgSymbol *,int );
extern SgStatement * make_clock(SgStatement *s, int flag);
extern SgStatement * make_set_global_major(SgStatement *s, int storage);
extern SgStatement * make_sub_group_grid(SgStatement *s, SgSymbol *subgrp, int rank, SgSymbol *l, SgSymbol *u, SgSymbol *ss, SgSymbol *grp);
extern SgStatement * make_set_io_process(SgStatement *s, SgExpression *grpExp);
// for OUT-OF-CORE
extern SgStatement * make_oc_init(SgStatement *s,SgSymbol *dad,int in_core_size);
extern SgStatement * make_oc_filemap(SgStatement *s,SgSymbol *dad, SgExpression *fileExp, int in_core_size);
extern SgStatement * make_out_of_core(SgStatement *s, SgSymbol *array, SgSymbol *dad, int in_core_size);
extern SgStatement * make_set_ic(SgStatement *s,SgSymbol *dad);
extern SgStatement * make_oc_read(SgStatement *s,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number);
extern SgStatement * make_oc_write(SgStatement *s,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number);
extern SgStatement * make_oc_prefetch_read(SgStatement *s,SgSymbol *handle,SgSymbol *dad,SgExpression * mem_position,SgExpression * start_position,SgSymbol *stride,SgSymbol *number);
extern SgStatement * make_prefetch_wait(SgStatement *s,SgSymbol *handle);

extern SgAssignStmt * make_stride_assign(SgStatement *s,SgSymbol * arraySym,char * stride ,int var_num,int i_dim);
extern SgAssignStmt * make_number_assign(SgStatement *s,SgSymbol * arraySym,SgExpression * numberExp,int i_dim );

extern void define_array_from_arrayref(SgStatement * s, SgArrayRefExp *arrayexp, SgSymbol *dadsymb);
extern void define_array_from_array(SgStatement * s, SgSymbol *arraysymb, SgSymbol *dadsymb);
extern Distribute_def * getDistributeDef(SgSymbol *arraysym);
extern int array_dim_collapsed(SgSymbol *,int );
extern int array_all_collapsed(SgSymbol *);
extern int array_collapsed(SgSymbol *);

extern int is_dummy(SgStatement *s, SgSymbol *symb);
extern int function_intrinsic(SgSymbol *func);
extern int is_out_of_core(SgSymbol *symb);
extern int is_dynamic(SgSymbol *symb);
extern int is_inherit(SgSymbol *symb);
extern int is_allocatable(SgSymbol *symb);
extern int is_pointer(SgSymbol *symb);
extern int is_resident(SgSymbol *symb);
extern int is_template(SgSymbol *symb);
extern int is_common(SgSymbol *symb);
extern int unknown_address(SgSymbol *symb);

extern int getArrayStorage(SgSymbol *array);
extern SgSymbol *getStorageSymb(SgStatement *s, int storage);
extern SgStatement *get_contains(SgStatement *s, int &no);
extern void change_array_by_stack(SgStatement *s, SgArrayRefExp *array);
extern void change_variable_by_stack(SgStatement *s, SgExpression *var);
extern int sub_processor(SgSymbol *sym1, SgSymbol *sym2);
extern int same_processor_group(SgSymbol *sym1, SgSymbol *sym2);

// frefop.cc
extern void inc_forall_level();
extern void dec_forall_level();

// optimize.cc
extern void optimize_glob_index(SgStatement *scope,Forall_header_def *fhdef);
extern int decide_optimize_major(SgStatement *s);
extern int column_major(SgStatement *s);
extern int row_major(SgStatement *s);

// in expr.cc
extern void inc_exp_level();
extern void dec_exp_level();
extern int expr_no_communication(SgStatement * s, SgSymbol *symb);

extern void collect_array_ref(SgStatement *,SgExpression *,int );
extern void analyse_array_ref_list(SgStatement *,SgStatement *);
extern void analyse_array_ref_in_exp(SgStatement *,SgExpression *,SgStatement *);

extern void collect_intrinsic(SgStatement *s,SgExpression *,int );
extern void analyse_intrinsic_list(SgStatement *,SgStatement *);
extern void analyse_intrinsic(SgStatement *,SgExpression *,SgStatement *);

extern void collect_scalar_pointer(SgStatement *s,SgExpression *exp,int level);
extern void analyse_pointer_list(SgStatement *s,SgStatement *where);
extern void analyse_pointer_in_exp(SgStatement *s,SgExpression *exp,SgStatement *where);

extern void collect_expression(SgStatement *s, SgExpression *exp, int level);
extern void analyse_expression_list(SgStatement *s, SgStatement *first, SgStatement *last);
extern void analyse_expression(SgStatement *, SgExpression *);

// indep.cc
void analyse_indepdo_header(SgStatement *s);
void get_indepdo_reference(SgStatement *s);
void index_indepdo_reference(int indexdim, SgSymbol * index, SgStatement * s);
void detect_indepdo_reference(SgStatement * s,int flag=1);
void display_indepdo_reference_communication(SgStatement *s);

//transindep.cc
void delete_indep_stmt(SgStatement * s);
void change_indepdo_body(SgStatement * s);
