// written by Zheng Gengbin
// wrap subroutines and functions by a new procedure which will do the 
// memory allocation for local array.

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
extern "C" void *malloc(...);
#endif

#include <stdio.h>
#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

#include "detect_comm.h"

/**************************************************************************
	PORT TO F2C Compiler
	some f2c compiler cannot accept in-constant non-dummy array 
	transform the subroutine and wrap it up by another subroutine 
    which will dynamic alloc array.
**************************************************************************/

SgStatement * wrap_subroutine(SgStatement *s)
{
    ASSERT(isSgFuncHedrStmt(s) || isSgProcHedrStmt(s) || isSgProgHedrStmt(s));
    SgProcHedrStmt *proc = (SgProcHedrStmt *)s;
    SgFunctionSymb *symb = (SgFunctionSymb *)s->symbol();
    SgFunctionSymb *wrapfuncSymb;
    SgFunctionSymb *oldfuncSymb;
    SgProcHedrStmt *newproc;
    SgCallStmt *call;
    int i;

if (!silentFlag)
printf("wrap_subroutine for '%s'.\n",symb->identifier());

      // create a new subroutine to wrap it up
    wrapfuncSymb = (SgFunctionSymb*)new SgSymbol(symb->variant(), symb->identifier());
    newproc = new SgProcHedrStmt(*wrapfuncSymb);
    SgStatement *global = s->controlParent();
    s->insertStmtBefore(*newproc, *global);
    wrapfuncSymb->setScope(*newproc);
    newproc->setName(*wrapfuncSymb);
    newproc->setVariant(proc->variant());
    for (i=0; i<symb->numberOfParameters(); i++)
    {
	SgSymbol *newarg = newproc->getSymbolInCurrentScope(symb->parameter(i)->identifier());
	newarg->setType(symb->parameter(i)->type());
	wrapfuncSymb->addParameter(i,*newarg);
    }

      // reset old function's name
    char *name = (char *)malloc(100);
    strcpy(name,"wrap_");
    strcat(name, s->symbol()->identifier());
    oldfuncSymb = symb;
    oldfuncSymb->thesymb->ident = name;
      // change program to subroutine
    if (isSgProgHedrStmt(s)) {
	s->setVariant(PROC_HEDR);
        oldfuncSymb->setVariant(PROCEDURE_NAME);
    }

      // have to change all the call symbol in the entire scope, may be some potential bug here?
    {
    SgStatement *global = proc;
    while (global->controlParent()) global = global->controlParent();
    for (SgStatement *stmt=global; stmt; stmt=stmt->lexNext())
	if (isSgCallStmt(stmt)) {
		SgCallStmt *call = isSgCallStmt(stmt);
		if (call->name() == oldfuncSymb)
			call->setName(*wrapfuncSymb);
	}
    }

	// create call statement
    call = new SgCallStmt(*s->symbol());
    for (i=0; i<wrapfuncSymb->numberOfParameters(); i++)
    {
  	SgExpression *arg = new SgVarRefExp(wrapfuncSymb->parameter(i));
  	call->addArg(*arg);
    }
    newproc->lastExecutable()->insertStmtAfter(*call, *newproc);

	// loop to deal with every array
    for (i=0; i<s->numberOfAttributes(INCONST_ARRAY); i++)
    {
    SgArrayRefExp *array = (SgArrayRefExp *)s->attributeValue(i,INCONST_ARRAY);

    // add argument to subroutine header.
    oldfuncSymb->addParameter(wrapfuncSymb->numberOfParameters(),*array->symbol());

      // malloc size and add arg to call
    SgSymbol * base = newproc->getSymbolInCurrentScope(mixstring, "tbs", array->symbol()->identifier());
    SgExpression *baseExp = new SgVarRefExp(base);
    SgSymbol * mHandle = newproc->getSymbolInCurrentScope(mixstring, "mh", array->symbol()->identifier());

      // caculate the alloc size of array
    SgSubscriptExp *ddot = (SgSubscriptExp *) array->lhs()->lhs();
    SgExpression *size = &( *ddot->ubound()->copyPtr() - *ddot->lbound()->copyPtr() + *new SgValueExp(1) );
    analyse_expression(call, size);
    size = calculateExp(size);

    SgStatement * alloc = make_stack_alloc_size(newproc,array->type(),base,
			          		size,mHandle);
    call->insertStmtBefore(*alloc, *call->controlParent());
      // pcrc_free_stack
#ifdef MEM_ALLOC
    base = mHandle;
#endif
    SgStatement * free = make_free(newproc,array->type(),base);
    call->insertStmtAfter(*free, *call->controlParent());

      // add argument of stack reference.
    SgSymbol * stack = newproc->getSymbolInCurrentScope(stackNameOfType(array->type()),"mb",NULL);
    SgArrayRefExp *stackExp = new SgArrayRefExp(*stack);
    stackExp->addSubscript(*baseExp);
    call->addArg(*stackExp);
    }

    return newproc;
}



