// rewritten by Zheng Gengbin
// translate forall statement to node program:
//    define_bounds(s);
//    check_ghost(s);
//    change_body(s,1);
//    add_ghost(s);
//    add_do_loop(s);

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"

#include "detect_comm.h"

extern int  comm_feq[15];

extern void add_coef(SgStatement * s, int counter, int * amount);
extern void add_coef_for_broadcast_dim(SgStatement * s, int counter, int * amount);
extern void create_section(SgStatement * s, int counter );
extern void append_index(SgStatement *s, int counter,int flag=0);
extern void create_lhs_section(SgStatement *s,SgArrayRefExp *array);

extern void change_scalar_reference(SgStatement * s, forall_reference * fref);
extern void change_array_reference(SgStatement * s, int counter, int nameflag, int indexflag,SgSymbol *stack=NULL,SgSymbol *stackBase=NULL);

extern void check_ghost(SgStatement * s);
extern void add_ghost(SgStatement * s);

extern void do_tshift(SgStatement * s, int counter, int * amount);
extern void do_remap(SgStatement * s, int counter);
extern void do_i_write_halo(SgStatement * s, int counter);
extern void do_gather(SgStatement * s, int counter);
extern void do_scatter(SgStatement * s, int counter);
extern void do_detect(SgStatement * s, int counter);

extern void do_broadcast_scalar(SgStatement * s, int counter);
extern void do_broadcast_dim(SgStatement * s, int counter);

extern void show_forall_reference(SgStatement * s);

extern int compare_data_mapping_pattern(forall_reference *, forall_reference *,Forall_header_def *);

// task.cc
extern SgStatement * wrap_by_on_clause(SgStatement *, SgStatement *);

//add by slx
//foralloc.cc
extern void trans_out_of_core(SgStatement *, SgStatement *);

void sum_communication(enum Communication comm)
{
	comm_feq[comm - NONE] ++;
}

void define_bounds(SgStatement * s)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);
  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *) body->attributeValue(0, Forall_Ref);
	  
  SgArrayRefExp * array = (SgArrayRefExp *) fref->getexpression();
  
  for (int i=0; i<fhdef->getindexnumber(); i++) {
    {
	// check if this index has been used by the lhs of the assign stmt, 
	// if not, then we will remain this loop and donot create the Loop Range.

      int lhsdim;	// use by the pcrc_range()
      for (lhsdim=0; lhsdim<array->numberOfSubscripts(); lhsdim++)
	   if (fref->getIndex(lhsdim, i)!=0)  break;
      if (lhsdim == array->numberOfSubscripts()) continue;

      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_loop", NULL);
      SgFunctionRefExp * functionexpr = new SgFunctionRefExp (*fname);

      {
	for (int j=0; j<3; j++) {
	  if (fhdef->getindextriple(i, j)!=NULL) {
	    SgExpression * strideExp = fhdef->getindextriple(i, j);
	    functionexpr->addArg(*strideExp);
	  }
	  else {
	    SgValueExp * strideexp = new SgValueExp(1);
	    functionexpr->addArg(*strideexp);
	    delete strideexp;
 	  }
	}

	{
	  int dim;

	  for (dim=0; dim<array->numberOfSubscripts(); dim++)
	    if (fref->getIndex(dim, i)!=0) {
	      break;
	    };

	  SgExpression * stride;

	  if (fref->getstride(dim, i)!=NULL) { 
	    if (fref->getstride(dim,i)->isInteger())
	      stride = new SgValueExp(fref->getstride(dim,i)->valueInteger());
	    else
	      stride = fref->getstride(dim,i);
	  }
	  else 
	    stride = new SgValueExp(0);
         analyse_expression(s, stride);
	  functionexpr->addArg(*stride);
	    
	  SgExpression * offset;
	  if (fref->getoffset(dim,i)!=NULL) {
	    if (fref->getoffset(dim,i)->isInteger())
	      offset = new SgValueExp(fref->getoffset(dim,i)->valueInteger());
	    else
	      offset = fref->getoffset(dim,i);
	  }
	  else 
	    offset = new SgValueExp(0);
         analyse_expression(s, offset);
	 functionexpr->addArg(*offset);

	};

	SgSymbol * distsym = s->childList1(0)->expr(0)->symbol();
	SgSymbol * distdad = s->getSymbolInCurrentScope(mixstring,"dad",distsym->identifier());
	SgValueExp * dimention = new SgValueExp (lhsdim+1);
	SgExpression *funcexpr = make_range(s,distdad,dimention);

	functionexpr->addArg(*funcexpr);
      }

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgVarRefExp (irng);

      SgAssignStmt * assign = new SgAssignStmt (*rngexp, *functionexpr);

      s->insertStmtBefore(*assign, *s->controlParent());

      
      SgSymbol * funname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*funname);
     
      afuncall->addArg(* rngexp);
      s->insertStmtAfter(*afuncall, *s->controlParent());
      
      delete functionexpr, rngexp, assign , afuncall;
    }

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "loop_bounds", NULL);

      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgVarRefExp (irng);

      afuncall->addArg(*rngexp);

      SgSymbol * index=fhdef->getindexsymb(i);
      Forall_index_def * fidef = (Forall_index_def *) index->getAttribute(0, FINDEX_Def)->getAttributeData();

      SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
      SgExpression * lbexpr = new SgVarRefExp(lb);
      afuncall->addArg(*lbexpr);
      SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
      SgExpression * ubexpr = new SgVarRefExp(ub);
      afuncall->addArg(*ubexpr);
      SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
      SgExpression * stepexpr = new SgVarRefExp(step);
      afuncall->addArg(*stepexpr);
    
      s->insertStmtBefore(*afuncall, *s->controlParent());

      delete afuncall,rngexp,lbexpr,ubexpr,stepexpr;
    };
  };
}

void add_do_loop(SgStatement * s)
{

  SgStatement * inner;
  SgStatement * assign_stmt;
  SgStatement * old_assign_stmt=s->childList1(0);

  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  if (fhdef->getcondition()==NULL) {
//    inner=s->childList1(0);
//    inner = s->childList1(0)->extractStmt();
    inner=s->childList1(0)->copyPtr();
    assign_stmt = inner;
  }
  else {
    //inner = new SgIfStmt(*fhdef->getcondition(), *s->childList1(0));
    assign_stmt=s->childList1(0)->copyPtr();
//    assign_stmt = s->childList1(0)->extractStmt();
    inner = new SgIfStmt(*fhdef->getcondition()->copyPtr(), *assign_stmt);
  };

  SgStatement * newbody=inner;

  int i;

  // for column major.

  forall_reference * lhsref = (forall_reference *) old_assign_stmt->attributeValue(0, Forall_Ref);

  for (i=0; i<fhdef->getindexnumber(); i++) {
     SgSymbol * index=fhdef->getindexsymb(i);

     SgExpression *lbexpr,*ubexpr,*stepexpr;
	// if this index is not used , then do not create it and use the old bound
     if (lhsref->refered(i) == 0)
     {
	lbexpr=fhdef->getindextriple(i,0);
	ubexpr=fhdef->getindextriple(i,1);
	stepexpr=fhdef->getindextriple(i,2);
	if (stepexpr==NULL) stepexpr=new SgValueExp(1);
     }
     else
     {
     Forall_index_def *fidef = (Forall_index_def *) index->attributeValue(0, FINDEX_Def);

     lbexpr = new SgValueExp(0);

     SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
     ubexpr = new SgVarRefExp(ub);
     {
       SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
       SgExpression * lbexpr = new SgVarRefExp(lb);
       (*ubexpr) =(* ubexpr) - (* lbexpr);
       
       SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
       SgExpression * stepexpr = new SgVarRefExp(step);
       (*ubexpr) =(* ubexpr) / (* stepexpr);
     }

     stepexpr = new SgValueExp(1);
     }
   
     SgStatement * loop = new SgForStmt(*index, *lbexpr, *ubexpr, *stepexpr, *inner);

     inner = loop;

     delete lbexpr,ubexpr,stepexpr;
  };

  // add those index definition statement;

//if ( opt_strategy & INDEX_OPTIMIZE )
//{
  for (i=0 ; i< s->childList1(0)->numberOfAttributes(INDEX_INIT) ; i++ )
  {
 	SgStatement *init_stmt = (SgStatement *)s->childList1(0)->attributeValue(i,INDEX_INIT);
  	s->insertStmtBefore(*init_stmt, *s->controlParent());
  }

	// zheng.gb  1997/6/12
	// add temporary index 
  int index_number=fhdef->getindexnumber();

//  if ( index_number > 1 )	zheng.gb 1998/2/1
  int level;

  for (level =0; level<=6 ; level++)
  for (i=0 ; i< s->childList1(0)->numberOfAttributes(INDEX_OPT+level) ; i++ )
  {
	SgStatement *push = (SgStatement *)s->childList1(0)->attributeValue(i,INDEX_OPT+level);
    	SgStatement * where = newbody;
    	for ( int j = 0; j < level ; j++ )
		where=where->controlParent();
    	where->insertStmtBefore(*push, *where->controlParent());
  }

	// add the increase of the index variable
  for (level =0; level<=6 ; level++)
  for (i=0 ; i< s->childList1(0)->numberOfAttributes(INDEX_ADD+level) ; i++ )
  {
	SgStatement *add_stmt = (SgStatement *)s->childList1(0)->attributeValue(i,INDEX_ADD+level);
    	SgStatement * where = newbody;
    	for ( int j = 0; j < level ; j++ )
		where=where->controlParent();
    	where->insertStmtAfter(*add_stmt, *where->controlParent());
  }
//}
//else // no optimize

// add these statement anyway
  for (i=1; i<s->numberOfChildrenList1();i++) {
    if (s->childList1(i)->variant() == CONTROL_END ) continue;
    SgStatement * index = &s->childList1(i)->copy();
    newbody->insertStmtBefore(*index, *newbody->controlParent());
  };


  // add procossor position detection;

  SgSymbol * lhssec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL, 0);
  SgExpression * grpexp = make_group(s,lhssec);

  SgExpression * onexp = make_on(s, grpexp);

  inner = new SgIfStmt(*onexp, *inner);

    // to see if we wrap the statement by on_block
  inner = wrap_by_on_clause(s, inner);

  s->insertStmtBefore(*inner, *s->controlParent());

	// add by zheng.gb
  if ( opt_strategy & LOCAL2GLOBAL )
  {
//  	optimize_index(assign_stmt,fhdef);
  	optimize_glob_index(newbody->controlParent(),fhdef);
  }

  // in transform_expression_in_reference , we already deal with the 
  // index-invariant intrinsic function , now we have to deal with the 
  // variant case, for example sum(i:10).
  analyse_intrinsic(assign_stmt,assign_stmt->expr(1),assign_stmt);

  //add by slx  98/3/1
  trans_out_of_core(s,inner);

  delete grpexp, onexp, inner;
}

//new function after analyse;
// flag==1, oprate left hand side
void change_body(SgStatement * s, int flag)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgStatement * body = s->childList1(0);
  forall_reference * leftref = (forall_reference *)body->attributeValue(0, Forall_Ref);
  
  int counter=0;

  int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};

  int irr = 0;
  {
//    if (lhsref->getCommunication() ==  IRREG) 
    if (leftref->getexpression()->numberOfAttributes(IRREG_ARRAY)>0)
    {
        if (debug_trans0)
        sum_communication(IRREG);

        irr = 1;
	if (!silentFlag)
	{
        printf("%s -- %s: ",leftref->getexpression()->symbol()->identifier(),leftref->getexpression()->symbol()->identifier());
	printf("irregular:scatter\n");
	for(int j=0; j<MAX_ALIGN_DIM;j++)
	      printf("%3d", leftref->getAmount(j));
	printf("\n");
	}

  	append_index(s, counter);
	do_scatter(s, 0);
  	add_coef(s, counter, amount);
	goto rhs_label;
      }
  }

  if (flag == 1)
  {
  create_section(s, counter);

  add_coef(s, counter, amount);
  }

  append_index(s, counter);

rhs_label:
  for (int i=1; i<body->numberOfAttributes(Forall_Ref); i++) {
    forall_reference * rightref = (forall_reference *) body->attributeValue(i, Forall_Ref);
    
    //    rightref->show();
    SgExpression *expr = rightref->getexpression();

    // already changed to node code.
    if (expr->haveAttribute(NODE_CODE_BIT)) continue;
    expr->setAttribute(NODE_CODE_BIT);

    switch (rightref->getexpression()->variant()) {
    case VAR_REF: {

      // not there in the old version:
      counter++;

      change_scalar_reference(s, rightref);
      break;
    }
    case ARRAY_REF: {

      counter++;

      SgArrayRefExp * rhsExp = (SgArrayRefExp *) rightref->getexpression ();
      SgSymbol * rhsSym=rhsExp->symbol();

      if (!silentFlag)
      printf("%s -- %s: ",leftref->getexpression()->symbol()->identifier(),rightref->getexpression()->symbol()->identifier());
      
      if (debug_trans0)
      sum_communication(rightref->getCommunication());

      switch (rightref->getCommunication()) {
      case IRREG: {
        if (!silentFlag)
	{
        printf("irregular:gather\n");
	for(int j=0; j<MAX_ALIGN_DIM;j++)
	printf("%3d", rightref->getAmount(j));
	printf("\n");
	}
	do_gather(s, counter);
	change_array_reference(s, counter, 1,1);
	break;
      }

      case NONE: {
        if (!silentFlag)
	{
	  printf("none\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}

	  // we can do better;
	int i;
	int same_mapping=1;
	// if every dimension is not collapsed , no need to deal specially
	for (i=0; i<rhsExp->numberOfSubscripts(); i++)
		if (array_dim_collapsed(rhsSym,i) != 0)  // 1 or -1 , collapsed and not sure all make things complex.
		     { same_mapping=0; break; }
	if (same_mapping)  ;
	else if (leftref->getexpression()->symbol() == rhsSym)
		same_mapping=1;
	else if (compare_data_mapping_pattern(leftref,rightref,fhdef))
		same_mapping=1;

	add_coef(s, counter, amount);
	if (same_mapping == 1) {
	  // we can do this only if src and dst have same data mapping
	  append_index(s, counter);
	}
	else {
	  append_index(s, counter,1);	            // flag == 1
//	  change_array_reference(s, counter, 0,3);  // deal with collapsed dimnesion.
	} 
	change_array_reference(s, counter, 0,0);

	break;
      }

      case SHIFT: { 	    

        if (!silentFlag)
	{
	  printf("shift\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	};

/*
	int gOverflow=0;
	int shiftDim=0;

	Shadow_def * actShadowDef;

	for (int numS=0; numS<rhsSym->numberOfAttributes(Shad_Def); numS++) {
	  actShadowDef = (Shadow_def *)rhsSym->attributeValue(numS, Shad_Def);
	  if (actShadowDef->getTag() == GNEEDED)
	    break;
	}
	ASSERT(actShadowDef->getTag() == GNEEDED);

	for (int j=0; j<rhsExp->numberOfSubscripts();j++) {
	  if (rightref->getAmount(j) != 0 ) {
	    shiftDim++;
	    if (rightref->getAmount(j) < 0 ) {
	      if (abs(rightref->getAmount(j)) > actShadowDef->getLower(j)) {
		gOverflow=1;
		break;
	      }
	    }
	    else {
	      if (abs(rightref->getAmount(j)) > actShadowDef->getUpper(j)) {
		gOverflow=1;
		break;
	      }
	    }
	  }
	}
	
	if (gOverflow==1) {
	  if (shiftDim>1) {
	    do_remap(s, counter);
	    change_array_reference(s, counter, 1,1);
	  } 
	  else { // seems should be get ride of.
	    int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};
	  
	    for (int j=0; j<MAX_ALIGN_DIM;j++)
	      amount[j]=rightref->getAmount(j);
	  
	    do_tshift(s, counter, amount);
	    add_coef(s, counter, amount);
	    append_index(s, counter);
	    change_array_reference(s, counter, 1,0);
	  }
	}
	else {
	  int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};
	  for (int j=0; j<MAX_ALIGN_DIM;j++)
	    amount[j]=rightref->getAmount(j);
	  
	  add_coef(s, counter, amount);
	  append_index(s, counter);
	  change_array_reference(s, counter, 0,0);
	}
*/
	int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};
	  
	for (int j=0; j<MAX_ALIGN_DIM;j++)
	  amount[j]=rightref->getAmount(j);
	  
	add_coef(s, counter, amount);
	  
	append_index(s, counter);
	  
	if (rightref->isIShift())
		do_i_write_halo(s,counter);

	change_array_reference(s, counter, 0,0);

	break;
      }
      
      case REMAP: {
	if (!silentFlag)
	{
	  printf("remap\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
      
	if (rightref->sameAs == -1)
		do_remap(s, counter);
	change_array_reference(s, counter, 1,1);
	break;
      }
	    
      case UNKNOW: {
	
	if (!silentFlag)
	{
	  printf("unknow\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	}
	
	if ( opt_strategy & RUNTIMECOMM )
	  do_detect(s, counter);	 // add runtime communication detection.
	else 
  	{	// deal like remap
	  do_remap(s, counter);
	  change_array_reference(s, counter, 1,1);
	}
	break;
      }

// add by zheng.gb
      case BROADCAST_SCALAR:  {
	  if (!silentFlag)
	  {
	  printf("broadcast to scalar\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	  }

	do_broadcast_scalar(s, counter);
	change_array_reference(s, counter, 1,2);
	break;
      }

// add by Zheng Gengbin  1997/12/11
      case BROADCAST_DIM:  {
	  if (!silentFlag)
	  {
	  printf("broadcast to dimension\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	  }

	if (rightref->sameAs == -1)
	{
	do_broadcast_dim(s, counter);
	add_coef_for_broadcast_dim(s, counter, amount);
	append_index(s, counter);
	}
	change_array_reference(s, counter, 1,0);
	break;
      }

      case BROADCAST: {
	{
	  if (!silentFlag)
	  {
	  printf("broadcast\n");
	  for(int j=0; j<MAX_ALIGN_DIM;j++)
	    printf("%3d", rightref->getAmount(j));
	  printf("\n");
	  }
	}
	printf("not supported\n");
	break;
      }

      default:
	printf("Return valuse of detect_communication is invalid!\n");
	break;
      }
      
      break;
    }

    default: {
      printf("In change_body, unexpected expression on line %d, id: %d, variant: %d\n", s->lineNumber(), rightref->getexpression()->id(), rightref->getexpression()->variant());
      break;
    }
    }
  }

  if (irr == 1) 
	change_array_reference(s, 0, 1,1);
  else
  	change_array_reference(s, 0, 0,0);

}

// transform expression in index
void collect_index(SgStatement *s,int level)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
  for (int i=0; i<fhdef->getindexnumber(); i++)
  {
	for (int j=0; j<3 ; j++)
	{
		SgExpression *exp = fhdef->getindextriple(i,j);
		if (exp) 
		{
		  collect_array_ref(s,exp,level++);
		  collect_intrinsic(s,exp,level++);
//		  collect_scalar_pointer(s,exp,level++);
		}
	}
  }
}

void collect_expression_in_reference(SgStatement *s, SgExpression * expr)
{
  Forall_header_def *fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  int  i;
  if (expr) {
    switch(expr->variant()) {

    case ARRAY_REF: {
	// we donot touch the array ref itself, but we process the subscripts 
	// of this array reference.
	SgArrayRefExp *array = isSgArrayRefExp(expr);
	// check every subscript
	for (int k=0; k<array->numberOfSubscripts(); k++) 
	{
	  SgArrayRefExp *dimExp = isSgArrayRefExp(array->subscript(k));
	  if (dimExp == NULL) continue;		// not an array ref
	  // this subscripts must not contain forall index
  	  for (i=0; i<fhdef->getindexnumber(); i++) {
		SgSymbol *index=fhdef->getindexsymb(i);
		if (dimExp->IsSymbolInExpression(*index)) break;
          }
	  if ( i==fhdef->getindexnumber() )
 	  {
		collect_array_ref(s,dimExp,1);
		collect_intrinsic(s,dimExp,1);
//		collect_scalar_pointer(s,dimExp,1);
	  }
	}
        break;
	}

    case VAR_REF: {
//	collect_scalar_pointer(s,expr,1);
	break;
	}

    case  FUNC_CALL:
    case  INTRIN_FUNC_CALL: {
  	for (i=0; i<fhdef->getindexnumber(); i++) 
	{
		SgSymbol *index=fhdef->getindexsymb(i);
		if (expr->IsSymbolInExpression(*index)) break;
	}
	if ( i==fhdef->getindexnumber()  && ! isSgArrayType(expr->type()) )
	{
		collect_array_ref(s,expr,1);
		collect_intrinsic(s,expr,1);
//		collect_scalar_pointer(s,expr,1);
	}
        break;
	}

    default:
	collect_expression_in_reference(s, expr->operand(1));
	collect_expression_in_reference(s, expr->operand(2));
	
      break;
    };
  };
}

void transform_invariant(SgStatement *s)
{
	collect_index(s,0);
	collect_expression_in_reference(s,s->childList1(0)->expr(0));
	collect_expression_in_reference(s,s->childList1(0)->expr(1));
	analyse_array_ref_list(s,s);
	analyse_intrinsic_list(s,s);
//	analyse_pointer_list(s,s);
}

void trans_forall(SgStatement * s)
{
  extern Flag profile_flag;
  extern int forall_counter ;

  forall_counter ++;

  if ( profile_flag == on) 
  {
   {
    SgSymbol * fname = s->getSymbolInCurrentScope("prof", "pushstack", NULL);
    SgCallStmt * afuncall = new SgCallStmt(*fname);
    SgExpression *val = new SgValueExp(forall_counter);
    afuncall->addArg(*val);
    s->insertStmtBefore(*afuncall, * s->controlParent());
   }
   {
    SgSymbol * fname = s->getSymbolInCurrentScope("prof", "popstack", NULL);
    SgCallStmt * afuncall = new SgCallStmt(*fname);
    s->insertStmtAfter(*afuncall, * s->controlParent());
   }
  }

  SgStatement * body=s->childList1(0);
  forall_reference *leftref = (forall_reference *) body->attributeValue(0, Forall_Ref);
  
  SgExpression * refexp=leftref->getexpression();

  if (refexp->variant()==VAR_REF) {
    printf("Assignment to scalar in loop!\n");
  }
  else if (refexp->variant()==ARRAY_REF) {

      // transform the index-invariant expression in forall index expression and Rhs expression
    transform_invariant(s);

      // deal irregular lhs array
    if (refexp->numberOfAttributes(IRREG_ARRAY)>0)
    {
	// if the lhs is irregular reference , create the temp DAD for it.
	create_lhs_section(s,(SgArrayRefExp*)refexp);
    }
    define_bounds(s);
    check_ghost(s);
    change_body(s,1);
    add_ghost(s);
    add_do_loop(s);	

  };

}

