#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
extern "C" int abs(...);
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"
#include "detect_comm.h"
#include "indep.h"

extern void sum_communication(enum Communication comm);
extern int array_dim_collapsed(SgSymbol *arraysym,int dim);

//foralloc.cc
extern SgStatement * loop_tile(SgStatement *s,int loop_num=0);

//indepcom.cc
extern void create_remap_target (SgStatement *s,int counter);
extern void add_indepdo_remap(SgStatement *s,int counter);
extern void check_indepdo_ghost(SgStatement *s);
extern void do_write_halo(SgStatement * s);
extern void analyse_intrinsic(SgStatement *s,SgExpression *exp,SgStatement *where);

//indepcall.cc
extern void dispose_indepdo_array_argument1(SgStatement *s,int counter);
extern int find_and_replace(SgExpression *exp,SgVarRefExp *old,SgExpression * new_exp=NULL);
SgStatement * forall2doloop(SgStatement *s);

void change_indepdo_scalar(SgStatement *s,SgExpression *scalar);
void change_indepdo_var_exp(SgStatement *s,SgExpression *exp);
void calcExp(SgExpression *exp);
extern SgExpression *ddot_sub_triplet(SgArrayRefExp *array,int i,int flag);

void delete_indep_stmt(SgStatement * s){
	SgStatement * s1=s;
	SgStatement * s2;
        while(s1->childList1(0)->variant()==INDEPENDENT_STMT&&s1->childList1(2)->variant()==ENDDO_STMT){
	s2=s1->childList1(1);
	s1->childList1(0)->deleteStmt();
	s1=s2;
	}
}

void change_indepdoref_in_doloop(SgArrayRefExp *array,stack &exp_stk){
	int i;
	for(i=0;i<array->numberOfSubscripts();i++){
		SgExpression *sub=array->subscript(i);
		if(sub->numberOfAttributes(SUB_RANGE)>0){
			exp_stk.push(sub->copyPtr());
			SgExpression *sub_range=(SgExpression *)sub->attributeValue(0,SUB_RANGE);
			sub->replace(*sub_range->lhs()->lhs()); 	
		}
	}
}

void resume_indepdoref_in_doloop(SgArrayRefExp *array,stack &exp_stk){
        int i;
        for(i=array->numberOfSubscripts()-1;i>-1;i--){
                SgExpression *sub=array->subscript(i);
                if(sub->numberOfAttributes(SUB_RANGE)>0){
                        sub->replace(*exp_stk.top_exp());
                }
        }
}

void define_indepdo_bounds(SgStatement * s)
{
  Indepdo_header_def * ihdef = (Indepdo_header_def *) s->attributeValue(0, IHEDR_Def);
  Indepdo_reference * iref = (Indepdo_reference *) s->attributeValue(0, Indepdo_Lref);

  SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();

  for (int i=0; i<ihdef->getindexnumber(); i++) {
    {
      int lhsdim;       // use by the pcrc_range()
      for (lhsdim=0; lhsdim<array->numberOfSubscripts(); lhsdim++)
           if (iref->getIndex(lhsdim, i)!=0)  break;
      if (lhsdim == array->numberOfSubscripts()) continue;
	Indepdo_index_def * iidef=(Indepdo_index_def *)s->attributeValue(i,IINDEX_Def);
	iidef->loop_cut=1;
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "new_range_loop", NULL);
      SgFunctionRefExp * functionexpr = new SgFunctionRefExp (*fname);
      {
        for (int j=0; j<3; j++) {
          if (ihdef->getindextriple(i, j)!=NULL) {
            SgExpression * strideExp = ihdef->getindextriple(i, j)->copyPtr();
            functionexpr->addArg(*strideExp);
          }
          else {
            SgValueExp * strideexp = new SgValueExp(1);
            functionexpr->addArg(*strideexp);
            delete strideexp;
          }
        }

        {
          int dim;

          for (dim=0; dim<array->numberOfSubscripts(); dim++)
            if (iref->getIndex(dim, i)!=0) {
              break;
            };

          SgExpression * stride;

          if (iref->getstride(dim, i)!=NULL) {
            if (iref->getstride(dim,i)->isInteger())
              stride = new SgValueExp(iref->getstride(dim,i)->valueInteger());
            else
              stride = iref->getstride(dim,i);
          }
          else
            stride = new SgValueExp(0);
         analyse_expression(s, stride);
          functionexpr->addArg(*stride);

          SgExpression * offset;
          if (iref->getoffset(dim,i)!=NULL) {
            if (iref->getoffset(dim,i)->isInteger())
              offset = new SgValueExp(iref->getoffset(dim,i)->valueInteger());
            else
              offset = iref->getoffset(dim,i);
          }
          else
            offset = new SgValueExp(0);
         analyse_expression(s, offset);
          functionexpr->addArg(*offset);

        };

        SgSymbol * distsym = iref->getexpression()->symbol();
        SgSymbol * distdad = s->getSymbolInCurrentScope(mixstring,"dad",distsym->identifier());
        SgValueExp * dimention = new SgValueExp (lhsdim+1);
        SgExpression *funcexpr = make_range(s,distdad,dimention);

        functionexpr->addArg(*funcexpr);
      }

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgVarRefExp (irng);

      SgAssignStmt * assign = new SgAssignStmt (*rngexp, *functionexpr);

      s->insertStmtBefore(*assign, *s->controlParent());


      SgSymbol * funname = s->getSymbolInCurrentScope(mixstring, "delete_range", NULL);
      SgCallStmt * afuncall = new SgCallStmt(*funname);

      afuncall->addArg(* rngexp);
      s->addAttribute(INSERTED_AFTER_STMT,afuncall, sizeof(SgCallStmt));

      delete functionexpr, rngexp, assign , afuncall;
    }

    {
      SgSymbol * fname = s->getSymbolInCurrentScope(mixstring, "loop_bounds", NULL);

      SgCallStmt * afuncall = new SgCallStmt(*fname);

      SgSymbol * irng = s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);
      SgExpression * rngexp = new SgVarRefExp (irng);

      afuncall->addArg(*rngexp);

      SgSymbol * index=ihdef->getindexsymb(i);

      SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
      SgExpression * lbexpr = new SgVarRefExp(lb);
      afuncall->addArg(*lbexpr);
      SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
      SgExpression * ubexpr = new SgVarRefExp(ub);
      afuncall->addArg(*ubexpr);
      SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
      SgExpression * stepexpr = new SgVarRefExp(step);
      afuncall->addArg(*stepexpr);

      s->insertStmtBefore(*afuncall, *s->controlParent());

      delete afuncall,rngexp,lbexpr,ubexpr,stepexpr;
    };
  };
}

void delete_indepdo_section(SgStatement *s,int counter)
{
   SgSymbol * sddsymb;
  if(counter==-1)
        sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd",NULL);
  else sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd",NULL,counter);
  SgStatement *funcall = make_delete_array(s,sddsymb);
  s->addAttribute(INSERTED_AFTER_STMT,funcall, sizeof(SgStatement));
}

void create_indepdo_section(SgStatement *s,int counter=-1)
{
  Indepdo_reference *iref;
  if(counter==-1){
	  iref=(Indepdo_reference *)s->attributeValue(0,Indepdo_Lref);
  }
  else   iref=(Indepdo_reference *)s->attributeValue(counter,Indepdo_Ref);

  Indepdo_header_def *ihdef = (Indepdo_header_def *)s->attributeValue(0, IHEDR_Def);

  SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();
  SgSymbol * arraysymb  = iref->getexpression()->symbol();
  SgSymbol * sddsymb;
  if(counter==-1) 
  	sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd",NULL);
  else sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd",NULL,counter);
  // pcrc_new_array_section
  {
    int dim;
    if (iref->getCommunication() == BROADCAST_SCALAR )
        dim = 0;
    else
        dim = iref->numberOfIndex();
    SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
    SgStatement *assign = make_new_array_section(s, sddsymb, dim , dad);
    s->insertStmtBefore(*assign, *s->controlParent());
  }

  stack exp_stk;
  if(array->numberOfAttributes(SUB_REFED_LOOPVAR)>0) {
	change_indepdoref_in_doloop(array,exp_stk);
  }

  int dimcounter = 0;

  for (int i=0; i<array->numberOfSubscripts(); i++) {
    int index = iref->indexNumberOnDim(i);

   if (index != -1)
    {   // pcrc_set_array_triplet
      dimcounter++;

      // the following subprogram add start, end, stride;
      SgExpression *start;
      start=iref->getoffset(i,index);
      start=start+((*iref->getstride(i,index))*(*ihdef->getindextriple(index,0)));
      analyse_expression(s, start);

      SgExpression *end;
      end=iref->getoffset(i,index);
      end=end+((*iref->getstride(i,index))*(*ihdef->getindextriple(index,1)));
      analyse_expression(s, end);

      SgExpression *stride;
      stride = ihdef->getindextriple(index, 2);
      stride = stride * (*iref->getstride(i,index));
      stride = calculateExp(stride);

      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());

      SgStatement *funcall = make_set_array_triplet(s,sddsymb,dimcounter,
                                                start,end,stride,dad,i+1);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
    else if(array->subscript(i)->variant()==DDOT){
      dimcounter++;
	SgExpression *start = get_lbound(arraysymb, i);
      analyse_expression(s, start);
      SgExpression *end  = get_ubound(arraysymb, i);
      analyse_expression(s, end);
      SgExpression *stride = new SgValueExp(1);
	SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());

      SgStatement *funcall = make_set_array_triplet(s,sddsymb,dimcounter,
                                                start,end,stride,dad,i+1);
      s->insertStmtBefore(*funcall, *s->controlParent());

    }
    else {
        // pcrc_set_array_scalar
	SgExpression *sub=array->subscript(i);
      SgExpression * scalar;
	if(sub->numberOfAttributes(SUB_RANGE))
		scalar = get_lbound(array->symbol(),i);
	else scalar=array->subscript(i);
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
      int dim = i+1;
      SgStatement *funcall = make_set_array_scalar(s,sddsymb,scalar,dad,dim);
      s->insertStmtBefore(*funcall, *s->controlParent());
    }
  }
    if(exp_stk.elements()) resume_indepdoref_in_doloop(array,exp_stk);

  delete_indepdo_section(s,counter);
}

void create_remap_source(SgStatement *s,int counter){
	create_indepdo_section(s,counter);
}

void add_indepdo_coef(SgStatement * s, int counter, int * amount)
{
  SgSymbol * coefsym = s->getSymbolInCurrentScope(mixstring, "coef", NULL);
  Indepdo_header_def * ihdef = (Indepdo_header_def *)s->attributeValue(0, IHEDR_Def);
  Indepdo_reference * iref;
  if(counter!=-1) iref=(Indepdo_reference *)s->attributeValue(counter, Indepdo_Ref);
  else iref=(Indepdo_reference *)s->attributeValue(0, Indepdo_Lref);

//  if ( ref->sameAs != -1 ) return ;

  SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();

  SgSymbol * arraysymb = array->symbol();

  stack exp_stk;

  if(array->numberOfAttributes(SUB_REFED_LOOPVAR)>0) {
	change_indepdoref_in_doloop(array,exp_stk);
  }
	
  for (int i=0; i<array->numberOfSubscripts(); i++) {

    SgCallStmt * call = new SgCallStmt(*coefsym);

    {
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
      SgVarRefExp * dadexp = new SgVarRefExp(dad);
      call->addArg(*dadexp);
    };

    {
      SgValueExp * dimexp = new SgValueExp(i+1);
      call->addArg(*dimexp);
    }
    {
      int index=iref->indexNumberOnDim(i);

      if (index==-1) {
          SgExpression *dimexp = array->subscript(i);
		if(dimexp->variant()==DDOT){
			SgExpression *start,*end,*stride;
			if(dimexp->lhs()->lhs()){
				start=dimexp->lhs()->lhs();
			}
			else start = get_lbound(arraysymb, i);
			if(dimexp->lhs()->rhs()){
				end = dimexp->lhs()->rhs();
			}
			else{
				end = get_ubound(arraysymb, i);
			}
			if(dimexp->rhs()) stride=dimexp->rhs();
			else stride=new SgValueExp(1);
			analyse_expression(s, start);
			call->addArg(*start);
			analyse_expression(s, end);
			call->addArg(*end);
			analyse_expression(s, stride);
			call->addArg(*stride);
		}
		else{
          		call->addArg(*dimexp->copyPtr());
		        call->addArg(*dimexp->copyPtr());
		        dimexp = new SgValueExp(1);
		        call->addArg(*dimexp);
		}
      }
      else {
        for (int j=0; j<3; j++) {
          if (ihdef->getindextriple(index, j)!=NULL) {
            SgExpression * strideExp = ihdef->getindextriple(index, j)->copyPtr();
            strideExp = calculateExp(strideExp);
            call->addArg(*strideExp);
          }
          else {
            SgValueExp * strideexp = new SgValueExp(1);
            call->addArg(*strideexp);
          }
        }
      }
    }

    {
      int index=iref->indexNumberOnDim(i);

      if (index==-1) {
            SgValueExp * dimexp = new SgValueExp(1);
            call->addArg(*dimexp);
            dimexp = new SgValueExp(0);
            call->addArg(*dimexp);
      }
      else {
            SgExpression *exp=coefficient(array->subscript(i), ihdef->getindexsymb(index));
            SgExpression * stride = exp->lhs();
            SgExpression * offset = exp->rhs();
            calcExp(stride);
            offset=calculateExp(offset);

          if (stride)
            call->addArg(*stride);
          else {
            SgValueExp * strideexp = new SgValueExp(0);
            call->addArg(*strideexp);
          }
          if (offset)
            call->addArg(*offset);
          else {
            SgValueExp * strideexp = new SgValueExp(0);
            call->addArg(*strideexp);
          }
      }
    }

    {
      SgExpression *amountexp;
      if (amount == NULL) {
        SgSymbol *symb=s->getSymbolInCurrentScope(mixstring, "amount", NULL,counter);
        amountexp = new SgArrayRefExp(*symb);
        SgExpression *sub = new SgValueExp(i+1);
        ((SgArrayRefExp *)amountexp)->addSubscript(*sub);
      }
      else
        amountexp = new SgValueExp(* (amount+i));
        call->addArg(*amountexp);
      }

      {
      SgSymbol * indexsym = s->getSymbolInCurrentScope(mixstring, "u", NULL, counter, i);
      SgVarRefExp * indexexp = new SgVarRefExp(indexsym);
      call->addArg(*indexexp);
      }

      {
      SgSymbol * indexsym = s->getSymbolInCurrentScope(mixstring, "v", NULL, counter, i);
      SgVarRefExp * indexexp = new SgVarRefExp(indexsym);
      call->addArg(*indexexp);
      }

      s->addAttribute(INDEPDO_Coef,call, sizeof(SgCallStmt));
    };
    if(exp_stk.elements()) resume_indepdoref_in_doloop(array,exp_stk);
}

SgExpression *get_indepdo_dim_collapse_offset(SgStatement *s, int counter,int dim,int flag=1)
{
  	Indepdo_reference * iref = (Indepdo_reference *)s->attributeValue(counter, Indepdo_Ref);
  	Indepdo_header_def * ihdef = (Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
  	SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();
  	SgSymbol * arraysym=iref->getexpression()->symbol();

	int indexNumber = iref->indexNumberOnDim(dim);
	if (indexNumber == -1) return NULL;
    	SgExpression *index2=NULL;
        if ((flag==1 && array_dim_collapsed(arraysym,dim)!=0))
	{
		  // calculate the first effective element in collapsed dim defined by the global index
		SgSymbol *usymb = s->getSymbolInCurrentScope(mixstring,"u",NULL,counter,dim);
		SgExpression * uexp = new SgVarRefExp(usymb);
		  // must change local to global for collapesd dimension
	  	SgSymbol * indexsym = ihdef->getindexsymb(indexNumber);
		SgExpression *strideExp = ihdef->getindextriple(indexNumber,2);
		if (strideExp == NULL) strideExp=new SgValueExp(1);
		else strideExp=strideExp->copyPtr();
		SgExpression *lbExp = ihdef->getindextriple(indexNumber,0);
		if (lbExp == NULL) lbExp=new SgValueExp(0);
		else lbExp=lbExp->copyPtr();
		SgExpression *dimexp = new SgVarRefExp(indexsym);
		  // create the pcrc_glob(irg,pcrc_lil) to replace the forall index;
		{
		SgSymbol * lil = s->getSymbolInCurrentScope(mixstring, "lil", indexsym->identifier());
		SgExpression *lilexp = new SgVarRefExp(lil);
		SgSymbol * irg = s->getSymbolInCurrentScope(mixstring, "irg", NULL, indexNumber);
		SgExpression *globexp = make_glob(s,irg,lilexp);
		dimexp->replace(*globexp);
		}
		dimexp = &( (*dimexp - *lbExp)/ *strideExp );  // the No of this global index relative to array section
		index2 = &( *uexp * *dimexp );
	}
	return index2;
}

void append_indepdo_index(SgStatement *s, int counter)
{
  Indepdo_reference * iref;
  if(counter!=-1) iref = (Indepdo_reference *)s->attributeValue(counter, Indepdo_Ref);
  else{
 	iref = (Indepdo_reference *)s->attributeValue(0, Indepdo_Lref);
	counter=s->numberOfAttributes(Indepdo_Ref);
  }

  Indepdo_reference * lhsref = (Indepdo_reference *)s->attributeValue(0, Indepdo_Lref);
  Indepdo_header_def * ihdef = (Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);

  SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();

  SgSymbol * arraysym=iref->getexpression()->symbol();

  SgExpression * init_index = NULL;
  SgExpression * add_index ;

  SgSymbol * index_sdx_sym = s->getSymbolInCurrentScope(mixstring, "sdx", NULL,counter);
  SgExpression * index_sdx_exp = new SgVarRefExp(index_sdx_sym);

  SgSymbol *indextmpsym;
  SgExpression *index_tmp_sdx_exp;

  int index_num=ihdef->getindexnumber();
  int   dimcounter=0;
  for (int j=1; j<index_num; j++)
  {
        int scalar_dim=1;
        for (int i = 0; i<array->numberOfSubscripts(); i++)
                if (iref->getIndex(i,index_num-1-j)!=0) scalar_dim=0;
        dimcounter++;
        // if not scalar , leave to next step
        if ( scalar_dim == 0) continue;
        // donot add in the innest level
        indextmpsym = s->getSymbolInCurrentScope(mixstring, "tmp","sdx", counter,dimcounter-1);
        index_tmp_sdx_exp = new SgVarRefExp(indextmpsym);

        SgAssignStmt *push_stmt = new SgAssignStmt( *index_tmp_sdx_exp , *index_sdx_exp);
        s->addAttribute(INDEX_OPT+j, push_stmt , sizeof(push_stmt) );

        SgAssignStmt *add_stmt = new SgAssignStmt( *index_sdx_exp , *index_tmp_sdx_exp );
        s->addAttribute(INDEX_ADD+j, add_stmt , sizeof(add_stmt) );
  }

  SgStatement *init_stmt = NULL;

  for (int i = 0; i<array->numberOfSubscripts(); i++)
  {
    dimcounter=0;

    SgSymbol * vsymb = s->getSymbolInCurrentScope(mixstring, "v", NULL, counter, i);
    SgExpression *index1 = new SgVarRefExp(vsymb);

    init_index = init_index + *index1;

    for (int j=0;j<index_num; j++) {

      dimcounter++;
      if (iref->getIndex(i,index_num-1-j)!=0) {

	if(j!=0){
	        indextmpsym = s->getSymbolInCurrentScope(mixstring, "tmp","sdx",counter,dimcounter-1);
        	index_tmp_sdx_exp = new SgVarRefExp(indextmpsym);

	        SgStatement *push_stmt = new SgAssignStmt( *index_tmp_sdx_exp ,
*index_sdx_exp);
        	s->addAttribute(INDEX_OPT+j, push_stmt , sizeof(push_stmt));
	}

	add_index = NULL;
	if(arraysym->numberOfAttributes(Outofcore_Def) > 0){
        	SgSymbol * usymb = s->getSymbolInCurrentScope(mixstring, "oc_number",arraysym->identifier(), counter);
	        SgArrayRefExp * uexp = new SgArrayRefExp(*usymb);
        	SgExpression * step=new SgValueExp(1);
		if(storageMode==COLUMN_MAJOR){
			for (int l = 0; l<array->numberOfSubscripts(); l++) {
				SgExpression * sub=new SgValueExp(l+1);
    	       			sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
	        		uexp->setSubscripts(*sub);
        	 		if (iref->getIndex(l,index_num-1-j)!=0) {
					step=calculateExp(step);
               				add_index = add_index + *step;
	          		}
      				step=&(*step* uexp->copy());
       			}
		}
		else{
	        	for (int l = array->numberOfSubscripts()-1;l>-1; l--) {
                        	SgExpression * sub=new SgValueExp(l+1);
	                        sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
        	                uexp->setSubscripts(*sub);
                	        if (iref->getIndex(l,index_num-1-j)!=0) {
                        	        step=calculateExp(step);
                                	add_index = add_index + *step;
                        	}
	                        step=&(*step* uexp->copy());
        		}
	    	}
	}
	else{
	        for (int l = 0; l<array->numberOfSubscripts(); l++) {
        	   if (iref->getIndex(l,index_num-1-j)!=0) {
                	SgSymbol * usymb = s->getSymbolInCurrentScope(mixstring, "u", NULL, counter, l);
	                SgExpression * uexp = new SgVarRefExp(usymb);
        	        add_index = add_index + *uexp;
	           }
        	}
	}

        if ( j !=0 )
                add_index = index_tmp_sdx_exp + (*add_index);
        else
                add_index = index_sdx_exp + (*add_index);

        SgStatement *add_stmt = new SgAssignStmt( *index_sdx_exp->copyPtr() , *add_index );
        s->addAttribute(INDEX_ADD+j, add_stmt , sizeof(add_stmt) );

      } // end of if
    }; // end for j
  }     // end for i

    for (int i = 0; i<array->numberOfSubscripts(); i++) 
    {
	// if the lhs is collapsed in corresponent dimension, no need to add get offset.
        int rindex = iref->indexNumberOnDim(i);
        if (rindex != -1) {
	   int lindex = lhsref->dimNumberOnIndex(rindex);
	   if (lindex != -1) {
           if (array_dim_collapsed(lhsref->getexpression()->symbol(), lindex) == 1)
		continue;
	   }
	}
	  // index2 hold the offset of this collapsed dimension.
	SgExpression *index2 = get_indepdo_dim_collapse_offset(s,counter,i);
	if (index2 == NULL) continue;
    	init_index = init_index + *index2;
    }  // end of for
  if(arraysym->numberOfAttributes(Outofcore_Def)>0){
  	init_stmt = new SgAssignStmt( *index_sdx_exp , *new SgValueExp(0));
	s->addAttribute(OC_INDEX_INIT, init_stmt , sizeof(init_stmt) );
  } 
  else{ 	
       	init_stmt = new SgAssignStmt( *index_sdx_exp , *init_index );
	s->addAttribute(INDEX_INIT, init_stmt , sizeof(init_stmt) );
  }
}

void create_base_get_stmt(SgStatement *s,int couter,SgSymbol * arraysym,SgSymbol *basesym){
	SgSymbol *mHandle=s->getSymbolInCurrentScope(mixstring,"mh",NULL,couter);
	SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",arraysym->identifier());
	SgStatement *base_get=make_stack_alloc(s,arraysym->type(),basesym,dad,mHandle);
	s->addAttribute(BASE_GET_STMT,base_get,sizeof(base_get));
	SgStatement *free=make_free(s,arraysym->type(),mHandle);
	s->addAttribute(STACK_FREE_STMT,free,sizeof(free));
	
}

void calc_oc_array_collapse_num(SgStatement *s,int i){
	Indepdo_reference * rightref = (Indepdo_reference *) s->attributeValue(i, Indepdo_Ref);
	SgArrayRefExp *array =(SgArrayRefExp *) rightref->getexpression();
	SgSymbol *arraysym=array->symbol();
	SgExpression *collapse=new SgValueExp(1);
	for(i=0;i<array->numberOfSubscripts();i++){
		SgExpression * sub, * range;
		if(array->subscript(i)->numberOfAttributes(SUB_RANGE)){
			range=(SgExpression *) array->subscript(i)->attributeValue(0,SUB_RANGE);
			sub=array->subscript(i)->copyPtr();
			array->subscript(i)->replace(*range);
		}
		if(array->subscript(i)->variant()==DDOT){
				SgExpression *l = ddot_sub_triplet(array,i,0);
                        	analyse_expression(s, l);
        	                SgExpression *u  = ddot_sub_triplet(array,i,1); 
                	        analyse_expression(s, u);
        	                SgExpression *st  = ddot_sub_triplet(array,i,2); 
				SgExpression * num=&((*u-*l) / *st + *new SgValueExp(1) );
                                analyse_expression(s, num);
                                collapse=&(*collapse * *num);
                }
		if(array->subscript(i)->numberOfAttributes(SUB_RANGE))
			array->subscript(i)->replace(*sub);
	}
        collapse=calculateExp(collapse);
	if(*collapse!=*new SgValueExp(1)){
		if(s->numberOfAttributes(COLLAPSE_NUM)>0){
			SgExpression *col=(SgExpression *)s->attributeValue(0,COLLAPSE_NUM);
			SgSymbol * maxSym= s->getSymbolInCurrentScope("max");
                	SgFunctionRefExp * maxExp = new SgFunctionRefExp (*maxSym);
			maxExp->addArg(col->copy());
			maxExp->addArg(*collapse);
			col->replace(*maxExp);
		}
		else{
			s->addAttribute(COLLAPSE_NUM,collapse,sizeof(SgExpression));
		}
	}
}

void get_exp(SgExpression *exp,stack &stk1,stack &stk2){
	if(!exp) return;
	switch(exp->variant()){
		case ADD_OP:
			get_exp(exp->lhs(),stk1,stk2);
			get_exp(exp->rhs(),stk1,stk2);
			break;
		case MINUS_OP:
			get_exp(exp->lhs(),stk1,stk2);
			get_exp(exp->rhs(),stk2,stk1);
			break;
		case MULT_OP:
		case DIV_OP:
			calcExp(exp->lhs());
			calcExp(exp->rhs());
			exp->replace(*calculateExp(exp));
			stk1.push(exp);
			break;
		case INT_VAL:
			if(exp->valueInteger()==0) ;
			else stk1.push(exp);
			break;
		default: 
			stk1.push(exp);
	}
}

SgExpression *compose_exp(stack stk1,stack stk2){
	SgExpression *exp=NULL;
	int i1=1,i2=1;
	if(stk1.elements()){
		exp=stk1.get_exp(0);
		while(i1<stk1.elements()){
			exp=new SgExpression(ADD_OP,exp,stk1.get_exp(i1),NULL);
			i1++;
		}
	}
	if(stk2.elements()){
		if(!exp) exp=new SgExpression(MINUS_OP,stk2.get_exp(0),NULL,NULL);
		else{ i2=0; }
		while(i2<stk2.elements()){
			exp=new SgExpression(MINUS_OP,exp,stk2.get_exp(i2),NULL);
			i2++;
		}
	}
	if(exp) return exp;
	else return new SgValueExp(0);
}

void compute_value(stack &stk1,stack &stk2){
	int i=0,value=0;
	while(i<stk1.elements()){
		if(stk1.get_exp(i)->variant()==INT_VAL){
			value+=stk1.get_exp(i)->valueInteger();
			stk1.delete_exp(i);
		}
		else i++;
	}
	while(i<stk2.elements()){
		if(stk2.get_exp(i)->variant()==INT_VAL){
			value-=stk2.get_exp(i)->valueInteger();
			stk2.delete_exp(i);
		}
		else i++;
	}
	if(value>0) stk1.push(new SgValueExp(value));
	else if(value<0) stk2.push(new SgValueExp(-value));
}

void mult_exp(SgExpression * & e,stack stk1,stack stk2){
	stack stk;
	for(int i=0;i<stk1.elements();i++){
		for(int j=0;j<stk2.elements();j++){
			stk.push(&(*stk1.get_exp(i) * *stk2.get_exp(j)));
		}
	}
	if(stk.elements()){
		e=stk.get_exp(0);
		int p=1;
		while(p<stk.elements()){
			e=new SgExpression(ADD_OP,e,stk.get_exp(p),NULL);
			p++;
		}
	}
	else e=NULL;
			
}

void break_exp(SgExpression *exp){
	if(!exp) return;
	stack stk1,stk2,stk3,stk4,stk5,stk6;
	SgExpression *e,*e1,*e2;
	switch(exp->variant()){
	case ADD_OP:
	case MINUS_OP:
		break_exp(exp->lhs());
		break_exp(exp->rhs());
		break;
	case MULT_OP:
		break_exp(exp->lhs());
		break_exp(exp->rhs());
		get_exp(exp->lhs(),stk1,stk2);
		get_exp(exp->rhs(),stk3,stk4);
		mult_exp(e1,stk1,stk3);
		mult_exp(e2,stk2,stk4);
		e=e1;
		if(e2) e=new SgExpression(ADD_OP,e1,e2,NULL);
		mult_exp(e1,stk1,stk4);
		if(e1) e=new SgExpression(MINUS_OP,e,e1,NULL);
		mult_exp(e2,stk2,stk3);
		if(e2) e=new SgExpression(MINUS_OP,e,e2,NULL);
		exp->replace(*e);
	}
}

void stride_ofvar(stack stk,SgVarRefExp *var,SgExpression * &e){
	e=NULL;
	for(int i=0;i<stk.elements();i++){
		SgExpression *tmp_e=stk.get_exp(i)->copyPtr();
		if(find_and_replace(tmp_e,var,new SgValueExp(1))){
			if(!e) e=tmp_e;
			else e=new SgExpression(ADD_OP,e,tmp_e,NULL);
		}
	}
}

void offset_ofvar(stack stk,SgVarRefExp *var,SgExpression * &e){
	e=NULL;
	for(int i=0;i<stk.elements();i++){
		SgExpression *tmp_e=stk.get_exp(i)->copyPtr();
		if(!find_and_replace(tmp_e,var,NULL)){
			if(!e) e=tmp_e;
			else e=new SgExpression(ADD_OP,e,tmp_e,NULL);
		}
	}
}

void get_coef(SgExpression *e,SgVarRefExp *var,SgExpression * &a,SgExpression * &b){
	if(!e){ a=new SgValueExp(0);b=new SgValueExp(0); return;}
	calcExp(e);
	break_exp(e);
	stack stk1,stk2;
	get_exp(e,stk1,stk2);
	SgExpression *e1,*e2;
	stride_ofvar(stk1,var,e1);
	stride_ofvar(stk2,var,e2);
	a=e1;
	if(!e1&&!e2) ;
	else if(e1&&!e2) ;
	else if(!e1&&e2) a=new SgExpression(MINUS_OP,e2,NULL,NULL);
	else a=new SgExpression(MINUS_OP,a,e2,NULL);
	if(a) calcExp(a);
	else a=new SgValueExp(0);
		
	offset_ofvar(stk1,var,e1);
	offset_ofvar(stk2,var,e2);
	b=e1;
	if(!e1&&!e2) ;
	else if(e1&&!e2) ;
	else if(!e1&&e2) b=new SgExpression(MINUS_OP,e2,NULL,NULL);
	else b=new SgExpression(MINUS_OP,b,e2,NULL);
	if(b) calcExp(b);
	else b=new SgValueExp(0);	
	return;
	
}

void get_factor(SgExpression *exp,stack & stk){
	if(!exp) return;
	switch(exp->variant()){
	case MULT_OP:
		get_factor(exp->lhs(),stk);
		get_factor(exp->rhs(),stk);
		break;
	default:
		stk.push(exp);
	}
}

void simplify_exp(SgExpression *exp){
	if(!exp) return;
	stack stk;
	get_factor(exp,stk);
	int tmp=1;
	int i=0;
	while(i<stk.elements()){
		if(stk.get_exp(i)->variant()==INT_VAL){
			tmp*=stk.get_exp(i)->valueInteger();
			stk.delete_exp(i);
		}
		else i++;
	}
	if(!stk.elements()||tmp==0){
		exp->replace(* new SgValueExp(tmp));
		return;
	}
	SgExpression *e=stk.get_exp(0);
	for(i=1;i<stk.elements();i++){
		e=new SgExpression(MULT_OP,e,stk.get_exp(i),NULL);
	}
	if(tmp!=1) { e=new SgExpression(MULT_OP,e,new SgValueExp(tmp),NULL); }
	exp->replace(*e);
}

int minus_exp(SgExpression *e1,SgExpression *e2,SgExpression * & e){
	stack stk1,stk2;
	get_factor(e1,stk1);
	get_factor(e2,stk2);
	SgExpression *tmp_e=NULL;
	int p1=0,p2;
	while(p1<stk1.elements()){
		if(!stk1.get_exp(p1)->isInteger()){
			if(!tmp_e) tmp_e=stk1.get_exp(p1);
			else tmp_e=new SgExpression(MULT_OP,tmp_e,stk1.get_exp(p1),NULL);
		}
		p1++;
	}
	p1=0; 
	while(p1<stk1.elements()){
		p2=0;
		while(p2<stk2.elements()){
			if(*stk1.get_exp(p1)==*stk2.get_exp(p2)&&(!stk1.get_exp(p1)->isInteger())){
				stk1.delete_exp(p1);
				stk2.delete_exp(p2);
				if(p1<stk1.elements()) p2=0;
				else break;
			}
			else p2++;
		}
		p1++;
	}
	int k1=1,k2=1;
	for(int i=0;i<stk1.elements();i++){
		if(!stk1.get_exp(i)->isInteger()) return 0;
		k1*=stk1.get_exp(i)->valueInteger();
	}
	for(int i=0;i<stk2.elements();i++){
		if(!stk2.get_exp(i)->isInteger()) return 0;
		k2*=stk2.get_exp(i)->valueInteger();
	}
	if(k1-k2==0){ e=NULL;return 1;}
	if(k1-k2>0){
		if(!tmp_e) e=new SgValueExp(k1-k2);
		else e=new SgExpression(MULT_OP,new SgValueExp(k1-k2),tmp_e,NULL);
		return 1;
	}	
	if(!tmp_e) e=new SgValueExp(k2-k1);
	else e=new SgExpression(MULT_OP,new SgValueExp(k2-k1),tmp_e,NULL);
	
	return -1;
				
}

int add_exp(SgExpression *e1,SgExpression *e2,SgExpression * & e){
	stack stk1,stk2;
	get_factor(e1,stk1);
	get_factor(e2,stk2);
	SgExpression *tmp_e=NULL;
	int p1=0,p2;
	while(p1<stk1.elements()){
		if(!stk1.get_exp(p1)->isInteger()){
			if(!tmp_e) tmp_e=stk1.get_exp(p1);
			else tmp_e=new SgExpression(MULT_OP,tmp_e,stk1.get_exp(p1),NULL);
		}
		p1++;
	}
	p1=0; 
	while(p1<stk1.elements()){
		p2=0;
		while(p2<stk2.elements()){
			if(*stk1.get_exp(p1)==*stk2.get_exp(p2)&&(!stk1.get_exp(p1)->isInteger())){
				stk1.delete_exp(p1);
				stk2.delete_exp(p2);
				if(p1<stk1.elements()) p2=0;
				else break;
			}
			else p2++;
		}
		p1++;
	}
	int k1=1,k2=1;
	for(int i=0;i<stk1.elements();i++){
		if(!stk1.get_exp(i)->isInteger()) return 0;
		k1*=stk1.get_exp(i)->valueInteger();
	}
	for(int i=0;i<stk2.elements();i++){
		if(!stk2.get_exp(i)->isInteger()) return 0;
		k2*=stk2.get_exp(i)->valueInteger();
	}
	if(k1+k2==0){ e=NULL;return 1;}
	if(!tmp_e) e=new SgValueExp(k1+k2);
	else e=new SgExpression(MULT_OP,new SgValueExp(k1+k2),tmp_e,NULL);
	return 1;
				
}

void merge_exp(stack &stk){
	int i1=0,i2;
	while(i1<stk.elements()){
		i2=i1+1;
		while(i2<stk.elements()){
			SgExpression *e1,*e2,*e;
			e1=stk.get_exp(i1);	
			e2=stk.get_exp(i2);	
			if(add_exp(e1,e2,e)){
				if(!e){
					stk.delete_exp(i2);
					stk.delete_exp(i1);
					if(i1<stk.elements()) i2=i1+1;
					else break;
				}
				else{
					stk.delete_exp(i2);
					stk.replace_exp(i1,e);
				}
			}	
			else i2++;
		}
		i1++;
	}
}
void computeExp(SgExpression *exp){
	if(!exp) return;
	exp->replace(*calculateExp(exp));
	break_exp(exp);
	stack stk1,stk2;
	get_exp(exp,stk1,stk2);
	merge_exp(stk1);
	merge_exp(stk2);
	int i1=0;
	while(i1<stk1.elements()){
		int i2=0;
		while(i2<stk2.elements()){
			SgExpression *e;
			int flag;
			if(flag=minus_exp(stk1.get_exp(i1),stk2.get_exp(i2),e)){
				if(!e){
					stk1.delete_exp(i1);
					stk2.delete_exp(i2);
				}
				else if(flag==-1){
					stk1.delete_exp(i1);
					stk2.replace_exp(i2,e);
				}
				else{
					stk2.delete_exp(i2);
					stk1.replace_exp(i1,e);
				}
				if(i1<stk1.elements()) i2=0;
				else break;
			}
			else i2++;
		}
		i1++;
	}
	compute_value(stk1,stk2);
	exp->replace(*compose_exp(stk1,stk2));
}

void calcExp(SgExpression *exp){
	if(!exp) return;
	exp->replace(*calculateExp(exp));
	stack stk1,stk2;
	get_exp(exp,stk1,stk2);
	merge_exp(stk1);
	merge_exp(stk2);
	int i1=0;
	while(i1<stk1.elements()){
		int i2=0;
		while(i2<stk2.elements()){
			SgExpression *e;
			int flag;
			if(flag=minus_exp(stk1.get_exp(i1),stk2.get_exp(i2),e)){
				if(!e){
					stk1.delete_exp(i1);
					stk2.delete_exp(i2);
				}
				else if(flag==-1){
					stk1.delete_exp(i1);
					stk2.replace_exp(i2,e);
				}
				else{
					stk2.delete_exp(i2);
					stk1.replace_exp(i1,e);
				}
				if(i1<stk1.elements()) i2=0;
				else break;
			}
			else i2++;
		}
		i1++;
	}
	compute_value(stk1,stk2);
	exp->replace(*compose_exp(stk1,stk2));
}

void divide_exp(SgExpression *exp){
	if(exp->variant()!=DIV_OP) return;
	stack stk1,stk2;
	get_factor(exp->lhs(),stk1);
	get_factor(exp->rhs(),stk2);
	int i1=0;
	int i2;
	while(i1<stk1.elements()){
		i2=0;
		while(i2<stk2.elements()){
			if(*stk1.get_exp(i1)==*stk2.get_exp(i2)){
				stk2.delete_exp(i2);
				stk1.delete_exp(i1);
				if(i1<stk1.elements()) i2=0;
				else break;
			}
			else i2++;
		}
		i1++;
	}
	SgExpression *lhs=NULL,*rhs=NULL;
	i1=0;i2=0;
	while(i1<stk1.elements()){
		if(!lhs) lhs=stk1.get_exp(i1);
		else lhs=&(*lhs * *stk1.get_exp(i1));
		i1++;
	}
	while(i2<stk2.elements()){
		if(!rhs) rhs=stk2.get_exp(i2);
		else rhs=&(*rhs * *stk2.get_exp(i2));
		i2++;
	}
	if((!lhs)&&(!rhs)) exp->replace(* new SgValueExp(1));
	else if(!rhs) exp->replace(*lhs);
	else if(!lhs) exp->replace(*new SgValueExp(1) / *rhs);
	else exp->replace(*lhs / *rhs);
}

void change_indepdo_array_reference(SgStatement * s, int counter,Communication com=NONE)
{
  Indepdo_reference * iref = (Indepdo_reference *)s->attributeValue(counter,Indepdo_Ref);
  SgArrayRefExp * oldarray = (SgArrayRefExp *)iref->getexpression();
  SgSymbol * arraysym = oldarray->symbol();

  SgArrayRefExp * newarray;

  if ( unknown_address(arraysym)&&com==NONE )
  {//means array is dynamic,allocateable,or pointer
      SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", NULL,counter);
      SgSymbol * dad = s->getSymbolInCurrentScope(mixstring, "dad", arraysym->identifier());
      SgStatement * index_stmt = make_index(s, basesym, dad, arraysym->type());
      s->insertStmtBefore(*index_stmt, *s->controlParent());
  }
  int address_unknown=unknown_address(arraysym);

  int incore_useable=!(address_unknown&&com==NONE||arraysym->numberOfAttributes(Outofcore_Def)>0&&iref->pre_sym_ref_num&&com==NONE||com!=NONE);

  if ( incore_useable ) {
    SgSymbol * local= s->getSymbolInCurrentScope(arraysym->identifier());
    newarray = new SgArrayRefExp(*local);
  }
  else {
    SgSymbol * rs;
#ifdef MEM_ALLOC
    rs= s->getSymbolInCurrentScope(stackNameOfType(arraysym->type()), "mb",NULL);
#else
    rs= s->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysym->type()), NULL);
#endif
  newarray = new SgArrayRefExp(*rs);
  };

  SgSymbol * indexsym;
  if(com==NONE) indexsym = s->getSymbolInCurrentScope(mixstring, "sdx", NULL,counter);
  else{
	Indepdo_reference *iref=(Indepdo_reference *)s->attributeValue(0,Indepdo_Lref);
	if(!(s->numberOfAttributes(Lref_created))){
		int index_count;
		if(iref->pre_ref==-1) index_count=iref->counter;
		else index_count=iref->pre_ref;
 		indexsym = s->getSymbolInCurrentScope(mixstring, "sdx", NULL,index_count);
	}
	else
		indexsym = s->getSymbolInCurrentScope(mixstring, "sdx", NULL,s->numberOfAttributes(Indepdo_Ref));

  }
  SgExpression * newindex = new SgVarRefExp(indexsym);

  if(oldarray->numberOfAttributes(SUB_REFED_LOOPVAR)){
  	int i;
	for(i=0;i<oldarray->numberOfSubscripts();i++){
		SgExpression *sub=oldarray->subscript(i);
		SgExpression *newsub=sub->copyPtr();
		if(sub->numberOfAttributes(SUB_RANGE)){
			SgSymbol * usymb = s->getSymbolInCurrentScope(mixstring, "u", NULL, counter, i);
	       		SgExpression * uexp = new SgVarRefExp(usymb);
			SgExpression *sub_range=(SgExpression *)sub->attributeValue(0,SUB_RANGE);
			SgExpression *tmp=&(*newsub-sub_range->lhs()->lhs()->copy());
			tmp=&(*tmp * *uexp);
			newindex=&(*newindex+*tmp); 
			calcExp(newindex);
			
		}
	}
 }

 if (incore_useable)
        newarray->addSubscript(*newindex);
 else {
        SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs",NULL, counter);
	if(com==NONE&&address_unknown==0) create_base_get_stmt(s,counter,arraysym,basesym);//only out of core do it
        SgExpression * baseexp= new SgVarRefExp(basesym);
        newindex=&(*baseexp+*newindex);
	ASSERT(newindex);
	ASSERT(newarray);
	newindex=new SgExpression(EXPR_LIST,newindex,NULL,NULL);
        newarray->setSubscripts(*newindex);
 }

  // modify it
 SgArrayRefExp *kept_array=(SgArrayRefExp *)oldarray->copyPtr();
 for(int i=0;i<oldarray->numberOfSubscripts();i++){
	if(oldarray->subscript(i)->numberOfAttributes(SUB_RANGE)){
		SgExpression *tmp=(SgExpression *)oldarray->subscript(i)->attributeValue(0,SUB_RANGE);
		kept_array->subscript(i)->addAttribute(SUB_RANGE,tmp,sizeof(tmp));
	}
 }
 oldarray->addAttribute(KEPT_ARRAY_REF,kept_array,sizeof(SgArrayRefExp));
  Indepdo_reference *lref=(Indepdo_reference *)(s->attributeValue(0,Indepdo_Lref));
  SgSymbol *lsym=lref->getexpression()->symbol();
  if(arraysym->numberOfAttributes(Outofcore_Def)>0&&(
	is_out_of_core(lsym)||com==NONE)){	
  	oldarray->addAttribute(KEPT_OC_REF,kept_array,sizeof(SgArrayRefExp));
	calc_oc_array_collapse_num(s,counter);
	if(s->numberOfAttributes(INDEPDO_Have_Oc_Array)==0)
		s->addAttribute(INDEPDO_Have_Oc_Array,NULL,0);
  }

  oldarray->replace(*newarray);
}

int prechangeable(SgStatement *s,int i){
	Indepdo_reference * iref = (Indepdo_reference *)s->attributeValue(i,Indepdo_Ref);
	SgExpression * oldarray = iref->getexpression();
	SgSymbol * arraysym = oldarray->symbol();
	if(iref->pre_ref!=-1){
		Indepdo_reference * pre_iref = (Indepdo_reference *)s->attributeValue(iref->pre_ref,Indepdo_Ref);
                oldarray->replace(pre_iref->getexpression()->copy());
		return 1;
	}
	return 0;
}

void change_loop_bounds(SgStatement *s){
	Indepdo_header_def * ihdef = (Indepdo_header_def *) s->attributeValue(0, IHEDR_Def);
	Indepdo_reference * iref = (Indepdo_reference *) s->attributeValue(0, Indepdo_Lref);

	SgArrayRefExp * array = (SgArrayRefExp *) iref->getexpression();

	for (int i=0; i<ihdef->getindexnumber(); i++) {
		Indepdo_index_def * iidef=(Indepdo_index_def *)s->attributeValue(i,IINDEX_Def);
		if(!iidef->loop_cut) continue;
		SgSymbol * index=ihdef->getindexsymb(i);
		SgSymbol * lb = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
		SgExpression * lbexpr = new SgVarRefExp(lb);
		SgSymbol * ub = s->getSymbolInCurrentScope(mixstring, "liu", index->identifier());
		SgExpression * ubexpr = new SgVarRefExp(ub);
		SgSymbol * step = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
		SgExpression * stepexpr = new SgVarRefExp(step);

		ihdef->getindextriple(i,0)->replace(* new SgValueExp(0));
		ihdef->getindextriple(i,1)->replace((*ubexpr-*lbexpr)/(*stepexpr));
		if(ihdef->getindextriple(i,2)) ihdef->getindextriple(i,2)->replace(* new SgValueExp(1));
	}
}

void wrap_indepdo(SgStatement *s,SgStatement * & new_s)
{
	int i;
	SgSymbol * lhssec = s->getSymbolInCurrentScope(mixstring, "sdd", NULL);
	SgExpression * grpexp = make_group(s,lhssec);

	SgExpression * onexp = make_on(s, grpexp);

	new_s=s->copyPtr();

	SgStatement *s1= new SgIfStmt(*onexp, *new_s);

	s->insertStmtBefore(*s1, *s->controlParent());

	for(i=0;i<s->numberOfAttributes(BASE_GET_STMT);i++){
                SgStatement *stmt=(SgStatement *)s->attributeValue(i,BASE_GET_STMT);
                new_s->insertStmtBefore(*stmt,*new_s->controlParent());
        }

	for(i=0;i<s->numberOfAttributes(INDEPDO_Coef);i++){
                SgStatement *stmt=(SgStatement *)s->attributeValue(i,INDEPDO_Coef);
                new_s->insertStmtBefore(*stmt,*new_s->controlParent());
        }

	for (i=0 ; i< s->numberOfAttributes(INDEX_INIT) ; i++ )
  	{
        	SgStatement *init_stmt = (SgStatement *)s->attributeValue(i,INDEX_INIT);        
		new_s->insertStmtBefore(*init_stmt, *new_s->controlParent());
  	}

	for(i=0;i<s->numberOfAttributes(INSERTED_AFTER_STMT);i++){
		SgStatement *stmt=(SgStatement *)s->attributeValue(i,INSERTED_AFTER_STMT);
		s1->insertStmtAfter(*stmt,*s1->controlParent());
	}
	s->deleteStmt();
}

SgStatement *get_old_s(int loop_num,SgStatement *s){
	SgStatement *s1=s;
	int i;
	for(i=0;i<loop_num;i++){
		s1=s1->childList1(0);
	}
	return s1;
}

int in_core_useable(SgStatement * s,int i){
	Indepdo_reference * iref=(Indepdo_reference *)s->attributeValue(i,Indepdo_Ref);
        SgArrayRefExp *array=(SgArrayRefExp *)iref->getexpression();
        SgSymbol * arraySym=array->symbol();
	int j;
	for(j=0;j<i;j++){
		Indepdo_reference * iref1=(Indepdo_reference *)s->attributeValue(i,Indepdo_Ref);
	        SgArrayRefExp *array1=(SgArrayRefExp *)iref->getexpression();
        	SgSymbol * arraySym1=array1->symbol();
		if(arraySym->identifier()==arraySym1->identifier()) return 0;
	}
	return 1;
}

void add_oc_access_stmt(SgStatement *s1,SgStatement *s2,SgStatement *s,int i){
	Indepdo_header_def *ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	Indepdo_reference * iref=(Indepdo_reference *)s->attributeValue(i,Indepdo_Ref);
	int communicated=0;
	int communicated_by=-1;
	if(iref->getCommunication()!=NONE&&iref->getCommunication()!=SHIFT){
		communicated=1;
		communicated_by=((Indepdo_reference *)s->attributeValue(0,Indepdo_Lref))->counter;
	}	
	SgArrayRefExp *newarray=(SgArrayRefExp *)iref->getexpression();
	SgArrayRefExp *array=(SgArrayRefExp *)newarray->attributeValue(0,KEPT_OC_REF);
	SgSymbol * arraySym=array->symbol();
	SgSymbol *dad;
	if(!communicated) dad=s1->getSymbolInCurrentScope(mixstring, "dad",arraySym->identifier());
	else dad=s1->getSymbolInCurrentScope(mixstring, "tdd",NULL,i);
        SgSymbol * memsym=s1->getSymbolInCurrentScope(arraySym->identifier());
        SgArrayRefExp * memExp;
	if(!communicated&&iref->pre_sym_ref_num==0) {
		if (unknown_address(arraySym))
		{
      		   SgSymbol * basesym = s1->getSymbolInCurrentScope(mixstring, "tbs", NULL, i);
      		   SgSymbol * stack = s1->getSymbolInCurrentScope(stackNameOfType(arraySym->type()),"mb", NULL); 
      		   memExp = new SgArrayRefExp(*stack);
      		   SgExpression *subs = new SgVarRefExp(basesym);
      		   memExp->addSubscript(*subs);
		}
		else
		   memExp = new SgArrayRefExp(*memsym);
	}
	else{
		SgSymbol *rs;
		#ifdef MEM_ALLOC
    		rs= s1->getSymbolInCurrentScope(stackNameOfType(arraySym->type()), "mb",NULL);
		#else
    		rs= s1->getSymbolInCurrentScope(mixstring, stackNameOfType(arraySym->type()), NULL);
		#endif
		memExp=new SgArrayRefExp(*rs);
		SgSymbol * basesym = s1->getSymbolInCurrentScope(mixstring, "tbs",NULL, i);
		memExp->setSubscripts(*new SgExpression(EXPR_LIST,new SgVarRefExp(basesym),NULL,NULL));
	}
        SgSymbol * stridesym=s1->getSymbolInCurrentScope(mixstring,"oc_stride",arraySym->identifier(),i);
        SgSymbol * numbersym=s1->getSymbolInCurrentScope(mixstring,"oc_number",arraySym->identifier(),i);
        SgSymbol * startsym=s1->getSymbolInCurrentScope(mixstring,"sdx","oc",i);
        SgExpression * startExp = new SgVarRefExp(*startsym);
	SgExpression * index=new SgValueExp(0);
	int j;
	for(j=0;j<array->numberOfSubscripts();j++){
		SgSymbol * vsymb = s1->getSymbolInCurrentScope(mixstring, "v", NULL,communicated?communicated_by:i,j);
		SgExpression *index1 = new SgVarRefExp(vsymb);
		index=&(*index+*index1);
		int l;
		SgSymbol * usymb = s1->getSymbolInCurrentScope(mixstring, "u", NULL,communicated?communicated_by:i,j);
                SgExpression *index2 = new SgVarRefExp(usymb);
		SgArrayRefExp *stride=new SgArrayRefExp(*stridesym);
		SgExpression * sub=new SgValueExp(j+1);
		sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
		stride->setSubscripts(*sub);
		SgStatement *stride_stmt=new SgAssignStmt(*stride,*index2);
		SgArrayRefExp *number=new SgArrayRefExp(*numbersym);
		number->setSubscripts(sub->copy());
		s1->insertStmtBefore(*stride_stmt,*s1->controlParent());
		SgExpression *index3=NULL;
		for(l=0;l<ihdef->getindexnumber();l++){
			if(iref->getIndex(j,l)!=0){
				SgSymbol * oldsym=ihdef->getindexsymb(l);
				SgSymbol * newsym = s1->getSymbolInCurrentScope(mixstring,"T",oldsym->identifier());
				SgExpression *newexp=new SgVarRefExp(*newsym);
				index=&(*index+(index2->copy()* *newexp));
				SgStatement *s3=s2;
				int k=l;
				while(k>0){
					int tmp=0;
					while(s3->childList1(tmp)->variant()!=FOR_NODE) tmp++;
					s3=s3->childList1(tmp);
					k--;
				}
				SgExpression *doHeaderExp=s3->expr(0);
				SgExpression *step=doHeaderExp->operand(2);
				SgExpression *lb=doHeaderExp->operand(1)->operand(1)->copyPtr();
				SgExpression *ub=doHeaderExp->operand(1)->operand(2)->copyPtr();
				index3=&(*ub-*lb);
				if(step) index3=&(*index3/(step->copy()));
				index3=&(*index3 + * new SgValueExp(1));
			}
		}
		SgStatement *number_stmt;
		if(index3)
			index3=calculateExp(index3);
		else{
			SgExpression * sub, * range;

			if(array->subscript(j)->numberOfAttributes(SUB_RANGE)){
				range=(SgExpression *) array->subscript(j)->attributeValue(0,SUB_RANGE);
				sub=array->subscript(j)->copyPtr();
				array->subscript(j)->replace(*range);
			}
			if(array->subscript(j)->variant()==DDOT){
				SgExpression *l = ddot_sub_triplet(array,j,0);
                        	analyse_expression(s, l);
        	                SgExpression *u  = ddot_sub_triplet(array,j,1); 
                	        analyse_expression(s, u);
        	                SgExpression *st  = ddot_sub_triplet(array,j,2); 
				index3=&((*u-*l) / *st + *new SgValueExp(1) );
				index3=calculateExp(index3);
			}
			else 	index3=new SgValueExp(1);
			if(array->subscript(j)->numberOfAttributes(SUB_RANGE))
				array->subscript(j)->replace(*sub);
		}
		number_stmt=new SgAssignStmt(*number,*index3);
		s2->insertStmtBefore(*number_stmt,*s2->controlParent());
	}
	index=calculateExp(index);
	SgStatement *start_stmt=new SgAssignStmt(*startExp,*index);
	s2->insertStmtBefore(*start_stmt,*s2->controlParent());
//	if(!in_core_useable(s,i)){
//		array2stack(s1,memExp,i);
//	}
	if((iref->get_ref_attrib()|iref->added_ref_attrib)&Right_Bit||iref->getexpression()->numberOfAttributes(CONDITION_ASSIGN)){
		SgStatement * ocreadStmt=make_oc_read(s2,dad,memExp,startExp,stridesym,numbersym);
		s2->insertStmtBefore(*ocreadStmt,*s2->controlParent());
	}
	if((iref->get_ref_attrib()|iref->added_ref_attrib)&Left_Bit){
		SgStatement * ocwriteStmt=make_oc_write(s2,dad,memExp,startExp,stridesym,numbersym);
                s2->insertStmtAfter(*ocwriteStmt,*s2->controlParent());
        }
}

void dispose_indepdo_oc_arrayref(SgStatement *s){
	SgStatement * new_s;
	wrap_indepdo(s,new_s);
	Indepdo_header_def * ihdef = (Indepdo_header_def *) s->getAttribute(0, IHEDR_Def)->getAttributeData();

	if(s->numberOfAttributes(INDEPDO_Have_Oc_Array)==0){
		for(int i=0;i<s->numberOfAttributes(INSERTED_AFTER_STMT_IN);i++){
			SgStatement *stmt=(SgStatement *)s->attributeValue(i,INSERTED_AFTER_STMT_IN);
			new_s->insertStmtAfter(*stmt,*new_s->controlParent());
		}
		return;
	}
//independent do has not oc array reference,so return

//else,begin dispose oc array references
	int loop_num=ihdef->getindexnumber();
	if(s->numberOfAttributes(COLLAPSE_NUM)>0) new_s->addAttribute(COLLAPSE_NUM,s->attributeValue(0,COLLAPSE_NUM),sizeof(SgExpression));
	SgStatement *s1=loop_tile(new_s,loop_num);
	SgStatement *s2=get_old_s(loop_num,s1);
	int i;
        for(i=0;i<s->numberOfAttributes(STACK_FREE_STMT);i++){
                SgStatement *stmt=(SgStatement *)s->attributeValue(i,STACK_FREE_STMT);
                s1->insertStmtAfter(*stmt,*s1->controlParent());
        }
        for(i=0;i<s->numberOfAttributes(OC_INDEX_INIT);i++){
                SgStatement *stmt=(SgStatement *)s->attributeValue(i,OC_INDEX_INIT);
                s2->insertStmtBefore(*stmt,*s2->controlParent());
        }
	for(i=0;i<s->numberOfAttributes(INSERTED_AFTER_STMT_IN);i++){
		SgStatement *stmt=(SgStatement *)s->attributeValue(i,INSERTED_AFTER_STMT_IN);
		s1->insertStmtAfter(*stmt,*s1->controlParent());
	}
	for(i=0;i<s->numberOfAttributes(Indepdo_Ref);i++){
		Indepdo_reference * iref = (Indepdo_reference *)s->attributeValue(i,Indepdo_Ref);
		if(iref->getexpression()->numberOfAttributes(KEPT_OC_REF)>0){
			add_oc_access_stmt(s1,s2,s,i);
		}
	}
}

SgStatement * get_first_stmt(SgStatement *s){
	Indepdo_header_def *ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	int loopnum=ihdef->getindexnumber();
	SgStatement *inmost_loop=s;
	for(int i=0;i<loopnum-1;i++){
		inmost_loop=inmost_loop->childList1(0);
	}
	int stmt_num=inmost_loop->numberOfChildrenList1();
	return inmost_loop->childList1(0);
}

SgStatement * get_last_stmt(SgStatement *s){
	Indepdo_header_def *ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	int loopnum=ihdef->getindexnumber();
	SgStatement *inmost_loop=s;
	for(int i=0;i<loopnum-1;i++){
		inmost_loop=inmost_loop->childList1(0);
	}
	int stmt_num=inmost_loop->numberOfChildrenList1();
	return inmost_loop->childList1(stmt_num-3);
}

void add_indepdo_index(SgStatement *s)
{
  int level, i;
  SgStatement *first=get_first_stmt(s);
  SgStatement *last=get_last_stmt(s);
  for (level =1; level<=6 ; level++)
  	for (i=0 ; i< s->numberOfAttributes(INDEX_OPT+level) ; i++ )
  	{
        	SgStatement *push = (SgStatement *)s->attributeValue(i,INDEX_OPT+level);
        	SgStatement * where = first;
        	for ( int j = 0; j < level ; j++ )
                	where=where->controlParent();
	        where->insertStmtBefore(*push, *where->controlParent());
  	}
  for (level =0; level<=6 ; level++){
	  for (i=0 ; i< s->numberOfAttributes(INDEX_ADD+level) ; i++ )
  	{
        	SgStatement *add_stmt = (SgStatement *)s->attributeValue(i,INDEX_ADD+level);
		SgStatement * where = last;
        	for ( int j = 0; j < level ; j++ )
                	where=where->controlParent();

	        where->insertStmtAfter(*add_stmt, *where->controlParent());
  	}
  }
}
void change_indepdo_call_stmt(SgStatement *s){
	if(s->numberOfAttributes(INDEPDO_Call_stmt_with_Arrayref)==0) return;
	int i;
	for(i=0;i<s->numberOfAttributes(INDEPDO_Call_stmt_with_Arrayref);i++){
		SgCallStmt *call=(SgCallStmt *)s->attributeValue(i,INDEPDO_Call_stmt_with_Arrayref);
		SgSymbol *newname=s->getSymbolInCurrentScope(call->expr(0)->symbol()->identifier(),"indepdo",NULL);
		SgCallStmt *newcall=new SgCallStmt(*newname);
		int j;
		for(j=0;j<call->numberOfArgs();j++){
			int refnum;
			if((refnum=call->arg(j)->numberOfAttributes(Array_Arg_num))==0){
			newcall->addArg(call->arg(j)->copy());
			continue;
			}
			newcall->addArg(call->arg(j)->copy());
			SgVarRefExp *dad=(SgVarRefExp *)call->arg(j)->attributeValue(0,Array_Arg_dad);
			newcall->addArg(* dad);
		}
		call->insertStmtBefore(*newcall,*call->controlParent());
		call->deleteStmt();	
	}
}

void change_indepdo_var_ref(SgStatement * s,Indepdo_reference *var_ref){
	change_indepdo_scalar(s,var_ref->getexpression());
}

void change_indepdo_var_exp(SgStatement * s,SgExpression *exp){
	if(exp){
		switch(exp->variant()){
		case MULT_OP: 
		case ADD_OP:
		case MINUS_OP:
			change_indepdo_var_exp(s,exp->lhs());
			change_indepdo_var_exp(s,exp->rhs());
			break;
		}
		 
	}
}

void change_indepdo_scalar(SgStatement * s,SgExpression *scalar){
  Indepdo_header_def * ihdef = (Indepdo_header_def *) s->getAttribute(0, IHEDR_Def)->getAttributeData();
  SgSymbol *scalarsym=scalar->symbol();

  for (int i=0; i<ihdef->getindexnumber(); i++) {
    if(ihdef->getindexsymb(i)->id()==scalarsym->id()) {

      Indepdo_index_def * iidef=(Indepdo_index_def *)s->attributeValue(i,IINDEX_Def);
      if(!iidef->loop_cut) continue;

      SgSymbol * rng=s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);

      // we need to get local index;
      SgExpression * indexexp= new SgVarRefExp(scalarsym);

      SgSymbol * indexs = s->getSymbolInCurrentScope(mixstring, "lis", scalarsym->identifier());
      SgExpression * indexsexp= new SgVarRefExp(indexs);

      indexexp = &((* indexsexp) * (* indexexp));

      SgSymbol * indexl = s->getSymbolInCurrentScope(mixstring, "lil", scalarsym->identifier());
      SgExpression * indexlexp= new SgVarRefExp(indexl);

      indexexp = &((* indexexp) + (* indexlexp));

      SgExpression *globFunc = make_glob(s, rng, indexexp);

      scalar->replace(*globFunc);

      break;
    }
  }
}

void create_indepdo_Lref_array(SgStatement *s){
	SgSymbol * dad_sym=s->getSymbolInCurrentScope(mixstring,"dad","indepdo_tmp_array");
	SgExpression *dad=new SgVarRefExp(dad_sym);
	SgSymbol * array_sym=s->getSymbolInCurrentScope("indepdo_tmp_array");
	SgArrayRefExp * array=(SgArrayRefExp *)(((Indepdo_reference *)s->attributeValue(0,Indepdo_Lref))->getexpression());
	SgStatement *s1=s->controlParent();
	int variant=s1->variant();
	while(variant!=PROG_HEDR&&variant!=PROC_HEDR&&variant!=FUNC_HEDR&&variant!=MODULE_STMT){
		s1=s1->controlParent();
		variant=s1->variant();
	}
	ASSERT(s1);
	if(!s1->numberOfAttributes(Proc_Def)) ASSERT(0);
	SgSymbol *procs=((Processor_def *)s1->attributeValue(0,Proc_Def))->getsymbol();
	SgSymbol *grp=s->getSymbolInCurrentScope(mixstring,"grp",procs->identifier());
	SgStatement *new_array_stmt=make_new_array_data(s,dad_sym,dad_sym->type(),-1,
array->numberOfSubscripts(),new SgVarRefExp(grp));
	s->insertStmtBefore(*new_array_stmt,*s->controlParent());
	SgStatement *delete_array_stmt=make_delete_array(s,dad_sym);
	s->addAttribute(INSERTED_AFTER_STMT,delete_array_stmt,sizeof(SgStatement));
	Indepdo_header_def * ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	for(int i=0;i<array->numberOfSubscripts();i++){
		SgSymbol *rng=s->getSymbolInCurrentScope(mixstring,"indepdo_rng","TEMP",i+1);
		SgExpression *rngExpr=new SgVarRefExp(rng);
		SgExpression *right=make_new_range_distribute(s,ihdef->getindextriple(i,0)->copyPtr(),
			ihdef->getindextriple(i,1)->copyPtr(),1,grp,i+1);
		SgAssignStmt *assign=new SgAssignStmt(*rngExpr,*right);
		s->insertStmtBefore(*assign,*s->controlParent());
		SgStatement *delete_rng=make_delete_range(s,rngExpr);
		s->addAttribute(INSERTED_AFTER_STMT,delete_rng,sizeof(SgStatement));
	}
	for(int i=0;i<array->numberOfSubscripts();i++){
		SgSymbol *rng=s->getSymbolInCurrentScope(mixstring,"indepdo_rng","TEMP",i+1);
		SgExpression *rngExpr=new SgVarRefExp(rng);
		SgStatement *align=make_set_array_align(s,dad_sym,i+1,ihdef->getindextriple(i,0)->copyPtr(),
			ihdef->getindextriple(i,1)->copyPtr(),new SgValueExp(1),
			new SgValueExp(0),new SgValueExp(0),new SgValueExp(0),rngExpr);
		s->insertStmtBefore(*align,*s->controlParent());
	}
}

void trans_func(SgStatement * s)
{
	for(int i=0;i<s->numberOfAttributes(INDEPDO_stmt_with_FUNC);i++){
		SgStatement *stmt=(SgStatement *) s->attributeValue(i,INDEPDO_stmt_with_FUNC);
		for(int j=0;j<stmt->numberOfAttributes(FUNC_EXP);j++){
			SgExpression *exp=(SgExpression *)stmt->attributeValue(j,FUNC_EXP);
			analyse_intrinsic(stmt,exp,stmt);
		}
	}
}

void change_indepdo_body(SgStatement * s)
{
check_indepdo_ghost(s);

Indepdo_header_def * ihdef = (Indepdo_header_def *) s->getAttribute(0, IHEDR_Def)->getAttributeData();

int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};

if(s->numberOfAttributes(Lref_created)){
	create_indepdo_Lref_array(s);
	add_indepdo_coef(s,-1,amount);
	append_indepdo_index(s,-1);
}

define_indepdo_bounds(s);

create_indepdo_section(s);

for (int i=0; i<s->numberOfAttributes(Indepdo_Ref); i++) {

Indepdo_reference * rightref = (Indepdo_reference *) s->attributeValue(i, Indepdo_Ref);

switch (rightref->getexpression()->variant()) {
	case VAR_REF: {
		if(!prechangeable(s,i)) change_indepdo_var_ref(s, rightref);
		break;
		}
	case ARRAY_REF: {
		SgArrayRefExp * rhsExp = (SgArrayRefExp *) rightref->getexpression ();
		SgSymbol * rhsSym=rhsExp->symbol();

		switch (rightref->getCommunication()) {
			case NONE: {
				if(!prechangeable(s,i)){
					if(rhsExp->numberOfAttributes(SUB_REFED_INDEXVAR)){
			        		add_indepdo_coef(s, i, amount);
					        append_indepdo_index(s, i);
			        		change_indepdo_array_reference(s, i);
					}
					else{
						if(!rightref->isScalar()){
							printf("not processed expression\n");
							ASSERT(0);
							exit(1);
						}
						collect_array_ref(s,rhsExp,0);
						analyse_array_ref_list(s,s);
					}
				}	
        			break;
				}

			case SHIFT: {
				int amount[MAX_ALIGN_DIM]={0,0,0,0,0,0,0,0,0,0};

				for (int j=0; j<MAX_ALIGN_DIM;j++)
					amount[j]=rightref->getAmount(0,j);

			        add_indepdo_coef(s, i, amount);

			        append_indepdo_index(s, i);

			        change_indepdo_array_reference(s,i);

			        break;
      			}

			case UNKNOW:
				printf("\tRealize UNKNOW with REMAP-------array:");
				rhsExp->unparsestdout();
				printf("\n");
			case REMAP: {
				create_remap_source(s,i);
				create_remap_target(s,i);
				add_indepdo_remap(s,i);
				change_indepdo_array_reference(s,i,REMAP);
			        break;
			}

			case BROADCAST_DIM:{
				printf("At line:%d.\n",rightref->body->lineNumber());
				printf("This array reference is BROADCAST_DIM communication ---> ");
				rightref->kept->unparsestdout();
				printf("\nNot touched currently,sorry!\n");
				exit(1);
				break;
			}

			case BROADCAST_SCALAR:{
				printf("At line:%d.\n",rightref->body->lineNumber());
				printf("This array reference is BROADCAST_SCALAR communication ---> ");
				rightref->kept->unparsestdout();
				printf("\nNot touched currently,sorry!\n");
				exit(1);
				break;
			}

			case IRREG: {
				printf("At line:%d.\n",rightref->body->lineNumber());
				printf("This array reference is IRREG communication ---> ");
				rightref->kept->unparsestdout();
				printf("\nNot touched currently,sorry!\n");
				exit(1);
			        break;
      			}

			default:
				printf("Return value of detect_communication is invalid!\n");
				exit(1);
				break;
		}//end of switch

		if(rhsExp->numberOfAttributes(Call_Stmt_of_arrayref)) dispose_indepdo_array_argument1(s,i);

		break;
		}
	default: {
		printf("In change_indepdo_body, unexpected expression on line %d, id: %d, variant: %d\n", s->lineNumber(), rightref->getexpression()->id(), rightref->getexpression()->variant());
		ASSERT(0);
		exit(1);
		break;
		}
}//end of switch
}//end of for

do_write_halo(s);
change_loop_bounds(s);
add_indepdo_index(s);
trans_func(s);
change_indepdo_call_stmt(s);
dispose_indepdo_oc_arrayref(s);

}

