#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include        "pass.h"

#include "compilerApi.h"

extern char * mixstring;

extern Flag in_main_program ;
extern Flag in_procedure ;
extern Flag in_function;

extern PassType passtype;

void function2subroutine(SgStatement *s)
{
	SgFuncHedrStmt *func_hedr = isSgFuncHedrStmt(s);

	SgArrayType *returnedType = isSgArrayType(((SgFuncHedrStmt *)s)->returnedType());
	if (returnedType == NULL ) return;

printf("Change function to subroutine :%s\n",s->symbol()->identifier());
	SgSymbol *newarg = func_hedr->resultName();
	if ( newarg == NULL )
	{
		newarg = s->getSymbolInCurrentScope(mixstring, "func",s->symbol()->identifier() );
		newarg->setType(returnedType);
	}

//	((SgFuncHedrStmt *)s)->addArg(*newarg);
	((SgFuncHedrStmt *)s)->addArg(0, *newarg);
	s->thebif->variant = PROC_HEDR;

	s->addAttribute(ARG_ADD,newarg,sizeof(newarg));
	s->symbol()->addAttribute(ARG_ADD,newarg,sizeof(newarg));
}


// replace all the symbol reference in an expression .
void replace_funcname(SgExpression *exp,SgSymbol *funcname,SgSymbol *newarg)
{
	SgSymbol *refSymb=NULL;
	switch ( exp->variant() )  {
	case VAR_REF:
	case ARRAY_REF:
		refSymb = exp->symbol();
	}
	if ( refSymb ) 
	   if (refSymb->thesymb == funcname->thesymb)
		exp->setSymbol(newarg);
	if (exp->lhs()) replace_funcname(exp->lhs(),funcname,newarg);
	if (exp->rhs()) replace_funcname(exp->rhs(),funcname,newarg);
		
}

// change the function call to procedure call 
// the function call is now unified to : A = func (X)
void funccall2sub(SgStatement *s)
{
	// check if it can be converted
	SgExpression *lhs = s->expr(0);
	SgExpression *rhs = s->expr(1);
	SgFunctionCallExp *funccall = isSgFunctionCallExp(rhs);
	if ( funccall == NULL ) return;

	// if return scalar type , then return
	SgArrayType *returnedType = isSgArrayType(funccall->type());
//	if (funccall->variant() == INTRIN_FUNC_CALL)
//		printf("Intrinsic function:%s\n",funccall->symbol()->identifier());
//	else
	if (funccall->variant() == INTRIN_FUNC_CALL) return;
		if (returnedType == NULL ) return;

//	funccall->addArg(*lhs);
	SgExpression *lexprlist = new SgExpression(EXPR_LIST,lhs, rhs , NULL);
	lexprlist->setRhs(funccall->lhs());
	funccall->setLhs(lexprlist);
	funccall->thellnd->variant = PROC_CALL;
	// set the BIF_LL1 & BIF_LL2 according to PROC_STAT
	s->setExpression(0,funccall);
	BIF_LL2(s->thebif) = LLNULL;
        // change the statement variant
	s->thebif->variant = PROC_STAT;
}

// contains 2 pass : FUNCTION_HEDR_PASS & FUNCTION_CALL_PASS
void analyse3(SgStatement * s)
{
  switch(s->variant())
  {
    case GLOBAL:
      {
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
             children[i]=s->childList1(i);
         for(i=0;i<org_numberOfChildren;i++)
             analyse3(children[i]);

        delete [] children;

        break;
      };

   case PROG_HEDR:
      { 
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
             children[i]=s->childList1(i); 

         in_main_program=on;

         for(i=0;i<org_numberOfChildren;i++)
             analyse3(children[i]);

         delete [] children;

         in_main_program=off;
  
         break;
       };

  case INTERFACE_STMT:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse3(children[i]);
      
      delete [] children;

      break;
    }

   case PROC_HEDR:
      {
        SgProcHedrStmt * prochedr=(SgProcHedrStmt *)s;
        int org_numberOfChildren=s->numberOfChildrenList1();
        SgStatement * * children=new SgStatement * [org_numberOfChildren];

        int i;

        for(i=0;i<org_numberOfChildren;i++)
            children[i]=s->childList1(i);

        in_procedure=on;

        for(i=0;i<org_numberOfChildren;i++)
             analyse3(children[i]);
     
        delete [] children;

        in_procedure=off;

        break;
      };

    case FUNC_HEDR:
      {
         SgFuncHedrStmt * prochedr=(SgFuncHedrStmt *)s;
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;

         for(i=0;i<org_numberOfChildren;i++)
              children[i]=s->childList1(i);

         in_function=on;

	 if (passtype == FUNCTION_HEDR_PASS)
	 	function2subroutine(s);

         for(i=0;i<org_numberOfChildren;i++)
             analyse3(children[i]);

         delete [] children;

         in_function=off;

         break;
      };  
    case FORALL_STMT:
      {
         int org_numberOfChildren=s->numberOfChildrenList1();
         SgStatement * * children=new SgStatement * [org_numberOfChildren];

         int i;
         for(i=0;i<org_numberOfChildren;i++)
              children[i]=s->childList1(i);

         for(i=0;i<org_numberOfChildren;i++)
             analyse3(children[i]);

         delete [] children;

         break;
      };  
  case VAR_DECL:
	if (passtype == FUNCTION_HEDR_PASS)
	 	((SgVarDeclStmt *)s)->deleteTheSymbol(*s->getScopeForDeclare()->symbol());
	break;
  case ASSIGN_STAT:
	if (passtype == FUNCTION_HEDR_PASS)
	{
		// in the function body , replace the function reference
		if ( in_function && s->getScopeForDeclare()->numberOfAttributes(ARG_ADD) )
		{
			SgSymbol *funcname = s->getScopeForDeclare()->symbol();
			SgSymbol *newarg = (SgSymbol *)funcname->attributeValue(0,ARG_ADD);
			replace_funcname(s->expr(0),funcname,newarg);
		}

		funccall2sub(s);
	}
	break;
  case COMBINED_DIRECTIVE_DECL:
  case DISTRIBUTE_DECL:
  case INHERIT_DECL:
  case DYNAMIC_DECL:
  case ALIGN_DECL:
	if (passtype == FUNCTION_HEDR_PASS) 
	{
		// change the var_list 's function name
		SgSymbol *funcname = s->getScopeForDeclare()->symbol();
		SgSymbol *newarg = (SgSymbol *)funcname->attributeValue(0,ARG_ADD);
		replace_funcname(s->expr(0),funcname,newarg);
	}
	break;
  }
 
}

