// written by Zheng Gengbin  1998/3/8
// For OUT OF CORE 

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <strings.h>

#define PASSION_CCPLUSPLUS
#include "passion_base.h"

#include "adlib.h"
#include "Seq.h"
#include "pcrc.h"
#include "Array.h"

#include "extname.h"

extern const void *ad_ToPointer(int idx);
extern int ad_FromPointer(const void* ptr);
extern void ad_RmPointer(int idx);

extern int runtimeDebug;
extern int storageMode;
extern int defaultStorageMode;

extern "C" {
   void pcrc_oc_init(int *dad, int *icla_size);
   void pcrc_oc_reset(int *dad);
   void pcrc_oc_filemap(int *dad, void *file, int *namelen, int *icla_size);
   void pcrc_set_ic(int *dad);
   void pcrc_oc_read(int *dad, void *array, int *start, int *stride, int *number);
   void pcrc_oc_write(int *dad, void *array, int *start, int *stride, int *number);
   int  pcrc_oc_prefetch_read(int *dad, void *array, int *start, int *stride, int *number);
   void pcrc_prefetch_wait(int *handle);

   int  pcrc_is_outofcore(int *dad);
}

#define DEBUG

#define FORTRAN_FILE_OFFSET 0


// init an out of core array
void pcrc_oc_init(int *dad, int *icla_size)
{
   pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;

     // try to close existing pfile
   if (ptr->status == OUT_OF_CORE) {
     if (ptr->pfile) {
	ptr->pfile->share->refcount --;
    	if (ptr->pfile->share->refcount == 0)
       		LAF_close(ptr->pfile);
     }
   }
   else
     ptr->status = OUT_OF_CORE;

   ptr->icla_size = ptr->size*5 < *icla_size? ptr->size*5 : *icla_size;
//   ptr->icla_size = *icla_size;

/*
     // create and init file.
   char name[1024];
   sprintf(name,"%s%d-%d","LAF",AdlibLocId());
   char *tempfile=tempnam(".",name);
   if (runtimeDebug == 1)
   	printf("pcrc_oc_int: file:%s\n",tempfile);

   ptr->pfile = PASSION_create(tempfile, FORTRAN_FILE_OFFSET, ptr->size*ptr->elementLen);
*/
   ptr->pfile = LAF_create(AdlibLocId(), FORTRAN_FILE_OFFSET, ptr->size*ptr->elementLen);	/* Fortran Data File have 4 offset */
   if (ptr->pfile == NULL) {
     fprintf(stderr, "** RUNTIME ERROR **  pcrc_oc_init: cannot create Oot-of-core file.\n");
   }
   else {
	// reset data => pfile
     ptr->data = (void *)ptr->pfile;
     ptr->pfile->share->refcount = 1;
     ptr->pfile->share->storage = ptr->major;
        // set ocla_size every rank
     for (int i=0; i<ptr->rank; i++) {
	Range x = ptr->rng(i).tem();
	ptr->pfile->ocla_size[i] = x.width();
     }
   }
}


// reset out of core array's LAF
void pcrc_oc_reset(int *dad)
{
   pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad);
   PFILE *oldPfile = ptr->pfile;

   ptr->pfile = LAF_create(AdlibLocId(), FORTRAN_FILE_OFFSET, ptr->size*ptr->elementLen);	/* Fortran Data File have 4 offset */
   if (ptr->pfile == NULL) {
     fprintf(stderr, "** RUNTIME ERROR **  pcrc_oc_reset: cannot create Oot-of-core file.\n");
   }
   else {
       // reset data => pfile
     ptr->data = (void *)ptr->pfile;
     ptr->pfile->share->refcount = 1;
     ptr->pfile->share->storage = oldPfile->share->storage;
     ptr->pfile->HeaderSize = oldPfile->HeaderSize;
     for (int i=0; i<10; i++) ptr->pfile->ocla_size[i] = oldPfile->ocla_size[i];
   }

     // try to close existing pfile
   oldPfile->share->refcount --;
   if (oldPfile->share->refcount == 0)
       	LAF_close(oldPfile);
}

// map file as LAF to a DAD.
void pcrc_oc_filemap(int *dad, void *file, int *namelen, int *icla_size)
{
   int i;
   pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
   char FileName[1024];

   strncpy(FileName, (char *)file, *namelen);
   for (i=*namelen-1; i>=0 ; i--)  if (FileName[i] != ' ') break;
   FileName[i+1]='\0';

     // try to close existing pfile
   if (ptr->status == OUT_OF_CORE) {
     if (ptr->pfile) {
	ptr->pfile->share->refcount --;
    	if (ptr->pfile->share->refcount == 0)
       		LAF_close(ptr->pfile);
     }
   }
   else
     ptr->status = OUT_OF_CORE;

   ptr->icla_size = ptr->size*5 < *icla_size? ptr->size*5 : *icla_size;
//   ptr->icla_size = *icla_size;

     // associate an existing file to LAF
   Group *grp = (Group*) ad_ToPointer(ptr->grpHandle);
   if ( grp->member() )
   {
/*
	// try to open , create it only when failed.
      ptr->pfile = PASSION_open(FileName, FORTRAN_FILE_OFFSET);
      if (ptr->pfile == NULL)
        ptr->pfile = PASSION_create(FileName, FORTRAN_FILE_OFFSET, ptr->size*ptr->elementLen);	
*/
      ptr->pfile = PASSION_create(FileName, FORTRAN_FILE_OFFSET, ptr->size*ptr->elementLen);	/* Fortran Data File have 4 offset */
      if (ptr->pfile) ptr->pfile->share->refcount = 2;	   // prevent this file from deleting
   }
   else {
      ptr->pfile = LAF_create(AdlibLocId(), FORTRAN_FILE_OFFSET, 0);
      if (ptr->pfile) ptr->pfile->share->refcount = 1;
   }

   if (ptr->pfile == NULL) {
     fprintf(stderr, "** RUNTIME ERROR **  pcrc_oc_filemap: cannot create Out-of-core file.\n");
     return;
   }
     // reset data => pfile
   ptr->data = (void *)ptr->pfile;
   ptr->pfile->share->storage = ptr->major;
   ptr->pfile->share->Status = PASSION_OLD;
     // set ocla_size every rank
   for (i=0; i<ptr->rank; i++) {
	Range x = ptr->rng(i).tem();
	ptr->pfile->ocla_size[i] = x.width();
   }
}

void pcrc_set_ic(int *dad)
{
   pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad) ;
     // try to close existing pfile
   if (ptr->status == OUT_OF_CORE) {
     if (ptr->pfile) {
	ptr->pfile->share->refcount --;
    	if (ptr->pfile->share->refcount == 0)
       		LAF_close(ptr->pfile);
     }
   }
   ptr->status = IN_CORE;
}

int  pcrc_is_outofcore(int *dad)
{
   pcrcfDAD* ptr    = (pcrcfDAD*) ad_ToPointer(*dad);
   return ptr->status == OUT_OF_CORE?1:0;
}

void pcrc_oc_read(int *dadh, void *array, int *start, int *stride, int *number)
{
   pcrcfDAD* dad    = (pcrcfDAD*) ad_ToPointer(*dadh) ;
   char *Array = (char *)array;
   int   StartIndex = *start;
   int   *u = stride;
   int   *DataLen = number;

   if (dad == (pcrcfDAD *)0) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_read) nonexisted dad. \n");
    return;
   }

   if (dad->status == IN_CORE) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_read) in_core array. \n");
    return;
   }

   PFILE *PFilePtr = dad->pfile;

   if (PFilePtr == (PFILE *)0)
   {
	fprintf(stderr,"** RUNTIME ERROR ** : (pcrc_oc_read) Invalid File(s)\n");
	return ;
   }

   if (PFilePtr->share->Status == PASSION_NEW)
	return;

   if (passion_validate_section(dad->rank, StartIndex, u, DataLen, PFilePtr->HeaderSize,
				PFilePtr->ocla_size, dad->icla_size) == -1)
   {
	fprintf(stderr,"PASSION : oc_read(passion_validate_section) Invalid Argument(s)\n");
	return;
   }

   ProfBeginRead();

   int storage = PFilePtr->share->storage;

   switch (dad->rank)
   {
   case 0:
	if (passion_read_block_section_1D(PFilePtr->share->Fd, Array, 0, 
                                         StartIndex, 1, 1,  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size))
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (0D)\n");
#endif
	    ProfEndRead();
	    return ;
	}
	break;
   case 1:
	if (passion_read_block_section_1D(PFilePtr->share->Fd, Array, 0, 
                                         StartIndex, u[0], DataLen[0],  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size))
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (1D)\n");
#endif
	    ProfEndRead();
	    return ;
	}
	break;
   case 2: 
	if (storage == COLUMN_MAJOR)
	{
          if (passion_read_block_section_2D(PFilePtr->share->Fd, Array, 0, StartIndex,
					u[0], u[1], DataLen[0], DataLen[1],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, DataLen[0], 
					PFilePtr->ocla_size[0]))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (2D COLUMN_MAJOR)\n");
#endif
	    ProfEndRead();
	    return ;
	  }
	}
        else {
          if (passion_read_block_section_2D(PFilePtr->share->Fd, Array, 0, StartIndex,
					u[1], u[0], DataLen[1], DataLen[0],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, DataLen[1], 
					PFilePtr->ocla_size[1]))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (2D ROW_MAJOR)\n");
#endif
	    ProfEndRead();
	    return ;
	  }
	}
	break;
   case 3:
	if (storage == COLUMN_MAJOR)
	{
          if (passion_read_block_section_3D(PFilePtr->share->Fd, Array, 0, 
					StartIndex, u[0], u[1], u[2],
					DataLen[0], DataLen[1], DataLen[2],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[0], DataLen[1],
					PFilePtr->ocla_size[0], PFilePtr->ocla_size[1]))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (3D)\n");
#endif
	    ProfEndRead();
	    return ;
	  }
        }
	else {
          if (passion_read_block_section_3D(PFilePtr->share->Fd, Array, 0, 
					StartIndex, u[2], u[1], u[0],
					DataLen[2], DataLen[1], DataLen[0],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[2], DataLen[1],
					PFilePtr->ocla_size[2], PFilePtr->ocla_size[1]))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_read) Read Error (3D)\n");
#endif
	    ProfEndRead();
	    return ;
	  }
	}
	break;
   }

     // adjust buffer to default storage mode
   buffer_storage_convert(Array, dad->rank, DataLen, dad->elementLen, 
                          storage, defaultStorageMode);

   ProfEndRead();
}


void pcrc_oc_write(int *dadh, void *array, int *start, int *stride, int *number)
{
   pcrcfDAD* dad    = (pcrcfDAD*) ad_ToPointer(*dadh) ;
   char *Array = (char *)array;
   int   StartIndex = *start;
   int   *u = stride;
   int   *DataLen = number;

   if (dad == (pcrcfDAD *)0) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_write) nonexisted dad. \n");
    return;
   }

   if (dad->status == IN_CORE) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_write) in_core array. \n");
    return;
   }

   PFILE *PFilePtr = dad->pfile;

   if (PFilePtr == (PFILE *)0)
   {
	fprintf(stderr,"** RUNTIME ERROR ** : (pcrc_oc_write) Invalid Argument(s)\n");
	return ;
   }

   if (passion_validate_section(dad->rank, StartIndex, u, DataLen,  PFilePtr->HeaderSize,
				PFilePtr->ocla_size, dad->icla_size) == -1)
   {
	fprintf(stderr,"PASSION : oc_write(passion_validate_section) Invalid Argument(s)\n");
	return;
   }

   int oldfile = 0;

   if (PFilePtr->share->Status == PASSION_NEW) oldfile=0;
   else oldfile=1;

   ProfBeginWrite();

   int storage = PFilePtr->share->storage;

     // adjust buffer to default storage mode
   buffer_storage_convert(Array, dad->rank, DataLen, dad->elementLen, 
                          defaultStorageMode, storage);

   switch (dad->rank)
   {
   case 0:
	if (passion_write_block_section_1D(PFilePtr->share->Fd, Array, 0, 
                                         StartIndex, 1, 1,  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size, 0))
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (0D)\n");
#endif
	    ProfEndWrite();
	    return ;
	}
	break;
   case 1:
	if (passion_write_block_section_1D(PFilePtr->share->Fd, Array, 0, 
                                         StartIndex, u[0], DataLen[0],  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size, oldfile))
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (1D)\n");
#endif
	    ProfEndWrite();
	    return ;
	}
	break;
   case 2: 
	if (storage == COLUMN_MAJOR)
	{
          if (passion_write_block_section_2D(PFilePtr->share->Fd, Array, 0, StartIndex,
					 u[0], u[1], DataLen[0], DataLen[1],
					 dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size, DataLen[0], 
					 PFilePtr->ocla_size[0], oldfile))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (2D)\n");
#endif
	    ProfEndWrite();
	    return ;
	  }
	}
	else {
          if (passion_write_block_section_2D(PFilePtr->share->Fd, Array, 0, StartIndex,
					 u[1], u[0], DataLen[1], DataLen[0],
					 dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size, DataLen[1], 
					 PFilePtr->ocla_size[1], oldfile))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (2D)\n");
#endif
	    ProfEndWrite();
	    return ;
	  }
	}
	break;
   case 3:
	if (storage == COLUMN_MAJOR)
	{
          if (passion_write_block_section_3D(PFilePtr->share->Fd, Array, 0, 
					StartIndex, u[0], u[1], u[2],
					DataLen[0], DataLen[1], DataLen[2],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[0], DataLen[1],
					PFilePtr->ocla_size[0], PFilePtr->ocla_size[1], 
					oldfile))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (3D)\n");
#endif
	    ProfEndWrite();
	    return ;
	  }
	}
	else {
          if (passion_write_block_section_3D(PFilePtr->share->Fd, Array, 0, 
					StartIndex, u[2], u[1], u[0],
					DataLen[2], DataLen[1], DataLen[0],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[2], DataLen[1],
					PFilePtr->ocla_size[2], PFilePtr->ocla_size[1], 
					oldfile))
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_write) Write Error (3D)\n");
#endif
	    ProfEndWrite();
	    return ;
	  }
	}
	break;
   }
   PFilePtr->share->Status = PASSION_OLD;

   ProfEndWrite();
}

int pcrc_oc_prefetch_read(int *dadh, void *array, int *start, int *stride, int *number)
{
   pcrcfDAD* dad    = (pcrcfDAD*) ad_ToPointer(*dadh) ;
   char *Array = (char *)array;
   int   StartIndex = *start;
   int   *u = stride;
   int   *DataLen = number;
   PREFETCH *PREFETCHp;

   if (dad == (pcrcfDAD *)0) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_read) nonexisted dad. \n");
    return -1;
   }

   if (dad->status == IN_CORE) {
    fprintf(stderr, "** RUNTIME ERROR **  (pcrc_oc_read) in_core array. \n");
    return -1;
   }

   PFILE *PFilePtr = dad->pfile;

   if (PFilePtr == (PFILE *)0)
   {
	fprintf(stderr,"** RUNTIME ERROR ** : (pcrc_oc_prefetch_read) Invalid File(s)\n");
	return  -1;
   }

   if (PFilePtr->share->Status == PASSION_NEW)
	return -1;

   if (passion_validate_section(dad->rank, StartIndex, u, DataLen, PFilePtr->HeaderSize,
				PFilePtr->ocla_size, dad->icla_size) == -1)
   {
	fprintf(stderr,"PASSION : oc_read(passion_validate_section) Invalid Argument(s)\n");
	return -1;
   }

   ProfBeginARead();

   int storage = PFilePtr->share->storage;

   switch (dad->rank)
   {
   case 0:
	PREFETCHp = passion_pread_block_section_1D(PFilePtr->share->Fd, 
					 Array, 0, StartIndex, 1, 1,  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size);
	if (PREFETCHp < 0)
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (0D)\n");
#endif
     	     ProfEndARead();
	     return -1;
	}
	break;
   case 1:
	PREFETCHp = passion_pread_block_section_1D(PFilePtr->share->Fd, Array,0,
                                         StartIndex, u[0], DataLen[0],  
                                         dad->elementLen, PFilePtr->HeaderSize, 
					 dad->icla_size);
	if (PREFETCHp < 0)
	{
#ifdef DEBUG
	    fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (1D)\n");
#endif
     	     ProfEndARead();
	     return -1;
	}
	break;
   case 2: 
	if (storage == COLUMN_MAJOR)
	{
          PREFETCHp = passion_pread_block_section_2D(PFilePtr->share->Fd, Array,
					0, StartIndex,
					u[0], u[1], DataLen[0], DataLen[1],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, DataLen[0], 
					PFilePtr->ocla_size[0]);
	  if (PREFETCHp < 0)
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (2D COLUMN_MAJOR)\n");
#endif
     	     ProfEndARead();
	     return -1;
	  }
	}
        else {
          PREFETCHp = passion_pread_block_section_2D(PFilePtr->share->Fd, Array,
					0, StartIndex,
					u[1], u[0], DataLen[1], DataLen[0],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, DataLen[1], 
					PFilePtr->ocla_size[1]);
	  if (PREFETCHp < 0)
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (2D ROW_MAJOR)\n");
#endif
     	     ProfEndARead();
	     return -1;
	  }
	}
	break;
   case 3:
	if (storage == COLUMN_MAJOR)
	{
          PREFETCHp = passion_pread_block_section_3D(PFilePtr->share->Fd, Array,
					0, StartIndex, u[0], u[1], u[2],
					DataLen[0], DataLen[1], DataLen[2],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[0], DataLen[1],
					PFilePtr->ocla_size[0], PFilePtr->ocla_size[1]);
	  if (PREFETCHp < 0)
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (3D)\n");
#endif
     	     ProfEndARead();
	     return -1;
	  }
        }
	else {
          PREFETCHp = passion_pread_block_section_3D(PFilePtr->share->Fd, Array, 0, 
					StartIndex, u[2], u[1], u[0],
					DataLen[2], DataLen[1], DataLen[0],
					dad->elementLen, PFilePtr->HeaderSize, 
					dad->icla_size, 
					DataLen[2], DataLen[1],
					PFilePtr->ocla_size[2], PFilePtr->ocla_size[1]);
	  if (PREFETCHp < 0)
	  {
#ifdef DEBUG
	     fprintf(stderr,"PASSION : (pcrc_oc_prefetch_read) Read Error (3D)\n");
#endif
     	     ProfEndARead();
	     return -1;
	  }
	}
	break;
   }

     // adjust buffer to default storage mode
   buffer_storage_convert(Array, dad->rank, DataLen, dad->elementLen, 
                          storage, defaultStorageMode);

   ProfEndARead();

   return PREFETCHp?ad_FromPointer(PREFETCHp):-1;
}

void pcrc_prefetch_wait(int *handle)
{
// 	MPIO_Request request;
//	request = (MPIO_Request)ad_ToPointer(*handle);
	PREFETCH *PREFETCHp ;
	if (*handle != -1)
		PREFETCHp = (PREFETCH*)ad_ToPointer(*handle);
	else
		return;

        ProfBeginARead();
	PASSION_prefetch_wait(PREFETCHp);
        ProfEndARead();

	ad_RmPointer(*handle);
}

